# ----------------------------------------------------------------------
# script     : l4c_gen_SOC_v7.4.2.py
#
# invocation : python27 <soc-gen-python-utility>  <new_SOC_file.h5>  <Input-List-Textfile.lis> <START_DAY-of-YEAR> <SOC-INPUT-ROOT-Directory>
# invocation : python27 l4c_gen_SOC_v7.4.1.py SMAP_L4_C_SOC_YD089_v7.4.1.h5  SMAP_L4_C_SOC_INPUTS_v7.4.1.lis  89  /pilv2/SMAP/L4CSYS/v7/ANC/SOC_DATA
# context    : NASA SMAP Mission 2017 Ops support.  Creates new soil-organic-carbon ancillary HDF5 file 
#              for SMAP Level 4 Carbon (L4C) operations for Vv2040/Vv2043 streams
# purpose    : Construct a new L4_C SOC daily intermediate file (SMAP_L4_C_SOC.h5)
#              that includes (2) "calendar" attributes of type int16 since a uint8 {0..255}
#              used previously does NOT represent needed 'day-of-year' values 1..366 adequately!
#
#              This script currently imports 1D-Sparse-Vector, raw binary datasets representing the
#              input HDF5 datasets, as:
#              -- all (3) SOC and NPP sum
#              -- the new FT_ROLLING_M03 daily intermediate

#  NOTE: After talking with Lucas we decided that the FT_ROLLING_AGE_3KM will replace the
#                 FT_ROLLING_QA_M03 raster. We'll still read the FT QA bit from the L3_SM_A
#        surface flag, but will use it to update the FT_ROLLING_AGE_3KM variable now
#              Specific layers imported:
#               # tcf_C[0-2]_M01.flt32, and (3) tcf_Cnum[0-2]_M01.flt32 files along
#              with the tcf_npp_sum_M01.flt32 file.
#
# author     : joe glassy
# maintainer : toby.kundig
# version    : 7.4.2
# revision   : 2021-09-22T09:30, read_store_BPLUT and MD5s modified after type error
# revision   : 2021-09-21T14:53, V6 update
# revision   : 2020-08-24T09:22, Rerun
# revision   : 2020-08-05T13:58, Update to NRv8.3
# revision   : 2020-07-17T11:15, Updated version stream identifiers to Vv5XXX
# revision   : 2020-07-11T12:34, Updates for NR v8.1, BPLUT_2020-07-01.csv
# revision   : 2018-05-20T07:18, Updated for Nature Run v7.2 spin-up/recalibration
# revision   : 2017-06-19T10:40, corrected build of SOC file to use FULL 2D and NOT SPARSE form of 
#              the 2 smrz ancillary binary images. Incorrectly using the 1D sparse forms of these
#              caused a very difficult to identify+correct bug, horizontal striping at 30 deg
#              of the emult_mean,gpp_mean,nee_mean outputs
# revision   : 2017-06-07T217:59, note error in use of sparse smrz
# revision   : 2017-05-31T11:43, update for major v7.2.1 2017 recalibration/spinup task.
# revision   : 2016-03-16T17:32, rebuild to eliminate SOC dataset names that include the start-date,
#              for stream Tv2001, to use start day 90 (March 31, 2015)
# revision   : 2016-03-09T12:42:00, rebuild for stream Tv2001 requiring start day 90 (March 31, 2015)
# revision   : 2015-09-24T16:49:00, after noting SOC monotonic size increases with each daily update
#              and documenting this as caused by GZIP compression, we are trying 0 compression here.
#              Early results show that when compression is disabled, file sizes stay constant with
#              multiple udpates.
# revision   : 2015-08-25T17:43:00, update for new Nature Run Aug 10 spinup, SOC,
#              where new code to trim off datatype (e.g. ".flt32") extensions is used
#
# revision   : 2015-08-20T11:25:00, update for new Nature Run Aug 10 spinup, SOC
# revision   : 2015-06-03T15:15:00, added R11Y2 tag and file attrib indicating which
#              spin-up starting day-of-year the embedded SOC pools are spun up to.
# revision   : 2015-05-29T6:27:00, adjusted ft_calendar,soc_calendar[] to reflect April 13 yearday 103 starting point
# revision   : 2015-05-20T17:24:00, freshen for Nature Run v4 spin calib exercise
# revision   : 2014-11-11T15:03, added ft_calendar_cumulative and soc_calendar_cumulative of
#              datatype int16 to accomodate SUMS in range {1..366} 
# revision   : 2014-07-17T17:44:00, added dynamic NatureRun rootzone soil moisture min,max 9km images
#              that will be updated each day with a new min,max...
# revision   : 2014-02-24T00:46, for D5
#
# Notes      :  xxx_M01land indicates these input files are 1D sparse vector format, where the M01
#               series equals LDOM1KM (134787240) pixels, and the M09 series is LDOM9KM (1664040) pixels
#               as well as (2) 3km (4872x11568) datasets full 2D grid FT_ROLLING_M03 and FT_QC_M03.
#               I note that even when "M01" are input, these end up with rank=1 when they 
#               should be rank 2! 
# Notes      : rootInputDir -- supply explicit rootInputDir so that np.fromfile() is provided with a
#              a full (unambiguous) path to input file
# Notes: To display the actual byte offset address of each dataset, we can use h5ls
# with -av something like this:
# h5ls -av SMAP_L4_C_SOC_YD103_NEW.h5 | grep "Dataset\|Address"
# Notes:  raw source input binary files located in:
# Notes:  raw source inputs: /pilv2/laj/smap/code/mkclimMet/output/land/

'''
   LDOM1KM        134787240ul       Total number of cells in EGv2 1km land domain (LDOM9KM*NCELL2)
   NROW1KM              14616       Number of grid rows in EGv2 1km 
   NCOL1KM              34704       Number of grid cols in EGv2 1km 

   LDOM3KM           56359296
   NROW3KM               4875
   NCOL3KM              11568

   LDOM9KM          1664040ul       Total number of cells in EGv2 9km land domain
   NROW9KM               1624       Number of grid rows in EGv2 9km */
   NCOL9KM               3856       Number of grid cols in EGv2 9km */

'''

AppVers          = "v7.4.2-2020-08-24T11:44:00-0700"
VersId           = "v7.4.2"
SOC_RevTag       = "v7.4.2-2020-08-24T11:45:00-0700"
L4C_OPSSTREAM    = "Vv6040"
L4C_OPSRELEASE   = "OPS_v5_Stream_Vv6040"

L4CSUBSYSTEM_TAG = "L4_C_SUBSYSTEM_OPS_v7.4.2-2021-09-22T14:57:00-0700"

BPLUT_UUID       = "4270d56e-2f2d-4c54-839c-4c19d5f8dd79"
BPLUT_MD5        = "cc2ea579966ab982ab74cd9cad20fe1f"

LDOM1KM = 134787240
LDOM3KM =  56359296
LDOM9KM =    166404

NROW1KM = 14616
NCOL1KM = 34704
NROW3KM =  4875
NCOL3KM = 11568
NROW9KM =  1624
NCOL9KM =  3856

# Original: compression set to 5, but caused increasing filesize over time
compressLevel = 5
# <JMG> 2015-09-24, now trying NO compression
compressLevel = 0
# override later with dataset specific values as desired.
chunkingTupl  = (0,0)

# Default INPUT SOURCE DIR is expected to be overridden, as needed, 
#  by the 3rd command line parameter
rootDefaultInputDir = '/anx_v2/SMAP/L4CSYS/ANC/SOC_DATA'
rootInputDir = '/anx_lagr3/arthur.endsley/SMAP_L4C/calibration/v6_Y2021'

import os,sys
import time
import random
import datetime
import uuid

import numpy as np
import h5py

# ---------------------------------------------------------------------
def logMsg(logf,msg):
    '''
    logMsg()
    inputs: logf -- open log file handle
            msg -- formatted text message to output to log file and console
    URL: for task logging techniques see:
     https://docs.python.org/dev/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes            
    '''
    logf.write(msg+"\n")
    if logf != sys.stdout:
        print(msg)
    return 0
    # end::logMsg()

# ----------------------------------------------------------------------    
def year_DOY_to_Date(yearIdx,doyIdx):
    '''
    function: year_DOY_to_Date(year,doy)
    inputs  : integer yearIdx, integer day-of-year, doyIdx 1..366
    purpose : Convert an integer year YYYY and day of year (yearday) 1..366 to
              a date in YYYY-MM-DD format.
    dependencies : import datetime    
    returns : string, in YYYY-MM-DD format
    notes   : Example:
        rslt = datetime.datetime.strptime('2010 120', '%Y %j')
        rsltDate = str(rslt)[0:10]
        
    notes   : This also works, assuming yr = 2001 and doy = 120
    date.fromordinal(date(yr,1,1).toordinal()+ doy -1)
          Out[14]: datetime.date(2001, 4, 30)

    '''
    thisDate = "UNDEFINED*DATE"
    # construct input string in "YYYY DOY" format
    inputDate = str(yearIdx) + " " + str(doyIdx).zfill(3)
    try:
        # Return 'thisDate' as string 'YYYY-MM-DD'
        thisDate = str(datetime.datetime.strptime(inputDate, '%Y %j'))[0:10]
    except:
        pass
    return thisDate

# ----------------------------------------------------------------------        
def date_to_DOY(yearIdx,monthIdx,dayOfMonthIdx):
    '''
    date_to_DOY() -- given a date in YYYY-MM-DD format, generate a DOY index {1..366}
    dayOfYear = (date.toordinal(date(2001,4,30)) - date.toordinal(date(2001,1,1)))+1
       Out[24]: 0
    # Note: this works when the 'yearIdx' stays 
    '''
    dayOfYear = toordinal(date(yearIdx,monthIdx,dayOfMonthIdx)) - date.toordinal(date(yearIdx,1,1))+1
    # Return the day-of-year integer value {1..365} if a common year, or {1..366} if a leap year
    return dayOfYear 
    # end:: date_to_DOY()


# ---------------------------------------------------------------------
def genISO8601_datetime(myTime=None,suppressSeconds=None,includeTimezone=None):
    """
    function: genISO8601_datetime()
    purpose : Generates a string, ISO 8601 compliant datetime for the current time.
      time.struct_time(tm_year=2012, tm_mon=4, tm_mday=2, tm_hour=14, tm_min=49,
       tm_sec=51, tm_wday=0, tm_yday=93, tm_isdst=1)
    inputs  : time.localtime()|None, True|False, True|False    
    returns : a date-string like: YYYYMMDDThhMM
    notes   : includeTimeZone option is not yet implemented
    notes   : the ":" appears to be an illegal delimiter in some contexts so
              we use the "-" as a time-parts delimiter
    """
    timeStmpDelimiter = '-'
    if myTime is None:
        rNow = time.localtime()
    else:
        rNow = myTime
    # First, generate the YYYY-MM-DD segment
    yearday = str(rNow.tm_year).zfill(4) +"-"+str(rNow.tm_mon).zfill(2)+"-"+str(rNow.tm_mday).zfill(2)
    # next, generate and glue on the 'T' and timestamp HH-MM segment (use dash, to avoid using illegal ':' delimiter)
    timeStmp= "T"+str(rNow.tm_hour).zfill(2) + timeStmpDelimiter + str(rNow.tm_min).zfill(2)
    # Now by default we seconds component, and only suppress it when requested
    if suppressSeconds is None:
        timeStmp = timeStmp + timeStmpDelimiter + str(rNow.tm_sec).zfill(2)
    # We currently omit the timezone specifier
    iso8601Datetime = yearday + timeStmp
    # return YYYYMMDDThhMM{ss}
    return iso8601Datetime 
    # end::genISO8601_datetime()

# ---------------------------------------------------------------------
def generate_uuid(granulePath):
    """
    generate_uuid()
    Inputs: 
        granulePath -- a basename of a file to generate this UUID for
    Methods:
        1) we use Python module UUID
        2) we use the uuid.uuid5() which generates a SHA-1 hash on the
        namespace plus the user provided seed information, such as a URI or URL
    Note:
    DOI for TGARS 2010 paper is DOI="10.1109/TGRS.2010.2070515"
        NOTE: check ATBD for proper CITATION for article below:
           Kim, Y., J. S. Kimball, K. C. McDonald, and J. Glassy. 2010.
           Developing a Global Data Record of Daily Landscape Freeze/Thaw Status
           using Satellite Microwave Remote Sensing.
           IEEE Transactions on Geoscience and Remote Sensing. 
        DOI: 10.1109/TGRS.2010.2070515. 
    """
    uuidString = ""
    DOI_Id = "no-DOI-assigned-yet"
    if(len(granulePath) <= 1):
        granulePath = "SMAP_L4_C_mdl_20150430T000000_V10001_001.h5"
    # we devised this URI seed just as an example...
    uuidSeed = "www.ntsg.umt.edu/SMAP/L4_C/ANC/" + granulePath + "/"+ str(time.time())
    granuleUUID = str(uuid.uuid5(uuid.NAMESPACE_URL,uuidSeed))
    uuidString = granuleUUID
    # return UUID as a string
    return uuidString
    # end::generate_uuid()

# ---------------------------------------------------------------------
def read_store_BPLUT(h5f,h5Varname,lutPath,logf):
    '''
    function: read_store_BPLUT
    purpose : Opens the BPLUT csv file, creates a string type HDF5 datatype
              and stores the BPLUT CSV text file as a string table inside
              the Ancillary file whose handle is h5f
    inputs  : h5f -- an open HDF5 file for this new SMAP_L4_C_Ancillary.h5 file
              h5BPUT_Varname -- the SDS name for this internal BPLUT stored instance.
              lutPath --the fully qualified path where the text LUT lives
              logf -- open log file handle
    returns : Populated string table containing all BPLUT data              
    revised : 2017-06-19T10:12, added to SOC build process    
    notes   : Uses NUMPY style string table handling, wrapped by h5py
    with open(inputPath) as inf:
     lis=[ln.strip("\n\r") for ln in inf if not ln.startswith("#")]
    '''
    strTbl=""
    # Create a special variable length STRING datatype
    thisStrType= h5py.special_dtype(vlen=str)
    if h5Varname not in hObj and os.path.exists(lutPath):
        with open(lutPath) as inf:
            strTbl = np.array([ ln.strip("\n\r") for ln in inf], dtype=thisStrType)
            logMsg(logf,"@A in read_store_BPLUT: storing %s now in %s"%(lutPath,h5Varname))
            dset= h5f.create_dataset(h5Varname,(len(strTbl),),dtype=thisStrType,data=strTbl)
            
        # NOTE: Here we show the idiom to read this BACK into a list, just for example.
        # Convert the h5py dataset object, dset, from array into a simple list
        # strTbl=[dset[i] for i in range(len(dset))]
    else:
        logMsg(logf,"@E (%s) Cannot open input OR HDF5 variable name is already defined -cannot overwrite!"%(h5Varname))
    # Returns string table containing the BPLUT, just in case its needed
    return strTbl
    # end::read_store_BPLUT()

# ---------------------------------------------------------------------
def read_store_MD5s(h5f,h5Varname,md5Path,logf):
    '''
    function: read_store_MD5s(h5f,md5Path,h5Varname,logf)
    purpose : Opens the MD5s csv file that contains MD5s for all raw 
              binary inputs, creates a string type HDF5 datatype
              and stores the MD5's CSV text file as a string table inside
              the Ancillary file whose handle is h5f
    inputs  : h5f -- an open HDF5 file for this new SMAP_L4_C_Ancillary.h5 file
              h5BPUT_Varname -- the SDS name for this internal MD5s stored instance.
              lutPath --the fully qualified path where the text MD5s live
              logf -- handle to open log file
    revised : 2017-06-19T10:12, added to SOC build process
    returns : Populated string table containing all MD5's for raw binary input layers
    notes   : Uses NUMPY style string table handling, wrapped by h5py
    with open(inputPath) as inf:
     lis=[ln.strip("\n\r") for ln in inf if not ln.startswith("#")]
    
    '''
    strTbl=""
    # Create a special variable length STRING datatype
    thisStrType= h5py.special_dtype(vlen=str)
    if h5Varname not in hObj and os.path.exists(md5Path):
        with open(md5Path) as inf:
            logMsg(logf,"@A in read_store_MD5: storing %s now in %s"%(md5Path,h5Varname))
            strTbl = np.array([ ln.strip("\n\r") for ln in inf], dtype=thisStrType)
            dset= h5f.create_dataset(h5Varname,(len(strTbl),),dtype=thisStrType,data=strTbl)
        # NOTE: Here we show the idiom to read this BACK into a list, just for example.
        # Convert the h5py dataset object, dset, from array into a simple list
        # strTbl=[dset[i] for i in range(len(dset))]
    else:
        logMsg(logf,"@E (%s) Cannot open input OR, HDF5 variable name for MD5 is already defined -cannot overwrite!"%(h5Varname))
    # Returns string table containing the MD5s, just in case its needed
    return strTbl
    # end::read_store_MD5s()


# ---------------------------------------------------------------------
def createCalendarArrays(hObj,spinUpDataDayIndex=None):
    '''
    createCalendarArray()
    
    CALENDAR checkoff arrays are now of datatype int16 since this is
    required to accomodate dayOfyear values {1..366} for both 
     ft_calendar[] and soc_calendar[].
     As each model day {1..366} inclusive is explicitly visited, a '1' value is 
     written to that day's slot. Assuming a base0 array, the first (0th)
     slot is conventionally left empty, Day 1 goes into array slot [1], etc.
    '''
    L4C_N_CALENDAR_SLOTS = 370
    
    ft_calendarAry = np.zeros(L4C_N_CALENDAR_SLOTS,dtype='uint8')
    soc_calendarAry= np.zeros(L4C_N_CALENDAR_SLOTS,dtype='uint8')

    ft_calendarCumulatives  = np.zeros(L4C_N_CALENDAR_SLOTS,dtype='int16')
    soc_calendarCumulatives = np.zeros(L4C_N_CALENDAR_SLOTS,dtype='int16')
    
    # If a mid-year "spin-up-to-this-day" index is supplied, set all
    # slots to 1 starting at day 1 and ending at day spinUpDataDayIndex
    if spinUpDataDayIndex is not None:
        print("@A createCalendarArrays: USING CUSTOM spin-up-to date %d"%(spinUpDataDayIndex))
        for dataDay in range(spinUpDataDayIndex):
            ft_calendarAry[dataDay]  = 1
            soc_calendarAry[dataDay] = 1
            # mark equiv. cumulatives all starting at 1
            ft_calendarCumulatives[dataDay] = 1
            soc_calendarCumulatives[dataDay]= 1
    else:
        print("createCalendarArrays Marking calendars with DEFAULT DAY 1 spin-up-to date")
    
    # Freeze-Thaw daily checkoff list
    hObj.attrs['ft_calendar' ]  = ft_calendarAry
    # same for SOC
    hObj.attrs['soc_calendar' ] = soc_calendarAry
    
    # Also write CUMULATIVE calendar values to allow cross-check 
    # such that as each model day is visited, we INCREMENT BY ONE
    # each slot. That way, if any slot is >1 we can tell that this 
    # particular day has been RE-RUN that number times, as can occur
    # in repetitive testing or GMAO Ops "restart" runs
   
    hObj.attrs['ft_calendar_cumulative' ]  = ft_calendarCumulatives
    # same for SOC
    hObj.attrs['soc_calendar_cumulative' ] = soc_calendarCumulatives
    
    return 0
    # end:: createCalendarArrays()
    
# ----------------------------------------------------------------------
def createFileAttributes(hObj,thisGranuleName,spinUpToDataDayIdx):
    '''
    function :createFileAttributes()
    inputs   : hObj,thisGranuleName,spinUpToDataDayIdx
    purpose  : create file level attributes for SMAP_L4_C_SOC.h5 using
    model of Ancillary.h5, currently written into the ROOT "/" group.
    
    '''
    #   create file level attributes (of type string)
    hObj.attrs['project'  ]   = "NASA SMAP Level 4 Carbon (L4_C) Model"

    hObj.attrs['TITLE']      = "SMAP L4_C Ancillary data -- EASEGRID2 1km (MO1),3km (M03) and 9km(M09) masks and land cover definitions"
    hObj.attrs['MISSION'  ]  = "NASA SMAP Mission, Operational phase"
    hObj.attrs['PROJECT'  ]  = "NASA SMAP Level 4 Carbon Model (L4_C) daily,global 9km"
    hObj.attrs['DELIVERY_VERSION'  ] = VersId
          
    # INCLUDE new MANDATORY subsystem release tag
    hObj.attrs['L4_C_SUBSYSTEM_RELEASE'  ] = L4CSUBSYSTEM_TAG
    
    hObj.attrs['L4_C_SW_VERSION'         ] = AppVers
    hObj.attrs['L4_C_OPS_STREAM'         ] = L4C_OPSSTREAM
    hObj.attrs['L4_C_OPS_RELEASE'        ] = L4C_OPSRELEASE
    hObj.attrs['L4_C_DATASET_VERSION'    ] = VersId
    hObj.attrs['L4_C_DATASET_DATETIME'   ] = AppVers
    hObj.attrs['L4_C_SW_DATETIME'        ] = genISO8601_datetime()
    hObj.attrs['L4_C_PRODUCTION_DATETIME'] = genISO8601_datetime()
    hObj.attrs['L4_C_SOC_GRANULE' ]        = thisGranuleName
    
    hObj.attrs['SMAP_L4_C_PSD_VERS'     ] = "1.8.1"
    hObj.attrs['SMAP_L4_C_PSD_DATE'     ] = "2015-06-22"

    # DOI is assigned for main L4CMDL
    hObj.attrs['L4_C_MDL_DOI'           ] = "10.5067/22TFAUSNLO9R"
    # DOI not yet assigned for SMAP_L4_C_Ancillary file itself
    hObj.attrs['L4_C_ANC_DOI'           ] = "DOI TBD for SMAP_L4_C_SOC.h5"
    
    hObj.attrs['granule_UUID' ]  = generate_uuid(thisGranuleName)
    
    hObj.attrs['software_used_to_generate'] = sys.argv[0]
    # hObj.attrs['software_author'          ] = "Joe Glassy,NASA SMAP L4_C team"
    # hObj.attrs['production_poc'         ]   = "Joe Glassy, Univ. Montana L4_C lead software engineer"
    hObj.attrs['research_algorithm'     ]   = "tcf_spin_v3.6,laj"
    hObj.attrs['research_science_PI'    ]   = "Dr. John Kimball, NTSG, Univ. Montana"
    hObj.attrs['research_code_author'   ]   = "Lucas Jones, Univ. Montana"
    hObj.attrs['data_source_host'       ]   = os.uname()[1]
    #
    hObj.attrs['projection_name'    ] = "EASEGRID v2"
    
    # <JMG> 2015-09-24T17:52:00 added for documentation
    hObj.attrs['compression_level_used' ] = compressLevel 
    # Store the chunking tuple used, as a string
    hObj.attrs['chunking_used' ]          = 'none'
    # <JMG> 2015-09-25T12:20:00, Store VERSION of h5py used and HDF5 library used
    hObj.attrs['h5py_version' ] = h5py.version.version
    hObj.attrs['HDF5_library_version' ] = h5py.version.hdf5_version
    hObj.attrs['Numpy_library_version' ] = np.version.version
    hObj.attrs['HDF5_library_version' ] = h5py.version.hdf5_version
    
    hObj.attrs['OPS_STREAM' ] = "Vv5022 with SOC DOY 89 start day March 30, 2015"
    
    
    # <JMG> added 6/3/2015, to make explict which DOY these SOC
    # pools represent, default is 1 (Jan 1) but may be any mid-year DOY
    hObj.attrs['SOC_SPINUP_START_DAYOFYEAR_DEFINITION' ] ='The START DOY reflects which DOY these SOC pools are spun up when first used'
    if spinUpToDataDayIdx == 0:
        hObj.attrs['SOC_SPINUP_START_DAYOFYEAR' ] = 1
    else:
        hObj.attrs['SOC_SPINUP_START_DAYOFYEAR' ] = spinUpToDataDayIdx
    
    hObj.attrs['soc_pool_labels']   = 'SOC 0=metabolic,fast 1=structural,medium, 2=recalcitrant,slow'
    hObj.attrs['soc_pool_0_fast'  ] = 'SOC code: 0 = metabolic,fast'
    hObj.attrs['soc_pool_1_medium'] = 'SOC code: 1 = structural,medium'
    hObj.attrs['soc_pool_2_slow'  ] = 'SOC code: 2 = recalcitrant,slow'
    
    hObj.attrs['npp_sum_and_soc_units'] = 'gC/m^-2'
      
    hObj.attrs['projection_1km_rows'           ] = NROW1KM
    hObj.attrs['projection_1km_columns'        ] = NCOL1KM
    hObj.attrs['projection_3km_rows'           ] = NROW3KM
    hObj.attrs['projection_3km_columns'        ] = NCOL3KM
    hObj.attrs['projection_9km_rows'           ] = NROW9KM
    hObj.attrs['projection_9km_columns'        ] = NCOL9KM
      
   
    # store counts of 1D sparse vector pixels for 1KM,9KM res
    # LDOM1KM is 134787240
    # LDOM9KM is 1664040
    # LDOM3KM = 56359296
    #
    hObj.attrs['LDOM1KM_1D_SparsePix' ] = LDOM1KM
    hObj.attrs['LDOM3KM_1D_SparsePix' ] = LDOM3KM
    hObj.attrs['LDOM9KM_1D_SparsePix' ] = LDOM9KM
    #
    # Fetch name of host we built this on, via uname -a parts...
    hObj.attrs['data_source_host'   ] = os.uname()[1]
    hObj.attrs['BPLUT_version'      ] = "2020-07-31"
    #  <JMG> matched to UUID inside of official CSV file
    hObj.attrs['BPLUT_UUID'  ] = BPLUT_UUID
    hObj.attrs['BPLUT_MD5'   ] = BPLUT_MD5
    
    hObj.attrs['L4CMET_FT_SOURCE_OPTIONS' ] = "(GEOS5_TS|ACTIVE_RADAR|ACTIVE_TSURF|PASSIVE_RADIOMETER)"
    
    # Include the QA legend as a convenience
    hObj.attrs['QA_BITFLAG_L01_HDR1'  ] = 'QA    Bitflag      Bit Position(s)  N.    Value        Description'
    hObj.attrs['QA_BITFLAG_L02_HDR2'  ] = 'flag  Name         (Start,End)      Bits  Range'
    hObj.attrs['QA_BITFLAG_L03_HDR3'  ] = '------------------------------------------------------------------------------------------'
    hObj.attrs['QA_BITFLAG_L04_FLAG01']="01    NEE bit       00 - 00         (1)   {0|1}        0=NEE within valid range,1=out of range"
    hObj.attrs['QA_BITFLAG_L04_FLAG02']="02    GPP bit       01 - 01         (1)   {0|1}        0=GPP within valid range,1=out of range"
    hObj.attrs['QA_BITFLAG_L04_FLAG03']="03    Rh  bit       02 - 02         (1)   {0|1}        0=Rh  within valid range,1=out of range"
    hObj.attrs['QA_BITFLAG_L04_FLAG04']="04    SOC bit       03 - 03         (1)   {0|1}        0=SOC within valid range,1=out of range"
    hObj.attrs['QA_BITFLAG_L04_FLAG05']="05    PFT dominant  04 - 07         (4)   {1..8}       Dominant PFT class, decimal integer {1..8} inclusive"
    hObj.attrs['QA_BITFLAG_L04_FLAG06']="06    QA score      08 - 11         (4)   {0,1,2,3}    0=best,3=worst; overall Quality score"
    hObj.attrs['QA_BITFLAG_L04_FLAG07']="07    GPP method    12 - 12         (1)   {0|1}        0=used MODIS FPAR,1=used FPAR climatology"
    hObj.attrs['QA_BITFLAG_L04_FLAG08']="08    NDVI method   13 - 13         (1)   {0|1}        0=used FPAR or FPAR clim, 1=used NDVI"
    hObj.attrs['QA_BITFLAG_L04_FLAG09']="09    F/T method    14 - 14         (1)   {0|1}        0=used RADAR F/T, 1=used Tsurf F/T"
    hObj.attrs['QA_BITFLAG_L04_FLAG10']="10    IsFill        15 - 15         (1)   {0|1}        0=is NOT a Fillvalue, 1=Is a Fillvalue" 
    
    # we repeate the PFT classification as a file-level attribute set for user convenience
    hObj.attrs['land_cover_classification' ] = "PFT Plant Function Type (PFT) from MODIS MCD Type-5"
    hObj.attrs['PFT_00_Water '        ] = "00 water"
    hObj.attrs['PFT_01_ENF'           ] = "01 ENF evergreen needleleaf tree"
    hObj.attrs['PFT_02_EBF'           ] = "02 EBF evergreen broadleaf tree"
    hObj.attrs['PFT_03_DNF'           ] = "03 DNF deciduous needleleaf tree"
    hObj.attrs['PFT_04_DBF'           ] = "04 DBF deciduous broadleaf tree"
    hObj.attrs['PFT_05_Shrub'         ] = "05 Shrub"
    hObj.attrs['PFT_06_Grass'         ] = "06 Grass"
    hObj.attrs['PFT_07_Cereal_crop'   ] = "07 Cereal crop"
    hObj.attrs['PFT_08_Broadleaf_crop'] = "08 Broadleaf crop"
    hObj.attrs['PFT_09_Urban'         ] = "09 Urban and built-up landuse (NOT-USED)"
    hObj.attrs['PFT_10_SNOW_ICE'      ] = "10 Snow and ice (NOT-USED)"
    hObj.attrs['PFT_11_BARREN  '      ] = "11 Barren or sparsely vegetated (NOT-USED)"
    hObj.attrs['PFT_254_UNCLASSIFIED' ] = "254 Unclassified (NOT-USED)"
       
    # end::createFileAttributes()
    
    
# ---------------------------------------------------------------------
def createNewSOC(hObj,srcPath,dType,socTypeIndex):
    '''
    createNewSOC(hObj,srcPath,dType,socType)
    Inputs: hObj --HDF5 handle
            srcPath -- path to SOC raw binary source file
            dType -- datatype 'float32'
            socTypeIndex {0,1,2} for: 'metabol','structu','recalcitrant'
    
    KEYNOTE 9/24/2015 to mitigate SOC "growing-file-size-with-ea-update" problem, I
     set compression level to 0 and disable chunking, resulting in a much
     larger SOC file, but one that doesn't get larger over time.
    '''
    fullWorkingPath = ''
    hDatasetName = 'undefined-hDatasetname'
    # define SOC layer labels 0,1,2 in order:
    socLabels = ['metab','struc','recal']
    # set hard defaults
    LDOM1KM = 134787240
    LDOM9KM = 1664040
    # LDOM3KM should be the size of 
    #  the official MCD12Q1_M03_lc_mask.uint8 file
    LDOM3KM = 56359296
    #
    socFillvalue = -9999.0
    ftFillvalue  = 254 
    fillVal      = socFillvalue  # set as default...

    # take early out if a bogus file is submitted
    if len(srcPath) < 4:
        return hObj
    #
    fullWorkingPath = rootInputDir + "/" + srcPath
    #   If file doesn't exist, take early out
    if not os.path.exists(fullWorkingPath):
        print("@E ERROR: file (%s) does NOT exist!"%(fullWorkingPath))
        return hObj

    # DEFINE HDF5 datasetname on basis of srcPath
    # for a path like: myPath.flt32 stores just "myPath"
    hDatasetName = os.path.splitext(srcPath)[0] ;
   
    print("@A SOURCE PATH (%s) HDF5-Datasetname %s"%(srcPath,hDatasetName))

    # establish baseline defaults if no "M01" or "M09" etc present
    rank = 2        
    nRows = nRows9km
    nCols = nCols9km
    nElem = nRows9km * nCols9km
    
    if srcPath.find("_M01land") >=0 :
        rank = 1
        nElem = LDOM1KM
        nRows = 1
        nCols = LDOM1KM
    elif srcPath.find("_M03land") >= 0:
        rank = 1        
        nElem = LDOM3KM
        nRows = 1 
        nCols = LDOM3KM
    elif srcPath.find("_M09land") >= 0:
        rank = 1        
        nElem = LDOM9KM
        nRows = 1
        nCols = LDOM9KM
    elif srcPath.find("_M01") >= 0:
        rank = 2        
        nRows = nRows1km
        nCols = nCols1km
        nElem = nRows1km * nCols1km
    elif srcPath.find("_M03") >=0:
        rank = 2        
        nRows = nRows3km
        nCols = nCols3km
        nElem =  nRows3km * nCols3km
    elif srcPath.find("_M09") >=0:
        rank = 2        
        nRows = nRows9km
        nCols = nCols9km
        nElem = nRows9km * nCols9km
    
    dType = "float32"
    # Convert raw binary file datatype descriptor to (numpy+H5Py) descriptor
    if srcPath.find(".flt64")   >=0 :
        dType = "float64"   
    elif srcPath.find(".flt32") >=0 :
        dType = "float32"
    elif srcPath.find(".uint32") >= 0:
        dType = "uint32"
    elif srcPath.find(".int32") >= 0:
        dType = "int32"
    elif srcPath.find(".uint16") >= 0:
        dType = "uint16"
    elif srcPath.find(".int16") >= 0:
        dType = "int16"
    elif srcPath.find(".uint8") >= 0:
        dType = "uint8"
    
    # Create SOC group if it does not already exist
    grpLis = hObj.keys()
    if not "SOC" in grpLis:
        socGrp = hObj.create_group("SOC")
    else:
        socGrp = hObj["SOC"]
        
    # M01land means: 1D sparse vector format!
    # SOC raw binary example: tcf_Cana0_M01land.flt32
    # NPP raw binary example: tcf_npp_sum_M01land.flt32
    #
    # isolate baseNam using srcPath in os.path.split() where
    # for example: "/this/is/a/path/todata.flt32" 
    # [0] is the part: "/this/is/a/path"
    # [1] is the part: "todata.flt32"
    # THEN, use os.path.splitext("todata.flt32") to further split out
    rootPath      = os.path.split(srcPath)[0]
    fullbaseName  = os.path.split(srcPath)[1] 
    baseNam       = os.path.splitext(fullbaseName)[0]

    # SMAP_NatureRun_v3_smrz_{min|max}_sat_M09.flt32 layers
    
    # NEW simplified form of soil rootzone min,max saturation files is 
    # <JMG> 2015-08-25, : smrz_min_sat_M09 |  smrz_max_sat_M09
    lblPath = srcPath 
    if srcPath.find("FT") >= 0:
        fillVal = ftFillvalue
        hDatasetName = hDatasetName.replace("tcf_","")      
    elif srcPath.find("_C") >= 0:
        fillVal = socFillvalue
        # <JMG> 2016-03-16, for SOC, DO STRIP OFF any name qualifier including
        # the SOC carbon spin-up date, to make these names generic again
        # L4C_INPUT_DATASET_SOC_C0  = SOC/SOC_C0_M01
        # L4C_INPUT_DATASET_SOC_C1  = SOC/SOC_C1_M01
        # L4C_INPUT_DATASET_SOC_C2  = SOC/SOC_C2_M01
        # and NOT this --> L4C_INPUT_DATASET_SOC_C0  = SOC/SOC_C0_M01_0001090
        hDatasetName = hDatasetName.replace("tcf_","SOC_")
        t = hDatasetName.split("_")
        hDatasetName = "_".join(t[0:3])
        
    print("@A ORIG name (%s) POST-ADJUSTED name (%s)"%(lblPath,hDatasetName))
        
    # <JMG> 2015-08-25, adopt new convention to strip off datatype identifier
    # TBD this could be fine tuned to strip off any trailing qualifier
    #  that begins at the rightmost "." symbol for a ".XYZ" etc...
    if hDatasetName.endswith(".uint8"):
        hDatasetName = hDatasetName.replace(".uint8","")
    if hDatasetName.endswith(".int16"):
        hDatasetName = hDatasetName.replace(".int16","")
    if hDatasetName.endswith(".uint16"):
        hDatasetName = hDatasetName.replace(".uint16","")
    if hDatasetName.endswith(".int32"):
        hDatasetName = hDatasetName.replace(".int32","")
    if hDatasetName.endswith(".uint32"):
        hDatasetName = hDatasetName.replace(".uint32","")
    if hDatasetName.endswith(".flt32"):
        hDatasetName = hDatasetName.replace(".flt32","")

    print("@A createSOC srcPath(%s) baseNam (%s) hdatasetName %s fillvalue %s"%(srcPath,baseNam,hDatasetName,str(fillVal)))    
        
    # skip any invalid or improperly formed dataset names...
    if len(hDatasetName) < 8:
        print("@E WARNING! Omitting improperly formed dataset(%s)"%(hDatasetName))
        return hObj

    print("@A np.fromfile dsName(%s) fullWorkingPath(%s) datatype(%s) nElem %d nRows %d nCols %d"%(hDatasetName,fullWorkingPath,dType,nElem,nRows,nCols))
    #
    
    if rank == 1:
        # handle as a 1D sparse vector array now
        dtaArray = np.fromfile(fullWorkingPath,dtype=dType)
    else:
        # handle as a 2D grid array. Only newer h5py versions support the fillvalue argument. 
        # e.g. on stout,missouri but NOT on barley or discover!
        dtaArray = np.fromfile(fullWorkingPath,dtype=dType).reshape(nRows,nCols)
        
    if (srcPath.find("FT") >= 0 or srcPath.find("SMAP_") >= 0) or srcPath.find("smrz")>=0 :
        print("@A writing (%s) into the root group"%(hDatasetName))
        # write this dataset into the ROOT group, "/"
        if rank == 1:
            dsObj = hObj.create_dataset(hDatasetName,(nElem,),dType,data=dtaArray,fillvalue=fillVal)
        else:
            dsObj = hObj.create_dataset(hDatasetName,(nRows,nCols),dType,data=dtaArray,fillvalue=fillVal)
            
    elif (srcPath.find("_C") >= 0 or srcPath.find("NPP")>= 0 or srcPath.find("npp")>=0 ) :
        print("@A writing (%s) into the /SOC group"%(hDatasetName))
        # write this dataset into the /SOC group
        if rank == 1:
            dsObj = socGrp.create_dataset(hDatasetName,(nElem,),dType,data=dtaArray,fillvalue=fillVal)
        else:
            dsObj = socGrp.create_dataset(hDatasetName,(nRows,nCols),dType,data=dtaArray,fillvalue=fillVal)

    # flush I/O buffers thus far
    hObj.flush()
    
    if not dsObj:
        print("@E createNewSOC: unexpected absence of dataset object dsObj for %s"%(hDatasetName))
        hObj.close()
        
    dsObj.attrs['valid_min'    ]  = 'TBD'
    dsObj.attrs['valid_max'    ]  = 'TBD'
    
    #   ADD: create dataset level attributes
    if hDatasetName.find("NPP_SUM")>=0 or hDatasetName.find("npp_sum")>=0 :
        dsObj.attrs['npp_sum_litterfall' ] = "NPP net primary productivity litterfall, daily intermediate state variable"
        dsObj.attrs['missing_value']  = socFillvalue
        dsObj.attrs['fill_value'   ]  = socFillvalue
        dsObj.attrs['units'        ]  = "gC/m^-2" 
        dsObj.attrs['data_resolution'    ] = '1km'  
    elif hDatasetName.find("smrz") >= 0 :
        dsObj.attrs['smrz_min_max_sat' ] = "SMAP Nature Run soil rootzone min or max saturated 9km daily updated intermediate state variable"
        dsObj.attrs['missing_value']  = -9999.0
        dsObj.attrs['fill_value'   ]  = -9999.0
        dsObj.attrs['units'   ]    = "saturation (dimensionless)"
        dsObj.attrs['data_resolution'    ] = '9km'  
        dsObj.attrs['valid_min'    ] = 0.0
        dsObj.attrs['valid_max'    ] = 1.0
    elif hDatasetName.find("SOC")>=0 or hDatasetName.find("_C")>=0:
        dsObj.attrs['soil_organic_carbon' ] = "Soil organic carbon daily intermediate spatio-temporal state variable "
        dsObj.attrs['missing_value']  = socFillvalue
        dsObj.attrs['fill_value'   ]  = socFillvalue
        dsObj.attrs['units'        ]  = "gC/m^-2" 
        dsObj.attrs['data_resolution' ] = '1km' 
        dsObj.attrs['valid_min'    ] = 0.0
        dsObj.attrs['valid_max'    ] = 25000.0
        if socTypeIndex >= 0 :
            dsObj.attrs['soc_pool'  ] = socLabels[socTypeIndex] 
    elif hDatasetName.find("FT") >= 0 :
            dsObj.attrs['freeze_thaw_rolling_state' ] = "Rolling 3km Freeze Thaw or Freeze Thaw QC FT_QC_M03 "
            dsObj.attrs['missing_value']   = ftFillvalue
            dsObj.attrs['fill_value'   ]   = ftFillvalue
            dsObj.attrs['units'        ]   = "dimensionless"
            dsObj.attrs['data_resolution'] = '3km'
        
    # CF convention like dataset attributes
    # Once confirmed, substitute in correct valid_min,valid_max
    
    dsObj.attrs['short_name']  = hDatasetName 
    dsObj.attrs['long_name ']  = hDatasetName 

    # VERIFY if we need the fullWorkingPath here...
    dsObj.attrs['data_source'        ] = fullWorkingPath
    #
    dsObj.attrs['data_source_label'  ] = hDatasetName
    
    
    # Return open handle to new HDF5 file in case more needs
    return hObj
    # end::createNewSOC()

# ---------------------------------------------------------------------
if __name__ == "__main__":
    '''
    main()
    '''
    # set default operating parameters
    nRows = 14616
    nCols = 34704
    #
    nRows1km = 14616
    nCols1km = 34704
    #
    nRows3km = 4872
    nCols3km = 11568
    #
    nRows9km = 1624
    nCols9km = 3856
    #
    dType = 'float32'
    inputSOCListPath = "soc-undefined"
    # SMAP_BPLUT_2018-05-03.csv  SMAP_SOC_MD5s.csv
    bplutPath = "SMAP_BPLUT_Vv60XX_rev_20210908.csv"
    md5sPath  = "SMAP_SOC_MD5s.csv"
    
    socOutHdf5Path   = "soc-output-hdf5-undefined" # SMAP_L4_C_SOC_DailyIntermediate_1km_D4.h5
    spinUpToDataDayIdx = 0  # default if not set
    
    # create session log file
    dateLbl = datetime.datetime.now().isoformat()[0:10]
    masterLogPath = "l4c_gen_SOC_" + VersId + "_"+ dateLbl + ".log"
    logf = open(masterLogPath,"wt")

    
    if len(sys.argv) >= 3:
        socOutHdf5Path = sys.argv[1]
        inputSOCListPath = sys.argv[2]
        if len(sys.argv) >= 4:
            spinUpToDataDayIdx = int(sys.argv[3])
            # Also, allow user to optionally override default rootInputDir directory
            if len(sys.argv) >= 5:
                rootInputDir = sys.argv[4]
    else:
        logMsg(logf,"%s invocation: <socOutHdfPath> <inputSOCListPath> {spinUpToYearDayIndex} {rootInputDir} " %(sys.argv[0]))
        sys.exit(-1)
    #
    strtSec = time.time()

    logMsg(logf,"\n@A %s started at %s"%(sys.argv[0],time.ctime()))
    logMsg(logf,"@A input list of target files to ingest (%s)"%(inputSOCListPath))
    logMsg(logf,"@A output SOC HDF5 file (%s)"%(socOutHdf5Path))
    logMsg(logf,"@A root input directory specified (%s)"%(rootInputDir))
    
    if spinUpToDataDayIdx >0:
        logMsg(logf,"@A Custom SPIN-UP! Will spin FT and SOC calendars to to data-day %d"%(spinUpToDataDayIdx))

    # Read in the list of files to import into this HDF5 file, from input list 
    # and 7/17/2014, added filter to skip over comment lines...
    inpLis =[ln.strip('\n\r') for ln in open(inputSOCListPath) if not ln.startswith('#') ]
    #
    logMsg(logf,"@A %d source input EASEGRIDv2 1km float 32 raw binary files will be ingested"%(len(inpLis)))
    # Immediately PRE-QUALIFY the existence of each target input file, so
    # that if any fail to exist, we stop early to correct it
    ithFile = 0 
    for candidateFile in inpLis:
        workingPath = candidateFile
        ithFile += 1
        # first try, verify if bare path supplied matches location
        if not os.path.exists(workingPath):
            workingPath = rootInputDir + "/" + workingPath
            # second try, verify if located relative to rootInputDir
            if not os.path.exists(workingPath):
                logMsg(logf,"@E %s Requested input file %3d (%s) does NOT exist as stated!"%(sys.argv[0],ithFile,workingPath))
                sys.exit(-1)
        else:
            logMsg(logf,"@A Requested input file %3d (%s) appears ready for ingest"%(ithFile,workingPath))

    hObj = h5py.File(socOutHdf5Path,"w")
    
    # CALENDAR checkoff arrays are now ft_calendar[] and soc_calendar[]
    #  Now optionally can supply 2nd param as a data-day index (1..365) to set the
    #  calendars as "spun up" to. Example if day 121 is supplied, it marks all
    #  days from 1..121 as '1' to indicate that day has been processed. This is
    #  used when we supply a new SOC defintion in mid-year e.g. April 13 for the
    #  Nature Run v4 re-processing in OPS era.
    if spinUpToDataDayIdx >0 and spinUpToDataDayIdx <= 365:
        createCalendarArrays(hObj,spinUpToDataDayIdx)
    else:
        createCalendarArrays(hObj,None)

    # Create file level attributes (of type string)
    logMsg(logf,"@A Creating all file level attributes now")
    createFileAttributes(hObj,socOutHdf5Path,spinUpToDataDayIdx)

    #  Open, read, store a copy of the BPLUT inside this Ancillary file
    h5_bplut_varname = "TEXT_BPLUT"
    read_store_BPLUT(hObj,h5_bplut_varname,bplutPath,logf)
    #  Open, read, store a copy of the MD5's inside this Ancillary file
    h5_md5s_varname = "TEXT_MD5S"
    read_store_MD5s(hObj,h5_md5s_varname,md5sPath,logf)

    hObj.flush()

    ithFile = 0
    for curPath in inpLis:
        # must exist, and skip over the missing counts file
        actualPath = rootInputDir + "/" + curPath
        ithFile += 1 
        inputPointerAttribName = ("INPUTPOINTER_%3.3d"%(ithFile))
        hObj.attrs[inputPointerAttribName] = actualPath
        
        # For the soil organic carbon layers, assign a POOL INDEX as {0,1,2}
        # NOTE: that .csv files where present in the target list are separately
        #  handled for ingest via read_store_BPLUT() or read_store_MD5s() ..
        if os.path.exists(actualPath) and actualPath.find(".csv") <= -1 :
            if curPath.find("_C0_") >= 0:  # metabolic,fast
                poolIndex =0
            elif curPath.find("_C1_") >= 0: # structural,medium
                poolIndex =1
            elif curPath.find("_C2_") >= 0: # recalcitrant,slow
                poolIndex =2
            else:
                poolIndex = -1  # NPP and the FT files have no relevant pool index
            # call the routine to create this new SOC dataset
            # pass in just the basename, as full path constructed again inside..
            hObj = createNewSOC(hObj,curPath,dType,poolIndex)   
        else:
            # skip over any CSV files here
            if actualPath.find(".csv")>=0:
                continue
            else:
                logMsg(logf,"@E cannot find SOC raw binary(%s) via actualPath(%s)"%(curPath,actualPath))
                continue
        logf.flush()
    #
    # update rev tags 

    hObj.close()
    endiSec = time.time()
    elapSec = endiSec - strtSec
    logMsg(logf,"@A task %s complete at %s requiring %.2f sec"%(sys.argv[0],time.ctime(),elapSec))
    logf.close()
    #
    sys.exit(0)
    # end::main()
