// ---------------------------------------------------------------------
// module       : frm_core.h
// purpose      : Implement CORE header storing common core symbols, for the "framework" API
// revision     : v1.9.6-2020-07-13T22:25, minor diag improvements.
// revision     : v1.9.5-2020-07-08T19:20, changed id's to int64_t for better hid_t compatibility
// revision     : v1.9.4-2017-05-18T18:11, added (non-mandatory) frm_ntsg_const.h for offline API testing
// revision     : v1.9.3-2017-04-27T17:06:00, added frm_StrTrimWhitespace() and added the trim
//                whitespace to nsbParseIntoStringblock() call, and also, 
//                corrected nsb_accessors.c to remove spurious warning when a macro
//                expansion e.g. !MyKeyword = /some/list_of_paths.lis was used.
//         
// revision     : v1.9.2-2017-03-22T13:21:00, doc updates only, no code changes or additions
//                where this API is now version matched to L4C Delivery v7.2
// revision     : v1.9.1-2016-02-15T14:29:00, {IsUint8..IsString, etc} enum ordering 
//                changed to categorical grouping convenience
// revision     : v1.9.1-2016-02-18T08:59:00, freshened docs for match to L4CMDL v7.1.0 delivery RMSE enabled
// revision     : v1.9.0-2016-01-07T17:18:00, minor doc added, verify other changes
// revision     : v1.8.9-2015-10-15T14:14:02, xLogClose() additional null test added, and
//                removed unused file open in NSB nsbGetMultiValueByKey() supporting new
//                macro option.
// revision     : v1.9.0-2016-01-07T17:18, added more checking in PkgMgr, added memCreateStringTable()
//                and memReleaseStringTable() in frm_mem_obj.c
// history      : v1.8.9-2015-09-24T12:46, Added capability to perform simple NSB
//                macro expansion of multi-valued keywords via reading a
//                text file containing a list of items, triggered by a
//                KEYWORD prefaced with a '!' symbol, e.g. !MyKey, and added
//                 int nsbGetMaxKeyWidth(NsbObj *nsb_p);
//                 int nsbGetMaxValueWidth(NsbObj *nsb_p);
// history      : v1.8.8-2015-09, minor changes, to be documented
// history      : v1.8.7-2015-08-05T11:38, added FRM_API_VERBOSE macro to curb excessive chatter
// history      : v1.8.6-2015-06-01T23:55:00-07:00, prep for V10002 OPS delivery only, no changes.
// history      : v1.8.5-2014-11-03T16:37:00, in frm_util_fx.c, added frm_utilGetHostnameAndUser()
//                and added frm_utilCalc_std_dev_and_var().
// history      : v1.8.4-2014-10-21T21:28:00-07:00, added frm_utilDiagImage() and related
//                code to frm_util_fx.c for better L4C support. 
// history      : v1.8.3-2014-09-30T17:02:00-07:00, added frm_hdf5_GetDatasetId() and
//                frm_hdf5_CloseDataset(), and revised frm_hdf5_DatasetExists() to now
//                use H5Dopen() which recognizes nested group datasets, and
//                changed frm_hdf5_ReadData() to use H5Dread() instead of "lite" variant
// history      : v1.8.2-2014-09-12T16:08:02-07:00, added frm_utilSummarizePixel() and 
//                frm_utilTestBoolean(), and added in pkg Objects, char *pkgMgr_GetObjH5NameByHdf5Id() 
//                and pkgMgr_GetObjH5NameByHdf5Id, and in themes, added
//                int pkgMgr_DoesThemes_ChildNameExist(), and made the ISO8601 UTC form
//                (in frm_util_fx.c) returned include a sub-second qualifier ".000Z"
// revision     : v1.8.1-2014-09-11T15:36:00, changed nsbParseIntoStringblock singleMultiValue
//                behavior to concatenate single-line multiple sub-values as now separated
//                by a back-tick "`" symbol, which allows the caller to identify these
//                compound fields later and re-translate them back to strings containing
//                embedded blanks using a call using frm_strReplaceChr() like this:
//                  frm_strReplaceChr("`",valueStr," ") ..which restores the blanks back into the string.
// history      : v1.8.0-2014-09-05T13:55:00, added frm_hdf5_GetLibraryVersion()
// history      : v1.7.9-2014-09-03T15:01, added in nsb_accessors.c,  nsbDoesKeyExist() boolean
// history      : v1.7.8-2014-08-20T09:54:00 correct minor message error memObjFree()
//                 and added  *frm_strAddDatetimeStamp() to frm_str_fx.c
//                and added pkgMgr_GetMgr_AllNames() and pkgMgr_GetThemes_ChildListOfNames()
//                and in memObjFree() relaxed name matching to allow match on 
//                user-name or on DNA label, and, added pkgMgr_Get|SetObjH5Datatype(),
//                Set|GetObjH5Dspace(), Set|GetObjH5Dcpl() to support L4_C better
// history      : v1.7.7-2015-08-12T10:33:00-07:00, added frm_utilPrivateUUID() 
//                in frm_util_fx, as work-around when NCCS' libuuid.a is
//                not available or difficult to link into
// history      : v1.7.6-2014-08-08T17:06:00, refined frm_hdf5_DatasetExists() and
//                frm_hdf5_GetDatasetInfo(), which occassionally reported a dataset
//                missing when it was present in a HDF5 (SOC) file.
// history      : v1.7.5-2014-08-06T15:10, added pkgMgr_GetValidRange(),pkgMgr_SetValidRange()
// history      : v1.7.4-2014-07-29T17:32::00, simplified frm_hdf5_Openfile() until further
//                work is done, and re-org frm_hdf5_WriteData() arg call to match
//                frm_hdf5_ReadData() better, and added frm_utilAssignBits() in 
//                frm_util_fx.c, to enable easier insertion of an integer into a
//                bitflag integer, based on mumMaskBits() from MODIS. and
//                corrected the K&R version of frm_utilGetBits().
// history      : v1.7.3-2014-07-01T13:47:00, added more robustness tests 
//                in pkgMgr_InitObj() and others.
// history      : v1.7.2-2014-06-26T19:08:00, added frm_hdf5_WriteAttributeCustomString
//                for better HDF5 attribute write support, and added pkgMgr_CreateTheme(),
//                pkgMgr_ReleaseTheme() to support un-bound themes.
// history      : v1.7.1-2014-06-17T11:21:00, added frm_hdf5_ClassifyHdf5Datatypes() and
//                frm_hdf5_GetObjectInfo(), frm_hdf5_GetObjectCount() as well as 
//                frm_hdf5_CreateGroup(),OpenGroup(),CloseGroup(), and frm_utilExecOsCmd()
//                and frm_hdf5_ErrorStackEnable() and frm_hdf5_ErrorStackDisable().
// history      : v1.7.0-2014-06-12T11:01:00, added pkgMgr_ReleaseAllObjects() and
//                improved robustness of pkgMgr_ReleaseThemes() and related
//                destructors.
// history      : v1.6.6-2014-05-13T18:04:00,added frm_utilGetISO8601_datetime() to frm_util_fx.c
// history      : v1.6.5-2014-03-09T20:55,added Read,Write attribute to frm_hdf_io.c
// history      : v1.6.4-2014-03-08T12:44:00, changed TI_ATOM* to TIV_ATOM to distinghish
//                from L4C's TI_ATOM_PTR etc.
// history      : v1.6.3-2014-03-07T16:39:00, added pkgMgr_SetChildIdByName(),pkgMgr_GetChildIdByName()
// history      : v1.6.2-2014-03-04T15:35:00, milestone & maintenance capstone.
// history      : v1.6.1-2014-02-26T17:04:00, added SetResolution,GetResolution
// history      : v1.6.0-2014-02-26T14:50:00, pkg_mgr_themes.c added error checking for invalid searches.
// history      : v1.5.9-2014-02-25T14:27:00, added pkgMgr_SetObjMetadata(),GetObjMetadata,SetThemeMetadata(), 
//                GetThemeMetadata(), SetObjCategoryCode(), GetObjCategoryCode etc
// history      : v1.5.8-2014-02-23T23:55:00, added frm_hdf5_GetDatasetInfo() and frm_hdf5_OpenfileAppend()
// history      : v1.5.7-2014-02-21T22:17:00, added frm_hdf_io.h into ./inc, and frm_hdf_io.c into
//                ./src to bring into main API Makefile build process. frm_hdf_io.h only includes
//                minimal functionality (open,close,read,write) until I can get more time
//                to bring the rest in. also added pkgMgr_GlobalReport()
// history      : v1.5.6-2014-02-19T12:58:00, added pkgThemeReport() prototype
// history      : v1.5.5-2014-02-18T16:12:00-07:00, cross-integrated PkgMgr+MemOb client apis
//                allowing the PkgMgr data model to directly access memObj entries
// history      : v1.5.4-2014-02-17T12:49:00,added in nsbAccessors.c,
//                nsbGetNumericByKey(), nsbGetNumericByIndex()
//                to directly retrieve a string value known to be a numeric, as a generic numeric
//                via a "pre-emptive upcast" to a float64 return value.
// history      : v1.5.3-2014-02-16T08:53:00, added memObjSetName(), memObjGetName() and brought
//                primary NSB client api into the Framework library proper. Note that nsb_propdb.c
//                services are NOT yet supported, pending further development.
// history      : v1.5.2-2014-02-10T17:13:00-07:00, added EASEGRID enums, FRM_API_ROUND(x),
//                and consolidated related const from individal codes elsewhere
// history      : v1.5.1-2014-02-10T10:02:00-07:00, added TIV_INT64_DEF test, minor clean ups.
// history      : v1.5-2014-02-09T20:26:00-07:00, relocated file related fx to frm_io_fx.c
//              : and added datatype string label array to fm_core.c.
// history      : v1.5-2014-02-09T16:07:00-07:00, added frm_Openfile,frm_Closefile() to frm_util_fx.c
// history      : v1.4-2014-02-01T10:08:00, deprecatedexperimental gAryAdt, not yet implemented.
// history      : 2014-01-30T13:24:00 added -DHAVE_PTHRD for those platforms that have it, and
//                changed prefix "R_" to "FRM_API_" for all xxx_RANDOM macros to avoid namespace
//                conflict with pre-existing uses
// history      : 2014-01-28T16:49:00-07:00, pulled all common core symbols out into this header
// history      : 2014-01-27T17:15:00-07:00, added filepath diagnostics to read,write 
//                binary functions
// history      : 2014-01-24T18:05:00-07:00, orig re-factor from MUM API
// history      : n/a
// author       : joe glassy, Lupine Logic Inc.
// dependencies : HDF5 libraries, headers, H5Tpublic.h datatype (hid_t) access at required.
// history      : 2014-01-007T15:38:00, orig consolidation of MUM fx into one file
// notes        : Significant code segments appeared originally in UM MODIS (MUM) API (1992-2001), 
//                and are adapted & modernized here for the NASA SMAP L4_C framework context
// -----------------------------------------------------


#ifndef FRM_COMMON_CORE_DEF
#define FRM_COMMON_CORE_DEF   1

#define FRM_CORE_VERS_TAG "v1.9.6-2020-07-13T22:25-07:00"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h>

// Defines some, not all of expected int types
#include <stddef.h>
// unistd.h is typically where size_t is defined
#include <unistd.h>
// stdint.h now defines int64_t, uint64_t ....
#include <stdint.h>
// Where WORDSIZE is correctly established
#include <sys/types.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>  // for getpid()
// for stat
#include <sys/stat.h>

#ifdef HAVE_UUID
// #include <sys/uuid.h>
#include <uuid/uuid.h>
#endif

#ifdef HAVE_HDF4
#include "hdf.h"
#include "mfhdf.h"
#endif

#ifdef HAVE_HDF5
#include "hdf5.h"
#include "hdf5_hl.h"
#endif

// THREAD SUPPORT: Not all platform+build combinations 
// of the FRAMEWORK API have access to ptheads, but for
//  those that do and want to enable it, define in the Makefile
#ifdef HAVE_PTHRD
#include <pthread.h>
#endif

#ifdef HAVE_UUID
// for UUID...
// Note: on Missouri,requires: 
// yum install libuuid-devel libuuid uuid uuid-devel
#include <uuid/uuid.h>
#endif

// borrow PI from math.h's M_PI, if not otherwise defined.
//  TBD: a problem with this implementation technique is that
//  when 'PI' is defined LATER IN THE CALL CHAIN, this will cause
//  a multiply-defined warning on the PI symbol. Keep working on 
//  a more robust solution, that will probably have to depend on
//  additional levels of indirection, e.g. define a FRM_API_PI etc
#ifndef PI
// Ultimately, we probably want to avoid redefining such a public 
// symbol at all !
#define PI M_PI
#endif

// <JMG> 2015-08-05T11:38 added qualifier to optionally trim excessive
// diagnostic chatter
// #define FRM_API_VERBOSE  1

// Define n. chars to size a string buffer to store any UUID.
//  This only requires from 32 (if bare) to 36 bytes (if hyphenated), 
//  and since we store it as a string, we round it up to 
//  nearest power of two to accomodate a null terminator, etc.
#ifndef FRM_API_UUID_MAXLEN
#define FRM_API_UUID_MAXLEN   64
#endif

// may be adjusted upward at build time if necessary
#ifndef FRM_API_MAXDIMS
#define FRM_API_MAXDIMS   4
#endif

#ifndef NULL_CH
#define NULL_CH   '\0'
// memObjCalloc() and memObjFree() support...
#endif

#ifndef STRING_MATCHED
#define STRING_MATCHED  0
#endif

// <JMG> changed prefix to match namespace FRM_API to avoid
// a namespace conflict with pre-existing uses
#define FRM_API_RANDOMIZE() srand((unsigned)time(NULL))

// return random probability in range {0.0<=r<=1.0} 
#define FRM_API_RANDOM_PROB() ((double)rand()/(double)RAND_MAX)
#define FRM_API_RANDOM_INT(limit) ((rand()%(abs(limit)+1)))
#define FRM_API_RANDOM_RANGE_1(lo,hi) ((lo)+(rand()%(abs((hi)-(lo))+1)))

// base0 index style range-wise PRNG
#define FRM_API_RANDOM_RANGE_0(lo,hi) (((lo)-1)+(rand()%(abs((hi)-(lo))+1)))

// base1 index style range-wise PRNG
#define FRM_API_RANDOM_RANGE_1(lo,hi) ((lo)+(rand()%(abs((hi)-(lo))+1)))

// floating point round() macro from "The Standard C", P.J.Plaugher 
#define FRM_API_ROUND(x) (((x)<0.50)?ceil(x-0.5):floor(x+0.50))

#ifdef JUST_FOR_REF
#define DFNT_FLOAT32     5
#define DFNT_FLOAT       5  // For backward compat; don't recommend using
#define DFNT_FLOAT64     6
#define DFNT_DOUBLE      6  // For backward compat; don't recommend using 
#define DFNT_FLOAT128    7  // No current plans for support 

#define DFNT_INT8       20
#define DFNT_UINT8      21
#define DFNT_INT16      22
#define DFNT_UINT16     23
#define DFNT_INT32      24
#define DFNT_UINT32     25
#define DFNT_INT64      26
#define DFNT_UINT64     27
#endif

// "TIV" (type-independent-variable) datatypes are the same as HDF4 datatype
// identifiers, same as 'FRM_API_DATATYPES' as:
#ifndef FRM_API_DATATYPES_DEF
#define FRM_API_DATATYPES_DEF   1

// Note: TIV datatype enum values are now sync'd with the
//   corresponding equivalent HDF4 DFNT_xxx numeric type values.
//   These old HDF4 conventions are adopted here since the HDF5 datatype
//   identifiers don't conventiently resolve to simple integer values.
// These below are hard-coded to obviate the need to reference HDF4
//   headers which may or may not be available, acknowledging this is risky
//   if/when the underlying HDF4 integer encodings change.
// <JMG> 2016-02-15T14:27, enum ordering changed to categorical grouping convenience
enum FrameworkObjDatatypes {
    IsNoType =0   , IsChar   =4 , IsInt8   = 20 ,  IsUint8 = 21 , IsInt16 =22, IsUint16 = 23,
    IsInt32  =24  , IsUint32 = 25  , IsInt64 = 26 , IsUint64=27, 
    IsFloat32= 5  , IsFloat64= 6,
    IsInteger = 50, IsReal     = 51    , IsBoolean  = 52, IsSizeT  = 53,
    IsTimeT   = 54, IsDatetime = 55    , IsDirectoryspec =56, IsFilespec = 57, 
    IsString  = 60, IsStringTable = 61 , IsCustomUserType = 70
} ;
#endif

// enum for FRM_API mapping to H5O_TYPE_xxx enum identifiers plus
// adds some of our own that they should have included IsHdf5Attribute
// uses: H5I_get_type().. very useful!
enum FrmHdf5ObjectTypes {
    IsHdf5File,          IsHdf5Group,     IsHdf5Dataset, 
    IsHdf5NamedDatatype, IsHdf5Attribute, IsHdf5Dataspace, 
    IsHdf5UnknownType,   IsHdf5InvalidObj
} ;

enum EndianStates {
    isMSBfirst, isLSBfirst, isVAXEndian, unknownEndian 
} ; 


// enum const for frm_ioReadTextIntoTable() and similar functions
enum FrmCommentHandling {
 DoIncludeComments, DoIgnoreComments
} ;

enum FrmIoModes {
 DoCreateNewText,   DoOpenTextToRead ,  DoOpenTextToWrite, DoOpenTextForAppend, 
 DoCreateNewBinary ,DoOpenBinaryToRead, DoOpenBinaryToWrite, DoOpenBinaryForAppend
} ;


// Verify that the ORDER of these is correct (should be..)
enum EASEGRID_Types {
  IsEG_Global_9km, IsEG_Global_3km ,IsEG_Global_1km, IsEG_N_9km, IsEG_N_1km
} ;


#ifndef HAVE_HDF4
// Define well-known-scalar datatypes, typically defined in HDFv4, but if HDF4
// is not used, define them here via stdint.h primitives
typedef  uint8_t   uint8   ;
typedef  int16_t   int16   ;
typedef  uint16_t  uint16  ;
typedef  int32_t   int32   ;
typedef  uint32_t  uint32  ;
typedef  float     float32 ;
typedef  double    float64 ;
// NOTE: this 'else' is critical due to need to handle INT64's separately
// so make sure this else and its clause is always retained
#else

#ifndef TIV_INT64_DEF
#define TIV_INT64_DEF 1
// added for more robust 64 bit integer support
typedef  int64_t   int64 ;
typedef  uint64_t  uint64;
#endif

#endif


// -----------------------------------------------------
//  FUNCTION PROTOTYPES: frm_core.c
// -----------------------------------------------------

char *frm_GetApiVersion(char *versionIdentifier_p) ;

// end:: idempotence outer most block

#endif


