// ---------------------------------------------------------------
// module  : frm_mem_obj.h
// purpose : implement function prototypes for memObjCalloc(), memObjFree() and friends
// version : 1.5.4
// revised : 2017-04-27, re-verify position of memObj_p->vOwner in structure, as 
//           the gen_cxref.shl reports a possible syntax error (line 282-283) of frm_mem_mgr.c
// revised : 2016-01-07, v1.5.4, added journaled version for creating, releasing "named" string tables
//           added "memCreateStringTable(n,wid,name)" and memReleaseStringTable(strTbl_p,n);
// revised ; 2015-08-03T12:24, added (memCallocWithName()) convenience function, to name when allocating
// revised : 2014-07-30, added void memObjAdjustVerbose(const int newSetting)
// revised : 2014-03-04T10:16:00, increased MAX_MEM_OBJ back up, to 256 (may go to 512)
// history : v1.5-2014-02-23T12:04, added memObjSetDatatype(),memObjGetDatatype()
//           and memObjGetNelem(), memObjSetMetadata(),memObjGetMetadata()
//           and in memObjFreeAll() finally added ability to release IsStringTable's
//           and in MemMgr, added boolean vDoReportOrphans{0|1} to show or suppress
//           full report on orphans detected & freed.
// history : v1.4-2014-02-16T08:37:00,(added memObjGetName(), memObjSetName()
// history : 2014-01-28T13:41:00
// history : 2014-01-28T13:41:00, added MemMgr ADT for future thread-safe extensions, but
//           the implementation of MemMgr is not currently implemented or tested.
// history : 2014-01-21T11:26:00, minor clean up, prep for deprecating auto dot.qualifier
//           as its better for caller to manually construct any dot.qualifier on name
//           needed.. and added memObjGetVersTag()
// history : 2014-01-06T03:09:00, added vDnaLabel[40] sized to fit %p or MD5 string
// history : 2014-01-03T13:42:00, corrected ability to now free IsStringTable
// author  : joe glassy, Lupine Logic Inc.
// notes   : n/a
// ---------------------------------------------------------------

#ifndef MEM_OBJ_MGR_DEF

#define MEM_OBJ_MGR_DEF 1

// include all common, core symbols for frm_api
#include "frm_core.h"

#define MEM_OBJ_API_VERSTAG "v1.5.4-2016-01-07T10:55:00-07:00"
// MEM_OBJ_API_VERSTAG "v1.5.3-2015-08-03T12:25:00-07:00"
// MEM_OBJ_API_VERSTAG "v1.5.2-2014-07-30T0310:00-07:00"

// Symbols for built in memory manager and garbage collector
#define MAX_MEM_OBJ                 512

#define MAX_MEMOBJ_CANDIDATES       MAX_MEM_OBJ

#define MAX_MODULENAME_LEN          512
#define MAX_MEMOBJ_VNAME_LEN        256
#define MAX_DATASET_NAME_LEN        MAX_MEMOBJ_VNAME_LEN
#define MAX_MEM_DNA_LEN             40


#ifndef MEM_OBJ_AVAILABILITY_DEFINED
#define MEM_OBJ_AVAILABILITY_DEFINED 1
enum MemAvailability {
    IsNotAvailable, IsAvailable 
} ;
#endif

#ifndef  MEM_OBJ_DEFINED
#define  MEM_OBJ_DEFINED 1

typedef struct tagMemObj {
    char vName[MAX_MEMOBJ_VNAME_LEN]; // unique identifying text label for this memory object
    char vMetadata[MAX_MEMOBJ_VNAME_LEN]; // optional addition text label for more context
    int  vRank     ; // rank, for future use
    int  vDatatype ; // datatype enum: ... IsUint8,IsFloat64,IsString,IsStringTable etc 
    int  vSizeOfOne; // size of 1 scalar instance
    int  vDataclass; // IsDataClassUndefined,IsDataClassStringTable,IsDataClassScalar
    int  vOwner    ; // ID of owner. <JMG> 2017-04-27T17:26 replaced, repositioned in struct.
    char vDnaLabel[MAX_MEM_DNA_LEN]; // adopts new default unique ID as addr assigned by malloc,calloc
    size_t vBytes  ; // n. bytes for this object
    size_t vNelem  ; // n. of elements and serves as vSlotState if >0
    FILE *vLogf    ; // ptr to whichever logf instance this MemObj will use
    void *vContents_p; // ptr to memory allocated for this variable
} MemObj ;
#endif


#ifndef MEM_MGR_DEFINED
#define MEM_MGR_DEFINED 1
// Each MemMgr encapsulates a private MemObj registry, to facilitate
//   thread-safe instantiation of concurrent, independent tables 
//   of MemObj objects. Note that a "registered" slot refers to any
//   slot that now has memory allocated to it and is in use. When a slot
//   is 'freed' it is un-registered and thus made available again into the
//   pool of candidate slots that can be used again (re-cycled)
typedef struct tagMemMgr {
    int    vMgrId ;   // unique ID for this MemMgr itself   
    int    vMgrOwner; // unique ID of owner of this MemMgr instance
    char   vMgrDna[MAX_MEM_DNA_LEN];  // DNA identifier of this MemMgr itself
    char   vMgrLabel[FILENAME_MAX] ;  // text label for this MemMgr instance
    char   vDoReportOrphans; // boolean 1=show full identity of orphans in reports 0=suppress
    // prior: for the single-instance model these USED to be statics
    int     vMgr_nMemObjRegistered    ; // current n. of slots in use (registered)
    int     vMgr_highestSlotReferenced; // highest candidate slot actually used
    size_t  vMgr_nMemObjBytesTracked  ; // total sum of bytes this memMgr manages
    size_t  vMgr_nMemHighWaterMark;     // memMgr's own high water mark
    size_t vMgrCount ; // count of independent MemObj instances this MemMgr owns
    size_t vMgrInstancesUsed; // count of vMgrTable_p[x] items actually used
    FILE   *vMgrLogf ;  // ptr to a logf, shared by parent MemMgr and all its children MemObj
#ifdef HAVE_PTHRD   
    pthread_t vMgrThreadOwner; // opaque ID of a pthread owner, ignored if not threaded
#endif  
    MemObj *vMgrMemRegistry_p; // a table of (vMgrCount) independent MemObj's   
} MemMgr ;

#endif

// prototypes have now been consolidated into one header ...

// --------- frm_mem_obj.c PROTOTYPES ----------------------------

char *memObjGetVersTag(void);
void    memObjRegisterLog(FILE *whichLogf);
void    memObjPrivateMsg(char *fmtMsg_p,...);

size_t  memObjInit(FILE *usersLogf);
int     memObjInitCondition(void);

// fetch object by name
void    *memObjGet(const char *objName_p);
// or by base0 slot index
void    *memObjGetBySlot(const int ithSlot);
// based on ptr itself, fetch its registry slot index
int memObjGetSlot(void *referent_p);

// new "lite" anonymous service functions, use objects own DNA
// to self-referentially identify itself!
void    *memCalloc(const size_t nElem,size_t sizeOfOne);
void    memFree(void *mem_p);
// <JMG> 2015-08-03T12:24, added convenience function, to name when allocating
void    *memCallocWithName(const size_t nElem,size_t sizeOfOne,const char *memObjName_p);

// legacy family: full, name-enabled service functions
void    *memObjCalloc(const char *objName_p,const size_t sizeOfOne,const size_t nElem,const int vDtype,const char *fxLbl,const char *callerLbl);
void    memObjFree(char *objName_p,void *mem_p);
int     memObjFreeAll(void);
size_t  memObjGetCount(const char *objName_p);
void    memObjGarbageCollect(void);
MemObj  *memObjExport(void) ;
size_t  memObjTotalBytesUsed(void);
void    memObjZeroOut(const char *objName_p);

int     memObjSlotsUsed(void);
size_t  memObjGetMetrics(size_t *highestSlot,size_t *totalBytesTracked,size_t *hwmMem);
int     memObjWrite(const char *objName_p,const char *userPath,const char *allOption);

// added 2/16/2014...
void    memObjSetName(void *mem_p,char *vNewLabel_p);
char   *memObjGetName(void *mem_p);
// added 2/23/2014...
void   memObjSetDatatype(void *mem_p,const int newDatatype);
int    memObjGetDatatype(void *mem_p);
void   memObjSetMetadata(void *mem_p,const char *newMetadata_p);
char   *memObjGetMetadata(void *mem_p);
size_t memObjGetNelem(void *mem_p);

char    *memObjQualify(char *fxNameStr,const char *memObjNameStr,char *buffer);
void    memTestRedundancy(void);
void    memTestCallocLite(const int nElem);

// <JMG> 2016-01-07, journaled version for creating, releasing "named" string tables
char   **memCreateStringTable(const int nMembers,const int strWidth,const char *strTblName_p);
void   memReleaseStringTable(char **strTbl_p,const int nMembers);

// non-journaled utility functions
char **memCreateStringTbl(const int nMembers,const int strWidth);
void   memFreeStringTbl(char **strTbl_p,const int nMembers);

// produce a memory registry report using each memObj's slot as its access key
int    memObjReport(FILE *userRptf,const char *reportTitle);;
// produce a memory registry report using each memObj's DNA as its access key
int    memObjReportDna(FILE *userRptf,const char *reportTitle);

// <JMG> 2014-07-30 added for better debugging
// call: 1=enable verbose 0=disable verbose.
void memObjAdjustVerbose(const int newSetting) ;
// void    memBaselineTestMemObjFacility(void);

#endif

