// ---------------------------------------------------------------
// module       : nsb_api.h
// purpose      : Implements header info for the "string-block" based 
//                name-value-pair (NVP) management package.
// project      : Lupine Logic Inc -- single instance, "stringblock" backing-store mechanism 
//                to manage name-value-pair symbol tables.
// author       : joe glassy, Lupine Logic Inc.
// dependencies : HDF4 only, not HDFEOS or HDF5 (yet)
// revised      : v2.1-2017-04-26T12:10, removed spurious warning in nsb_accessor.c when
//                a macro expansion performed (e.g. a !KEY prefaced with a "!") references
//                a filename which contained an arbitrary long list of entries that are
//                expanded into a stringtable returned by nsbGetMultivalueByKey()
// revised      : v2.0-2015-09-30, added several functions to explictly store and retrieve
//                delimiter strings, for separating KEYWORD|VALUE and separating the
//                multi-value items in a single value string such as the properties of a
//                datasets; all of these are backward compatible with prior semantics.  e.g.
//                nsbGetDelimiters(), vs. nsbSetDelimiters()
//                nsbSetDelimiterMultiValue() vs. nsbGetDelimiterMultiValue()
// history      : v2.0-2015-09-28T16:43, added nsbCreateDatabase() for custom user sized NSBs
// history      : v2.0-2015-09-24T11:25:00, Started prototyping "!" macro expansion
//                syntax, allowing a single KEY's value as a path to ONE text
//                file that itself contains a list of items (e.g. multiple paths).
//                This syntax uses a KEYWORD prefaced with a "!" (e.g. !MyKey) 
//                to trigger expansion (macro expansion)of a VALUE specifying a text file
//                where the referent (text file) contains one item per line., AND,
//                added (2) 'get' functions, nsbGetKeyMaxWidth() and nsbGetValueMaxWidth()
// history      : v1.9-2014-07-31T15:00, added nsbGetKeywordCount() to count either
//                if a scalar is present or count (=1), or count the multi-valued
//                list style keyword entries. Note that on nsbGetMulti(..) both
//                strict keyword match and partial (substring) match now supported
// history      : v1.8-2014-02-23T10:57:00, re-certified readiness for L4_C, and
//                migrated nsbKeyStrTbl to use un-registered memCreateStringTbl(),memFreeStringTbl()
// history      : v1.7.5-2014-02-22T16:32:00, simplified memory using memCalloc for main nsb_p
// history      : v1.7.4-2014-02-16T16:12:00, added to NsbObj "nsbKeyPolicy" to allow
//                for cases where a predefined list of supported keys is NOT available
//                or is not applicable. When nsbKeyPolicy=0, no supported key table is required,
//                by the main parser, nsbParseIntoStringblock(). When nsbKeyPolicy=1,
//                a supported key (table of just the keywords supported) is required.
// history     : v1.7.3-2014-02-12T22:13:00, added optional schema path parsing
// history     : v1.7.2-2014-02-09T20:09:00-7:00, deprecated use of property NSB field, went back
//                to using just (2) NSB fields (keyword,value).
// history     : v1.7.1-2014-02-05:12:26:00-7:00,wheat, jmg --removed nsb_util.c dependency
//                and removed HDFEOS ref from Makefile, migrated to use FRM_API_RANDxxx 
//                and updated to use newest FRM API, -lfrm.a
// history      : 1.7.0 -- major clean up.
// history      : v1.6-2014-01-06T10:22:00:00-7:00,jmg, -- changed to use canonicals in mem_obj.c
//              : 2014-01-03T12:49:00 - see history below.
//              : 2014-01-02T18:07:00,jmg --refactored structure of a stringblock
//              : 2014-01-03T12:49:00,jmg --started NSB namespace encapsulated object representation of
//                the original single stand-alone Nvp stringblock implementation
//              : 2014-01-02T13:33:00
// notes        : Supported TIV datatypes eventually will be :
//
// enum TIV_Datatypes {
//        IsNoType, IsInt8, IsUint8, IsInt16, IsUint16, IsInt32, IsUint32,
//        IsInt64, IsUint64, IsFloat32, IsFloat64, IsSizeT, IsTimeT,
//        IsBoolean, IsDatetime, IsDirectorySpec, IsFilespec, 
//        IsInteger, IsReal, 
//        IsString, IsStringTable, IsStringList, IsCustomUserType
// } 
//

#ifndef NSB_API_DEF

#define NSB_API_DEF 1

// note this version tag is distinct from test harnesses' VERS_TAG
#define NSB_VERS_TAG "v2.1-2017-04-26T12:10:00-07:00"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stddef.h>
#include <stdint.h>
#include <stdarg.h>
#include <assert.h>
#include <limits.h>
#include <math.h>
#include <sys/types.h>  // for getpid()
#include <unistd.h>     // for gethostname() and getpid()

#ifdef HAVE_UUID
#include <uuid/uuid.h>  // for uuid services, uuid_generate() etc..
#endif

// import HDF4 and HDFEOS required constants
#include "hdf.h"
#include "mfhdf.h"

#include "frm_api.h"

// special mark indicating a NSB DB record is un occupied
#define NSB_RECORD_UNUSED    "NSB~RECORD~UNUSED"

// max length of an NsbObj name-label
#define NSB_LBL_MAXLEN                    256

// changed to NSB_ for consistency.
// <JMG> 2014-09-11, increased from 1024 to 4096 from orig 256 to 512 here
// Note: the WIDTH of an NSB data is dynamically determined as the longest
//  string length of the set of input text lines encountered upon first
//  parsing, when the NSB database is created.
#define NSB_STRBLK_RIPLINE_MAXWIDTH      4096
#define NSB_STRBLK_KEYWORD_MAXWIDTH       512
#define NSB_UUID_MAXWIDTH                  64

// options for types of NSB backingstore import, exports
enum NsbBackingstoreOptions {
    IsNsbText, IsNsbSqlite, IsNsbMySql, IsNsbMongdb, IsNsbHdf5
} ;

// <JMG> 20134-07-31 added, for new nsbGetKeywordCount() fx
enum MatchPolicyOptions {
  DoCountStrictMatch, DoCountPartialMatch
} ; 


// Derive the byte offset into single NSB buffer where the RECORD
// starts, as consisting of one KEYWORD and one VALUESTRING).
#define NSB_OFFSET_TO_RECD(ithNvp,recWidth) (ithNvp*recWidth)

// Derive byte offset in single NSB buffer where the KEYWORD of the
//   ithNvp entry starts
#define NSB_OFFSET_TO_KEY(ithNvp,recWidth) ((ithNvp)*(recWidth))

// Derive the byte offset in single NSB buffer where the VALUE STRING
//   of the ithNvp entry starts
#define NSB_OFFSET_TO_VAL(ithNvp,recWidth,keyWidth) ((ithNvp*recWidth)+(keyWidth))


// NSB schema object (not fully supported yet as of 2/12/2014)
typedef struct tagNsbSchemaObj {
    char keyField[NSB_STRBLK_KEYWORD_MAXWIDTH];
    char valField[NSB_STRBLK_KEYWORD_MAXWIDTH];
    char nsbLocationPath[FILENAME_MAX]; // directory location for import, or export
    char nsbUnoccupiedMark[16]; // string indicating record is unoccupied
    int  nsbFieldcount ;        // default is 2, for {keyword,value}
    int  nsbBackingstoreCode;   // enum, identifying backing store if defined
    size_t nsbRecordCount ;     // count parsed from imported DB text file      
    char   nsbDbDelim[16] ;     // delimiters allowed inside NSB stringblock itself
} NsbSchemaObj ;


// NSB parent container object, implements a self-contained, concurrent instance of a
//  "NVP string block" -- as a high-level object that encapsulates a replicable
//  instance of a NVP stringblock instance
// <JMG> 2015-09-30, added nsbMarkedAsDeletedCount, for future implementation of
// "mark+recycle" of deleted records, Added nsbDelimSym[] , internally stored string of delimiter symbols
typedef struct tagNsbObj {
    char   nsbLabel[NSB_STRBLK_KEYWORD_MAXWIDTH] ; // identifying label for whole, enclosing stringblock 
    char   nsbSchemaPath[FILENAME_MAX]; // optional, references a NSB schema or other metadata file
    size_t nsbActualCount;    // count of ACTUAL NVPs (e.g. records) defined at parse-time for stringblock
    size_t nsbMaxCount;       // count of max no. NVPs storable, may be greater than nsbNvpCount...
    size_t nsbStanzaCount;    // 0=if no stanzas, otherwise is count of stanzas or sub-groups. 
    size_t nsbRecsize ;       // size of any given "record" in the stringblock (sum of all field widths)
    size_t nsbMarkedAsDeletedCount; // TBD, count of records marked as deleted, e.g. N. slots ready to recycle
    int    nsbKeyMaxwidth ;   // dynamic, set at runtime as longest keyword encountered on 1st parse
    int    nsbValMaxwidth ;   // dynamic, set at runtime as longest 'value' string encountered on 1st parse. 
    char   nsbUUID[64] ;      // UUID identifying this stringblock instance
    int    nsbOwnerPID ;      // unique ID of OS process owning this instance
    int    nsbKeyPolicy;      // option: when 1=forces use of supported key table,0=allows any key
    int    nsbKeyStrTblCount; // count of pre-registered supported keys, if defined in nsbLocalUsersCustomKeyStrTbl_p
    // nsbDelimCompund[] stores special delimiters used to separate multiple items in ONE value string such
    // as the many properties of a dataset (HDF5 or other), where this delimiter is intentionaly DIFFERENT than
    // the delimiter used to separate a KEYWORD from its VALUE
    char   nsbDelimCompound[16];  // Example: "," --indicates delimiters for items: "myVar,float32,minVal,maxVal,units"
    char   nsbDelimSymbols[16];   // Recognized,valid delimiter symbols " \t=," between KEYWORD and VALUEs
    char  *nsbCoreStringblock_p ; // ptr to the stringblock (memory) itself
    // TBD: will be bound to result of a call to nvpLoadSupportedKeyCatalog()
    char  **nsbKeyStrTbl_p;   // ptr to StringTable, listing of pre-registered supported keywords
    int   nsbKeyStrTblWidth;  // width of strings in nsbKeyStrTbl_p
    NsbSchemaObj nsbSchema;   // schema info for this NSB instance
} NsbObj;


#endif

