// ------------------------------------------------------------
// module       : frm_io_fx.c
// purpose      : Implement various I/O functions not found elsewhere, 
//                as part of L4_C "framework"
// revision     : 2014-01-28T12:59:00, added frmIoModeLabels()
// history      : 2014-01-27T17:14:00-07:00, added filepath diagnostics to read,write binary functions
// history      : 2014-01-27T17:04:00,wheat,jmg
// version      : see macro FRM_IO_VERS_TAG in header file frm_io_fx.h
// history      : n/a
// author       : joe glassy
// dependencies : verify, HDF5, HDF5 etc
// notes        : n/a


#include "frm_api.h"


// text labels for various I/O file opening modes,matching enum indices, as:
// enum FrmIoModes {
// DoCreateNewText,   DoOpenTextToRead ,  DoOpenTextToWrite, DoOpenTextForAppend, 
// DoCreateNewBinary ,DoOpenBinaryToRead, DoOpenBinaryToWrite, DoOpenBinaryForAppend
//} ;

const char *frmIoModeLabels[] = {
 "DoCreateNewText      " ,
 "DoOpenTextToRead     " ,
 "DoOpenTextToWrite    " ,
 "DoOpenTextForAppend  " ,
 "DoCreateNewBinary    " ,
 "DoOpenBinaryToRead   " ,
 "DoOpenBinaryToWrite  " ,
 "DoOpenBinaryForAppend"
} ;

// ---------------------------------------------------------------
// function : frm_ioQualifyPathsWithRoot()
// purpose  : Check all input, output paths for a "/", and if 
//            missing, prepend the appropriate ROOTDIR in front
// returns  : char *, fully qualified (unambiguous) origPath_p
// notes    : '/' is ASCII char 47. '\' is ASCII 92. "=34.
// notes    : 7/29/2014, use caution when using this with pre-allocated
//            stringtables, whose widths were custom tailored to the
//            initial payload. Use frm_strExpandStringTbl() to widen out
//            a string table so it can accomodate a pre-pended ROOTDIR.
char *frm_ioQualifyPathWithRoot(const char *rootDirPath_p,char *origPath_p)
{
 char workPath[FILENAME_MAX] = {0} ;
 char fxName[256] = {0}  ;
 
    strcpy(fxName,__func__);
    
    // possible bug in this logic. Pending...
    if(strchr(origPath_p,'/') == NULL)
    {
        // Compose new fully qualified path
        sprintf(workPath,"%s/%s",rootDirPath_p, origPath_p);
        strcpy(origPath_p, workPath);
        memset(workPath,'\0',FILENAME_MAX);
    } else {
        xLogMsg("@W In(%s) Blocked concatenate op -- since a forward-slash was already present in (%s)~ \n",
                fxName,origPath_p);
    }
 return origPath_p; 
} // end:: frm_ioQualifyPathWithRoot()


// ------------------------------------------------------------
// function     : frm_ioGetFilesize()
// purpose      : Retrieve an existing files size in bytes
// inputs       : const char *inputPath_p --path to an existing file
//                const char *callerlabel
// returns      : size_t -- filesize in bytes, or 0 if failed
// revision     : 2014-02-09T16:22, moved from frm_util_fx.c to frm_io_fx.c
// history      : 2014-01-24T13:49:00,wheat,jmg
// author       : joe glassy
// dependencies : n/a
// notes        : TBD future, may also return st_{times} last modified, etc
size_t frm_ioGetFilesize(const char *inputPath_p,const char *callerLabel)
{
 size_t fileSiz = 0;
 struct stat statbuf ;
 FILE *inf = NULL;
 int  fileHandle = -1 ;

        // open file, fetch std stream
        if((inf=fopen(inputPath_p,"rb"))==NULL)
        {
            fprintf(stdout,"@E frm_utilGetFilesize: cannot open(%s) caller(%s)!\n",
                inputPath_p, callerLabel);
                exit(-1);
            return fileSiz ;
        }
        
        // fetch integer handle via open stream
        fileHandle = fileno(inf);
        if(fstat(fileHandle, &statbuf) < 0)
        {
            fprintf(stderr,"@E frm_utilGetFilesize stat failed on(%s)hndl(%d) caller(%s)!\n",
                    inputPath_p,fileHandle, callerLabel);
            exit(-1);
        } else {
            fileSiz = statbuf.st_size ;
        }
        // close file
        fclose(inf);
    return fileSiz;
} // end::frm_ioGetFilesize()


// ------------------------------------------------------------
// function : frm_ioGetIOModeLabel
// purpose  : return text label assoc with a given I/O mode 
char *frm_ioGetIOModeLabel(const int ioModeIndex)
{
 static char modeLabel[32] = {0} ; // protect original as a const ....
 char *ioModeLbl_p = NULL;

    if(ioModeIndex >= DoCreateNewText && ioModeIndex <= DoOpenBinaryForAppend)
    {
        strcpy(modeLabel, frmIoModeLabels[ioModeIndex]);
        ioModeLbl_p = modeLabel;
    }

 return ioModeLbl_p ; 
} // end:: frm_ioGetIOModeLabel()


// ------------------------------------------------------------
// function     : frm_ioOpenfile()
// inputs       : const char *filePath_p,char char *callerLabel,const int tivIoMode
// purpose      : generic function to open a text or binary file for read,write or append
// returns      : FILE *filHndl -- ptr to opened file or if failed, relies on xLogMsg()
//                semantics for final exception handling (e.g. NULL if failed)
//                A failure is currently treated as a FATAL ERROR.
// revision     : 2014-01-24T13:49:00,wheat,jmg
// history      : n/a
// author       : joe glassy
// dependencies : verify, HDF5, HDF5 etc, and enum with DoXXXX modes
//     { DoCreateNewText,DoOpenTextToRead,DoOpenTextToWrite,DoOpenTextForAppend,
//       DoCreateNewBinary,DoOpenBinaryToRead,DoOpenBinaryToWrite,DoOpenBinaryForAppend}
//
// notes        : HDF4,HDF5 are NOT yet supported but MAY be in the future,
//                just to consolidate this operation in one file.
FILE *frm_ioOpenfile(const char *filePath_p,const char *callerLabel,const int thisIoMode)
{
 FILE *filHndl = NULL;
  
    switch(thisIoMode)
    {
    case DoCreateNewText      : filHndl = fopen(filePath_p,"wt") ; break;
    case DoOpenTextToRead     : filHndl = fopen(filePath_p,"rt") ; break;
    case DoOpenTextToWrite    : filHndl = fopen(filePath_p,"wt") ; break;
    case DoOpenTextForAppend  : filHndl = fopen(filePath_p,"wt+"); break;
    case DoCreateNewBinary    : filHndl = fopen(filePath_p,"wb") ; break;
    case DoOpenBinaryToRead   : filHndl = fopen(filePath_p,"rb") ; break;
    case DoOpenBinaryToWrite  : filHndl = fopen(filePath_p,"wb") ; break;
    case DoOpenBinaryForAppend: filHndl = fopen(filePath_p,"wb+"); break;
    xLogMsg("@E frm_ioOpenfile: unsupported Framework IO mode %d ~\n",
        thisIoMode);
    } // end:: switch()
    
    if(filHndl == NULL)
    {
        xLogMsg("@F frm_ioOpenfile: FAILED to open(%s) from caller(%s)~\n",
            filePath_p, callerLabel);
    }

    return filHndl ; // return file-handle
    
} // end:: frm_ioOpenfile()


// ------------------------------------------------------------
// function     : frm_ioClosefile(file)
// inputs       : FILE *filHndl, const char *filePath_p,char char *callerLabel
// purpose      : generic function to close a text or binary file.
// returns      : int, as fclose status
// revision     : 2014-01-24T13:49:00,wheat,jmg
// history      : n/a
// author       : joe glassy
// dependencies : verify, HDF5, HDF5 etc
// notes        : HDF4,HDF5 are NOT yet supported but MAY be in the future,
//                just to consolidate this operation in one file.
// notes        : A failure is current treated as a FATAL error.
int frm_ioClosefile(FILE *filHndl,const char *filePath_p,const char *callerLabel)
{
 int ioStatus = 0;

    // FLUSH first 
    fflush(filHndl);

    // CLOSE it
    if((ioStatus = fclose(filHndl)) != 0)
    {
        xLogMsg("@F frm_ioClosefile: FAILED to close file(%s) from caller(%s)~\n",
            filePath_p, callerLabel);
    }
    return ioStatus;
} // end:: frm_ioClosefile()


// ------------------------------------------------------------
// function     : frm_ioReadbinary()
// purpose      : Using atomic operation, reads data starting at byte 0
//                from a  binary file into a user supplied buffer, 
//                of size nElem in bytes.The user is responsible for 
//                later interpreting the data type.
// inputs       : const char *inFilepath_p -- path of file being read
//                const int tivDatatype --identifier for datatype IsFloat32 etc
//                const size_t nElem -- n. items of tivDatatype to read
//                void *imgData_p -- ptr to buffer to read data into
// revision     : 2014-01-24T13:49:00,wheat,jmg
// history      : n/a
// author       : joe glassy
// dependencies : verify, HDF5, HDF5 etc
// notes        : This is not a simple wrapper for fread; an
//                atomic operation {Open,Read,Close} idiom is used.
void *frm_ioReadbinary(const char *inFilepath_p,const int tivDatatype,const size_t nElem,void *imgData_p)
{
 FILE *binf = NULL;
 char fxName[256] = {0} ;
  
 size_t ioStatus = 0;
 size_t sizOfOne = (size_t)tiv_SizeofTivType(tivDatatype);
 
 TIV_ATOM_PTR tivObj ; // declare a working TIV instance

    strcpy(fxName,(char *)__func__);
    
    // Interpret user's buffer as a uint8 for convenience. User is
    // responsible later for correctly interpreting results as the
    // proper datatype, typically using the TIV interface.
    tivObj.aUint8_p = (uint8 *)imgData_p ;

    // OPEN file to read from
    binf = frm_ioOpenfile(inFilepath_p, fxName,DoOpenBinaryToRead);

    // Read the data starting at byte 0
    if((ioStatus=fread(tivObj.aUint8_p, sizOfOne,nElem, binf)) != nElem)
    {
        xLogMsg("@F In(%s) FAILED on fread file(%s) of %zu elements\n",
            fxName, inFilepath_p, nElem);
    
    } // end:: block to read a given image or vector 

    // CLOSE raw binary file read
    frm_ioClosefile(binf,inFilepath_p, fxName);

  // caller must sort out what its real datatype needs to be interpreted as
  return tivObj.aUint8_p  ;
  
} // end:: frm_ioReadbinary()


// ------------------------------------------------------------
// function     : frm_ioWritebinary()
// purpose      : Write a row-major-order image of specified TIV datatype into
//                a prior opened binary file.
// inputs       : const char *outFilepath_pj -- path of file being written to
//                const int tivDatatype -- identifier, like IsFloat32, IsUint8, etc.
//                const size_t nElem --n. of items of tivDatatype being written
//                void *imgData_p -- ptr to data to be written to an openraw binary file
// revision     : 2014-01-27T17:20:00-07:00,wheat,jmg
// history      : n/a
// author       : joe glassy
// dependencies : verify, HDF5, HDF5 etc
// notes        : n/a
int frm_ioWritebinary(const char *outFilepath_p,const int tivDatatype,const size_t nElem,void *imgData_p)
{
 FILE *binf = NULL;
  int hdf4Datatype = tiv_TivTypeToHdf4Type(tivDatatype);
  size_t ioStatus = 0;
  size_t sizOfOne =  DFKNTsize(hdf4Datatype) ;
 
 TIV_ATOM_PTR tivObj ;
 
    tivObj.aUint8_p = (uint8 *)imgData_p ;
 
    // OPEN file to write to
    binf = frm_ioOpenfile(outFilepath_p,"frm_ioWritebinary",DoOpenBinaryToWrite);

    // WRITE it
    if((ioStatus=fwrite(tivObj.aUint8_p, sizOfOne, nElem,binf)) != nElem)
    {
        xLogMsg("@F frm_ioWriteBinary FAILED on fwrite file(%s) of %zu elements\n",
            outFilepath_p, nElem);
    } // end:: block to write a given image or vector 

    // CLOSE it
    // flush & close raw binary file just written
    frm_ioClosefile(binf, outFilepath_p,"frm_ioWriteBinary");

  return (int)ioStatus;
  
} // end:: frm_ioWritebinary()


// -----------------------------------------------------
// function : frm_ioWriteText
// purpose  : write into a text file, using sprintf style formatting
// inputs   : FILE *outf, char *fmtSpecStr_p,...
// returns  : int, status
// notes    : n/a
int frm_ioWriteText(FILE *outf,char *fmtSpecStr_p,...)
{
 va_list argptr ;
 char workBuffer[2048] = {0} ;// longest supported text line is 2047
 int status =0 ;
 
    // start up
    va_start(argptr,fmtSpecStr_p);
    // format it
    vsprintf(workBuffer, fmtSpecStr_p,argptr);
    // close it down
    va_end(argptr);
        
    status = fprintf(outf,"%s",workBuffer);
    
    return status;
    
} // end:: frm_ioWriteText()



// --------------------------------------------------------------------------
// function : frm_Openfile() 
// Purpose  : open a file, labelled with caller
// Inputs   : const char *filePath_p,const char *modeStr_p,char *caller_p
// Returns  : pid_t -- process id, unsigned long via type pid_t
// Author   : joe glassy
// Revised  : 2014-02-09T15:41:00
// Notes    : n/a
FILE *frm_Openfile(const char *filePath_p,const char *modeStr_p,char *caller_p)
{
  FILE *filf = NULL;
  
    if((filf=fopen(filePath_p,modeStr_p))==NULL)
    {
        // TBD: change to xLogMsg() asap
        xLogMsg("@F frm_Openfile FAIL opening(%s) mode(%s) called-by(%s)~\n",
                filePath_p, modeStr_p, caller_p);
        exit(-1);
    } // end:: block if file create or open failed

  return filf ;
} // end:: frm_Openfile()


// --------------------------------------------------------------------------
// function : frm_Closefile() 
// Purpose  : close an open file, labelled with caller
// Inputs   : 
// Returns  : pid_t -- process id, unsigned long via type pid_t
// Author   : joe glassy
// Revised  : 2014-02-09T15:41:00
// Notes    : n/a
int   frm_Closefile(FILE *filf,char *caller_p) 
{
 int ioStatus = -1 ;
 
    if( filf == NULL)
    {
        xLogMsg("@F frm_Closefile: File at addr(%p) already closed? [%s]\n",
            filf, caller_p);
        exit(-1);
    }
    
    if((ioStatus = fclose(filf)) != 0)
    {
        xLogMsg("@F frm_Closefile: FAILED closing file at addr(%p) caller(%s)\n",
            filf, caller_p);
        exit(-1);
    }
 
 return ioStatus ;
    
} // end:: frm_Closefile()

// -----------------------------------------------------
// function : frm_ioWriteTextTable
// purpose  : Write a stringtable into a text file one call,
// inputs   : char *outputPath_p -- pointer to input files path
//            char **stringTbl_p -- pointer to stringtable to write into file
//            int nLines -- n. of items to write,starting at top
//            opMode, if 0 then writes newlines, if 1 or other int, writes raw content only
// revised  : 2015-09-24T13:23
// notes    : Does not yet validate nLines against actual length of
//            string table.
int frm_ioWriteTextTable(char *outputPath_p,char **strTbl_p,int nLines,int opMode)
{
 char fxName[256] = {0} ;
 char outFmtStr[8] = {0} ;
 int status =0 ;
 int kWritten = 0 ;
 FILE *outf = NULL;
 
 strcpy(fxName,__func__);
 xLogMsg("@A In(%s) writing stringtable of %d lines to text file %s~\n",
  fxName, nLines, outputPath_p);
 
 if(opMode == 0)
   strcpy(outFmtStr,"%s\n");
 else
   strcpy(outFmtStr,"%s");
 
  // OPEN new file   
  outf =  frm_Openfile(outputPath_p, "wt", fxName);

  // WRITE all indicated stringtable lines where nLines might be < total
  // available in the string table.
  // TBD: plan to later add a custom (start,stop,columnList[k])
  // functionality for fuller CSV read/write capabililty
 
  for(kWritten=0; kWritten < nLines; kWritten++)
        fprintf(outf, outFmtStr, strTbl_p[kWritten]);
 
 // flush then close...       
 frm_Closefile(outf,fxName);
 // return n. of stringtable lines written out, back to caller
 return kWritten ;
  
} // end::frm_ioWriteTextTable()



// -----------------------------------------------------
// function : frm_ioReadTextIntoTable
// purpose  : read whole text file using binary I/O in one call,
//            loading it into an IsStringTable 
// inputs   : char *inputPath_p -- pointer to input files path
//            const int options -- opCommand: 1=excludes comment lines. If 0,includes ALL lines
// returns  : int *maxWidthChars  --n. characters in widest qualifying line encountered
// returns  : int *nLines, --count of lines loaded into string table returned
// returns  : char **stringTable_p -- ptr to populated IsStringTable object
// outputs  : int *nLines --count of lines loaded into string table returned

// notes    : Options: DoLoadAllLines, DoIgnoreComments -- may be use as 'options'
//          : This method reads the entire text file into a single buffer in one raw binary
//            read call, then parses it. It is faster for very large text files.
// warning  : the DoLoadAllLines,DoIgnoreComments semantics are NOT yet
//            fully implemented (no enum for these is defined yet, this is tbd)
char **frm_ioReadTextIntoTable(char *inputPath_p,const int options,int *maxWidthChars,int *nLines)
{
 char   *buffer_p = NULL;
 char   **strTbl_p = NULL;
 FILE   *inf      = NULL;
 int    ret       = -1 ;
 char   fxName[256] = {0} ;
 char   *line_p = NULL;
 char   *brkp = NULL;
 
 int    ithLine = 0 ;
 int    sLen = 0 ;
 int    maxWidth = 0 ;
 size_t filSiz    = 0;
 time_t  benchSec = time(0);

    strcpy(fxName,__func__);
    // Start: retrieve the text files size in bytes
    filSiz = frm_ioGetFilesize(inputPath_p,fxName);
    
    if((buffer_p=(char *)calloc(filSiz+1,sizeof(char)))==NULL)
    {
        fprintf(stderr,"@E In(%s)failed to open(%s)!\n",
                fxName,inputPath_p);
        exit(-1);
    }
    
    inf= frm_Openfile(inputPath_p,"rb",fxName);
    
    // ingest entire file into a raw binary buffer
    if((ret = fread(buffer_p,sizeof(char),filSiz,inf))!= (int)filSiz)
    {
        fprintf(stderr,"@F In(%s)  PASS-1 FAILED on fread of (%s) filesize %zu\n",
            fxName, inputPath_p, filSiz);
        exit(-1);
    } // end:: fread block
    
    // PASS 1: determine nLines and widest-char count
    //         Uses thread-safe variant of strtok, allow for Windows, MacOS \r or Linux \n
    for(line_p=strtok_r(buffer_p,"\n\r",&brkp);
        line_p != NULL;
        line_p=strtok_r(NULL,"\n\r",&brkp))
    {
            sLen = strlen(line_p);
            
            // SKIP COMMENT lines if options == 1
            if(options == 1 && line_p[0] == '#')
                continue ;
            
            if(ithLine== 0)
            {
                maxWidth = sLen;
            } else {
                if(sLen > maxWidth) maxWidth = sLen;
            }
            // load into string table record
            ithLine++ ;
    } // end:: line_p parser 

    benchSec = time(0);
    // fprintf(stderr,"@A PASS 2 starting read of(%s) at %s!\n",
    //    inputPath_p, ctime(&benchSec));

    rewind(inf);  // start over at top of file...
    // pre-init buffer 
    memset(buffer_p, '\0', sizeof(char)* filSiz );

    // reminder: fread() arg order: (buffer,sizeof(obj),nEleme,stream)
    if((ret = fread(buffer_p,sizeof(char),filSiz,inf)) != (int)filSiz)
    {
        fprintf(stderr,"@F In(%s) PASS-2 FAILED on fread of (%s) filesize %zu\n",
            fxName,inputPath_p, filSiz);
        exit(-1);
    } // end:: fread block

    // close the input file itself
    frm_Closefile(inf, fxName);
    
    // Update for return: number of qualifying text lines countered
    *nLines = ithLine;
    
    // Update for return: max width, in character units, of widest line encountered
    *maxWidthChars = maxWidth ;
    
    // Create IsStringTable
    strTbl_p = memCreateStringTbl(ithLine, maxWidth+2);

    // PASS 2: parse into lines, using real buffer now
    ithLine = 0 ;
    for(line_p=strtok_r(buffer_p,"\n\r",&brkp);
        line_p != NULL;
        line_p=strtok_r(NULL,"\n\r",&brkp))
    {
            // SKIP comment lines if options say so
            if(options == 1 && line_p[0] == '#')
                continue ;
                
            strcpy(strTbl_p[ithLine], line_p);
            ithLine++ ;
    } // end:: line_p parser 

        fprintf(stdout,"@A In(%s) DIAG: nLines(%d) width(%d)\n",fxName,ithLine,maxWidth); 

    // release local (LINE) level buffer 
    free(buffer_p);

    // return strTbl_p[] containing all data, parsed into lines
   return strTbl_p;
} // end:: frm_ioReadTextIntoTable()



// -----------------------------------------------------
// function : frm_ioReadTextIntoBuffer
// purpose  : read whole text file using binary I/O in one call,
//            loading it into an IsStringTable using
// inputs   : char *inputPath_p -- pointer to input files path
// returns  : char *buffer --containing raw data without line interpretation
// notes    : Useful for NSB style processing
char *frm_ioReadTextIntoBuffer(char *inputPath_p)
{
 char   *buffer_p = NULL;
 FILE   *inf      = NULL;
 int    ret       = -1 ;
 size_t filSiz    = 0;

    // Start: retrieve the text files size in bytes
    filSiz = frm_ioGetFilesize(inputPath_p,"frm_ioReadTextIntoBuffer");
    
    if((buffer_p=(char *)calloc(filSiz+2,sizeof(char)))==NULL)
    {
        fprintf(stderr,"@E frm_ioReadTextIntoBuffer failed to open(%s)!\n",
                inputPath_p);
        exit(-1);
    }
    
    inf= frm_Openfile(inputPath_p,"rb","frm_ioReadTextIntoBuffer");
    
    // ingest entire file into a raw binary buffer
    if((ret = fread(buffer_p,sizeof(char),filSiz,inf))!= (int)filSiz)
    {
        fprintf(stderr,"@F frm_ioReadTextIntoBuffer PASS-1 FAILED on fread of (%s) filesize %zu\n",
            inputPath_p, filSiz);
        exit(-1);
    } // end:: fread block
    
    // close the input file itself
    frm_Closefile(inf,"frm_ioReadTextIntoBuffer");  
    
    return(buffer_p);
    
} // end:: frm_ioReadTextIntoBuffer()


// -----------------------------------------------------
// function : frm_ioGetTextfileWidth()
// purpose  : read a text file, scans to establish the widest text file
//            line, and returns the character count of the widest line.
// inputs   : const char *inputTextPath_p,int *lineCount
// returns  : int, count in characters of widest line
// notes    : n/a
int  frm_ioGetTextfileWidth(const char *inputTextPath_p,int *lineCount)
{
 int  thisCh =0,
    nLines =0 ,
    maxWidth =0,
    k=0,currLineWidth =0;

  FILE *inf = NULL;

    if((inf=fopen(inputTextPath_p,"rt")) == NULL)
    {
        fprintf(stderr,"@E frm_strGetTextfileWidth: Cannot open text(%s)\n",
            inputTextPath_p);
        return -1 ; // returns -1 if cannot open
    }

    // do while...
    do 
    {
        currLineWidth = 0 ; // reset count for each line
        // read current line, count width
        k =0;
        while( (thisCh=getc(inf)) != '\n' && thisCh != EOF)
        {
            // testbuffer[k++] = thisCh;
            currLineWidth++; // increment count of chars on this line
        } // end:: while thisCh..iterating through chars on this line
        
        // fprintf(stdout,"RAWLINE %d <%s>\n",nLines, testbuffer);
        // memset(testbuffer,'\0',1024);
        
        nLines++ ; // increment count of lines encountered
        if(nLines == 1)
        {
            maxWidth = currLineWidth;
        } else {
            if(currLineWidth > maxWidth)
                maxWidth = currLineWidth ;
        }
    } while(thisCh != EOF) ; // end:: while in feof() loop

    // close the file examined
    fclose(inf);
    // update line count to return as well
    *lineCount = nLines-1;

    fprintf(stdout,"@A frm_strGetTextfileWidth: DIAG Opened(%s) n. lines(%d) maxWidth(%d)\n",
                        inputTextPath_p,nLines-1, maxWidth+1);

    // return character count of widest line
    // add one for null term
    return maxWidth+1 ;
} // end:: frm_ioGetTextfileWidth()


// --------------------------------------------------------------------
// Purpose: boolean tests if a directory actually exists at a given path
int frm_ioDoesDirectoryExist(const char *testDirectoryPath)
{
 // bool isdir =0 ;
 int  isDirectoryPresent = 0;
 struct stat st; 

    // call stat() supplying this path
    isDirectoryPresent =(stat(testDirectoryPath,&st) == 0)? 1 :0;
    
    // alternate logic...using the S_ISDIR() macro
    // isdir = S_ISDIR(st.st_mode);
    
 return isDirectoryPresent;
}  // end::frm_ioDoesDirectoryExist()

// --------------------------------------------------------------------
// Purpose : boolean tests if a directory actually exists at a given path
// inputs  : const char *testFilePath
// returns : int, boolean 1 when true,  0 when false.
int frm_ioDoesFileExist(const char *testFilePath)
{
 // bool isdir =0 ;
 int  isFilePresent = 0;
 struct stat st; 

    // call stat() supplying this path
    isFilePresent = (stat(testFilePath,&st) == 0)? 1: 0 ;
    // alternate c++ style logic...using the S_ISDIR() macro
    // isdir = S_ISDIR(st.st_mode);
 return isFilePresent;
}  // end::frm_ioDoesFileExist()

