// ---------------------------------------------------------------
// module  : frm_mem_mgr
// purpose : Implement the new MemMgr multiple-instance interface, where one (private)
//            MemMgr is defined per process, or per thread, when using a multi-threaded paradigm
//            each MemMgr "overseer" instance represents its own independent memObj registry
//            to provide for thread-safe task partitioning. Note this module file
//            is NOT yet fully implemented nor included in the standard API makefile.
// version : 1
// revised : 2015-05-29T12:27, added missing closing paren in memMgrInit() as above.
// revised : 2014-01-30T12:42:00, implemented initial MemMgr version.
// history : 2014-01-28T13:41:00, added MemMgr ADT for future thread-safe extensions, but
//           the implementation of MemMgr is not currently implemented or tested.
// history : 2014-01-21T11:26:00, minor clean up, prep for deprecating auto dot.qualifier
//           as its better for caller to manually construct any dot.qualifier on name
//           needed.. and added memObjGetVersTag()
// history : 2014-01-06T03:09:00, added vDnaLabel[40] sized to fit %p or MD5 string
// history : 2014-01-03T13:42:00, corrected ability to now free IsStringTable
// author  : joe glassy
// notes   : There are really TWO ways to create a first MemMgr instance:
//           1) If you call memMgrInit() with nMemObjInitialCount=0, no MemMgr parents are created
//              but this call DOES minimally initialize the facility. After you init MemMgr facility
//              this way, you MUST then make individual calls to memMgrCreate() to actually create
//              any given parent MemMgr object.
//           2) If you call mgrMem_p = memMgrInit with a positive count of MemObjs, this call
//              creates a SINGLE memMgr_p, and within it, creates nMemObjInitialCount children MemObj instances.
// notes    : Once a MemMgr parent exists and owns K "ready-to-use" MemObj children, you allocate
//            memory to any of the parent MemMgr's children using 
// ---------------------------------------------------------------

#include "frm_api.h"

static int memMgrFacilityInitialized = 0 ;


// ---------------------------------------------------------------
// function  : memMgrInit()
// purpose   : Mandatory init function for MemMgr; this creates ONE MemMgr, and sets this MemMgr's fixed length 
//             registry size - the number of MemObj slots it owns. It does NOT allocate any memory to any
//             MemObj slot.
// inputs    : const char *memMgrInitialLabel,const int nMemRegistrySize,const int ownerId,FILE *usersLogf)
// returns   : MemMgr *m --a ptr to a ready-to-use MemMgr instance if nMemObjInitialCount >0, or NULL if 
//             the user decided to wait and create new MemMgr instances manually later.
// notes     : It is mandatory to call this function ONCE (early in app startup) for any app
//             that uses any of the MemMgr (hence, memObj) functionalities. 
// notes     : <JMG> 2015-05-29, corrected syntax bug (missing closing paren added)
//             on line for: if((memMgr_p = (MemMgr_p *)calloc(1,sizeof(MemMgr))==NULL))
//             Noting this was missed since this module is NOT yet in Makefile and
//             wasn't compiled as part of API yet. Integration pending, not yet used.
MemMgr *memMgrInit(const char *memMgrInitialLabel,const int nMemRegistrySize,
                    const int ownerId,FILE *usersLogf)
{
 MemMgr *memMgr_p = NULL;
 int nMemObjects = nMemRegistrySize;
 int m =0;
 
    // Caller MUST specify a positive count registry size as nMemRegistrySize
    //   This is the count of MemObj this MemMgr parent will manage.
    assert(nMemRegistrySiz >0) ;

    // Create a new instance of a MemMgr, hence a new MemObj registry
    memMgr_p = calloc(sizeof(MemMgr),1) ;
    if(memMgr_p == NULL)
    {
        fprintf(stderr,"@F memMgrInit: FAILED creating MemMgr (%s)\n",memMgrLabel);
        exit(-1);
    }

    // Initialize member fields of this new MemMgr instance
    memMgr_p->vMgrId = FRM_API_RANDOM_INT(10000); // placeholder for now
    
    // Register the logf used by this MemMgr and all its MemObj children
    memMgr_p->vMgrLogf = usersLogf ; 
    
    // Store this MemMgr's user provided text label or name
    strcpy(memMgr_p->vMgrLabel,memMgrLabel);
    
    // self-register my own DNA
    sprintf(memMgr_p->vMgrDna,"%p",memMgr_p); 
    
    // Memory journaling mechanism: these will be dynamically
    //  adjusted later...
    memMgr_p->vMgr_highestSlotReferenced =  0;
    memMgr_p->vMgr_nMemObjRegistered     =  0;  
    
    // This is the TOTAL SUM of bytes, allocated across
    //     ALL MemObj children, that just THIS MemMgr instance owns
    memMgr_p->vMgr_nMemObjBytesTracked   =  0;
    
    // Create this MemMgr's private registry, consisting of nMemObjects slots
    if((memObjPrivate_p = (MemObj *)calloc(nMemObjects,sizeof(MemObj)))==NULL)
    {
        fprintf(stderr,"@F memMgrInit: FAILED creating (%s) private MemObj registry of %d slots\n",
            memMgrLabel, nMemObjects);
        exit(-1);
    }
    memMgr_p->vMgrCount = nMemObjects; // store the count of MemObj's this MemMgr will own
        
    // iterate through each MemObj child object, initializing each
    for(m=0; m < nMemObjects; m++)
    {
        // Initialize this 'm'th slot, as the newest child instance.
        //  Remember that THIS step only declares a slot as available for use, it does
        //  NOT actually allocate any memory to the 'm'th child --to do this use
        //  memObjCalloc() or memCalloc() as per usual convention
        memMgr_memObjInit(memMgr, m, mMgrLogf);
    }
        
    // Note that none of the valid candidate slots are registered yet
    memMgr_p->vMgr_nMemObjRegistered =  0; 
        
    // set critical flag to TRUE to indicate the memMgr facility at large is ready to use
    memMgrFacilityInitialized = 1 ;

    // return pointer to ONE new MemMgr.
     return memMgr_p ;
     
} // end::memMgrInit()


// ---------------------------------------------------------------
// function  : memMgr_InitRegistry()
// purpose   : just an alias for memMgrInit()
// inputs    : const char *memMgrInitialLabel,const int nMemRegistrySize,
//                  const int ownerId,FILE *usersLogf
// returns   : MemMgr ptr
// notes     : n/a
MemMgr *memMgr_InitRegistry(const char *memMgrInitialLabel,const int nMemRegistrySize,
                    const int ownerId,FILE *usersLogf)
{
 MemMgr memMgr_p = NULL;
 
     memMgr_p = memMgrInit(memMgrInitialLabel, nMemRegistrySize,
                    ownerId, usersLogf);
 return memMgr_p; 
 
} // end::memMgr_InitRegistry()



// ---------------------------------------------------------------
// function: memMgrFindNextAvailSlot()
// purpose : finds the index of the next availble slot owned by this MemMgr
// returns : base0 slot index (>=0), or if NO slots available returns -1
int memMgrFindNextAvailSlot(MemMgr *memMgr_p)
{
 MemObj memObj_p = NULL;    
 int m=0;
 int slotIndexFound = -1 ;
 size_t siz = 0;
 size_t nElem=0;
 
    // locate first available un-registered (candidate) slot 
    for(m=0; m < memMgr_p->vMgrCount; m++)
    {
        memObj_p = &memMgr_p->vMgrMemRegistry_p[m];
        if(memObj_p->vBytes ==0 && strlen(memObj_p->vDnaLabel)==0)
            break;
    } // end:: m loop through all candidate slots
    
    if( m < memMgr_p->vMgrCount)
    {
                slotIndexFound = m ;
    } else {
                slotIndexFound = -1;
    }
    
 return slotIndexFound ;
    
} // end:: memMgrFindNextAvailSlot()


// ---------------------------------------------------------------
// function: memMgrFindNextAvailSlot()
// purpose : finds the index of the next availble slot owned by this MemMgr
// returns : base0 slot index (>=0), or if NO slots available returns -1
int memMgrFindSlotByDna(MemMgr *memMgr_p,void *vPtr_p)
{
 MemObj memObj_p = NULL;    
 char dnaName[32] = {0}  ;
 int m=0;
 int slotIndexFound = -1 ;
 size_t siz = 0;
 size_t nElem=0;
 
    // capture DNA name of arg passed in
    sprintf(dnaName,"%p",(uint8 *)vPtr_p);
 
    // use it to locate first available un-registered (candidate) slot 
    for(m=0; m < memMgr_p->vMgrCount; m++)
    {
        memObj_p = &memMgr_p->vMgrMemRegistry_p[m];
        if(strcmp(dnaName,memObj_p->vDnaLabel)==0)
            break;
    } // end:: m loop through all candidate slots
    
    if( m < memMgr_p->vMgrCount)
    {
                slotIndexFound = m ;
    } else {
                slotIndexFound = -1;
    }
 return slotIndexFound ;
} // end:: memMgrFindSlotByDna()


// ---------------------------------------------------------------
// function: memMgr_FindSlot()
// purpose : find an existing slot (in use) e.g. as registerd using its slotIndex or user name
// inputs  : MemMgr *memMgr_p,
//           const char *objName_p -- user specified text label or name for a registered MemObj child 
//           const int slotIndex -- a valid base0 slot index mapped to a priorregistered slot
// returns : int, slotIndex (>=0) if it WAS found , or -1 if it was NOT found.
// notes   : must either specify a valid slotIndex, or a user name, but typically not both. If
//           both are supplied, then the search is done first on slot then on name.
//         : To search on objName_p, set slotIndex = -1, 
//           To search on slotIndex, set objName_p= "".
int memMgrFindSlot(MemMgr *memMgr_p,const char *objName_p,const int slotIndex)
{
 MemObj memObj_p = NULL;
 int m;
 int foundSlot = -1;    

    for(m=0; m < memMgr_p->vMgrCount; m++)
    {
        // set ptr to next child...
        memObj_p = &memMgr_p->vMgrMemRegistry_p[m];
        
        if(slotIndex >0 && slotIndex < megMgr_p->vMgrCount) 
        {
                foundSlot = m ;
                break;
        } else if( strlen(objName_p)>0 && strcmp(memObj_p->vName,objName_p)==0) {
                foundSlot = m ;
                break;
        }
    } // end:: m loop thorough candidates
    
 // existing slot found, so return its base0 index  
 return foundSlot ; 
} // end:: memMgrFindSlot()


// ---------------------------------------------------------------
// function: memMgr_Calloc()
// purpose : memMgr_Calloc()
// inputs  :
// returns :
void *memMgr_Calloc(MemMgr *memMgr_p,const int tivDatatype, const size_t nElemForMemObj)
{
 MemObj memObj_p = NULL;    
 int m=0;
 int foundSlot = -1;
 size_t siz = 0;
 size_t nElem=0;
 
    // locate first available un-registered (candidate) slot 
    if((foundSlot =  memMgrFindNextSlot(memMgr_p)) >= 0)
    {
        memObj_p = &memMgr_p->vMgrMemRegistry_p[foundSlot];
    }
    
    // found an available slot.. set it up
    siz = tivSizeOfOneDatatype(tivDatatype);
    memObj_p->vContents_p = (uint8 *)calloc(nElem,siz) ;
    if(memObj_p->vContents_p == NULL)
    {
      fprintf(stderr,"@F memMgr_Calloc: FAILED allocating slot(%d) nElme(%d) sizOfOne(%d)\n",m, nElem,siz);
      exit(-1);
    }
    // ok, so can now proceed...
    
    // first, retrieve and store my own DNA
    sprintf(memObj_p->vDnaLabel,"%p",memObj_p->vContents_p);
    // fetch "size of one" for this TIV datatype
    siz = tiv_SizeofTivType(tivDatatype);

    // set all the properties
    memObj_p->vRank = 1 ; 
    memObj_p->vSizeOfOne = siz;
    memObj_p->vDatatype  = tivDatatype;
    memObj_p->vDataclass = -1 ;
    memObj_p->vNelem     = nElem ;
    memObj_p->vOwner     = 0 ;
    memObj_p->vBytes     = nElem * siz ;
    // Next, adjust all PARENT's (memMgr) metadata
    memMgr_p->vMgr_nMemObjBytesTracked += memObj_p->vBytes ;
    // Increment count of successfully registered slots
    memMgr_p->vMgr_nMemObjRegistered++ ;
    // Bump up the sessions high-water-mark if a new high was hit
    if(memMgr_p->vMgr_nMemObjBytesTracked > memMgr_p->vMgr_nMemHighWaterMark)
        memMgr_p->vMgr_nMemHighWaterMark = memMgr_p->vMgr_nMemObjBytesTracked;
    
 return memObj_p->vContents_p ;

} // end:: memMgr_Calloc()


// ---------------------------------------------------------------
// function  : memMgr_CallocWithName()
// purpose   : Identifying a MemMgr's child MemObj slot by Name (memObjName), 
//             this function allocates the child of tivDatatype with nElements and assigns the
//             callers chosen "name" to it.
// inputs    : MemMgr *memMgr_p -- pointer to the parent who owns the MemObj we're working with
//             const char *childName -- the text name to assign to this new object.
//             const int tivDatatype -- the datatype applying to the child, e.g. IsFloat32, IsInt16 etc.
//             const size_t nElem -- the number of elements of tivDatatype to allocate memory for
// returns   : void * --a ptr to the allocated memory
// notes     : n/a
void *memMgr_CallocWithName(MemMgr *memMgr_p,const char *childName_p,const int tivDatatype,const size_t nElem)
{
 void *vPtr_p = NULL;
 int thisSlotIdx = -1;
 
  // allocate the memory. So far, this object has no user assigned name.
  vPtr_P = memMgr_Calloc(memMgr_p,tivDatatype,nElem);
  
  // find the base0 slot index was just mapped to this newest entry
  thisSlotIdx =  memMgrFindSlotByDna(memMgr_p,vPtr_p);
  
  // By supplying a valid slot index, we force the assignment of the name
  // to the correct slot just allocated...
  memMgr_SetName(memMgr_p,childName_p,thisSlotIdx);

  return vPtr_p ;
    
} // end::memMgr_CallocWithName()


// ---------------------------------------------------------------
// function : memMgr_FreeAndUnregisterOneSlot()
// purpose  : low level (private) function to free and un-register a slot
//            assoc with memObj_p.
// inputs   : 
// notes    : apps and callers do NOT ever actually call this. It is
//            a private function only, called by memMgr_Free() or
//            memMgr_FreeBySlot() or memMgr_FreeByName() 
void memMgr_FreeAndUnregisterOneSlot(MemObj *memObj_p)
{
        free(memObj_p->vContents_p);
        // All these next steps are what we do to 
        // "un-register" an existing slot.
        strcpy(memObj_p->vName,"");
        strcpy(memObj_p->vDnaLabel,"");
        memObj_p->vRank = 0 ;
        memObj_p->vSizeOfOne = 0;
        memObj_p->vDatatype = -1;
        memObj_p->vDataclass= -1;
        memObj_p->vNelem    =  0 ;
        memObj_p->vOwner    = -1 ;
        // can now zero out this objects vBytes used
        memObj_p->vBytes = 0 ;
        // explicitly Re-MARK as NULL 
        memObj_p->vContents_p = NULL;
    
} // end::memMgr_FreeAndUnregisterOneSlot()


// ---------------------------------------------------------------
// function  : memMgr_Free()
// purpose:  frees 1 memObj owned by MemMgr, using its DNA to identify which one
//           to free. Equivalent to original memFree()
void memMgr_Free(MemMgr *memMgr_p,void *mPtr_p)
{
 MemObj *memObj_p = NULL;
 char internalDNA_Name[32] = {0} ;
 char thisDNA_Name[32] = {0} ;
 int m= 0;

        // isolate DNA of entity provided by caller
        sprintf(thisDNA_Name,"%p",mPtr_p);
        
        for(m=0; m < memMgr_p->vMgrCount ; m++)
        {
            memObj_p = &memMgr_p->vMgrMemRegistry_p[m];
            // Fetch a sample of its DNA
            sprintf(internalDNA_Name,"%p",memObj_p->vContents_p);
            // Test for a DNA match
            if(strcmp(internalDNA_Name,thisDNA_Name)==0)
            {
                // fetch vBytes to use to decrement count of managed bytes
                memMgr_p->vMgr_nMemObjBytesTracked -= memObj_p->vBytes;
                // now FREE the memory and "un-register" this slot
                memMgr_FreeAndUnregisterOneSlot(memObj_p);
                // Adjust memMgr's count of used vs. free slots
                // (decrement count of slots in use)
                memMgr_p->vMgr_nMemObjRegistered-- ;
                break;
            } // end:: block if matched DNA 
        } // end:: m loop through memObj owned by this MemMgr
        
} // end::memMgr_Free()


// ---------------------------------------------------------------
// function  : memMgr_FreeBySlot()
// purpose:  : frees 1 memObj owned by MemMgr, using its base0 slot index
//             to identify which one to free. By 'un-registering' a slot
//             this way, it is put back into a free pool of slots that
//             can be re-cycled again within the session.
// inputs    : MemMgr *memMgr_p,
//           : const int ithSlot
void memMgr_FreeBySlot(MemMgr *memMgr_p,const int ithSlot)
{
 MemObj *memObj_p = NULL;
 int m= 0;

    if(ithSlot < memMgr_p->vMgrCount)
    {
        memObj_p = &memMgr_p->vMgrMemRegistry_p[ithSlot];
        
        // fetch vBytes to use to decrement count of managed bytes
        memMgr_p->vMgr_nMemObjBytesTracked -= memObj_p->vBytes;
        // now FREE the memory and "un-register" this slot
        memMgr_FreeAndUnregisterOneSlot(memObj_p);
        // Adjust memMgr's count of used vs. free slots
        // (decrement count of slots in use)
        memMgr_p->vMgr_nMemObjRegistered-- ;
    } // end:: block if slot qualified as a candidate
    
} // end::memMgr_FreeBySlot()


void memMgr_FreeByName(MemMgr *memMgr_p,const char *thisObjName_p)
{
 MemObj *memObj_p = NULL;
 int m= 0;

        // isolate DNA of entity provided by caller
        sprintf(thisDNA_Name,"%p",mPtr_p);
        
        for(m=0; m < memMgr_p->vMgrCount ; m++)
        {
            memObj_p = &memMgr_p->vMgrMemRegistry_p[m];
            
            // Test for a user provided name match
            if(strcmp(thisObjName_p, memObj_p->vName)==0)
            {
                // fetch vBytes to use to decrement count of managed bytes
                memMgr_p->vMgr_nMemObjBytesTracked -= memObj_p->vBytes;
                // now FREE the memory and "un-register" this slot
                memMgr_FreeAndUnregisterOneSlot(memObj_p);
                // Adjust memMgr's count of used vs. free slots
                // (decrement count of slots in use)
                memMgr_p->vMgr_nMemObjRegistered-- ;
                break;
            } // end:: block if name matched 
        } // end:: m loop through memObj owned by this MemMgr
    
} // end::memMgr_FreeByName()


// ---------------------------------------------------------------
// function  :
// purpose   : memMgrRelease() deletes a parent MemMgr, after first
//             releasing all children etc.
// inputs    :
// returns   :
// notes     : n/a
void memMgrRelease(MemMgr *memMgr_p)
{
 MemMgr memMgr_p = NULL;
 MemObj memObj_p = NULL;
 
 int nChildren = memMgr_p->vMgrCount;
 
    for(m=0; m < nChildren; m++)
    {
            memObj_p = &memMgr_p->vMgrMemRegistry_p[m];
            // free this child, by slot index
            memMgr_FreeBySlot(memMgr_p,m);
    } // end:: m loop iterating through children to free first
    
    // free the parent
    free(memMgr_p);

} // end::memMgrRelease()




// ---------------------------------------------------------------
// function  :memMgr_SetId()
// purpose   : assign a MemMgr with an unique integer ID
// inputs    : MemMgr *memMgr_p,const int memMgrId
// returns   :
// notes     : n/a
void memMgr_SetId(MemMgr *memMgr_p,const int thisId)
{
    memMgr_p->vMgrId = thisId ;
} // end::memMgr_SetId()


// ---------------------------------------------------------------
// function  :
// purpose   : assign a MemMgr with an unique integer Owner ID
// inputs    :
// returns   :
// notes     : n/a
void memMgrSetOwnerId(MemMgr *memMgr_p,const int memMgrOwnerId)
{

} // end::memMgrSetOwner()


// ---------------------------------------------------------------
// function  :
// purpose   : return a pointer to this MemMgr's private MemObj registry
// inputs    :
// returns   :
// notes     : n/a
MemObj *memMgrGetRegistry(MemMgr *memMgr_p)
{

    return memObj_p->vMgrMemRegistry_p ;
    
} // end::memMgrGetRegistry()


// ---------------------------------------------------------------
// function  :
// purpose   : run a full report all MemMgr and its K children MemObj's 
// inputs    :
// returns   :
// notes     : n/a
void memMgrReport(MemMgr *memMgr_p,FILE *destf)
{

} // end::memMgrReport()

// ---------------------------------------------------------------
// function: memMgr_SetName()
// purpose : assigns a text name to a MemObj child owned by memMgr_p.
// inputs  : MemMgr *memMgr_p,const char objName
// returns :
// notes   : must either specify a valid slotIndex, or a dnaLabel
void memMgr_SetName(MemMgr *memMgr_p,const char *objName_p,const int slotIndex)
{
 MemObj memObj_p = NULL;
 int m;
 int foundSlot = -1;    

    if(slotIndex > 0) {
        foundSlot = memMgrFindSlot(memMgr_p,"",slotIndex);
    }else if(strlen(objName_p)>0) {
        foundSlot = memMgrFindSlot(memMgr_p,objName_p,-1);
    }else {
        fprintf(stderr,"@E memMgr_SetName: FAILED --no valid name or slot-index specified!\n");
    }
    
    // slot found, so assign the (name) property
    if(foundSlot >= 0)
        strcpy(memObj_p->vName, objName_p);
    
} // end:: memMgr_SetName()


// ---------------------------------------------------------------
// function: memMgr_SetRank()
// purpose : assigns a rank to a MemObj child owned by memMgr_p.
// inputs  : MemMgr *memMgr_p,const char objName
// returns :
// notes   : must either specify a valid slotIndex, or a dnaLabel
void memMgr_SetRank(MemMgr *memMgr_p,const char *objName_p,const int slotIndex,const int thisRank)
{
 MemObj memObj_p = NULL;
 int m;
 int foundSlot = -1;    

    if(slotIndex > 0) {
        foundSlot = memMgrFindSlot(memMgr_p,"",slotIndex);
    }else if(strlen(objName_p)>0) {
        foundSlot = memMgrFindSlot(memMgr_p,objName_p,-1);
    }else {
        fprintf(stderr,"@E memMgr_SetRank: FAILED --no valid name or slot-index specified!\n");
    }
    
    // slot found, so assign the (rank) property
    if(foundSlot >= 0)
        memObj_p->vRank = thisRank;
    
} // end:: memMgr_SetRank()


// ---------------------------------------------------------------
// function  :memMgrSetDataclass()
// purpose   : assign a MemMgr's memObj with an unique integer ID
// inputs    : MemMgr *memMgr_p,const int thisDataclass
// returns   :
// notes     : n/a
void memMgr_SetDataclass(MemMgr *memMgr_p,const char *objName_p,const int slotIndex,const int thisDataclass)
{
 MemObj memObj_p = NULL;
 int m;
 int foundSlot = -1;    

    if(slotIndex > 0) {
        foundSlot = memMgrFindSlot(memMgr_p,"",slotIndex);
    }else if(strlen(objName_p)>0) {
        foundSlot = memMgrFindSlot(memMgr_p,objName_p,-1);
    }else {
        fprintf(stderr,"@E memMgr_SetDataclass: FAILED --no valid name or slot-index specified!\n");
    }
    
    // slot found, so assign the (dataclass) property
    if(foundSlot >= 0)
        memObj_p->vDataclass = thisDataclass;
    
} // end:: memMgr_SetDataclass()

// ---------------------------------------------------------------
// function  : memMgr_GetChildByName()
// purpose   :
// inputs    :
// returns   :
// notes     : n/a
MemObj *memMgr_GetChildByName(MemMgr memMgr_p,const char *memObjName_p)
{
 MemMgr memMgr_p = NULL;
 MemObj memObj_p = NULL;
 
 int foundSlot = -1 ;  
   
    foundSlot = memMgrFindSlot(memMgr_p, memObjName_p, -1);
    
    // slot found
    if(foundSlot >= 0)
        memObj_p = &memMgr_p->vMgrMemRegistry_p[foundSlot];
 
 // return CHILD object located at this slot
 return memObj_p ;
 
} // end:: memMgrGetChildByName()


// ---------------------------------------------------------------
// function  : memMgr_GetChildBySlot()
// purpose   :
// inputs    :
// returns   :
// notes     : n/a
MemObj *memMgr_GetChildBySlot(MemMgr memMgr_p,const int thisSlot)
{
 MemMgr memMgr_p = NULL;
 MemObj memObj_p = NULL;
 
 int foundSlot = -1 ;  
   
    foundSlot = memMgrFindSlot(memMgr_p, "", thisSlot);
    
    // slot found
    if(foundSlot >= 0)
        memObj_p = &memMgr_p->vMgrMemRegistry_p[foundSlot];
 
 // return CHILD object located at this slot
 return memObj_p ;
 
} // end:: memMgrGetChildBySlot()
