// ------------------------------------------------------------
// module       : frm_util_fx.c
// purpose      : Implement misc utility functions as part of L4_C "framework"
// revision     : 2014-11-03T16:37:00, added frm_utilGetHostnameAndUser()
// revision     : 2014-02-09T22:08:38:00,migrated frm_Openfile(), frm_Closefile() to frm_io_fx.c
//                and corrected labelling of frm_utilGenerateUUID() issue.
// history      : 2014-01-30T11:37:00
// history      : 2014-01-30T13:23:00,changed R_ to FRM_API_ prefix for all _RANDOM macros
// history      : 2014-01-24T13:49:00,wheat,jmg
// history      : n/a
// author       : joe glassy
// dependencies : verify, HDF4, HDF5 etc
// notes        : the "frm_util_fx.c" codebase is a "catch-all" repository of
//                useful Framework that isn't classified (yet) as belonging to a
//                larger themed group.
//
// Notes: Intel byteorder is isLSBfirst
//  enum EndianStates { isMSBfirst, isLSBfirst, isVAXEndian, unknownEndian } 
//

#include "frm_api.h"

// endian labels 
char *endianLabels[] = { "big", "little", "vaxEndian","unknownEndian"} ;  

#define NUM_BITS_LIMIT 32

// <JMG> added for univar calc (placeholder)
static size_t n_uni_errors =0;


int BIT32[] = 
{
  0x1,        0x3,        0x7,        0xF,
  0x1F,       0x3F,       0x7F,       0xFF,
  0x1FF,      0x3FF,      0x7FF,      0xFFF,
  0x1FFF,     0x3FFF,     0x7FFF,     0xFFFF,
  0x1FFFF,    0x3FFFF,    0x7FFFF,    0xFFFFF,
  0x1FFFFF,   0x3FFFFF,   0x7FFFFF,   0xFFFFFF,
  0x1FFFFFF,  0x3FFFFFF,  0x7FFFFFF,  0xFFFFFFF,
  0x1FFFFFFF, 0x3FFFFFFF, 0x7FFFFFFF, 0xFFFFFFFF
} ;


// ---------------------------------------------------------------------
// function : frm_utilTestBooleanStr()
// inputs   : const char *booleanOptionSwitch_p
// returns  : int, boolean, 1=if true 0=false
// revision : 2014-09-12T15:37:00
// author   : joe glassy, Lupine Logic Inc.
// purpose  : Optionally rescales output array, and, derives and returns
int frm_utilTestBooleanStr(const char *optionSwitch_p)
{
 int resultBoolean = 0 ;
 char evalBuffer[32] = {0} ;
 
  strcpy(evalBuffer,optionSwitch_p);
  // normalize...
  frm_strToUpper(evalBuffer);
  // test
  
  // allow for a variety of responses to equal true...
  if(strcmp(evalBuffer,"YES")  ==0 || 
     strcmp(evalBuffer,"ON")   ==0 || 
     strcmp(evalBuffer,"TRUE") ==0 || 
     strcmp(evalBuffer,"1")    ==0 ||   
     strstr(evalBuffer,"ENABLE")!=NULL)
  {
        // set TRUE if any of these match
        resultBoolean = 1 ;
        
  } else if (strcmp(evalBuffer,"NO")  ==0 || 
     strcmp(evalBuffer,"OFF")   ==0 || 
     strcmp(evalBuffer,"FALSE") ==0 || 
     strcmp(evalBuffer,"0")    ==0 ||   
     strstr(evalBuffer,"DISABLE")!=NULL)
  {
        // set FALSE when explicitly told to
        resultBoolean = 0 ;
  } else {
      // set FALSE conservatively
        resultBoolean = 0 ; 
  }
  
 return resultBoolean ;
} // end:: frm_utilTestBoolean()

// ---------------------------------------------------------------------
// function : frm_utilGetPid() 
// Purpose  : retrieve this process' id
// Inputs   : void
// Returns  : pid_t -- process id, unsigned long via type pid_t
// Author   : joe glassy
// Revised  : 2013-11-04T17:09:00-07:00,wheat
// Notes    : we could also use sys/sysctl.h and sysctl() etc etc.
pid_t frm_utilGetPid(pid_t *parent_pid)
{
    pid_t pid = 0;

    // simplify, just fetch PID using system call 
    pid = getpid();
    // fetch parent process'es PID while we're at it
    *parent_pid = getppid();

    // return the process ID 
    return pid ;
} // end:: frm_utilGetPid()

// ---------------------------------------------------------------------
// function : frm_utilGetHostnameAndUser() 
// Purpose  : retrieve the hostname process is running on and user name
// Inputs   : void
// Returns  : int status 0= ok, -1 otherwise
// Author   : joe glassy
// Revised  : 2014-11-03T16:32:00
// Notes    : thisHostname buffer must be >= 256 ch and thisUsername buffer
//            must be >= 128 ch.
// Notes    : Currently only guaranteed for use with non-daemon processes
// 
int frm_utilGetHostnameAndUser(char *thisHostname,char *thisUsername)
{
 int status = -1 ;
 
    // clear buffer, then capture name of host we're running on
    memset(thisHostname,'\0',256);
    memset(thisUsername,'\0',128);
    gethostname(thisHostname,255) ;

    // Capture name of user controlling terminal session, limited to 128 ch.
    // This WILL fail if this process is daemonized, so note this constraint
    getlogin_r(thisUsername, 128);

    if(strlen(thisHostname)>0 && strlen(thisUsername)>0)
        status = 0 ;
        
    // return the status
    return status;
} // end:: frm_utilGetHostnameAndUser()

// ---------------------------------------------------------------------
// function : frm_utilGetCtime() 
// Purpose  : retrieve a formatted instance of current time using ctime.
// Inputs   : char *timBuf,const time_t usersTime
// Author   : joe glassy
// Revised  : 2013-11-04T17:09:00-07:00,wheat
// Notes    : if a valid 'usersTime' is submitted, this is used otherwise
//            the current time(0) at the time of the call is used.
//            This could be extended to support localtime, UTC time, etc
char *frm_utilGetCtime(char *timBuf,const time_t usersTime)
{
 time_t secNow = time(0L);

    if(usersTime > 0)
    {
        sprintf(timBuf,"%s",ctime(&usersTime));
    } else {
        sprintf(timBuf,"%s",ctime(&secNow));
    }

   // trim off the obnoxious last char we never need or want
   timBuf[strlen(timBuf)-1] = NULL_CH;

  // return string containing formatted ctime result
  return timBuf ;
} // end::frm_utilGetCtime()


// ---------------------------------------------------------------------
// function   : frm_utilEventCompletionMsg()
// Purpose    : Announces that function 'thisFunc_p' has
//              completed, with a snapshot of current time,
// Inputs     : const char *thisFunction_p,const time_t startSec,const int enableCompletionMsg
// Returns    : double, cumulative sec since session began
// Notes      :  n/a
// revision   : 2014-07-31T16:21:00, added
// Author     : joe glassy
double frm_utilEventCompletionMsg(const char *thisFunction_p,const time_t startSec,const int enableCompletionMsg)
{
 time_t secNow = time(0);
 char timeStr[32] = {0};
 
 // update a snapshot of cumulative running time
 double cumulativeSec = difftime(secNow,startSec) ;
 
    // <JMG> 4/28/2014, only display completion-message when
    // explicitly enabled, to reduce clutter...
    if(enableCompletionMsg == 1)
        xLogMsg("@A <%s> COMPLETED %s (%.2f elap sec)~\n",
            thisFunction_p,  frm_utilGetCtime(timeStr,0), cumulativeSec );
        
    return cumulativeSec;
    
} // end:: frm_utilEventCompletionMsg()


// ---------------------------------------------------------------------
// function: frm_utilTestByteorder()
// purpose : perform boolean test for byte ordering, retrieve platform's endian descriptor
// author  : joe glassy
// revised : 2013-09-01
// returns : 0=Big Endian (enum isMSBfirst), 1=Little endian (enum isLSBfirst)
// notes   : The ORDER of the enum constants in the enum itself is CRITICAL here, 
//           so these MUST conform to this ordering:
//             enum EndianStates { isMSBfirst, isLSBfirst, isVAXEndian, unknownEndian } ; 
//             char *endianLabels[] = {"big","little"} ;
uint8 frm_utilTestByteorder(void)
{
 uint8 rtn = 0;
 union {
   int i ;
   char c[sizeof(int)] ;
 } foo ;
 foo.i = 1 ;

 if(foo.c[0] == 1)
   rtn = isLSBfirst ;
 else
   rtn = isMSBfirst ;
 return rtn;
} // end::frm_utilTestByteorder() 


// ---------------------------------------------------------------------
// function   : frm_utilGetBits()
// inputs     : const int32 strt_at --base0 bit position to start fetch at
//              const int32 end_at -- base0 bit position to end fetch at
//              uint32 sourceData  -- integer holding bit pattern to retrieve from
// purpose    : implements getbits() from K&R C Language, to extact nBits starting at pStart
//                from a input unsigned integer bit word. Input may be 16|32 bits depending on platform
//              but on 64 bit Linux, is 32 bits by default.
// notes      : bit positions supplied "end_at" must be >= "strt_at" 
// revision   : Updated 2014-07-21T16:10:00
// author     : joe glassy
uint32 frm_utilGetBits(const int32 strt_at,const int32 end_at,uint32 sourceData)
{
 int32 p = end_at,             // anchor the start here 
       n = (end_at-strt_at)+1; // sum of bits (width) to retrieve 
       
 uint32 resultWord = (sourceData >> (p+1-n)) & ~(~0 << n) ;
 
 return resultWord;
 
} // frm_utilGetBits()


// -----------------------------------------------------------------------------
// function: frm_utilRetrieveBits() adapted from MUM API for MODIS
// purpose : from an integer 'data' of size 'dataSize', retrieve bits from 
//           base0 bit positions 'startPos' to 'endPos', returning the result
//           as a 32 bit unsigned integer.
// return  : uint32, dest integer with JUST the specified bitpattern requested
// notes   : This function returns identical results as frm_utilGetBits() but
//           has the advantage of requiring explicit (sizeof) parameters
//           which removes the ambiguity of the void *data passed in.
uint32 frm_utilRetrieveBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)
{
  int16  byteOrder = frm_utilTestByteorder();
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData;

  if(byteOrder == isMSBfirst)
   memcpy(localDataPtr  + (sizeof(uint32) - dataSize), data, dataSize);
  else
   memcpy(localDataPtr, data, dataSize);

  return((uint32)((localData >> startPos) & BIT32[endPos - startPos]));
  
} // end::frm_utilRetrieveBits()



// -----------------------------------------------------------------------------
// function   : frm_utilSetBits() ...adapted from mumSetBits()
// purpose    : Sets individual bits, or a range of bits explicitly
//              to '1' in 'data', starting at base0 bit 
//              position 'startPos' and ending at bit position endPos.
// invocation : mumSetBits(strtBit,strtBit+(nBits-1),&myWord,sizeof(myWord));
// history    : from Petr Votava, Joe Glassy from NASA Terra MODIS MUM API (MODIS code)
uint32 frm_utilSetBits(const int16 startPos,const int16 endPos, void *data, size_t dataSize)
{
  int16  byteOrder = frm_utilTestByteorder(); // renamed from mumTestByteorder()...
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData;

  if(byteOrder == isMSBfirst)
    memcpy(localDataPtr  + (sizeof(uint32) - dataSize), data, dataSize);
  else
    memcpy(localDataPtr, data, dataSize);

  localData = localData | (BIT32[endPos - startPos] << startPos);

  return(localData);
  
} // frm_utilSetBits()() 


// ---------------------------------------------------------------------
// function: frm_utilAssignBits()
// purpose : This function is typically used to ASSIGN (INSERT) an integer 
//           value (occupying multiple bits) into a specified bit position
//           range within a fixed sized destination integer variable.  The
//           origin (multiple-bit) integer value is as supplied via an integer
//          'data' of size 'dataSize', retrieve bits from base0 bit positions 
//           'startPos' to 'endPos' inclusive, returning the result
//           as a 32 bit unsigned integer, which may itself be down-cast to a smaller
//           result scalar such as a byte,int16, or uint16
// return     : uint32 -- final modified value of destination after the insertion.
// inputs     :  int16 startPos --the base0 starting bit index of the (data) to retrieve
//            :  int16 endPos -- the base0 ending bit index of the (data) to retrieve
//            :  void *destData -- pointer to the data (word) to retrieve
//            :  size_t dataSize -- the size in bytes of the data (word) to 
//               retrieve, e.g. sizeof(uint8) etc.
//               where dataSize is in the supported range of 1 (byte) to 4 (uint32)
//            : void *newValueMask -- the pointer to the NEW integer (multiple-bit)
//              value to INSERT (e.g. ASSIGN) into the destination word (void *destData)
//            : size_t maskSize -- the size of the new integer variable containing
//              the new value to insert into the word.
// Note that the enum defining IsMSBfirst and IsLSBFirst is located in frm_core.h
// Notes: (Petr Votava): All byte orderings are normalized to a standard
//           big endian (most significant byte first) ordering.
//
// EXAMPLE: from MODIS pge33, Petr Votava_ 8/15/02 - Yet new QQ schema:
//  isQaBest = 0,
//  isQaGood = 1,
//  isQaFair_BadGeom = 2,
//  isQaPoor = 3,
//  isQaNotPossible = 4
//  The bits are now {5..7}
//
//   NOTES: the SOURCE of the integer is uint8 "lai_obj_p->q_montana"
//          the DESTINATION (word) is qa_field variable (a uint8) 
//          the resulting value will occupy 3-bits, at base0 indices {5,6,7}
//          within the 8-bit result variable named 'qa_field'.
//
//   Fetch the final result (modified word) via "qa_field" as returned:
//
//     qa_field = mumMaskBits( 5,7, &qa_field, sizeof( uint8 ),
//               &lai_obj_p->q_montana, sizeof( uint8 ) );
//
uint32 frm_utilAssignBits(const int16 startPos,const int16 endPos, void *destData, size_t dataSize,
                      void *newValueMask, size_t maskSize)
{
  int16  byteOrder = frm_utilTestByteorder();
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData; // set pointer to addressable entity
  uint32  localMask = 0;
  uint8   *localMaskPtr = (uint8 *)&localMask; // set pointer to addressable entity


  if(byteOrder == isMSBfirst)
  {
    // offset is (sizeof(uint32) minus the size of the destData variable
    memcpy(localDataPtr  + (sizeof(uint32) - dataSize), destData, dataSize);
    memcpy(localMaskPtr  + (sizeof(uint32) - maskSize), newValueMask, maskSize);
  } else {
    memcpy(localDataPtr, destData, dataSize);  
    memcpy(localMaskPtr, newValueMask, maskSize);
  }  

    // first set 0's to the fields that we want to set, using logical AND  
    localData = localData & (~(BIT32[endPos - startPos] << startPos));

    // now mask it, using logical OR  
    localData = localData | ((localMask & BIT32[endPos - startPos]) << startPos);
  
    // Retrieve the new modified (destination) value
    // as a returned uint32 integer.  Cast to appropriate type if not a uint32.
    return(localData);
  
} // end:: frm_utilAssignBits()


// ---------------------------------------------------------------------
// function: frm_utilDecToBinarystr(uint32 num,int32 width,char *result_str)
// inputs  : uint32 num -- source decimal integer holding value to display
//           as a binary bit pattern of 0's and 1's.
//           int32 width -- width of string returned (padding factor)
//           char *result_str -- result string of width, must be allocated
//           as NUM_BITS_LIMIT+1  or larger 
// purpose : convert a decimal integer value in "num" into a string that
//           displays the binary bit pattern as 0's and 1's
// return  : char *string of '0's and '1's like '01101111'
//      This function is typically used for diagnostic purposes only.
//    See Deitel & Deitel "C/C++ How To Program" for another example,
//    ...display_bits()...
//
// notes   : typical practice is to allocate: result_str[64] = {0} .
//
char *frm_utilDecToBinarystr(uint32 num,int32 width,char *result_str)
{
  int16 i   = 0, 
        pow = 0 ;

  char work[NUM_BITS_LIMIT *2] = {0};

// ---------------------------------------------------------------------
// STEP 01) clear buffer each time, iteratively build up bit pattern...
//----------------------------------------------------------------------

  // explictly clear each time 
  memset(result_str, '\0', NUM_BITS_LIMIT+1);

  for(i = NUM_BITS_LIMIT - 1; i >= 0; i--)
  {
   result_str[i] = (char)(((num >> pow) & 0x1) + (int16)'0');
   pow++;
  }
  result_str[NUM_BITS_LIMIT] = '\0';

  // of the 32 candidate bits represented, we pass through 
  //   just the right-most WIDTH bits 
  strcpy(work, result_str +(NUM_BITS_LIMIT - width) );
  strcpy(result_str, work);
//----------------------------------------------------------------------
// STEP 02) Return pointer to buffer containing final binary bit pattern
//----------------------------------------------------------------------
  return result_str;
  
} // end:: frm_utilDecToBinarystr() 


//----------------------------------------------------------------------
// function    : frm_utilPrivateUUID()
// inputs      : char *resultPrivateUUID_Str --ptr to prealloc buffer 64 ch
//               to hold resulting UUID. 
// purpose     : Generate a quick-and-dirty, unofficial, 
//               time-based hyphenated UUID string
// revision    : 2014-08-12T10:26:00-07:00
// author      : joe glassy
// returns     : char *uuidPrivateStr - ptr to completed hyphenated UUID string
// attribution : Directly adapted from URL: 
//     http://stackoverflow.com/questions/2174768/generating-random-uuids-in-linux
// notes       : BUG! DO NOT US AS-IS! The original routine is not correct
//               as it produces 60 char UUID's.e.g.
//               
//----------------------------------------------------------------------
char *frm_utilPrivateUUID(char *resultPrivateUUID_Str)
{
    FRM_API_RANDOMIZE(); // srand(time(NULL));

    // init, presuming buffer is at least 64 char (intentionally oversize)
    memset(resultPrivateUUID_Str,'\0',64);
    // generate a private, proxy hyphenated UUID in case system library is not available
    sprintf(resultPrivateUUID_Str, "%x%x-%x-%x-%x-%x%x%x", 
        rand(), rand(),                 // Generates a 64-bit Hex number
        rand(),                         // Generates a 32-bit Hex number
        ((rand() & 0x0fff) | 0x4000),   // Generates a 32-bit Hex number of the form 4xxx (4 indicates the UUID version)
        rand() % 0x3fff + 0x8000,       // Generates a 32-bit Hex number in the range [0x8000, 0xbfff]
        rand(), rand(), rand());        // Generates a 96-bit Hex number
    
    return resultPrivateUUID_Str ;
    
} // end:: frm_utilPrivateUUID()


// ---------------------------------------------------------------------
// function : frm_utilGenerateUUID()
// Purpose  : generate a 5-part, 36-character, hyphenated type-1 UUID using libuuid's services
// Author   : joe glassy
// Inputs   : char *resultUUID_String
//          : displayMode (options:
//             0 --> displays real UUID as a  packed string
//             1 --> displays real UUID as a normal hypenated couplet form of output, as from uuid_unparse()
//             2 --> forces generation of a pseudo UUID and displays it in hyphenated couplet form
//             3 --> forces generation of a pseudo UUID and displays it as a packed string
//           or if displayMode==2, produces a dummy random placeholder (not authentic) UUID
// Example  : uuidgen command line utility returns: "398e667e-9e43-4f53-9aa4-718b0e5d1881"
//           "f93185bb-0ad4-4978-93cc-415736f3a5c0", "2393e78f-6272-4a98-bfe7-7849c925b215"
// History  : 2014-01-03T14:10:00-07:00, clone of l4c_GenerateUUID()
// Notes    : if the UUID library (libuuid.a or libuuid.so) is NOT available, a pseudo UUID is generated.
// Revision : 2014-01-03T14:10:00-07:00,jmg 
// Dependencies: -luuid must be linked in and header <uuid/uuid.h> must be available
//               on the build platform!
char *frm_utilGenerateUUID(char *resultUUID_String,const int displayMode)
{
 int  a,c,r,k=0;
 const int uLims[5] = {8,4,4,4,12} ;
 // define loop limits for the sub-parts of a synthetic UUID constuctor
 char uuidPart[64] = {0};
 char uuidPrivateStr[64] = {0};
 char thisPart[64] = {0};
 char uuidProxy[64]= {0} ;
 
    // note that a hyphenated couplet UUID is 36 ch so 37 req to terminate it
    // pre-initialize a 32 character UUID string
    memset(resultUUID_String,   '\0',48);
    memset(uuidPrivateStr   ,   '\0',64); // experimental, not used yet
    memset(uuidProxy        ,'\0',64)   ; // home-made proxy of a UUID
    
    // pre-emptively construct a time based proxy for 
    // an artificial UUID 5-part string 
    memset(uuidPart,'\0',64);   
    for(a=0; a < 5; a++)
    {
            memset(thisPart,'\0',64);
            for(c=0; c < uLims[a]; c++)
            {
                if(FRM_API_RANDOM_INT(100) <= 75)
                    thisPart[c]=(char)FRM_API_RANDOM_RANGE_1(97,122); // a..z
                else
                    thisPart[c]=(char)FRM_API_RANDOM_RANGE_1(48,57);  // 0..9
            }
            strcat(uuidPart,thisPart);
            if(a < 4) strcat(uuidPart,"-");
    }
    strcat(uuidProxy,uuidPart);
    
    memset(resultUUID_String,   '\0',48);
    memset(uuidProxy,           '\0',64);
    
    // if a real UUID was requested and service available, do it that way
#ifdef HAVE_UUID
        uuid_t uuid ;
        uuid_generate(uuid); // generate a Type-I uuid
        // load it into the result string
        uuid_unparse(uuid, resultUUID_String);    
        // fprintf(stdout,"\nfrm_utilGenerateUUID: uuid_unparse result[%s]\n",resultUUID_String);
#else

    // Systems /usr/lib64/libuuid.so or libuuid.a not able to link to,
    // so as an alternative, supply a locally generated timebased proxy 
    // of a UUID.
    
    // Method 1: experimental, not used yet (generates an incorrect, 60 char UUID )
    // strcpy(resultUUID_String, frm_utilPrivateUUID(uuidPrivateStr));
    
    // Method 2: generates a homemade timebase proxy UUID. This is used.
    strcpy(resultUUID_String, uuidProxy);
#endif    

    if(displayMode == 0 || displayMode == 3)
    {
        if(displayMode == 3)
        {
            memset(uuidProxy,'\0',64);
            // len is 36 chars, but add one for term null
            strncpy(resultUUID_String,uuidProxy,37);
        }
        
         for(r=0; r < (int)strlen(resultUUID_String);r++)
          if(resultUUID_String[r] != '-')
                resultUUID_String[k++] = resultUUID_String[r];
         // re-insert terminating char
         resultUUID_String[k] = '\0';
    } else if (displayMode == 1) {
        r = 0 ; // pass through, UUID already done
    } else if (displayMode == 2) {
        // pass through as a hyphenated couplet form
        memset(resultUUID_String,'\0',37);
        strncpy(resultUUID_String,uuidProxy,32);
    } else {
        
    }

   return(resultUUID_String);
} // end::frm_utilGenerateUUID()


// ---------------------------------------------------------------------
// function   : frm_utilGetISO8601_datetime()
// inputs     : time_t usersSec -- optional, 0, or user provided seconds =time(0L)
//            : char *loc_dtStr_p -- resulting localtime string YYYY-MM-DDThh:mm:ss
//              char *utc_dtStr_p -- resulting UTC string YYYY-MM-DDThh:mm:ssZ
// returns    : int, timezone difference in hours between UTC and local time (not adjusted for sign)
// purpose    : Retrieve an ISO8601 conformal datetime string, YYYY-MM-DDThh:mm:ss.000Z for UTC
//              as well as a simplified local time zone ISO8601 compliant datetime stamp
// invocation : frm_utilGetISO8601_datetime(secNow,localDatetimeString,utcDateTimeString)
// revised    : 2014-09-12T11:13:00, added ".000Z' qualifier for UTC form returned
// history    : 2014-05-13T18:02:00,jmg

// notes      : As of 2014-09-12, the UTC form returned includes a dummy
//              sub-second interval qualifier before the 'Z'
int frm_utilGetISO8601_datetime(time_t usersSec,char *loc_dtStr_p,char *utc_dtStr_p)
{
 char fmtSpecStr[64]={0} ;
 char tzBuf1[16] = {0} ;
 char tzBuf2[16] = {0} ;
 char fxName[256] = {0};
  
 time_t dtSecUsed ;
 struct tm *loc_tmPtr = NULL ; 
 struct tm  *gm_tmPtr = NULL ;
 int    tz_offset_hrdiff = 0;
 
    strcpy(fxName,__func__);
    
    dtSecUsed = (usersSec <= 0)? time(0L): usersSec;
    
    // capture and translate BOTH a local time and UTC {GMT}
    gm_tmPtr  = gmtime(&dtSecUsed);
    // <JMG> 2014-09-12, now returns UTC in NASA SMAP preferred format
    // that includes a dummy sub-second interval qualifier prior to the 'Z'
    strcpy(fmtSpecStr,"%Y-%2m-%2dT%H:%M:%S.000Z" );
    strftime(utc_dtStr_p, 64-1,fmtSpecStr,gm_tmPtr);
    strftime(tzBuf1,15,"%H", gm_tmPtr);
    // printf("\ngetISO8601datetime UTC [%s] tzBuf %s\n",utc_dtStr_p,tzBuf1);

    // Next, express datetime as a ISO 8601 local time as well 
    // the sub-second qualifier is NOT included in the local form returned
    loc_tmPtr = localtime(&dtSecUsed);  
    strcpy(fmtSpecStr,"%Y-%2m-%2dT%H:%M:%S" );
    strftime(loc_dtStr_p,64-1,fmtSpecStr,loc_tmPtr);
    strftime(tzBuf2,15,"%H", loc_tmPtr);
    // fprintf(stdout,"\nfrm_utilGetISO8601datetime MST [%s] tzBuf %s\n", loc_dtStr_p, tzBuf2);
    
    tz_offset_hrdiff = abs(atoi(tzBuf2) - atoi(tzBuf1))+1;

 // as a convenience, returns absolute value of the timezone difference in integer hours 
 // for this datetime, between local timezone and GMT (UTC) timezone
 return tz_offset_hrdiff ;
 
} // end:: frm_utilGetISO8601_datetime()


// ---------------------------------------------------------------------
// function: frm_utilCalcFrequencies()
// inputs  : char *taskLabel_p,uint8 *sourceImg_p,const int32 nElem,FILE *destf
// returns : int, population count.
// purpose : given an uint8 image of nElem pixels, generate abs frequency
//           table and output it to open file, destf.
// limitations : currently only works with uint8 images; could be extended using TIV
//           to a much wider set of image types.
// notes   : n/a
int frm_utilCalcFrequencies(char *taskLabel_p,uint8 *sourceImg_p,const int32 nElem,FILE *destf)
{
    int32 p =0; 
    // nElem usually is EGv2_GLOBAL_1KM_NS_PIXELS * EGv2_GLOBAL_1KM_EW_PIXELS;
    uint16 *frqVar_p= NULL;
    float64 vPix=0.0,vSum =0.0, vMean=0.0,vMin=0.0,vMax=0.0, vN =0.0;
    
    // Note: could be made to work with other integer datatypes, up to uint16...
    frqVar_p = (uint16 *)calloc(65536,sizeof(uint8));
    
    // construct frequency table (absolute values) of fpar
    for(p=0; p < nElem; p++)
    {
        vPix = (float64)sourceImg_p[p];
        // this is intentionally hard-wired for now since we 
        // want to filter out either OUT_DOMAIN or FILL_VALUE..
        if( vPix < 254.0 )
        {
            vN += 1.0 ;
            vSum += vPix ;
            if(p==0)
            {
                    vMin = vMax = vPix;
            } else {
                    if(vPix < vMin) vMin = vPix;
                    if(vPix > vMax) vMax = vPix;
            }
        } // end:: block if pixel is NOT a fill value or out of domain
        frqVar_p[sourceImg_p[p]]++ ;
    } // end:: p loop through all pixels of image
    
    // calculate simple mean
    vMean = vSum / vN ;
    
    fprintf(destf,"\n@A @D Diagnostic check of [%s] image frequencies~\n",taskLabel_p); 

    for(p=0; p < 256; p++)
        fprintf(destf,"@A @D bin %5.5d freq %d\n",p+1,(int)frqVar_p[p]);

    fprintf(destf,"\n@A @D-DIAG [%s] N(%g) min(%.3lf) max(%.3lf) mean(%.3lf)~\n",
            taskLabel_p, vN, vMin,vMax,vMean);  
    fprintf(destf,"\n@A Diagnostic task (%s) complete~\n",taskLabel_p); 
    
    // release memory
    free(frqVar_p); 
    
    // return count
    return (int)vN ;
    
} // end::frm_utilCalcFrequencies()


// ---------------------------------------------------------------------
// function  : frm_utilCalcUpdates()
// purpose   : Update sums and intermediates for univariate statistics
// inputs    : void
// returns   : double, mean
// author    : joe glassy, Lupine Logic Inc.
// revised   : 2014-08-13:04:00, added
// limitations n/a
// notes     : n/a
size_t frm_utilCalcUpdates(FrmApiUniObj *uni_p,const double value)
{

     if(uni_p->ulim_init == 0)
     {   
       uni_p->ulim_min = uni_p->ulim_max = value ;
       // set initialization flag true 
       uni_p->ulim_init     = 1;
     } else {
       // adjust objects internal minimum,maximum fields 
        if(value < uni_p->ulim_min)
               uni_p->ulim_min = value ;
        if(value > uni_p->ulim_max)
               uni_p->ulim_max = value ;
     }
     
    uni_p->sum      += value ;
    uni_p->sum_sq   += (value * value);
    // increment integer count of times this object updated 
    uni_p->n_samples++ ;
     
    return uni_p->n_samples ;
    
} // end:: frm_utilCalcUpdates()


// ---------------------------------------------------------------------
// function  : frm_utilCalcUnivariates()
// purpose   : Calculate univariate statistics on a dataset
// inputs    : void
// returns   : double, mean
// author    : joe glassy, Lupine Logic Inc.
// revised   : 2014-08-13:04:00, added
// limitations n/a
// notes     : n/a
double frm_utilCalcUnivariates(FrmApiUniObj *uni_p)
{
 double n = uni_p->n_samples ;
 
   if(n <= 0.0)
   {
    // adjust population size, to avoid div by zero error 
    n += 0.00000001 ; 
    // increment count of univ. adjustments required 
    n_uni_errors++ ;
   }

   uni_p->range    = uni_p->ulim_max - uni_p->ulim_min ;
   uni_p->avg      = uni_p->sum / n ;
   uni_p->stddev   = sqrt( (uni_p->sum_sq-(uni_p->sum * uni_p->sum)/n)/(n-1.0));
   uni_p->std_err  = uni_p->stddev / sqrt(n) ;

   uni_p->pctserr = 0.0 ;
   uni_p->coef_var= 0.0 ;

   if(uni_p->avg > 0.0)
   {
    uni_p->pctserr  = (uni_p->std_err / uni_p->avg) * 100.0 ;
    uni_p->coef_var = uni_p->stddev / uni_p->avg ; 
   } 
    
  return uni_p->avg ;
    
} // end:: frm_utilCalcUnivariates()


// --------------------------------------------------------------------
// function: frm_utilCalc_std_dev_and_var()
// purpose : Given an array's sum,sum-of-squares,and count, calculate
//           parametric statistic's: variance and standard-deviation,
//           protecting from pathological cases where n <= 0.0.

// inputs  : const size_t vCount,
//           const double vSum,
//           const double vSumSquares,
//           double *vVariance_p,
//           int *errStatus, ptr to int, returns 1 if error occurred, 0=OK
//
// returns : sample standard deviation, and ptr to variance
// revised : 2014-11-05T16:40:00
// author  : Joe Glassy, Lupine Logic Inc.
//
// notes   : TBD: In some cases where invalid n=0 is passed in, it may be
//           acceptable to substitute a small positive value
//           e.g. 0.0000001 for the faulty n, if avoiding a NaN/Inf 
//           result is deemed absolutely critical for robustness , 
//           though we need to verify any implication of this for
//           downstream round-off or other numeric errors.
//
// notes   : Checking errStatus value afterwards allows caller to 
//           identify and count any suspect or invalid calculations.
// 
double frm_utilCalc_std_dev_and_var(const size_t vCount,const double vSum,
        const double vSumSquares,double *vVariance_p,int *errStatus)
{
 double n = (double)vCount;
 double vStdDev = 0.0 ;
 double vStdErr = 0.0 ;
 double vPctStdErr = 0.0 ;
 double vVariance = 0.0 ;
 double vCoefVar  = 0.0;
 double vMean = 0.0 ;
 static size_t nErrCalcStdDev =0;

   
   if(n <= 0.0)
   {
       nErrCalcStdDev++ ;
       *errStatus = 1;
       vStdDev      = 0.0 ;
       *vVariance_p = 0.0;
       return vStdDev ;
   } // end:: block if invalid 0 count
   
   // Calculate std dev via classic method for now
   vVariance    = (vSumSquares-( vSum * vSum )/n)/(n-1.0);
   vStdDev      = sqrt(vVariance);
   vStdErr      = vStdDev / sqrt(n);
   vPctStdErr   = (vStdErr / vMean) * 100.0 ;
   vCoefVar     = vStdDev / vMean ;
   
   *vVariance_p =  vVariance;
   *errStatus   = 0;
   
    // return unbiased, corrected sample standard deviation 
    return vStdDev ;
    
} // end:: frm_utilCalc_std_dev_and_var()


// ---------------------------------------------------------------------
// function : frm_utilSummarizePixels(GenericMetrics *postMetrics_p,float32 *sciAry_p)
// inputs   : GenericMetrics *postMetrics_p, -- ptr to obj where min,max already
//            loaded, etc.
//          const char *dsetName_p -- label only, of dataset
//          const hid_t hdf5Id, -- open HDF5 loc_id, or if 0, is ignored
//          const int doWriteMetrics, --boolean, if true, writes HDF5 attributes
//          const size_t pixelCount -- size of sciAry_p to analyze
//          float32 *sciAry_p --- science data variable
//          
// returns  : size_t nQualified (non-fill) pixels
// revision : 2014-09-12T15:53:00, added to UM Framework
// author   : joe glassy, Lupine Logic Inc.
// purpose  : Optionally rescales output array, and, derives and returns
//            a set of very basic univariable statistics: observed min,max,
//            range,mean, std deviation, coefficient of variation etc.
//             for non-fillvalue pixels. Added to help supply granule level
//            ISO metadata terms.
// notes    : Currently, any 'fillvalue' MUST lie OUTSIDE of 
//            the qualification range (min,max).
// notes    : When 'doWriteMetrics' is false, the output is still written
//            to an open log file. When true, summaries are also written
//            as simple HDF5 attributes to hid_t hdf5Id assumed active.
// notes    : Currently limited to support for float32 arrays only, will
//            relax this constraint later with TIV style semantics.
float32 frm_utilSummarizePixels(GenericMetrics *genericMetrics_p,
        const char *dsetName_p,const hid_t hdf5Id,
        const int doWriteMetrics,const size_t pixelCount,float32 *sciAry_p)
{
 // SesCtl *sesCtl_p = l4c_GetSesCtl();
 
 char   fxName[256] = {0} ;
 char   finalMsg[256] = {0};
 char   timeBuf[64] = {0};
 herr_t ret = -1;
 size_t  ithPix = 0;
 size_t  nPixels = pixelCount ;
 size_t  nUnqualified = 0 ; // count of dis-qualified pixels
 float32 sciPix   = 0.0 ; // pixel DN for post-proc transform
 float32 minAllowValue = 0.0 ; // later allow to parameterize
 float32 maxAllowValue = 100.0;// later allow to parameterize
 float64 sum     = 0.0 ;  
 float64 sumCount= 0.0 ;
 float64 sumSq   = 0.0 ;
 float64 stdError= 0.0   ; // standard error
 float64 pctStdError=0.0 ; // percent standard error
 float64 coef_var =0.0 ; // coefficient of variation 
 
 int     doesQualify = 0 ; // boolean true when within valid_min,valid_max
 // boolean true when valid_min,valid_max have been established
 int     userValidRangeAvailable  = 0; 
 int     nElem = 0;
 
 time_t calcStrtSec = time(0);
 time_t calcEndiSec = time(0);
 double calcElapSec = calcEndiSec - calcStrtSec;
 
    strcpy(fxName,__func__);
    
    // always initialize back at 0
    genericMetrics_p->obs_CountQualified =0;
    // user must have pre-established these limits
    minAllowValue = genericMetrics_p->minAllow;
    maxAllowValue = genericMetrics_p->maxAllow;

    // NOTE: native value scaling is {0.0 <= pixel <= 1.0} but PSD requires
    // value scaling of {0.0 <= pixel <= 100.0} as a percentage, so transform these
    // by multiplying each (non-fillvalue) pixel by 100
    for(ithPix=0; ithPix < nPixels; ithPix++)
    {
        sciPix = sciAry_p[ithPix];
        
        // constrain to a sensible sub-range as available
        doesQualify = (sciPix >= genericMetrics_p->minAllow && genericMetrics_p->maxAllow);
        
        if( userValidRangeAvailable && doesQualify )
        {   
            // if supplied, multiply by scale factor, like 100.0
            if(genericMetrics_p->optionScaleFactor > 0.0)
                sciPix *= genericMetrics_p->optionScaleFactor;
                
            // accumulate basic granule level metrics
            if(genericMetrics_p->obs_CountQualified == 0)
            {
                // initialize observed min,max
                genericMetrics_p->obs_min = genericMetrics_p->obs_max = sciPix;
            } else {
                // conditionally update observed min,max
                if(sciPix < genericMetrics_p->obs_min) genericMetrics_p->obs_min = sciPix;
                if(sciPix > genericMetrics_p->obs_max) genericMetrics_p->obs_max = sciPix;
            }
            // accumulate sums, sum-of-squares
            sum += (float64)sciPix ; 
            sumSq += (float64)(sciPix * sciPix);
            // re-assign with transformed value
            sciAry_p[ithPix] = sciPix;
            // increment count of qualified pixels
            genericMetrics_p->obs_CountQualified++ ;
        } else {
            nUnqualified++ ;
        }
    } // end:: ithPix loop to scale up to range {0.0<=pixel<=100.0}

    if( genericMetrics_p->obs_CountQualified > 0)
    {
        // derive basic univariate summary statisics
        genericMetrics_p->obs_range = (genericMetrics_p->obs_max -genericMetrics_p->obs_min);
        genericMetrics_p->obs_mean  = (float32)sum / (float32)genericMetrics_p->obs_CountQualified;
        sumCount =(float64)genericMetrics_p->obs_CountQualified ;
        // Done in float64, but demoted to float32 after all calcs
        // uni_p->stddev   = sqrt( (uni_p->sum_sq-(uni_p->sum * uni_p->sum)/n)/(n-1.0));
        genericMetrics_p->obs_stddev= (float32)sqrt((sumSq-(sum * sum)/sumCount)/(sumCount-1.0));

        // unpublished statistics
        if(sumCount >0)
        {
            stdError    = genericMetrics_p->obs_stddev / sqrt(sumCount);
            if(genericMetrics_p->obs_mean > 0.0)
            {
                pctStdError = (stdError / genericMetrics_p->obs_mean)* 100.0;
                coef_var    = genericMetrics_p->obs_stddev / genericMetrics_p->obs_mean;
                // added later
                genericMetrics_p->obs_stderror = stdError; // observed percent std error (non-fillvalue)
                genericMetrics_p->obs_pctstderror = pctStdError; // observed percent std error (non-fillvalue)
                genericMetrics_p->obs_coef_var    = coef_var; // observed percent std error (non-fillvalue)
            }
            xLogMsg("@A In(%s) %24s Observed minimum        %.4f ~ \n",fxName,dsetName_p,genericMetrics_p->obs_min);
            xLogMsg("@A In(%s) %24s Observed maximum        %.4f ~ \n",fxName,dsetName_p,genericMetrics_p->obs_max);
            xLogMsg("@A In(%s) %24s Observed range          %.4f ~ \n",fxName,dsetName_p,genericMetrics_p->obs_range);
            xLogMsg("@A In(%s) %24s Observed mean           %.4f ~ \n",fxName,dsetName_p,genericMetrics_p->obs_mean);
            xLogMsg("@A In(%s) %24s Observed std_deviation  %.4f ~ \n",fxName,dsetName_p,genericMetrics_p->obs_stddev);
            xLogMsg("@A In(%s) %24s Observed coef. of var   %.4f ~ \n",fxName,dsetName_p,coef_var);
            xLogMsg("@A In(%s) %24s Observed std_error      %.4f ~ \n",fxName,dsetName_p,stdError);
            xLogMsg("@A In(%s) %24s Observed pct_std_error  %.4f ~ \n",fxName,dsetName_p,pctStdError);
            xLogMsg("@A In(%s) %24s served N. Unqualified %zu  ~ \n",fxName,dsetName_p,nUnqualified);
        } // end:: sumCount is positive...

        // write out as dataset level attributes if a valid dataset_id is supplied
        // attributes: observed_count,observed_min,observed_max,observed_mean
        
        //int frm_hdf5_WriteAttributeAsType(hid_t loc_id,const char *grpQualifier,
        // const char *attrName_p,const int frmApiDatatype,const int nElem,
        // void *attrData_p)

        
        // for now suppress writing out to HDF5 as metrics
        if(hdf5Id >0  && doWriteMetrics)
        {
            // TBD work-around, until the nElem datatype 
            // in frm_hdf5WriteAttributeAsType() is changed to size_t as it should be!
            nElem = (int)genericMetrics_p->obs_CountQualified;
            
            ret = frm_hdf5_WriteAttributeAsType(hdf5Id, ".", "observed_count" ,IsFloat32,nElem,&sciPix);
            sciPix = (float32)genericMetrics_p->obs_min;
            ret = frm_hdf5_WriteAttributeAsType(hdf5Id, ".", "observed_min" ,IsFloat32,nElem,&sciPix);
            sciPix = (float32)genericMetrics_p->obs_max;
            ret = frm_hdf5_WriteAttributeAsType(hdf5Id, ".", "observed_max" ,IsFloat32,nElem,&sciPix);
            sciPix = (float32)genericMetrics_p->obs_range;
            ret = frm_hdf5_WriteAttributeAsType(hdf5Id, ".", "observed_range" ,IsFloat32,nElem,&sciPix);
            sciPix = (float32)genericMetrics_p->obs_mean;
            ret = frm_hdf5_WriteAttributeAsType(hdf5Id, ".", "observed_mean" ,IsFloat32,nElem,&sciPix);         
            sciPix = (float32)genericMetrics_p->obs_stddev;
            ret = frm_hdf5_WriteAttributeAsType(hdf5Id, ".", "observed_std_deviation" ,IsFloat32,nElem,&sciPix);
        }

    } // end:: CountQualified block tests true...
    
    calcEndiSec = time(0);
    calcElapSec = calcEndiSec - calcStrtSec;
    // char *frm_utilGetCtime(char *timBuf,const time_t usersTime)
    sprintf(finalMsg,"@A In(%s) for %s %zu qualified %zu disqualified started %s ended %s requiring %.2f sec",
        fxName, dsetName_p, 
        genericMetrics_p->obs_CountQualified, nUnqualified,
        frm_utilGetCtime(timeBuf,calcStrtSec),
        frm_utilGetCtime(timeBuf,calcEndiSec),
        calcElapSec);
        
    xLogMsg("@A %s ~\n",finalMsg);

 // return count of qualified pixels
 return genericMetrics_p->obs_CountQualified ;
 
} // end:: frm_utilSummarizePixels()



// ---------------------------------------------------------------------
// function  : frm_utilExecOsCmd()
// purpose   : Execute an OS command in a private shell, using
//             custom user defined contents, up to 4kb in size.
// inputs    : char *fmtSpecStr_p -- user defined sprintf style spec string
// returns   : int , os cmd status returned
// author    : joe glassy, Lupine Logic Inc.
// revised   : 2014-06-13T16:27:00, added orig.
// limitations and assumptions: must restrict to 4Kb info
// notes     : n/a
int frm_utilExecOsCmd(char *fmtSpecStr_p,...)
{
 char fxName[256] = {0} ;
 
  char *userMessage= NULL; // POSIX...
  va_list argptr ;   // ptr to list of var args
  int     fmt_count =0;
  int osCmdStatus = -1 ;
  
    strcpy(fxName,(char *)__func__);
    
    va_start(argptr,fmtSpecStr_p);

    // construct a local buffer to build 
    // OS command line info into
    userMessage = (char *)calloc(4096,sizeof(char));
    // format it according to users needs
    fmt_count= vsprintf(userMessage,fmtSpecStr_p, argptr);
    va_end(argptr);
    // submit the command to the OS
    osCmdStatus = system(userMessage) ;
    // free local 4Kb buffer 
    free(userMessage);
    // completed...
    return osCmdStatus ;
   
} // end::frm_utilExecOsCmd()


// ---------------------------------------------------------------------
// function: frm_utilDiagImage()
// purpose : accept a raw binary (2D) numeric array of known TIV type and 
//           generate a simple diagnostic image (pgm if uint8, or
//           jpeg if float32) using a python support utility that
//           calls numpy,scipy to produce the actual image)
//
// inputs  : char *imgPrefix, 
//           char *outputPath,
//           int tivDatatype,
//           size_t nPixels,
//           int *dims,
//           void *imgData_p
// revised      : 2014-10-21T10:23:00, adapted from l4c_DiagImage() in L4CMDL
// returns      : status 0
// dependencies : HDF4 libs for DFKNTsize(),and genDiagImg.py python script in current directory
//
int frm_utilDiagImage(char *imgPrefix, char *outputPath,int tivDatatype,size_t nPixels, size_t *dims,void *imgData_p)
{
 char osCmd[FILENAME_MAX] = {0};    
 char inpAryPath[FILENAME_MAX] = {0} ;
 char tivTypeStr[32]      = {0} ;
 char fxName[256] = {0} ;
 char pyDiagPath[FILENAME_MAX] = {0} ; // full unambiguous path to genDiagImg.py script
 char utilDirPath[FILENAME_MAX] = {0} ; // prefix directory where genDiagImg.py located 
 
 // size_t nImgBytes  = DFKNTsize(tivDatatype) * nPixels;
 uint8 *imgBuf_p = (uint8 *)imgData_p;
 int status = -1 ;
 int sizOfOne = 0 ;
 
 strcpy(fxName,__func__);
 
 // first, verify that genDiagImg.py exists
 strcpy(utilDirPath,"."); // default for now...
 sprintf(pyDiagPath,"%s/%s",utilDirPath,"genDiagImg.py");
 
 if(!frm_ioDoesFileExist(pyDiagPath))
 {
    xLogMsg("@E In(%s) utilDirPath(%s). DiagnosticImage capability requires python support utility [%s] ~\n",
        fxName,  utilDirPath,pyDiagPath);
    return -1 ;
 }
 
 // convert numeric TIV (HDF4) datatype identifier to its string equivalent
 // PRIOR: strcpy(tivTypeStr, l4c_TIVDatatypeToDatatypeString(tivDatatype,typeBuffer));
 // NOW USING FRM API's tiv_NameofTivType()
 //    char *tiv_NameofTivType(const int tivType,int *sizOfOne)
 strcpy(tivTypeStr, tiv_NameofTivType(tivDatatype,&sizOfOne));
 
 
 // write out the raw numeric array as a file, "{imgLabel}.{datatype}"
 sprintf(inpAryPath,"%s.%s",imgPrefix,tivTypeStr);
 frm_ioWritebinary(inpAryPath,tivDatatype, nPixels, (uint8*)imgBuf_p);
 
 // the genDiagImg.py utility generates an image as well as displaying it via gthumb viewer
 // or (eventually) a user defined image viewer
 //  Invocation EXAMPLE:
 //  python ../../l4c_DAILY/src/genDiagImg.py float32 1624 3856 smrz_mx.bin smrz_TEST.png 
 
 // convert the numeric TIV datatype to its string representation
 // like "float32" etc
 sprintf(osCmd,"python %s %s %zu %zu %s %s \n",
        pyDiagPath, tivTypeStr, dims[0], dims[1], inpAryPath, outputPath);
 xLogMsg("@A In(%s) osCmd[%s] ~\n",fxName,osCmd);

 system(osCmd); 

 return status ;
} // end:: frm_utilDiagImage()


// ---------------------------------------------------------------------
// function: frm_utilGenFreqDistrib()
// purpose : Accepts a raw binary (2D) numeric array of known TIV type and 
//           generate a simple diagnostic frequency distribution and 
//           (when available) produces an image of it. Uses numpy and
//           scipy.
// revised : 2014-10-21T10:29:00, adapted from l4c_GenFreqDistrib() in
//           L4CMDL l4c_DAILY_sci.c module
// inputs       : char *imgPrefix, char *outputPath,int tivDatatype,
//              size_t nPixels, void *imgArray_p
// returns      : int, status
// dependencies : Python with numpy, matplotlib, later scipy as well
int frm_utilGenFreqDistrib(char *imgPrefix, char *outputPath,int tivDatatype,
            size_t nPixels, void *imgArray_p)
{
 char osCmd[FILENAME_MAX] = {0};    
 char inpAryPath[FILENAME_MAX] = {0} ;
 char tivTypeStr[32]      = {0} ;
 char fxName[256] = {0} ;
 size_t nImgBytes   = DFKNTsize(tivDatatype) * nPixels;
 uint8 *imgBuf_p = (uint8 *)imgArray_p;
 int status = -1 ;
 int sizOfOne = 0 ; // required by tiv_NameOfTivType() conversion
 FILE *outf = NULL;
 
 strcpy(fxName,__func__);
 
 // first, verify that genDiagImg.py exists
 if(!frm_ioDoesFileExist("genDiagImg.py"))
 {
    xLogMsg("@E In(%s) requires python utility [%s] ~\n",
        fxName,"genDiagImg.py" );    
    return -1 ;
 }
 
 // Convert numeric TIV (HDF4) datatype identifier to its string equivalent
 // DEPRECATED: strcpy(tivTypeStr,l4c_TIVDatatypeToDatatypeString(tivDatatype,typeBuffer));
 strcpy(tivTypeStr, tiv_NameofTivType(tivDatatype,&sizOfOne));
 
 // Write out the raw numeric array as a file, "{imgLabel}.{datatype}"
 sprintf(inpAryPath,"%s.%s",imgPrefix,tivTypeStr);
 outf=fopen(inpAryPath,"wb");
 status = fwrite(imgBuf_p,sizeof(uint8),nImgBytes,outf);
 fclose(outf);
 
 // the genFreqDistrib.py utility uses numpy to generate a 
 // frequency distribution.
 // python ../../l4c_DAILY/src/genFreqDistrib.py float32 smrz_mx_1624x3856.flt32 smrz_TEST.png 
 
 // convert the numeric TIV datatype to its string representation
 // like "float32" etc
 sprintf(osCmd,"python genFreqDistrib.py %s %s %s\n",
        tivTypeStr, inpAryPath, outputPath);
 xLogMsg("@A In(%s) osCmd[%s] ~\n",fxName,osCmd);

 system(osCmd); 

 return 0 ;
} // end:: frm_utilGenFreqDistrib()
