// ---------------------------------------------------------------
// module       : nsb_parsers.c
// purpose      : Implements the 2-stage parser for the replicable (thread-safe)
//                NVP "string-block" instances
// project      : Lupine Logic Inc, generalized name-value-pair stringblock package
// dependencies : n/a
// revision     : 2017-04-27T17:04:00 added frm_StrTrimWhitespaceToBuffer() calls to
//                nsbParseIntoStringBlock, to add robustness, in case key
//                or value string still has leading or trailing whitespace.
// revision     : 2014-09-11T15:36, changed nsbParseIntoStringblock singleMultiValue
//                behavior to concatenate single-line multiple sub-values as now separated
//                by a back-tick "`" symbol, which allows the caller to identify these
//                compound fields later and re-translate them back to strings containing
//                embedded blanks using a call using frm_strReplaceChr() like this:
//                  frm_strReplaceChr("`",valueStr," ") ..which restores the blanks back into the string.
// history      : <JMG> TBD: 2014-08-19, verify when delimiter is "\t=" if embedded blanks
//                are allowed and supported as they should be!
// history      : 2014-01-06T10:30:00,jmg --migrated to use xxLogMsg() and mem_obj.[ch]
// history      : 2014-01-03T13:32:00,jmg --refactored API to provide for 
//                multiple, concurrent, thread-safe of a NVP stringblock instances.
// history      : 2013-12-20T12:22:00, pending, eliminated all SLL dependencies
// history      : 2013-12-20T02:00:00,milestoned version
// history      : 2013-11-19T12:33:00-07:00, original code. 
// author       : joe glassy

// notes        : WORKFLOW to use the Name-value-pair (NVP) stringblock mechanism:
//                nsbCreateInstance()
//                  nsbSetSupportedKeyDictionary()
//                     nsbParseIntoStringblock()..
//                        ...calls nsbScanRip() ...which calls nsbInitStringblock()
//                  use it: "get"|"set"
//                 nsbReleaseInstance()
// ---------------------------------------------------------------

#include "nsb_api.h"
#include "nsb_api_prot.h"

// ---------------------------------------------------------------
// Function: nsbScanSchemaRip()

// Inputs  : char *ancillaryRipPath_p,
//           int *keyMaxwidth -- prior max key width, to update if needed
//           int *valMaxwidth -- prior max value string width, to update
//
// Returns : int, n. of  additional NSB records to size the Stringblock 
//           database to accomodate.
// Revision: 2014-02-12T23:03:00-07:00
// History : n/a
// Purpose : Passively scan an additional system level @schema(), @instance(0
//           RIP file, in order to size a Stringblock to additionally 
//           accomodate these NVP.s orginating outside a RIP file.
//
// Notes   : In NSB, a schema RIP contains special NSB syntax statements
//           explicitly declaring the internal structure of a NSB 
//           stringblock NVP database, as well as optionally pre-declaring 
//           a set of NVP variable properties such as 
//           (datatype,units,fillvalue,defaultValue,minValid,maxValid,etc)
//           using AMP language syntax, @instance(referent) property=referentValue
// Notes   : nsbScanSchemaRip() is called by: nsbScanRip()
//
int nsbScanSchemaRip(char *ancillaryRipPath_p,int *keyMaxwidth,int *valMaxwidth)
{
    
  FILE *ripf ;// schema file RIP 
    
 char lineBuffer[NSB_STRBLK_RIPLINE_MAXWIDTH]   = {0} ;
 char keywordToken[NSB_STRBLK_KEYWORD_MAXWIDTH] = {0} ;
 char prevKeyword[NSB_STRBLK_KEYWORD_MAXWIDTH]  = {0} ;
 char valueToken[NSB_STRBLK_RIPLINE_MAXWIDTH]   = {0} ;
 char fxName[256] = {0} ;
 char *ch_p = NULL   ; // qualified token pointer
 char *brkb_p = NULL ; // pointer required for strtok_r thread safety
 char tokenSepChars[8] = "=\t " ; // allowable token delimiters: blank,tab,space,'=' symbol

 int  nLines =0 ;
 int  hitKeyword = 0 ;
 int  ithColumn  = 0 ;

 int  nMultivalued =0 ;
 int  isSingleLineMulti = 0 ;
 int  keyLen =0, valLen =0;
 int  ithNvp = 0 ;
 
 // int  local widths...
 int  keywidth = 0   ; // tracks longest keyword string encountered
 int  valwidth= 0    ; // tracks longest value string encountered
 size_t nvpCount = 0 ; // tracks count of parsed and registered NVPs

    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 

    strcpy(prevKeyword,"");
    hitKeyword = 0 ;
    
    // Open the schema RIP file if present. If failed, just exit
    if((ripf=fopen(ancillaryRipPath_p,"rt"))==NULL)
    {
        xLogMsg("@W nvpScanSchemaRip: leaving prior key+value widths intact\n");
        
        xLogMsg("@W nvpScanSchemaRip: NO schema or ancillary RIP file(%s) found\n",
                ancillaryRipPath_p);
        // return a count of 0 extra records to size it for
        return 0 ;
    } // end:: fopen block on ancillary, schema RIP file
    

    // Parse RIP file
    while(fgets(lineBuffer, NSB_STRBLK_RIPLINE_MAXWIDTH-1, ripf)!= NULL)
    {
        lineBuffer[strlen(lineBuffer)-1]  = '\0'; // clip newline
        nLines++; //  Increment total count of parsed lines
        // RESET classifier with each new line
        isSingleLineMulti = 0 ;

        if(lineBuffer[0] != '#')  // if NOT a comment line, proceed
        {
            ithColumn  = 0 ;
            //  Spin through the delimited tokens on this RIP file line
            for(ch_p = strtok_r(lineBuffer,tokenSepChars,&brkb_p);
                ch_p != NULL;
                ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
            {
                if(ithColumn == 0)
                {
                    strcpy(keywordToken,ch_p);
                    // fprintf(stdout,"@A nvpScanRip NEW-keyword[%s]\n",keywordToken);
                    
                } else if(ithColumn >= 1) {
                    if(ithColumn == 1)
                    {
                        strcpy(valueToken,ch_p);
                        ithNvp++ ; // count NVP elements defined
                        // fprintf(stdout,"@A nvpScanRip %4.4d NEW-VALUE[%s]\n",ithNvp,valueToken);
                    } else {
                        isSingleLineMulti = 1 ; 
                        // strcat(valueToken,","); ...may want to separate single-line multi-valued tokens?
                        strcat(valueToken,ch_p);
                        // fprintf(stdout,"@A nvpScanRip %4.4d MULTI-VALUE[%s]\n",ithNvp,valueToken);                       
                    }

                    // NOTE: when current keyword matches previous keyword, we 
                    //  have a line-spanned, multi-valued NVP, so in this case, do NOT increment
                    //  hitKeyword! Only increment when they DO NOT match.
                    if(strcmp(keywordToken,prevKeyword) != 0) 
                    {
                        hitKeyword++ ; // increment count of unique keywords encountered
                    } else {
                        nMultivalued++ ;
                    } // end:: block if keyword matched previous keyword
                    
                } // end:: block if ithColumn >= 1
                
                ithColumn++ ; // increment column counter
            } // end:: for ch_p tokenizing loop
            
            // track widest keyword string encountered
            keyLen = strlen(keywordToken);
            valLen = strlen(valueToken)  ;
            if(ithNvp == 1)
            {
                keywidth = keyLen;
                valwidth = valLen;
            } else {
                if(keyLen > keywidth) // update maximum keyword width?
                    keywidth = keyLen;
                if(valLen > valwidth) // update maximum value string width?
                    valwidth = valLen;
            } // end:: block adjusting max string widths
        
        } // end:: block if not a comment
        
        // push a 1-deep stack, so we can detect line-spanned,multi-valued NVPs.
        strcpy(prevKeyword,keywordToken);
    } // end:: while loop parsing RIP file
    
    // Close the ancillary or schema RIP file itself
    fclose(ripf);
    
    // POST, update the max widths of encountered keywords,values
    if(keywidth > *keyMaxwidth)
        *keyMaxwidth = keywidth;
        
    if(valwidth > *valMaxwidth )
        *valMaxwidth = valwidth;

 // update count of name-value-pairs encountered    
 nvpCount = ithNvp;
 
 // return count of extra records...
 return nvpCount;   
    
} // end:: nsbScanSchemaRip()


// ---------------------------------------------------------------
// Function: nsbScanRip
// Purpose : On an initial pass, this function passively scans the
//           runtime input parameters (RIP) text file, deriving the count
//           of all qualifying name-value-pairs (NVP) along with the
//           width of longest string encountere forkeys and values. Then 
//           we initialize the NVP stringblock's "core" stringblock memory
//           with these two controlling counts via a call to
//            "nsbInitStringblock(nsb_p, nvpCount, keyMaxwidth, valMaxwidth)"
// Inputs  : NsbObj *nsb_p -- ptr to parent Nsb object
//         : FILE *ripf -- file assoc with primary RIP to scan
//           int *nPrimaryNvps -- count of NVPs from primary RIP
// Returns : int, sum of BOTH n. primary NVPs, and any additional NVPs
//           added after parsing an optional schema RIP
// Outputs : *nPrimaryNvps -- count of JUST the primary RIP's NVPs
// History : 2014-02-12T23:09:00-07:00, added 'nPrimaryNvps' count returned
// History : 2013-12-20T13:49:00-07:00
// Notes   : nsbScanRip() is called by: nsbParseIntoStringblock()
// Notes   : It is critical to understand that after running nsbScanRip(,
//           the maximum widths for keywords and value strings encountered
//           in the text file are only stored INTERNALLY inside the 
//           nsb_p parent object.  This is why they are not (yet) 
//           explicitly returned on the call interface in the way that
//           *nPrimaryNvps is returned.

int nsbScanRip(NsbObj *nsb_p,FILE *ripf,int *nPrimaryNvps)
{
 // reminders: NSB_STRBLK_RIPLINE_MAXWIDTH is 1024
 //            NSB_STRBLK_KEYWORD_MAXWIDTH is  256
 char keywordToken[NSB_STRBLK_KEYWORD_MAXWIDTH] = {0} ;
 char prevKeyword[NSB_STRBLK_KEYWORD_MAXWIDTH]  = {0} ;
 char valueToken[NSB_STRBLK_RIPLINE_MAXWIDTH]   = {0} ;
 char lineBuffer[NSB_STRBLK_RIPLINE_MAXWIDTH]   = {0} ;

 char fxName[256] = {0} ;
 char *ch_p = NULL   ; // qualified token pointer
 char *brkb_p = NULL ; // pointer required for strtok_r thread safety
 char tokenSepChars[16] = "=\t " ; // allowable token delimiters: blank,tab,space,'=' symbol

 int  nLines =0 ;
 int  hitKeyword = 0 ;
 int  ithColumn  = 0 ;

 int  nMultivalued =0 ;
 int  nToAdd = 0 ; // n.of optional schema or ancillary records to also accomodate
 int  isSingleLineMulti = 0 ;
 int  keyLen =0, valLen =0;
 int  ithNvp = 0 ;
 
 // DEPRECATED nsbPropMaxwidth..
 // int  prpMaxwidth = nsb_p->nsbPropMaxwidth; // width of properties field
 int  keyMaxwidth = 0   ; // tracks longest keyword string encountered
 int  valMaxwidth= 0    ; // tracks longest value string encountered
 
 size_t nBytes =0;
 size_t nvpCount = 0 ; // tracks count of parsed and registered NVPs

    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 

    strcpy(prevKeyword,"");
    hitKeyword = 0 ;
    
    // Default delimiters already defined in nsbCreateInstance() BUT
    // here we allow use-case where a prior call to nsbSetDelimiters() 
    // may have overridden these
    
    // parse RIP file
    while(fgets(lineBuffer, NSB_STRBLK_RIPLINE_MAXWIDTH-1, ripf)!= NULL)
    {
        lineBuffer[strlen(lineBuffer)-1]  = '\0'; // clip newline
        nLines++; //  Increment total count of parsed lines
        
        // fprintf(stdout,"@A nvpScanRip %4.4d lineBuffer[%s]\n",nLines,lineBuffer);
        
        // RESET classifier with each new line
        isSingleLineMulti = 0 ;

        if(lineBuffer[0] != '#')  // if NOT a comment line, proceed
        {
            ithColumn  = 0 ;
            //  Spin through the delimited tokens on this RIP file line
            //  now using internally stored nsbDelimSymbols
            for(ch_p = strtok_r(lineBuffer, nsb_p->nsbDelimSymbols,&brkb_p);
                ch_p != NULL;
                ch_p = strtok_r(NULL, nsb_p->nsbDelimSymbols, &brkb_p))
            {
                if(ithColumn == 0)
                {
                    strcpy(keywordToken,ch_p);
                    // fprintf(stdout,"@A nvpScanRip NEW-keyword[%s]\n",keywordToken);
                    
                } else if(ithColumn >= 1) {
                    if(ithColumn == 1)
                    {
                        strcpy(valueToken,ch_p);
                        ithNvp++ ; // count NVP elements defined
                        // fprintf(stdout,"@A nvpScanRip %4.4d NEW-VALUE[%s]\n",ithNvp,valueToken);
                    } else {
                        isSingleLineMulti = 1 ; 
                        // strcat(valueToken,","); ...may want to separate single-line multi-valued tokens?
                        strcat(valueToken,ch_p);
                        // fprintf(stdout,"@A nvpScanRip %4.4d MULTI-VALUE[%s]\n",ithNvp,valueToken);                       
                    }

                    // NOTE: when current keyword matches previous keyword, we 
                    //  have a line-spanned, multi-valued NVP, so in this case, do NOT increment
                    //  hitKeyword! Only increment when they DO NOT match.
                    if(strcmp(keywordToken,prevKeyword) != 0) 
                    {
                        hitKeyword++ ; // increment count of unique keywords encountered
                    } else {
                        nMultivalued++ ;
                    } // end:: block if keyword matched previous keyword
                    
                } // end:: block if ithColumn >= 1
                
                ithColumn++ ; // increment column counter
            } // end:: for ch_p tokenizing loop
            
            // track widest keyword string encountered
            keyLen = strlen(keywordToken);
            valLen = strlen(valueToken)  ;
            if(ithNvp == 1)
            {
                keyMaxwidth = keyLen;
                valMaxwidth = valLen;
            } else {
                if(keyLen > keyMaxwidth)
                    keyMaxwidth = keyLen;
                if(valLen > valMaxwidth)
                    valMaxwidth = valLen;
            } // end:: block adjusting max string widths
        
        } // end:: block if not a comment
        
        // push a 1-deep stack, so we can detect line-spanned,multi-valued NVPs.
        strcpy(prevKeyword,keywordToken);
    } // end:: while loop parsing RIP file

    // Store scan of RIP results: kMaxwidth, vMaxwidth in master session controls
    // adding 1 for the null terminator of each implicit NVP record
    keyMaxwidth++ ;
    valMaxwidth++ ;
    
    // Lock in, register the keyword and valuestring widths
    // that represent the 'one-size-fits-all' record width for ALL
    // NVP records in the stringblock.
    nvpCount = ithNvp ; // just for readability 
    nsb_p->nsbActualCount = nvpCount;
    
    nsb_p->nsbKeyMaxwidth = keyMaxwidth ;
    nsb_p->nsbValMaxwidth = valMaxwidth ;
    
    // Combine memory required for all (3) subfields times the number of
    //   NVP entries parsed from the RIP file that we need to accomodate,
    //   adding 1 at the end of each record as a buffer
    nBytes = (nvpCount * (keyMaxwidth + valMaxwidth+1));

    // set n. of primary NVPs to distinguish this subset from the potential
    // total added to (below) via a schema RIP
    *nPrimaryNvps = nvpCount ;

    // Based on the count of NVPs and the maximum widths of both the
    //   set of keys and values, create a single NVP buffer (catalog) to
    //   hold NVPs are records of "back-to-back" (prop+keyword+value) strings
    
    xLogMsg("@A In(%s): %d primary NVPs will require %zu bytes (Record width= %zu)~ \n",
        fxName, nvpCount, nBytes, (keyMaxwidth+valMaxwidth));
    xLogMsg("@A In(%s): Recognized KEY|VALUE delimiters (%s) Multi-Value delimiters (%s)~ \n",
        fxName, nsb_p->nsbDelimSymbols, nsb_p->nsbDelimCompound);
    
    // Next, prior to establishing a final NSB stringblock's summed (key+val) widths
    //  we check if the user specified an auxillary initialization RIP file such
    //  as used to define a set of @schema(), @instance() and related
    //  metadata, to allow the NSB database to ALSO accomodate these additional
    //  NVPs.
    if(strlen(nsb_p->nsbSchemaPath)>0)
    {
        nToAdd = nsbScanSchemaRip(nsb_p->nsbSchemaPath,&keyMaxwidth,&valMaxwidth);
        nvpCount += nToAdd ;
        
        nBytes = (nvpCount * (keyMaxwidth + valMaxwidth+1));
        
        xLogMsg("@A nvpScanRip: after ancillary RIP %d additional NVPs registered (TotalRec: %zu, Record width= %zu)\n",
                        nToAdd,nvpCount, (keyMaxwidth+valMaxwidth));
    } else {
        xLogMsg("@A nvpScanRip: No schema or ancillary RIP scan was requested\n");
    }
    
    // Use memObjCalloc() to allocate memory for single master instance
    //   of stringlblock to manager NVPs this session. Note that this
    //   function also effectively calls nsbSetStringblockRecordCount()
    nsbInitStringblock(nsb_p, nvpCount, keyMaxwidth, valMaxwidth);
    
    // Rewind the RIP file for actual pass using nvpParseIntoCatalog()
    rewind(ripf);
    
  // Return count of NVPs required to parse
  return nvpCount;
  
} // end::nsbScanRip()


// ---------------------------------------------------------------
// Function: nsbParseIntoStringblock() 
// Inputs  : NsbObj nsb_p --ptr to nvp stringblock parent object
//         : delimStr -- ptr to string of delimiters, any of which delimits key from value
//         : FILE *ripf -- RIP file handle
// Returns : int, n. of unique NVP's parsed
// Revision: 2017-04-27T17:04:00 added frm_StrTrimWhitespaceToBuffer() calls to
//           add robustness, in case key or value string still has leading or 
//           trailing whitespace.
// Revision: 2014-01-02T13:17:00,jmg --renamed to nsbParseIntoStringblock
// History : 2013-11-01
// Purpose : Parses a RIP file, loads all supported (keys) as name-value-pairs
//           into a Stringblock
// Notes   : nsbScanRip() is called by: nsbParseIntoStringblock()
// Notes   : TBD, in future, allow user to specify the delimiter chars to parse on!!
//
int nsbParseIntoStringblock(NsbObj *nsb_p,char *delimStr_p,FILE *ripf)
{
 // reminders: NSB_STRBLK_RIPLINE_MAXWIDTH is 1024
 //            NSB_STRBLK_KEYWORD_MAXWIDTH is  256
 char keywordToken[NSB_STRBLK_KEYWORD_MAXWIDTH] = {0} ;
 char prevKeyword[NSB_STRBLK_KEYWORD_MAXWIDTH]  = {0} ;
 char valueToken[NSB_STRBLK_RIPLINE_MAXWIDTH]   = {0} ;
 char lineBuffer[NSB_STRBLK_RIPLINE_MAXWIDTH]   = {0} ;
 
 // DEFAULT set of key <delimiter> value is '=',a tab, or a space
 // Token delimiters: {blank/space,tab,'='} symbols, BUT no comma by default
 char tokenSepChars[32] = "=\t " ; 
 char fxName[256] = {0} ;
 char *ch_p   = NULL   ; // qualified token pointer
 char *brkb_p = NULL ; // pointer required for strtok_r thread safety
  
 int  nLines =0 ;
 int  hitKeyword = 0 ;
 int  ithColumn  = 0 ;
 int  ithNvp = 0 ;
 
 int  nMultivalued =0 ;
 // int  nSchemaRip =0 ; // count of schema RIP's nvps
 int  nPrimaryNvps =0 ; // count of primary RIP's nvps
 
 int  nUnsupportedKeywords =0 ;
 int  k =0;
 int  isSingleLineMulti = 0 ;
 int  keyLen =0, valLen= 0 ;
 
 size_t nvpCount = 0; // sum of BOTH primary RIP's NVPs and schema's NVPs
  
 size_t byteOffsetToRec   = 0 ;
 size_t prevOffset =0;
 size_t deltaOffset=0;

    // REGISTER fx 
    strcpy(fxName,(char *)__func__); 
    strcpy(prevKeyword,"");
    hitKeyword = 0 ;
    
    // First, store default delimiters...
    strcpy(nsb_p->nsbDelimSymbols,tokenSepChars);
    // Allow user to override the default delimiters 
    if(delimStr_p != NULL && strlen(delimStr_p)>= 1)
    {
        strcpy(tokenSepChars,delimStr_p);
        // <JMG> 2015-09-30, also now updates the
        //  long-lived internal copy of these 
        strcpy(nsb_p->nsbDelimSymbols,delimStr_p);
    }
    
    // PASS 1: perform pre-scan, create master NVP catalog, returns k, n. of NVPs.
    xLogMsg("@A nsbParseIntoStringblock: calling nvpScanRip now \n");
    nvpCount = nsbScanRip(nsb_p,ripf,&nPrimaryNvps);
    
    xLogMsg("@A nsbParseIntoStringblock: nvpScanRip returns %zu nLines\n",nvpCount);
    
    // PASS 2: perform actual parse of the RIP file, storing NVPs in catalog
    xLogMsg("@A nsbParseIntoStringblock: iterating through %zu nvps to parse\n",nPrimaryNvps);
    
    while(fgets(lineBuffer, NSB_STRBLK_RIPLINE_MAXWIDTH-1, ripf)!= NULL)
    {
        lineBuffer[strlen(lineBuffer)-1]  = '\0'; // clip newline
        nLines++; //increment total count of parsed lines
        
        // RESET classifier with each new line
        isSingleLineMulti = 0 ;
        // TBD: make more robust in future, such that the '#' could be the
        // first non-whitespace char on the line, instead of strictly beginning
        // at lineBuffer[0] here!!
        if(lineBuffer[0] != '#')  // if NOT a comment line, proceed
        {
            ithColumn  = 0 ;
            // spin through the delimited tokens on this RIP file line
            for(ch_p = strtok_r(lineBuffer,tokenSepChars,&brkb_p);
                ch_p != NULL;
                ch_p = strtok_r(NULL,tokenSepChars,&brkb_p))
            {
                if(ithColumn == 0)
                {
                    strcpy(keywordToken,ch_p);
                    // <JMG> 2017-04-27T17:00:00, added for robustness, clip off possible leading or trailing whitespace
                    // where keywordToken is the modified (destination) string
                    frm_StrTrimWhitespaceToBuffer(ch_p,keywordToken);
                    keyLen = strlen(keywordToken);
                    // xLogMsg("@A keyword %5.5d (%s)\n",ithNvp,keywordToken);
                    
                } else if(ithColumn >= 1) {
                    
                    if(ithColumn == 1)
                    {
                        // <JMG> TBD: 2014-08-19, 2014-09-11, still a pending issue -- when the
                        //  default delimiter is "\t=" is used, this works OK for 
                        //  single-valued NVPs, but, when the value string is a compound-value one,
                        //  containing embedded blanks (like a comma-delim properties string) all 'tokens'
                        //  end up packed together since the NVP is classified as a multi-valued NVP
                        //  with all sub-values on one line. Solution being investigated but is not
                        //  yet resolved.
                        
                        // <JMG> 2017-04-27T17:00:00, added for robustness, clip off possible leading or trailing whitespace
                        // where valueToken is the resulting, modified (destination) string
                        frm_StrTrimWhitespaceToBuffer(ch_p,valueToken);
                        // xLogMsg("@A Value %5.5d (%s)\n",ithNvp,valueToken);
                    } else {
                        // a "Single Line Multi" is an NVP where its valueString contains multiple
                        // sub-values instead of a single value. This still needs work to correctly support
                        // all the different use cases.
                        isSingleLineMulti = 1 ; 
                        // TBD: In the future we may want to support single-line,multi-valued NVPs
                        //    as a "delimiter-packed" (CSV) collection too
                        // strcat(valueToken,",")  ..or.. could use a distinctive entry
                        // <JMG> 2014-08-11T14:52
                        strcat(valueToken,"`"); // special back-tick delimiter is used here
                        // <JMG> 2017-04-27T17:03 here we do NOT yet use frm_StrTrimWhitespaceToBuffer() since
                        // we are iteratively concatening a built-up result...
                        strcat(valueToken,ch_p);
                        
                        // xLogMsg("@A MULTI value %5.5d (%s)\n",ithNvp,ch_p);
                    } // end:: block if a multi-valued value trapped

                    // NOTE: when current keyword matches previous keyword, we 
                    //  have a line-spanned, multi-valued NVP, so in this case, 
                    //  do NOT increment hitKeyword! Only increment when they DO NOT match.
                    if(strcmp(keywordToken,prevKeyword) != 0) 
                    {
                        hitKeyword++ ; // increment count of unique keywords encountered
                    } else {
                        nMultivalued++ ;
                    } // end:: block if keyword matched previous keyword
                    
                } // end:: block if ithColumn >= 1
                
                ithColumn++ ; // increment column counter
            } // end:: for ch_p tokenizing loop
            
            if( ithColumn >= 1) 
            {
                // Using 'keywordToken', search the fixed,built-in dictionary, 
                //   and if matched, then is classified as a supported keyword,
                //   and it is thereby added to the NSB database.
                // If keywordToken is NOT matched, we just count the n. of 
                //   unsupported keyword encounters. Remember that 'k' here is just 
                //   the base0 index of this keywords slot in pre-defined dictionary
                //   of supported (recognized) keywords.
                // <JMG> 2/16/2014, now nsbKeyPolicy is FIRST CONSULTED inside the 
                // filter: nsbSearchForSupportedKey() and if nsbKeyPolicy==0, then
                // NO CHECK is made and NO predefined table of supported keys is required.
                //  If nsbKeyPolicy==1, a table of predefined keywords MUST be present
                //  and is enforced (e.g. a key MUST be in the set of keys listed).
                //  This allows a call to nsbSearchForSupportedKey() to be
                //  called in both cases, with or without a table of predefined keywords
                //  being present where nsbKeyPolicy==1.
                
                // nsbSearchForSupportedKey() returns -1 on failure, or 0 on success.
                if((k = nsbSearchForSupportedKey(nsb_p, keywordToken)) != -1)
                {
                    // Note that 'k' is not functionally used as an index inside this
                    //   this block. It is only used to indicate PASS/FAIL that the key
                    //   submitted was a supported one. k=0 whenever nsbKeyPolicy is 0,
                    //   where NO supported keyword table is required at all.
                    valLen = strlen(valueToken);
                    
                    byteOffsetToRec = NSB_OFFSET_TO_RECD(ithNvp, nsb_p->nsbRecsize);
                    
                    // For diagnostic only, stores record-to-record change in record wise byte offsets
                    deltaOffset = abs(byteOffsetToRec - prevOffset);
                    
                    // Store both the KEYWORD + VALUE Strings at ith record of the Stringblock (database)
                    nsbStoreKeyAndValue(nsb_p, ithNvp, keywordToken, valueToken);
                    
                    // increment count of the NVP's parsed and registered, stored in the stringblock
                    ithNvp++ ; 
                    // Track previous offset to start of overall record, just for diagnostics though
                    prevOffset = byteOffsetToRec;
                    
                } else { // end:: block if k is VALID
                    nUnsupportedKeywords++ ;
                    xLogMsg("@A  nsbParseIntoStringblock: [%s] Suppressed load of (%s) unsupported keyword~\n",
                        keywordToken);
                }           
            } // end:: if ithColumn is at least 1
            
        } // end:: block if not a comment
        
        // Push a 1-deep 'cache' stack, so we can detect line-spanned,multi-valued NVPs.
        strcpy(prevKeyword, keywordToken);
        
    } // end:: while loop parsing RIP file

    // if any multiValued entried encountered, the actual count is 1 plus this.
    if(nMultivalued >0)
         nMultivalued++ ;

    xLogMsg("\n");
    xLogMsg("@A nsbParseIntoStringblock: Post-RIP: %d lines parsed containing %d keywords [%d were multi-valued]~\n",
        nLines, hitKeyword, nMultivalued );
    
    // UPDATE the count of (primary) NVP's parsed and stored
    nsb_p->nsbActualCount = ithNvp;
    
    // Return count of TOTAL number of NVP's parsed from runtime input parameter (RIP) file
    return ithNvp;
    
} // end::nsbParseIntoStringblock()
