// --------------------------------------------------------------------
// module   : pkg_mgr.c
// purpose  : prototype implementatoin of PkgMgr service functions 
// revision : 2020-07-08T19:11, modified for HDF5 v1.10.x use, using int64_t for IDs
// revision : 2014-02-21T12:03:00, added pkgMgr_GlobalReport()
// revision : 2014-02-01T14:24:00,jmg (original)
// version  : 1.0
// author   : joe glassy
// assertions : The PkgMgr must define at least (1) PkgTheme
//              A PkgTheme may define 0,1 or more PkgObj up to the limit allowed.
//              By default unless overridden, a PkgTheme creates K PkgObj empty-instances
//              All PkgThemes have a unique integer ID, >=0, and a label
//              All PkgObj have a unique integer ID >=0, and a label
//              No data per say is ever stored in a PkgObj, as this is just a registry
//              Data for a variable is created elsewhere; the PkgObj just points to it.
// --------------------------------------------------------------------

#include "pkg_mgr.h"
#include "pkg_mgr_prot.h"

// --------------------------------------------------------------------
// function : pkgMgr_GetVersion()
// purpose  : return PkgMgr sub-API version identifier
// revision : 1.0-2014-07-31T15:22:00,jmg (original)
// author   : joe glassy
// notes    : To match ability to get UM Framework and NSB api versions
char *pkgMgr_GetVersion(char *versBuffer_p)
{
    strcpy(versBuffer_p,(char *)PKGMGR_VERS_TAG);
    
    return versBuffer_p ;
    
} // end:: pkgMgr_GetVersion()

// --------------------------------------------------------------------
// function : pkgMgr_CreateMgr()
// purpose  : create (1) new PkgMgr instance
// revision : 2014-09-11T11:23, added integrity check
// revision : 2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Caller may specify nThemes=0 to create these later.
PkgMgr *pkgMgr_CreateMgr(const char *pkgmgrLabel,const int nThemes,const char *commonCategory_p)
{
    PkgMgr  *pkgMgrNew_p = NULL;
    PkgTheme *pThm_p = NULL;
    char fxName[256] = {0};
    char uuidBuf[64] = {0} ;
    char autoLabel[256] = {0} ;
    char localLbl[256];
    int ithTheme =0;
    
    strcpy(fxName, __func__);
    // Create one PkgMgr: allocate using anonymous memCalloc()
    pkgMgrNew_p = memCalloc(1, sizeof(PkgMgr));
    pkgMgrNew_p->pkMgrId = -1 ; // assign a sentinal pre-ID, to override
    
    // to aid tracing later, update low level info in memObj registry too
    strcpy(localLbl, pkgmgrLabel); // this is to protect 'const' in arg above
    memObjSetName(pkgMgrNew_p, localLbl);
    memObjSetDatatype(pkgMgrNew_p,IsCustomUserType);

    // assign a unique UUID right away
    strcpy(pkgMgrNew_p->pkMgrUUID, frm_utilGenerateUUID(uuidBuf,1));
        
    xLogMsg("@A In(%s): initial theme count %d (if 0 may be init later)~\n",
        fxName, nThemes);
    
    // if available add the common category descriptor as metadata
    if(strlen( commonCategory_p)>0)
        memObjSetMetadata(pkgMgrNew_p, commonCategory_p);

    if(strlen(pkgmgrLabel)>0)
        strcpy(pkgMgrNew_p->pkMgrLabel,pkgmgrLabel); // assign users label
        
    // Store a common category label that may apply to entire collection of themes
    strcpy(pkgMgrNew_p->pkMgrThemeCommonCategory, commonCategory_p);
    
    pkgMgrNew_p->pkMgrThemeCount = 0;
    pkgMgrNew_p->pkMgrThemeAry_p = NULL; // none defined yet...
    
    // Auto-construct a name-label for the PkgTheme collection itself
    sprintf(autoLabel,"%s.pkgThemeArrayOf_%d",pkgmgrLabel,nThemes); 
    
    if( nThemes > 0 )
    {
        pkgMgrNew_p->pkMgrThemeCount = nThemes ;
        
        // Use callers count of themes to drive allocation
        pkgMgrNew_p->pkMgrThemeAry_p = memCalloc(nThemes,sizeof(PkgTheme));
        // Provide a default name (autoLabel0 to the theme array container itself
        memObjSetName(pkgMgrNew_p->pkMgrThemeAry_p    , autoLabel);
        memObjSetDatatype(pkgMgrNew_p->pkMgrThemeAry_p, IsCustomUserType);

        // Iterate through candidate PkgThemes, minimally initializing each
        // Ok if nThemes is 0 here, as will simply not iterate...
        for(ithTheme=0; ithTheme < nThemes ; ithTheme++)
        {
            pThm_p = &pkgMgrNew_p->pkMgrThemeAry_p[ithTheme];
            
            // int pkgMgr_InitTheme(PkgTheme *pkgTheme_p,const int themeId, const int ownerId,
            //      const char *themeLbl_p,const char *themeCategory_p)
            
            // mark with: id, owner, label, categoryDescriptor
            sprintf(autoLabel,"%s.pkgTheme_%-5.5d",pkgmgrLabel, ithTheme+1);    
            
            // minimally initialize this first set of themes. 
            // A UUID is now generated for each Theme inside here.
            pkgMgr_InitTheme(pThm_p, ithTheme+1, pkgMgrNew_p->pkMgrId,
                            autoLabel, "no-category-yet");
            // NOTE: while nThemes have been created, each of the
            // themes HAS NOT YET been given any child PkgObj to own
            
        } // end:: ithTheme loop, zeroing out candidate themes
        
    } else { // end:: block if a pos count of themes was specified
        xLogMsg("@W In(%s) %d themes must be initialized later~ \n",fxName,nThemes);
    }
    
    // return created, initialized PkgMgr instance
    return pkgMgrNew_p;
    
} // end::pkgMgr_CreateMgr()


// --------------------------------------------------------------------
// function : pkgMgr_SetThemeCount()
// purpose  : assign the count of this PkgMgr's max n. of themes.
// revision : 2014-06-11T18:06:00,jmg (original)
// author   : joe glassy
// notes    : Added 6/11/2014, as this was omitted before.
void pkgMgr_SetThemeCount(PkgMgr *pkgMgr_p,const int newThemeCount)
{
    pkgMgr_p->pkMgrThemeCount = newThemeCount;
    
} // end:: pkgMgr_SetThemeCount()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeCount()
// purpose  : return count of this PkgMgr's max n. of themes.
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Some PkgThemes are defined but not yet used (e.g. they
//            are marked 'IsUnoccupied'. This occurs when preemptive
//            overshoot is intentional, or after a PkgTheme has been
//            marked for deletion, freeing up a slot for recycled use.
int pkgMgr_GetThemeCount(PkgMgr *pkgMgr_p)
{
    return pkgMgr_p->pkMgrThemeCount ;
    
} // end:: pkgMgr_GetThemeCount()


// --------------------------------------------------------------------
// function : pkgMgr_ReleaseMgr()
// purpose  : delete and free (1) new PkgMgr instance
// revision : 2014-06-11T18:52:00, improved by only releasing memory
//            assoc with the PkgMgr overseer itself. We want to require
//            the system or user to first explictly call pkgMgr_ReleaseThemes()
//            prior to this call.
// revision : 1.1-2014-02-23T13:03, fixed bug, changed free(pkgMgr_p) to
//            memFree(pkgMgr_p) to enable better life-cycle tracing
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// issues   : <JMG> found,corrected bug where nObjPkg SHOULD have been 
//            used but nThemes was mistakenly used. Now nObjPkg is used.
// notes    : As usual the order of the free()'s is from inside to outside
//            so that inner components don't get orphaned!
// notes    : TBD, planning to add a callerLbl_p ...
// --------------------------------------------------------------------
void pkgMgr_ReleaseMgr(PkgMgr *pkgMgr_p)
{
 PkgTheme *pThm_p = NULL;
 int ithTheme= 0;
 int nObjPkg =0 ;
 
 // int ithObj  = 0;
 int nThemes = pkgMgr_p->pkMgrThemeCount;
 
    if(nThemes > 0)
    {
        // fprintf(stderr,"@W pkgMgr_ReleaseMgr: releasing %d themes~\n",nThemes);
        
        // free all existing themes, regardless if they are in use or not
        for(ithTheme=0; ithTheme < nThemes; ithTheme++)
        {
            pThm_p = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
            nObjPkg = pThm_p->pkThemeObjCount ;
            
            if(nObjPkg  >0)
            {
                fprintf(stderr,
                "@A pkgMgr_ReleaseMgr: releasing Theme %5d (%s) %d PkgObj~\n",
                    ithTheme, pThm_p->pkThemeLabel, nObjPkg);

                memFree(pThm_p->pkTheme_PkgObjAry_p);
                pThm_p->pkThemeObjCount = 0 ;
                
            } else {
                fprintf(stderr,
                "@W pkgMgr_ReleaseMgr: Theme %d(%s) PkgObj count %d so cannot release PkgObj!~\n",
                ithTheme, pThm_p->pkThemeLabel, pThm_p->pkThemeObjCount);
            }
        } // end:: ithTheme loop
        
        // Next, free the fixed array of PkgTheme's
        memFree(pkgMgr_p->pkMgrThemeAry_p) ;
        pkgMgr_p->pkMgrThemeCount = 0 ;

    } else {
        fprintf(stderr,"@W pkgMgr_ReleaseMgr: theme count %d so cannot release themes!~\n",
            nThemes);
    }
    
    // Finally, free the master overseer itself
    memFree(pkgMgr_p);
    
} // end::pkgMgr_ReleaseMgr()


// --------------------------------------------------------------------
// function : pkgMgr_SetMgrCommonCategoryLabel()
// inputs   : PkgMgr *pkgMgr_p,const char *commonCategory_p
// returns  : void
// purpose  : assign common category label that may be inherited by all
//            PkgThemes..
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
void pkgMgr_SetMgrCommonCategoryLabel(PkgMgr *pkgMgr_p,const char *commonCategory_p)
{
    strcpy(pkgMgr_p->pkMgrThemeCommonCategory , commonCategory_p);
    
} // end::pkgMgr_SetMgrCommonCategoryLabel()

// --------------------------------------------------------------------
// function : pkgMgr_SetMgrCommonCategoryLabel()
// inputs   : PkgMgr *pkgMgr_p,const char *commonCategory_p
// returns  : char *, common category label string
// purpose  : Retrieves PkgMgr's common category label that may be 
//            inherited later by all PkgThemes..
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
char *pkgMgr_GetMgrCommonCategoryLabel(PkgMgr *pkgMgr_p)
{
    return pkgMgr_p->pkMgrThemeCommonCategory;
    
} // end::pkgMgr_GetMgrCommonCategoryLabel()


// --------------------------------------------------------------------
// function : pkgMgr_SetMgrName()
// inputs   : PkgMgr *pkgMgr_p,const char *newPkgMgrLabel_p
// returns  : void
// purpose  : assign name or label descriptor to a PkgMgr
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
void pkgMgr_SetMgrName(PkgMgr *pkgMgr_p,const char *newPkgMgrLabel_p)
{
    strcpy(pkgMgr_p->pkMgrLabel , newPkgMgrLabel_p);
    
} // end::pkgMgr_SetMgrName()


// --------------------------------------------------------------------
// function : pkgMgr_GetMgrName()
// inputs   : PkgMgr *pkgMgr_p
// returns  : char *, pointer to PkgMgr label or name
// purpose  : retrieve name of a PkgMgr
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
char *pkgMgr_GetMgrName(PkgMgr *pkgMgr_p)
{
    return pkgMgr_p->pkMgrLabel ;
} // end::pkgMgr_GetMgrName()


// --------------------------------------------------------------------
// function : pkgMgr_SetThemeNameByIndex()
// inputs   : PkgMgr *pkgMgr_p,const int themeIndex
// returns  : char *, pointer to theme's name
// purpose  : assigns a theme's name, as located by its base0 index 
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
void pkgMgr_SetThemeNameByIndex(PkgMgr *pkgMgr_p,const int themeIndex,char *fmtMsg_p,...)
{
 PkgTheme *pThm_p = NULL;
 int t=0;
 char pkgThemeNameString[FILENAME_MAX] = {0} ;
  
  va_list argptr;
  int     fmt_count =0;
  
    va_start(argptr,fmtMsg_p);
    fmt_count = vsprintf(pkgThemeNameString, fmtMsg_p,argptr);
    va_end(argptr);
 
    for(t=0; t < pkgMgr_p->pkMgrThemeCount; t++)
    {
        if(themeIndex == t)
        {
            pThm_p = &pkgMgr_p->pkMgrThemeAry_p[t];
            // matched, so assign this themes name now
            strcpy(pThm_p->pkThemeLabel,pkgThemeNameString);
        } // end:: block when matched t with themeIndex
    } // end:: t lop
        
} // end:: pkgMgr_SetThemeNameByIndex()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeNameByIndex()
// inputs   : PkgMgr *pkgMgr_p,const int themeIndex
// returns  : char *, pointer to theme's name
// purpose  : retrieve name of a theme, located by its base0 index 
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
char *pkgMgr_GetThemeNameByIndex(PkgMgr *pkgMgr_p,const int themeIndex)
{
 PkgTheme *pThm_p = NULL;
    
 int t=0;
 
        for(t=0; t < pkgMgr_p->pkMgrThemeCount; t++)
        {
            if(themeIndex == t)
            {
                // get pointer to ith theme
                pThm_p = &pkgMgr_p->pkMgrThemeAry_p[t];
                break;
            }
        } // end:: i loop through pkgObj's owned by this PkgTheme
        
 // returns pointer to theme's name as matched , or NULL if not found
 return pThm_p->pkThemeLabel ;
 
} // end:: pkgMgr_GetThemeNameByIndex()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeIndexByName()
// inputs   : PkgMgr *pkgMgr_p,const int themeIndex
// returns  : char *, pointer to theme's name
// purpose  : retrieve base0 index (slot) of a theme, located by its text label
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
int pkgMgr_GetThemeIndexByName(PkgMgr *pkgMgr_p,const char *themeName_p)
{
 PkgTheme *pThm_p = NULL;
    
 int t=0;
 
        for(t=0; t < pkgMgr_p->pkMgrThemeCount; t++)
        {
            // set pointer to current theme PkgMgr owns
            pThm_p = &pkgMgr_p->pkMgrThemeAry_p[t];
            
            if(strcmp(themeName_p,pThm_p->pkThemeLabel)==0)
            {
                break;
            }
        } // end:: i loop through pkgObj's owned by this PkgTheme
 // returns index to match theme
 return t ;
} // end:: pkgMgr_GetThemeIndexByName()

// --------------------------------------------------------------------
// function : pkgMgr_GetMgr_ThemesChildDataByName()
// inputs   : PkgMgr *pkgMgr_p,const char *themeLabel,const char *childLabel
// returns  : void *, ptr to a DATA from a Theme's child, or NULL if not found
// purpose  : retrieve ptr to a theme's child's data, identified by
//            both the Theme's label and the child (PkgObj) label.
// revision : // revision : 2014-02-28T11:01:00,jmg (original)
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child data
// --------------------------------------------------------------------
void *pkgMgr_GetMgr_ThemesChildDataByName(PkgMgr *pkgMgr_p,const char *themeLabel_p,
                    const char *childLabel_p)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 void *resultData_p = NULL;
 
  // first, fetch Theme's handle
  if((pThm_p = pkgMgr_GetThemeByName(pkgMgr_p,themeLabel_p))!=NULL)
  {
    // next, give this theme, fetch child object
    
    if((pObj_p = pkgMgr_GetObjByName(pThm_p, childLabel_p))!= NULL)
    {
        // finally, fetch the data itself
        resultData_p = pkgMgr_GetObjData(pObj_p);

    } else {
        xLogMsg("@E In(%s) FAILED locating child named(%s) in theme(%s)~\n",
            (char *)__func__, childLabel_p, themeLabel_p);
    }
      
  } else {
        xLogMsg("@E In(%s) FAILED locating theme(%s)~\n",
                (char *)__func__, themeLabel_p);
  }
  
 // returns pointer to PkgObj child's data owned by theme, owned by PkgMgr
 return resultData_p ;
 
} // end:: pkgMgr_GetMgr_ThemesChildDataByName()


// --------------------------------------------------------------------
// function : pkgMgr_GetMgr_ThemesChildDataByIndex()
// inputs   : PkgMgr *pkgMgr_p,const int themeIdx,const int childIdx
// returns  : void *, ptr to a DATA from a Theme's child, or NULL if not found
// purpose  : retrieve ptr to a theme's child's data, identified by
//            both the Theme's index and the child (PkgObj) index.
// revision : 2014-02-28T11:01:00,jmg (original)
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child data
// --------------------------------------------------------------------
void *pkgMgr_GetMgr_ThemesChildDataByIndex(PkgMgr *pkgMgr_p,
        const int themeIdx, const int childIdx)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 void *resultData_p = NULL;
 
  // first, fetch Theme's handle
  if((pThm_p = pkgMgr_GetThemeByIndex(pkgMgr_p,themeIdx)) != NULL)
  {
    // next, give this theme, fetch child object
    if((pObj_p = pkgMgr_GetObjByIndex(pThm_p,childIdx))!= NULL)
    {
        // finally, fetch the data itself
        resultData_p = pkgMgr_GetObjData(pObj_p);
    } else {
        xLogMsg("@E In(%s) FAILED locating child index(%d) in theme(%d)~\n",
        (char *)__func__, childIdx, themeIdx);
    }
      
  } else {
        xLogMsg("@E In(%s) FAILED locating theme(%d) owning child index(%d)~\n",
        (char *)__func__, themeIdx, childIdx);
  }
  
 // returns pointer to PkgObj child's data owned by theme, owned by PkgMgr
 return resultData_p ;
 
} // end:: pkgMgr_GetMgr_ThemesChildDataByIndex()


// --------------------------------------------------------------------
// function : pkgMgr_GetMgr_AllNames()
// inputs   : PkgMgr *pkgMgr_p,
//            FILE *destf_p --optional ptr to dest file to also route
//            the resulting stringtable to, OR, may be 'stdout|stderr' or
//            if NULL, omits routing results to a file like object
// returns  : char **stringTable_p --list of ALL theme-names and ALL child names
//            owned by each theme
// purpose  : retrieve ptr to a string table containing ALL the theme
//            and child names, as a comma sep value list of 'records'
//            like:
//             themeName1,themesChildName1
//             themeName1,themesChildName2
//             themeName1,themesChildName{K}
//             themeName2,themesChildName1
//             themeName2,themesChildName2
//             themeName2,themesChildName{K}.. etc for the rest
//
// revision : 2014-08-14T14:30:00:00, <JMG> added pkgMgr_GetMgr_AllNames()
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child data
// --------------------------------------------------------------------
char **pkgMgr_GetMgr_AllNames(PkgMgr *pkgMgr_p,FILE *destf_p,int *masterCount)
{
 PkgTheme *pThm_p = NULL;
 // PkgObj   *pObj_p = NULL;
 
 char *themeName_p = NULL; // ptr to a given theme's name
 char *childName_p = NULL; // ptr to a given child PkgObj name
 char **resultStrTblList_p = NULL;
 char entryRecord[(PKGMGR_MAX_NAMELEN*2)+2] = {0};
 char listName[PKGMGR_MAX_NAMELEN] ; // name of whole list
 char fxName[256] = {0};
 // default delimiter is a comma, though could allow user defined one later
 int  chDelimiter = ',' ;
 
 int  ithTheme =0;
 int  ithObj   =0;
 int  ithEntry =0;
 int  nThemes = 0 ; // count of ALL themes owns by this PkgMgr
 int  nChildren=0 ; // local count of a given themes children
 
    // Step 1: count up the total number of themes and within each
    //   theme, the count of its children to sum up, enabling creation
    //   of a single list that can contain ALL of these joint names
    nThemes = pkgMgr_p->pkMgrThemeCount ;
    nChildren = 0 ;
    
    for(ithTheme=0; ithTheme < nThemes ; ithTheme++)
    {
      if((pThm_p = pkgMgr_GetThemeByIndex(pkgMgr_p,ithTheme)) != NULL)  
      {
            // accumulate sum of ALL children owned by ALL themes
            nChildren += pkgMgr_GetThemeObjectCount(pThm_p);
      }
     } // end:: ithTheme outer counting loop

    // update master count of ALL children owned across ALL themes
    *masterCount = nChildren ;

    // Give this master name list a name prefixed using pkgMgr's label...
    sprintf(listName,"%s_MasterNameList",pkgMgr_GetMgrName(pkgMgr_p));
    
    // Construct one very large stringtable to hold ALL names. 
    // NOTE: When memObjCalloc() is used to create a IsStringTable then
    // support for object naming IS supported.
    
    // resultStrTblList_p = memCreateStringTbl(nChildren,PKGMGR_MAX_NAMELEN+2);
    resultStrTblList_p = memObjCalloc(listName,PKGMGR_MAX_NAMELEN+2,nChildren,IsStringTable,fxName,fxName);

    // WARNING!! if memCreateStringTbl() is used, then memObjSetName
    // is NOT yet configured to name a stringtable
    // memObjSetName(resultStrTblList_p, listName);
     
     // Step 2: load stringtable with ALL theme and childnames
    for(ithEntry=0,ithTheme=0; ithTheme < pkgMgr_p->pkMgrThemeCount ; ithTheme++)
    {
      if((pThm_p = pkgMgr_GetThemeByIndex(pkgMgr_p,ithTheme)) != NULL)
      {
            nChildren = pkgMgr_GetThemeObjectCount(pThm_p);
            themeName_p=pkgMgr_GetThemeName(pThm_p);
            for(ithObj=0; ithObj < nChildren; ithObj++)
            {
                childName_p = pkgMgr_GetThemes_ChildName(pThm_p,ithObj);
                // Finally, load stringtable with this entry, using delimiter 'chDelimiter'
                sprintf(entryRecord,"%s%c%s",themeName_p, chDelimiter, childName_p);
                
                // store and increment master count of all entrys, ithEntry...
                strcpy(resultStrTblList_p[ithEntry++], entryRecord);
                // then, route to destination file like object if requested
                if(destf_p != NULL)
                {
                    fprintf(destf_p,"%s%c%s\n",themeName_p,chDelimiter, childName_p);
                }
            }  // end:: ithObj inner loop
      } // end:: block for a valid theme ptr
    } // end:: ithTheme loop through all themes
 
 // returns pointer to stringtable list of ALL children in ALL themes
 return resultStrTblList_p;
} // end:: pkgMgr_GetMgr_AllNames()


// --------------------------------------------------------------------
// function : pkgMgr_SetMgr_ThemesChildIdByName()
// inputs   : PkgMgr *pkgMgr_p,const char *themeLabel,const char *childLabel
// returns  : void
// purpose  : assign new ID a theme's child's ID, identified by
//            both the Theme's label and the child (PkgObj) label.
// revision : 2020-07-13, hid_t compat fix.
// revision : 2014-03-07(original)
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child's ID
// --------------------------------------------------------------------
void pkgMgr_SetMgr_ThemesChildIdByName(PkgMgr *pkgMgr_p,const char *themeLabel_p,
                    const char *childLabel_p,const int64_t newId)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 
  // first, fetch Theme's handle
  if((pThm_p = pkgMgr_GetThemeByName(pkgMgr_p,themeLabel_p))!=NULL)
  {
    // next, give this theme, fetch child object
    
    if((pObj_p = pkgMgr_GetObjByName(pThm_p, childLabel_p))!= NULL)
    {
        // finally, assign the ID data itself
        pkgMgr_SetObjId(pObj_p,newId);

    } else {
        xLogMsg("@E In(%s) FAILED locating child named(%s) in theme(%s)~\n",
            (char *)__func__, childLabel_p, themeLabel_p);
    }
      
  } else {
        xLogMsg("@E In(%s) FAILED locating theme(%s)~\n",
                (char *)__func__, themeLabel_p);
  }
  
} // end:: pkgMgr_SetMgr_ThemesChildIdByName()


// --------------------------------------------------------------------
// function : pkgMgr_GetMgr_ThemesChildIdByName()
// inputs   : PkgMgr *pkgMgr_p,const char *themeLabel,const char *childLabel
// returns  : void *, ptr to a DATA from a Theme's child, or NULL if not found
// purpose  : retrieve ptr to a theme's child's data, identified by
//            both the Theme's label and the child (PkgObj) label.
// revision : 2020-07-013, hid_t compat. fix.
// revision : 2014-02-28T11:01:00,jmg (original)
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child's ID
// --------------------------------------------------------------------
int64_t pkgMgr_GetMgr_ThemesChildIdByName(PkgMgr *pkgMgr_p,const char *themeLabel_p,
                    const char *childLabel_p)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 int thisId = -1 ;
 
  // first, fetch Theme's handle
  if((pThm_p = pkgMgr_GetThemeByName(pkgMgr_p,themeLabel_p))!=NULL)
  {
    // next, give this theme, fetch child object
    
    if((pObj_p = pkgMgr_GetObjByName(pThm_p, childLabel_p))!= NULL)
    {
        // finally, fetch the data itself
        thisId = pkgMgr_GetObjId(pObj_p);

    } else {
        xLogMsg("@E In(%s) FAILED locating child named(%s) in theme(%s)~\n",
            (char *)__func__, childLabel_p, themeLabel_p);
    }
      
  } else {
        xLogMsg("@E In(%s) FAILED locating theme(%s)~\n",
                (char *)__func__, themeLabel_p);
  }
  
 // returns ID of item
 return thisId ;
 
} // end:: pkgMgr_GetMgr_ThemesChildIdByName()


// --------------------------------------------------------------------
// function : pkgMgr_SetMgr_ThemesChildOwnerByName()
// inputs   : PkgMgr *pkgMgr_p,const char *themeLabel,const char *childLabel
// returns  : void
// purpose  : assign new ID a theme's child's ID, identified by
//            both the Theme's label and the child (PkgObj) label.
// revision : 2014-03-07(original)
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child's ID
// --------------------------------------------------------------------
void pkgMgr_SetMgr_ThemesChildOwnerByName(PkgMgr *pkgMgr_p,const char *themeLabel_p,
                    const char *childLabel_p,const int newOwnerId)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 
  // first, fetch Theme's handle
  if((pThm_p = pkgMgr_GetThemeByName(pkgMgr_p,themeLabel_p))!=NULL)
  {
    // next, give this theme, fetch child object
    
    if((pObj_p = pkgMgr_GetObjByName(pThm_p, childLabel_p))!= NULL)
    {
        // finally, assign the OWNER ID 
        pkgMgr_SetObjOwner(pObj_p,newOwnerId);

    } else {
        xLogMsg("@E In(%s) FAILED locating child named(%s) in theme(%s)~\n",
            (char *)__func__, childLabel_p, themeLabel_p);
    }
      
  } else {
        xLogMsg("@E In(%s) FAILED locating theme(%s)~\n",
                (char *)__func__, themeLabel_p);
  }
  
} // end:: pkgMgr_SetMgr_ThemesChildOwnerByName()


// --------------------------------------------------------------------
// function : pkgMgr_GetMgr_ThemesChildOwnerByName()
// inputs   : PkgMgr *pkgMgr_p,const char *themeLabel,const char *childLabel
// returns  : int, owner id of a Theme's child, or -1 if not found
// purpose  : retrieve ptr to a theme's child's owner ID, identified by
//            both the Theme's label and the child (PkgObj) label.
// revision : // revision : 2014-03-07(original)
// author   : joe glassy
// notes    : A shortcut wrapper that wraps calls into 2 levels down,
//            to allow just a PkgMgr_p reference and two labels to 
//            obtain child's ID
// --------------------------------------------------------------------
int pkgMgr_GetMgr_ThemesChildOwnerByName(PkgMgr *pkgMgr_p,const char *themeLabel_p,
                    const char *childLabel_p)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 int ownerId = -1 ;
 
  // first, fetch Theme's handle
  if((pThm_p = pkgMgr_GetThemeByName(pkgMgr_p,themeLabel_p))!=NULL)
  {
    // next, give this theme, fetch child object
    
    if((pObj_p = pkgMgr_GetObjByName(pThm_p, childLabel_p))!= NULL)
    {
        // finally, fetch the owner ID itself
        ownerId = pkgMgr_GetObjOwner(pObj_p);

    } else {
        xLogMsg("@E In(%s) FAILED locating child named(%s) in theme(%s)~\n",
            (char *)__func__, childLabel_p, themeLabel_p);
    }
      
  } else {
        xLogMsg("@E In(%s) FAILED locating theme(%s)~\n",
                (char *)__func__, themeLabel_p);
  }
  
 // returns pointer to PkgObj child's data owned by theme, owned by PkgMgr
 return ownerId;
 
} // end:: pkgMgr_GetMgr_ThemesChildOwnerByName()


// --------------------------------------------------------------------
// function : pkgMgr_GlobalReport()
// inputs   : PkgMgr *pkgMgr_p,FILE *outf,char *reportLabel_p,
// returns  : char *, pointer to theme's name
// purpose  : Generate a global PkgMgr, annotated, sent to outf
// revision : 2014-02-21:00,jmg (added)
// author   : joe glassy
// notes    : Plan to convert 'reportLabel_p' to sprintf() va_args style
//            formatted spec string method
// --------------------------------------------------------------------
int pkgMgr_GlobalReport(PkgMgr *pkgMgr_p,FILE *outf,char *reportLabel_p)
{
 PkgTheme *pThm_p = NULL;
 PkgObj   *pObj_p = NULL;
 int ithTheme = 0;
 int ithObj   = 0;
 int taskItemCount = 0;
 

    assert(outf != NULL);

    // OVERSEER level report: at the PkgMgr (overseer) level
    // <JMG> changed to %ld for hid_t compat
    fprintf(outf,"\n");
    fprintf(outf,"@A ~PkgMgr Global Report %s for ID %ld Label %s\n",
                    reportLabel_p, (long)pkgMgr_p->pkMgrId, pkgMgr_p->pkMgrLabel);
                    
    fprintf(outf,"@A ~UUID(%s) Theme CommonCategory (%s)\n",
                    pkgMgr_p->pkMgrUUID, pkgMgr_p->pkMgrThemeCommonCategory);
                    
    fprintf(outf,"@A ~Themes declared: %d\n", pkgMgr_p->pkMgrThemeCount );

        for(ithTheme=0; ithTheme< pkgMgr_p->pkMgrThemeCount; ithTheme++)
        {
            // set pointer to current theme PkgMgr owns
            pThm_p = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
            
            // THEME level report
            fprintf(outf,"@A ~Theme %d ID %d label (%s) UUID(%s) category(%s) referent-count %d\n",
                ithTheme, pThm_p->pkThemeId, pThm_p->pkThemeLabel,
                pThm_p->pkThemeUUID, pThm_p->pkThemeCategory,
                pThm_p->pkThemeObjCount);
                
                // REFERENT level report, at PkgObj (child) innermost level
                //  only a subset of items are reported. Later a filter-base scheme is used
                //  to allow caller to choose just the subset they want to see
                for(ithObj=0; ithObj < pThm_p->pkThemeObjCount ; ithObj++)
                {
                    pObj_p = &pThm_p->pkTheme_PkgObjAry_p[ithObj];
                    
                    taskItemCount= ( pObj_p->pkObjSubtaskEndIdx - pObj_p->pkObjSubtaskStartIdx) + 1;
                    
                    fprintf(outf,
                        "@A referent %d ID %ld ownerId %ld key(%s) categoryCode(%d) datatype %d Rank %d Nelem %zu Taskstart %zu end %zu N %d~\n",
                        ithObj,
                        pObj_p->pkObjId,
                        pObj_p->pkObjOwner,
                        pObj_p->pkObjKey,
                        pObj_p->pkObjCategoryCode,
                        pObj_p->pkObjDatatype,
                        pObj_p->pkObjRank,      
                        pObj_p->pkObjNelem, 
                        pObj_p->pkObjSubtaskStartIdx, 
                        pObj_p->pkObjSubtaskEndIdx,
                        taskItemCount);

                } // end:: ithObj loop
            
        } // end:: i loop through pkgObj's owned by this PkgTheme
 // returns index to match theme
 return 0 ;
} // end:: pkgMgr_GlobalReport()




