// --------------------------------------------------------------------
// module   : pkg_mgr_objects.c
// purpose  : implementatin of PkgObj services
// revision : 2020-07-08T19:13, modified for HDF5 v1.10.x compatibility using int64_t owner Id's
// revision : 2014-08-14T17:12, added pkgMgr_SetObjH5Datatype(), GetObjH5Datatype()
// history  : 2014-08-14T17:00:00, added in pkgMgr_CreateObj(), default assignment of
//            the internal memory objects name via memObjSetName(), same as PkgObj's name.
// history  : 2014-08-13, added pkgMgr_GetNameByIndex()
// history  : 2014-06-11T18:14:00, changed init values to 0 for
//            size_t types, as the -1 used before was an error causing
//            integer wrap for unsigned ints...
// history : 2014-01-15T14:24:00,jmg (original)
// version  : 1.0
// author   : joe glassy
// assertions : The PkgMgr must define at least (1) PkgTheme
//              A PkgTheme may define 0,1 or more PkgObj up to the limit allowed.
//              By default unless overridden, a PkgTheme creates K PkgObj empty-instances
//              All PkgThemes have a unique integer ID, >=0, and a label
//              All PkgObj have a unique integer ID >=0, and a label
//              No data per say is ever stored in a PkgObj, as this is just a registry
//              Data for a variable is created elsewhere; the PkgObj just points to it.
// notes     : Future-- separate out the common 'search' functionality 
//             now used in pkgMgr_GetXXX() functions, into 1 to 3
//             specialized 'search' functions all returning a common 
//             binary state result (DidFind|CannotFind), to allow for 
//             selective, late-bound use of a linear vs. binary search
//             algorithm on a sorted set. For larger collections (>= 1000
//             or 10,000 etc) a binary search could/should be implemented.
// --------------------------------------------------------------------

#include "frm_api.h"

#include "pkg_mgr.h"
#include "pkg_mgr_prot.h"


// --------------------------------------------------------------------
// function : pkgMgr_ObjBindj()
// purpose  : binds (1) PkgObj to a pre-existing referent, e.g. usually
//            a heap variable already allocated.
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_ObjBind(PkgObj *pkgObj_p,void *referent_p)
{

 pkgObj_p->pkObjContent_p = (uint8 *)referent_p ;
 
} // end:: pkgMgr_ObjBindj()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjProperties()
// purpose  : assign a common set of PkgObj properties AND performs "BIND"
//            of raw data
// inputs   : PkgObj *pkgObj_p --ptr to Obj
//            char *keyName_p  --ptr to string name,label, or keyword
//            const int tivDatatype -- ptr to TIV datatype identifier like IsFloat32
//            const int newRank -- rank (n. of dimensions)
//            const int nElem  --n. elements across all dimensions
//            void *referent_p --ptr to prior created raw scalar data variable of TIV datatype
// returns  : int, ID (as a reminder only)
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : limited in length to PKGMGR_MAX_NAMELEN (256)
int pkgMgr_SetObjProperties(PkgObj *pkgObjNew_p,char *keyName_p,const int tivDatatype,
                    const int newRank,const int nElem, void *referent_p)
{
    assert(strlen(keyName_p) > 0);
    strcpy(pkgObjNew_p->pkObjKey,keyName_p);
    
    pkgObjNew_p->pkObjDatatype = tivDatatype ;
    pkgObjNew_p->pkObjRank     = newRank ;
    pkgObjNew_p->pkObjNelem    = nElem   ;
    
    // for 1D vectors, this is a convenience assigment to dims[]
    if(newRank == 1 && nElem > 0)
     pkgObjNew_p->pkObjDims[0] = nElem ;
     
    // Perform the critical BIND of raw data to this element
    pkgObjNew_p->pkObjContent_p = referent_p;
    
    // returns ID
    return pkgObjNew_p->pkObjId;
    
} // end:: pkgMgr_SetObjProperties()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjName()
// purpose  : assign PkgObj's name using sprintf style formatter
// inputs   : PkgObj *pkgObj_p,newObjKey_p
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : limited in length to PKGMGR_MAX_NAMELEN (256)
void pkgMgr_SetObjName(PkgObj *pkgObj_p,char *fmtStr_p,...)
{
  char pkgObjNameString[FILENAME_MAX] = {0} ;
  va_list argptr;
  int     fmt_count =0;
  
    va_start(argptr,fmtStr_p);
    fmt_count = vsprintf(pkgObjNameString, fmtStr_p,argptr);
    va_end(argptr);
    // load formatted name string into PkgMgr's key, label or name area
    strcpy(pkgObj_p->pkObjKey, pkgObjNameString);
    
}  // end::pkgMgrSetObjName()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjName()
// purpose  : retrieve PkgObj keyword (name)
// inputs   : PkgObj *pkgObj_p
// returns  : char *, pointer to PkgObj's name or keyword
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
char *pkgMgr_GetObjName(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjKey ;
 
}  // end::pkgMgr_GetObjName()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjId()
// purpose  : assign PkgObj unique ID 
// inputs   : PkgObj *pkgObj_p,const int64_t newObjId
// assigns  : int64_t used now,  for compatibility with HDF5's hid_t
// revision : 2020-07-13T15:13
// revision : 2020-06-24T15:32
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_SetObjId(PkgObj *pkgObj_p,const int64_t newObjId)
{
    
  pkgObj_p->pkObjId = newObjId ;

} // end:: pkgMgr_SetObjId()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjId()
// purpose  : retrieve PkgObj unique ID 
// returns  : int64_t for compatibility with HDF5's hid_t, unique ID assigned to this PkgObj
// inputs   : PkgObj *pkgObj_p
// revision : 2020-07-13T15:13, return int64_t now for better hid_t compat.
// revision : 2020-06-24T15:32
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
int64_t pkgMgr_GetObjId(PkgObj *pkgObj_p)
{
    return pkgObj_p->pkObjId ;
} // end:: pkgMgr_GetObjId()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjRank()
// purpose  : assign PkgObj rank
// inputs   : PkgObj *pkgObj_p,const int newRank
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjRank(PkgObj *pkgObj_p,const int newRank)
{
    
  pkgObj_p->pkObjRank = newRank ;

} // end:: pkgMgr_SetObjRank()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjRank()
// purpose  : retrieve PkgObj rank
// inputs   : PkgObj *pkgObj_p,const int newRank
// returns  : int, rank
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
int pkgMgr_GetObjRank(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjRank;
} // end:: pkgMgr_GetObjRank()


// --------------------------------------------------------------------
// function : pkgMgr_SetCategoryCode()
// purpose  : assign PkgObj category code
// inputs   : PkgObj *pkgObj_p,const int newCategoryCode
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_SetObjCategoryCode(PkgObj *pkgObj_p,const int newCategoryCode)
{
  pkgObj_p->pkObjCategoryCode = newCategoryCode ;
} // end:: pkgMgr_SetObjCategoryCode()

// --------------------------------------------------------------------
// function : pkgMgr_GetCategoryCode()
// purpose  : fetch PkgObj category code
// inputs   : PkgObj *pkgObj_p
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
int pkgMgr_GetObjCategoryCode(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjCategoryCode ;
} // end:: pkgMgr_GetObjCategoryCode()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjNelem()
// purpose  : assign PkgObj nElem --count of elements across ALL dimensions
// inputs   : PkgObj *pkgObj_p,const size_t nElem
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjNelem(PkgObj *pkgObj_p,const size_t nElem)
{
  pkgObj_p->pkObjNelem = nElem;
} // end:: pkgMgr_SetObjNelem()

// --------------------------------------------------------------------
// function : pkgMgr_SetObjResolution()
// purpose  : assign PkgObj dataset resolution code or value
// inputs   : PkgObj *pkgObj_p,const int resolutionCode
// returns  : void
// revision : 2014-02-26T17:02:00,jmg (original)
// notes    : Use 1=1km, 3=3km, 9=9km etc for now
void pkgMgr_SetObjResolution(PkgObj *pkgObj_p,const int resolutionCode)
{

  pkgObj_p->pkObjResolution = resolutionCode;

} // end:: pkgMgr_SetObjResolution()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjSelectFilter()
// purpose  : assign PkgObj selection filter boolean {0|1} value
// inputs   : PkgObj *pkgObj_p,const uint8 selectionFilterBoolea
// returns  : void
// revision : 2014-03-05T16:14,jmg (original)
// notes    : Allows dis-contiguous subset selection from within a theme
void pkgMgr_SetObjSelectFilter(PkgObj *pkgObj_p,const uint8 selectionFilterBoolean)
{

  pkgObj_p->pkObjSelectFilter = selectionFilterBoolean;

} // end:: pkgMgr_SetObjSelectFilter()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjSelectFilter()
// purpose  : retrieves the PkgObj selection filter boolean {0|1} value
// inputs   : PkgObj *pkgObj_p
// returns  : void
// revision : 2014-03-05T16:14,jmg (original)
// notes    : Allows dis-contiguous subset selection from within a theme
uint8 pkgMgr_GetObjSelectFilter(PkgObj *pkgObj_p)
{

  return pkgObj_p->pkObjSelectFilter ;

} // end:: pkgMgr_GetObjSelectFilter()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjResolution()
// purpose  : retrieve PkgObj dataset resolution code or value
// inputs   : PkgObj *pkgObj_p
// returns  : int --resolution code or value 1=1km, 3=3km, 9=9km etc
// revision : 2014-02-26T17:02:00,jmg (original)
// notes    : Use 1=1km, 3=3km, 9=9km etc for now
int pkgMgr_GetObjResolution(PkgObj *pkgObj_p)
{

  return pkgObj_p->pkObjResolution;

} // end:: pkgMgr_GetObjResolution()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjNelem()
// purpose  : retrieve PkgObj nElem --count of elements across ALL dimensions
// inputs   : PkgObj *pkgObj_p,const int nElem
// returns  : size_t, nElem as count of elements across all dimensions
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
size_t pkgMgr_GetObjNelem(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjNelem ;
} // end:: pkgMgr_GetObjNelem()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjUnits()
// purpose  : assign PkgObj units string
// inputs   : PkgObj *pkgObj_p,const char *unitString
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjUnits(PkgObj *pkgObj_p,const char *unitString_p)
{
  strcpy(pkgObj_p->pkObjUnitStr, unitString_p);
} // end:: pkgMgr_SetObjUnits()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjUnits()
// purpose  : retrieve PkgObj units string
// inputs   : PkgObj *pkgObj_p
// returns  : char *unitsString_p -- ptr to units string
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
char *pkgMgr_GetObjUnits(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjUnitStr;
} // end:: pkgMgr_GetObjUnits()

void pkgMgr_SetObjGain(PkgObj *pkgObj_p,const double gainVal)
{
  pkgObj_p->pkObjGain = gainVal ;   
} // end:: pkgMgrSetObjGain()

double pkgMgr_GetObjGain(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjGain ;
} // end:: pkgMgrGetObjGain()

void pkgMgr_SetObjOffset(PkgObj *pkgObj_p,const double offsetVal)
{
  pkgObj_p->pkObjOffset = offsetVal ;   
} // end:: pkgMgrSetObjOffset()

double pkgMgr_GetObjOffset(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjOffset;
} // end:: pkgMgrGetObjGain()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjFillvalue()
// purpose  : assign PkgObj fillvalue
// inputs   : PkgObj *pkgObj_p,const double fillVal
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : Fillvalues are always stored as double (IsFloat64) and
// are interpreted as actual TIV datatype at runtime
void pkgMgr_SetObjFillvalue(PkgObj *pkgObj_p,const double fillVal)
{
  pkgObj_p->pkObjFillval = fillVal;
} // end:: pkgMgr_SetObjFillvalue()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjFillvalue()
// purpose  : retrieve PkgObj fill value
// inputs   : PkgObj *pkgObj_p
// returns  : double fillValue
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : Caller interprets canonical (double) fillvalue as its
//            actual TIV datatype later
double pkgMgr_GetObjFillvalue(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjFillval;
} // end:: pkgMgr_GetObjFillvalue()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjValidRange()
// purpose  : assign PkgObj valid min AND valid max value for object
// inputs   : PkgObj *pkgObj_p,const double valid_min,const double valid_max
// returns  : void
// revision : 1.0-2014-08-06,jmg (original)
// notes    : n/a
void pkgMgr_SetObjValidRange(PkgObj *pkgObj_p,const double validMin,const double validMax)
{
  pkgObj_p->pkObjValidMin = validMin;
  pkgObj_p->pkObjValidMax = validMax;
} // end:: pkgMgr_SetObjValidRange()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjValidRange()
// purpose  : retrieve PkgObj valid min AND valid max value for object
// inputs   : PkgObj *pkgObj_p,const double valid_min,const double valid_max
// returns  : double, difference between valid_min and valid_max
// revision : 1.0-2014-08-06,jmg (original)
// notes    : n/a
double pkgMgr_GetObjValidRange(PkgObj *pkgObj_p,double *valMin,double *valMax)
{
  // first, retrieve individual valid min,max
  *valMin = pkgObj_p->pkObjValidMin ;
  *valMax = pkgObj_p->pkObjValidMax ;
  // derive and return simple difference between valid max and min
  return( pkgObj_p->pkObjValidMax - pkgObj_p->pkObjValidMin ) ;
} // end:: pkgMgr_GetObjValidRange()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjValidMin()
// purpose  : assign PkgObj valid min value
// inputs   : PkgObj *pkgObj_p,const double valid_min
// returns  : void
// revision : 1.0-2014-07-23,jmg (original)
// notes    : n/a
void pkgMgr_SetObjValidMin(PkgObj *pkgObj_p,const double validMin)
{
  pkgObj_p->pkObjValidMin = validMin;
} // end:: pkgMgr_SetObjValidMin()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjValidMin()
// purpose  : assign PkgObj valid min value
// inputs   : PkgObj *pkgObj_p,const double valid_min
// returns  : double, valid_min
// revision : 1.0-2014-07-23,jmg (original)
// notes    : n/a
double pkgMgr_GetObjValidMin(PkgObj *pkgObj_p)
{
  return  pkgObj_p->pkObjValidMin ;
} // end:: pkgMgr_GetObjValidMin()

// --------------------------------------------------------------------
// function : pkgMgr_SetObjValidMax()
// purpose  : assign PkgObj valid min value
// inputs   : PkgObj *pkgObj_p,const double valid_min
// returns  : void
// revision : 1.0-2014-07-23,jmg (original)
// notes    : n/a
void pkgMgr_SetObjValidMax(PkgObj *pkgObj_p,const double validMax)
{
  pkgObj_p->pkObjValidMax = validMax;
} // end:: pkgMgr_SetObjValidMax()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjValidMax()
// purpose  : Retrieve PkgObj valid min value
// inputs   : PkgObj *pkgObj_p,const double valid_min
// returns  : void
// revision : 1.0-2014-07-23,jmg (original)
// notes    : n/a
double pkgMgr_GetObjValidMax(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjValidMax ;
} // end:: pkgMgr_GetObjValidMax()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjDatatype()
// purpose  : assign PkgObj TIV datatype identifier
// inputs   : PkgObj *pkgObj_p,const int tivDatatype
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjDatatype(PkgObj *pkgObj_p,const int objDatatype)
{
    
  pkgObj_p->pkObjDatatype = objDatatype ;
  
  // DIAG only 2/27/2014
  // xLogMsg("@D (%s) name(%s) new datatype(%d)~\n",__func__,
  //    pkgObj_p->pkObjKey, pkgObj_p->pkObjDatatype);

} // end:: pkgMgr_SetObjDatatype()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjDatatype()
// purpose  : retrieve PkgObj TIV datatype
// inputs   : PkgObj *pkgObj_p
// returns  : int, TIV datatype, like: IsFloat32, IsInt32, etc.
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
int pkgMgr_GetObjDatatype(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjDatatype;
} // end:: pkgMgr_GetObjDatatype()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjH5Datatype()
// purpose  : assign PkgObj HDF5 datatype like H5T_NATIVE_FLOAT
// inputs   : PkgObj *pkgObj_p, hid_t newHdf5Datatype
// returns  : void
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjH5Datatype(PkgObj *pkgObj_p,hid_t newHdf5Datatype)
{
  pkgObj_p->pkObjH5dtype = newHdf5Datatype;
} // end:: pkgMgr_SetObjH5Datatype()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjH5Datatype()
// purpose  : retrieve PkgObj HDF5 datatype, like H5T_NATIVE_UINT8
// inputs   : PkgObj *pkgObj_p
// returns  : hid_t, HDF5 datatype identifier, like: H5T_NATIVE_USHORT, etc.
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
hid_t pkgMgr_GetObjH5Datatype(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjH5dtype ;
} // end:: pkgMgr_GetObjH5Datatype()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjH5DatasetId()
// purpose  : Assign a PkgObj HDF5 dataset ID
// inputs   : PkgObj *pkgObj_p, hid_t newId
// returns  : hid_t, HDF5 dataset ID identifier
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjH5DatasetId(PkgObj *pkgObj_p,hid_t newId)
{
  pkgObj_p->pkObjH5dsetId = newId;
} // end:: pkgMgr_SetObjH5DatasetId()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjH5DatasetId()
// purpose  : retrieve PkgObj HDF5 dataset ID
// inputs   : PkgObj *pkgObj_p
// returns  : hid_t, HDF5 dataset ID identifier
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
hid_t pkgMgr_GetObjH5DatasetId(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjH5dsetId ;
} // end:: pkgMgr_GetObjH5DatasetId()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjH5NameByHdf5Id()
// purpose  : given a PkgTheme and a HDF5 ID assign prior, use the HDF5 ID
//            as the key by which to identify and return the PkgObj's name
// inputs   : PkgTheme *pkgTheme_p --ptr to theme owning this child
//            const hid_t h5Id -- HDF5 ID to use a query key
//            char *resultName_p -- ptr to name buffer assigned upon return
// returns  : ptr to name buffer assigned upon return
// revision : 1.0-2014-09-12T10:55:00,jmg (added)
// author   : joe glassy
// notes    : Added to support name lookups by HDF5 unique ID
char *pkgMgr_GetObjH5NameByHdf5Id(PkgTheme *pkgTheme_p,const hid_t h5Id,char *resultName_p)
{
 PkgObj   *thisObj_p   = NULL;
 int ithObj = 0;
 int nObj   = pkgTheme_p->pkThemeObjCount;
 
    // initialize
    strcpy(resultName_p,"");  
    
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        thisObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        if( h5Id == thisObj_p->pkObjH5dsetId )
        {
            strcpy(resultName_p,thisObj_p->pkObjKey);
            break;
        }
    } // end:: ithObj loop
    // return ptr to retrieved name or "" if not found
   return resultName_p;
} // end:: pkgMgr_GetObjH5NameByHdf5Id()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjH5IdByName()
// purpose  : given a PkgTheme and a name (key) use the name to retrieve
//            the HDF5 ID assigned prior.
// inputs   : PkgTheme *pkgTheme_p,const hid_t h5Id,char *resultName
// returns  : hid_t childHdf5Id -- the prior assigned HDF5 ID assoc. with this name
//            or, -1 if no such name exists in this themes collection of children
// revision : 1.0-2014-09-12T10:55:00,jmg (added)
// author   : joe glassy
// notes    : Added to support name lookups by HDF5 unique ID
hid_t pkgMgr_GetObjH5IdByName(PkgTheme *pkgTheme_p,const hid_t h5Id,char *targetName_p)
{
 PkgObj   *thisObj_p   = NULL;
 hid_t resultHdf5Id = -1;
 int ithObj = 0;
 int nObj   = pkgTheme_p->pkThemeObjCount;
 
    // initialize
    resultHdf5Id = -1;
    
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        thisObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        if(strcmp(targetName_p,thisObj_p->pkObjKey)==0)
        {
            resultHdf5Id = thisObj_p->pkObjH5dsetId ;
            break;
        }
    } // end:: ithObj loop
    // return ptr to retrieved hid_t HDF5 unique ID found, or -1 if failed
   return resultHdf5Id;
} // end:: pkgMgr_GetObjH5IdByName()

// --------------------------------------------------------------------
// function : pkgMgr_SetObjH5Dcpl()
// purpose  : assign a PkgObj HDF5 dcpl
// inputs   : PkgObj *pkgObj_p,hid_t newDcpl
// returns  : void
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
void  pkgMgr_SetObjH5Dcpl(PkgObj *pkgObj_p,hid_t newDcpl)
{
    pkgObj_p->pkObjH5dcpl = newDcpl ;
} // end:: pkgMgr_SetObjH5Dcpl()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjH5Dcpl()
// purpose  : retrieve PkgObj HDF5 dcpl
// inputs   : PkgObj *pkgObj_p
// returns  : hid_t, HDF5 dcpl identifer
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
hid_t pkgMgr_GetObjH5Dcpl(PkgObj *pkgObj_p)
{
    return pkgObj_p->pkObjH5dcpl ;
} // end:: pkgMgr_GetObjH5Dcpl()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjH5Dspace()
// purpose  : assign a PkgObj HDF5 dataset's dspace
// inputs   : PkgObj *pkgObj_p, hid_t newDspace
// returns  : void
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
void  pkgMgr_SetObjH5Dspace(PkgObj *pkgObj_p,hid_t newDspace)
{
    pkgObj_p->pkObjH5dspace = newDspace;
} // end:: pkgMgr_SetObjH5Dspace()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjH5Dspace()
// purpose  : retrieve PkgObj HDF5 dataset dspace
// inputs   : PkgObj *pkgObj_p
// returns  : hid_t, HDF5 dataset dspace
// revision : 2014-08-14T17:11:00,jmg (original)
// notes    : n/a
hid_t pkgMgr_GetObjH5Dspace(PkgObj *pkgObj_p)
{
    return pkgObj_p->pkObjH5dspace ;
} // end::  pkgMgr_GetObjH5Dspace()



// --------------------------------------------------------------------
// function : pkgMgr_SetObjDims()
// purpose  : assign PkgObj TIV datatype identifier
// inputs   : PkgObj *pkgObj_p,const int tivDatatype
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
void pkgMgr_SetObjDims(PkgObj *pkgObj_p,const size_t *pDims)
{
 int d=0;
 int nElem = 1;
    // enforce limits
    assert( pkgObj_p->pkObjRank >0 && pkgObj_p->pkObjRank < PKGMGR_MAX_DIMS);
    
    // initialize the dims array
    memcpy(pkgObj_p->pkObjDims,pDims,sizeof(size_t)* pkgObj_p->pkObjRank) ;
    
    // derive the nElem factor now that we have dims[] to work with
    for(nElem=1,d=0; d < pkgObj_p->pkObjRank; d++)
    {
        nElem *= pkgObj_p->pkObjDims[d] ;
    }
    pkgObj_p->pkObjNelem  = nElem;

} // end:: pkgMgr_SetObjDims()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjDims()
// purpose  : Retrieve a PkgObj assigned dimensions array 
//             of [0..PKGMGR_MAX_DIMS]
// inputs   : PkgObj *pkgObj_p
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// notes    : n/a
size_t *pkgMgr_GetObjDims(PkgObj *pkgObj_p)
{
  return pkgObj_p->pkObjDims;
  
} // end:: pkgMgr_GetObjDims()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjSubtaskStart()
// purpose  : assign PkgObj's  subtask start index
// inputs   : PkgObj *pkgObj_p,size_t newStartIndex
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Typically used for pthread task partitioning
void pkgMgr_SetObjSubtaskStart(PkgObj *pkgObj_p,const size_t newStartIndex)
{

  pkgObj_p->pkObjSubtaskStartIdx = newStartIndex;
 
}  // end::pkgMgr_SetObjSubtaskStart()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjSubtaskStart()
// purpose  : retrieve PkgObj's  subtask start index
// inputs   : PkgObj *pkgObj_p
// returns  : size_t, subtask end index
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Typically used for pthread task partitioning
size_t pkgMgr_GetObjSubtaskStart(PkgObj *pkgObj_p)
{

  return  pkgObj_p->pkObjSubtaskStartIdx ;
 
}  // end::pkgMgr_GetObjSubtaskStart()

// --------------------------------------------------------------------
// function : pkgMgr_SetObjSubtaskEnd()
// purpose  : assign PkgObj's  subtask ending index
// inputs   : PkgObj *pkgObj_p,size_t newEndIndex
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Typically used for pthread task partitioning
void pkgMgr_SetObjSubtaskEnd(PkgObj *pkgObj_p,const size_t newEndIndex)
{

  pkgObj_p->pkObjSubtaskEndIdx = newEndIndex;
 
}  // end::pkgMgr_SetObjSubtaskEnd()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjSubtaskStart()
// purpose  : retrieve PkgObj's  subtask end index
// inputs   : PkgObj *pkgObj_p
// returns  : size_t subtask end index
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Typically used for pthread task partitioning
size_t pkgMgr_GetObjSubtaskEnd(PkgObj *pkgObj_p)
{

  return pkgObj_p->pkObjSubtaskEndIdx;
 
}  // end::pkgMgr_GetObjSubtaskEnd()


// --------------------------------------------------------------------
// function : pkgMgr_SetObjOwner()
// purpose  : assign PkgObj owner ID 
// inputs   : PkgObj *pkgObj_p,const int64_t newOwnerId
// notes    : accepts: int64_t for compatibility with HDF5's hid_t
// revision : 1.2-2020-07-08T19:14, using int64_t for hid_t compatibility
// revision : 1.1-2020-06-24T15:16
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_SetObjOwner(PkgObj *pkgObj_p,const int64_t newOwnerId)
{
    
  pkgObj_p->pkObjOwner = newOwnerId ;

} // end:: pkgMgr_SetObjOwner()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjOwner()
// purpose  : retrieve PkgObj owner's unique ID 
// inputs   : PkgObj *pkgObj_p
// returns  : int64_t used now for compatibility with HDF5's hid_t
// revision : 1.2-2020-07-08T19:14, using int64_t for hid_t compatibility
// revision : 1.1-2020-06-24T15:16
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
int64_t pkgMgr_GetObjOwner(PkgObj *pkgObj_p)
{
    return pkgObj_p->pkObjOwner ;
    
} // end:: pkgMgr_GetObjOwner()
 
 
// --------------------------------------------------------------------
// function : pkgMgr_CreateObj()
// purpose  : create (1) PkgObj owned by this PkgTheme
// revision : 1.1-2020-07-13T14:21, change to int64_t for hid_t compat.
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// --------------------------------------------------------------------
PkgObj *pkgMgr_CreateObj(const int64_t objOwnerId,const char *keywordName_p,
        const int64_t objID,const int tivDatatype)
{
 PkgObj *pkgObjNew_p = NULL;
 char localName[256] = {0} ;
 
    // Allocate memory for this theme, using memCalloc() ...lite
    pkgObjNew_p = memCalloc(1,sizeof(PkgObj));
    
    // Pre-inialize with specific baseline metadata
    pkgObjNew_p->pkObjId  = objID ; 
    // <JMG> 2014-06-26, changed to simple objOwnerId instead of PkgTheme
    pkgObjNew_p->pkObjOwner = objOwnerId;
    
    // a -1 category means its created but NOT used
    pkgObjNew_p->pkObjCategoryCode = -1 ; 
    // Assign its TIV (sync'd now with HDF4) datatype identifier
    pkgObjNew_p->pkObjDatatype = tivDatatype ;
    // Assign its rank (as unspecified)
    pkgObjNew_p->pkObjRank = -1;
    // Assign its dimensionality (as unspecified)
    memset(pkgObjNew_p->pkObjDims,'\0', PKGMGR_MAX_DIMS);
    
    // <JMG> 6/11/2014, since these are size_t, assign each of
    //  these as 0 upon initialization. Prior, these were init as -1
    //  which caused illegal wrap around bug..
    pkgObjNew_p->pkObjSubtaskStartIdx = 0;
    pkgObjNew_p->pkObjSubtaskEndIdx   = 0;
    pkgObjNew_p->pkObjNelem = 0 ;
    // init its name...
    strcpy(pkgObjNew_p->pkObjKey,"");
    // assign its name
    if(strlen(keywordName_p)>0)
    {
        strcpy(pkgObjNew_p->pkObjKey, keywordName_p);
        // <JMG> 2014-08-14T17:00, added default assignment of
        // the internal memory objects name, same as PkgObj's name.
        strcpy(localName,keywordName_p);
        memObjSetName(pkgObjNew_p, localName);
    }
    
    // Initialize the pointer to this PkgObj's data as NULL
    // since this should be defined and bound later
    pkgObjNew_p->pkObjContent_p = NULL;
    
    // return ptr to new, PkgObj created
   return pkgObjNew_p ;
   
} // end:: pkgMgr_CreateObj()


// --------------------------------------------------------------------
// function : pkgMgr_InitObj()
// purpose  : iniialize the properties of (1) 
// inputs   : PkgObj *pkgObj_p -- ptr of object to init
// revision : 2014-07-01, added oRank test to skip over memcpy when rank = -1
// history  : 2014-06-11, changed init to 0 for size_t members
// history  : 2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : use separate function (pkgMgr_ObjBind) to associate the
//            PkgObj with the pre-existing data variable registered to it.
// --------------------------------------------------------------------
int pkgMgr_InitObj(PkgObj *pkgObj_p,const char *objKeyword_p,
    const int64_t objID,const int64_t objOwnerId,
    const int tivDatatype,const int newCategory, const int oRank, const size_t *newDims)
{
    char fxName[256] = {0} ;
    size_t r=0, 
           nElem = 0 ;
    
    strcpy(fxName,__func__);
    
    pkgObj_p->pkObjId    = objID ; 
    pkgObj_p->pkObjOwner = objOwnerId ;
    
    pkgObj_p->pkObjCategoryCode = newCategory ;
    pkgObj_p->pkObjDatatype = tivDatatype ;
    
    pkgObj_p->pkObjRank = oRank;

    // <JMG> 6/11/2014, changed init to '0' (can't use -1) since
    // these are all size_t! ..unsigned :)
    pkgObj_p->pkObjSubtaskStartIdx = 0;
    pkgObj_p->pkObjSubtaskEndIdx   = 0;
    pkgObj_p->pkObjNelem = 0 ;
    
    // zero out dimensions until explicitly assigned
    memset(pkgObj_p->pkObjDims, '\0', PKGMGR_MAX_DIMS*sizeof(size_t));
    
    // Establish this PkgObj's name or keyword (label)
    if(strlen(objKeyword_p) > 0)
        strcpy(pkgObj_p->pkObjKey,objKeyword_p);
    else 
        strcpy(pkgObj_p->pkObjKey,"");

    // <JMG> 2014-07-01, added newDims[]validity check, and
    // oRank test to skip over memcpy of dims if an invalid or
    // out or range RANK is supplied
    if( newDims != NULL)
    {
        if(oRank >= 1 && oRank <= PKGMGR_MAX_DIMS )
        {
            memcpy(pkgObj_p->pkObjDims, newDims, oRank * sizeof(int) );
            // auto-derive nElem by iterating across all dimensions
            for(nElem=1,r=0; r < (size_t)oRank; r++)
                nElem *= newDims[r];
            // assign nElem
            pkgObj_p->pkObjNelem = nElem ;
            
        } else {
            xLogMsg("@E In(%s) Invalid rank(%d) so cannot initialize pkObjDims[] for (%s)~\n",
                fxName, oRank, objKeyword_p);
        }
    }
    
#ifdef PKGMGR_DIAG
    xLogMsg("@A In(%s) id(%ld) owner(%id) name(%s)~\n",
        fxName, pkgObj_p->pkObjId, pkgObj_p->pkObjOwner, pkgObj_p->pkObjKey);
#endif  
    
    return 0 ;
    
} // end::pkgMgr_InitObj()



// --------------------------------------------------------------------
// function : pkgMgr_ReleaseObj()
// purpose  : free up (1) PkgObj owned by this PkgTheme, by first
//            clearing the member values then releasing memory.
// revision : 2014-06-11T18:16:00, changed free() to memFree()
// revision : 2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_ReleaseObj(PkgObj *pkgObj_p)
{
 
    pkgMgr_ClearObject(pkgObj_p);
    // <JMG> 6/11/2014 changed to memFree() instead of free()
    memFree(pkgObj_p);
 
} // end:: pkgMgr_ReleaseObj()




// --------------------------------------------------------------------
// function : pkgMgr_ClearObj()
// purpose  : reset individ field members of one PkgObj
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_ClearObject(PkgObj *pkgObj_p)
{
    pkgObj_p->pkObjId    = -1  ; // -1 means currently NOT used or occupied
    pkgObj_p->pkObjOwner = -1  ; // -1 means no one owns me now (or yet)
    pkgObj_p->pkObjCategoryCode = -1 ; // a -1 category means its created, but NOT used
    pkgObj_p->pkObjDatatype = -1 ;
    pkgObj_p->pkObjRank = -1;
    
    memset(pkgObj_p->pkObjDims,'\0',PKGMGR_MAX_DIMS);
    // these next three as size_t cannot be assigned -1, so use 0
    pkgObj_p->pkObjNelem = 0 ;
    pkgObj_p->pkObjSubtaskStartIdx = 0;
    pkgObj_p->pkObjSubtaskEndIdx   = 0;
    
    // clear ALL of the string fields
    strcpy(pkgObj_p->pkObjKey,"");
    strcpy(pkgObj_p->pkObjMetadata,"");
    strcpy(pkgObj_p->pkObjGenericPath,"");
    strcpy(pkgObj_p->pkObjShortname,"");
    strcpy(pkgObj_p->pkObjLongname,"");
    strcpy(pkgObj_p->pkObjDna,"");
    strcpy(pkgObj_p->pkObjUnitStr,"");
    
    pkgObj_p->pkObjContent_p = NULL;
    
} // end:: pkgMgr_ClearObject()

#ifdef PKGMGR_FUTURE
// --------------------------------------------------------------------
// function : pkgMgr_AddObj()
// purpose  : "grows" a collection of PkgObj such that the final count
//            if newTotalObj, which is expected to be >= current count
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void  pkgMgr_AddObj(PkgTheme *pkgTheme_p,const int newTotalObj)
{

    if(newTotalObj < pkgTheme_p->pkTheme_NobjUsed)
    {
        return ;
    }

    // Alternative method (slightly more efficient)
    // Step 1: Construct an oversize archive, of newTotalObj empty PkgObj's
    // Step 2: Copy (backup/clone) all current PkgObj into the pkgObjNew_p archive
    // Step 3: Carefully re-bind any void pointer referents from old into new.
    // Step 4: Mark all entries beyond PRIOR count as IsUnoccupied
    // Step 5: Carefully re-bind any void pointer referents from old into new.
    // Step 6: Free the original pkgObjects_p array, since we have a copy
    // Step 7: REBIND new ptr to archive to PkgTheme, pkgTheme_p->pkgObjects_p = pkgObjNew_p
    
    // Brute force (slow) method:
    // Step 1: Construct a second archive of all current PkgObj
    // Step 2: Copy all current PkgObj into this new archive
    // Step 3: Carefully re-bind any void pointer referents in old, into new
    // Step 4: Create a NEW PkgObj pkgObj_p array of newTotalObj
    // Step 5: Copy all OLD PkgObj contents from archive into NEW PkgObj archive
    // Step 6: Mark all entries beyong PRIOR count as IsUnoccupied
    // Step 7: Free the original PkgObj pkgObj_p, since it is replaced
    //         using new pkgObjNew_p
    // Step 8: rebind PkgTheme, pkgTheme_p->pkgObjects_p = pkgObjNew_p
    // Step 9: free the archive
    
    
} // end::pkgMgr_AddObj()
#endif


// --------------------------------------------------------------------
// function : pkgMgr_GetNameByIndex()
// purpose  : given a PkgTheme and a base0 internal index, return PkgObj's name
// inputs   : 
// revision : 1.0-2014-08-013T16:58:00,jmg (added)
// author   : joe glassy
// notes    : n/a
char *pkgMgr_GetNameByIndex(PkgTheme *pkgTheme_p,const int internalIndex,char *resultObjName)
{
 PkgObj   *thisObj_p   = NULL;
 int ithObj = 0;
 int nObj   = pkgTheme_p->pkThemeObjCount;
 int hit = 0 ; 
    // initialize
    strcpy(resultObjName,"");  
    
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        thisObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        // match on base0 internal index...
        if( internalIndex == ithObj)
        {
            strcpy(resultObjName,thisObj_p->pkObjKey);
            hit = 1 ;
            break;
        }
    } // end:: ithObj loop
    // return ptr to retrieved name or "" if not found
   return resultObjName;
} // end:: pkgMgr_GetNameByIndex()


// --------------------------------------------------------------------
// function : pkgMgr_GetNameById()
// purpose  : given a PkgTheme and a ID, return the PkgObj's name
// inputs   : 
// revision : 1.1-2020-07-13T18:10, change to int64_t for hid_t compat
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
char *pkgMgr_GetNameById(PkgTheme *pkgTheme_p,const int64_t uniqueId,char *resultObjName)
{
 PkgObj   *thisObj_p   = NULL;
 int ithObj = 0;
 int nObj   = pkgTheme_p->pkThemeObjCount;
 
    // initialize
    strcpy(resultObjName,"");  
    
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        thisObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        if( uniqueId == thisObj_p->pkObjId )
        {
            // do logic here... TBD
            strcpy(resultObjName,thisObj_p->pkObjKey);
            break;
        }
    } // end:: ithObj loop
    // return ptr to retrieved name or "" if not found
   return resultObjName;
} // end:: pkgMgr_GetNameById()


// --------------------------------------------------------------------
// function : pkgMgr_GetIdByName()
// purpose  : given a PkgTheme and a name, return the PkgObj's unique id
// inputs   : PkgTheme *pkgTheme_p -- ptr to theme that owns child
//            const char *targetObjName -- name of child whose ID we've requested
// returns  : int, childs unique integer id
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
int pkgMgr_GetIdByName(PkgTheme *pkgTheme_p,const char *targetObjName)
{
 PkgObj *thisObj_p = NULL;
 int uniqueId = -1 ; // sentinel, no valid id is -1...
 int ithObj = 0;
 int nObj   = pkgTheme_p->pkThemeObjCount;
 
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        thisObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        if(strcmp(targetObjName,thisObj_p->pkObjKey) ==0)
        {
            uniqueId = thisObj_p->pkObjId ;
            break;
        }
    } // end:: ithObj loop
  
    // return ptr to the selected PkgObj owned by PkgTheme
   return uniqueId ;
} // end:: pkgMgr_GetIdByName()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjByName()
// purpose  : retrieves a PkgObj as identified by name, owned by this PkgTheme
// returns  : PkgObj *p --ptr to child object owned by theme identified by name
// revision : 1.1-2020-07-13T13:47, changed target_id to int64_t for hid_t compat.
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
PkgObj *pkgMgr_GetObjByName(PkgTheme *pkgTheme_p,const char *targetObjName)
{
 PkgObj   *pkgObjThis_p   = NULL;
 int     ithObj   = 0;
 int64_t targetId = 0 ;
 int     didFind = 0;
 int     nObj     = pkgTheme_p->pkThemeObjCount; 
 
    // iterate through this PkgTheme's PkgObj collection
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        // set convenience ptr to current PkgObj instance
        pkgObjThis_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        
        // Based on pkgObjId, find its name, to fetch targetId
        targetId = pkgMgr_GetIdByName(pkgTheme_p,targetObjName);
        // exit upon first match
        if(pkgObjThis_p->pkObjId == targetId)
        {
            didFind = 1;
            break;
        }
    } // end:: ithObj loop
  
    if(!didFind)
    {
        xLogMsg("@E In(%s) FAILED to find name(%s) last-ID(%ld) within theme(%s)! ~\n",
            __func__, targetObjName,targetId,pkgTheme_p->pkThemeLabel);
    }
  
    // return ptr to the selected PkgObj owned by PkgTheme
   return pkgObjThis_p ;
   
} // end:: pkgMgr_GetObjByName()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjIndexByName()
// purpose  : retrieves the internal storage INDEX associated with 
//            a child PkgObj, as identified by name, owned by this PkgTheme
// returns  : int, index of PkgObj named 'targetObjName' or -1 if not existing
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
int pkgMgr_GetObjIndexByName(PkgTheme *pkgTheme_p,const char *targetObjName)
{
 PkgObj   *pkgObjThis_p   = NULL;
 int ithObj   = 0;
 int resultIndex = -1 ;
 
 int nObj     = pkgTheme_p->pkThemeObjCount; 
 
    // iterate through this PkgTheme's PkgObj collection
    for(ithObj=0; ithObj< nObj; ithObj++)
    {
        // set convenience ptr to current PkgObj instance
        pkgObjThis_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithObj];
        // search for 'targetObjName' if found stop and return its index
        if(strcmp(pkgObjThis_p->pkObjKey, targetObjName)==0)
        {
            resultIndex = ithObj;
            break;
        }
    } // end:: ithObj loop
  
    // return positive valued index (>=0) or -1 if failed
   return resultIndex;
} // end:: pkgMgr_GetObjIndexByName()


// --------------------------------------------------------------------
// function : pkgMgr_GetObjByIndex()
// inputs   : PkgTheme *pkgTheme_p,const int ithIndex (base0 index)
// purpose  : retrieves a PkgObj as identified by base0 index, as owned by this PkgTheme
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
PkgObj *pkgMgr_GetObjByIndex(PkgTheme *pkgTheme_p,const int ithIndex)
{
 PkgObj   *pkgObjThis_p   = NULL;

 
    if(ithIndex >=0 && ithIndex < pkgTheme_p->pkThemeObjCount)
    {
        pkgObjThis_p = &pkgTheme_p->pkTheme_PkgObjAry_p[ithIndex];
    } else {
        xLogMsg("@E In(%s) FAILED search on index (%d) in theme(%s)! ~\n",
            __func__, ithIndex, pkgTheme_p->pkThemeLabel);
    }
    
 // Return matched object, or NULL if no match (not found)
 return pkgObjThis_p;
} // end:: pkgMgr_GetObjByIndex()

// --------------------------------------------------------------------
// function : pkgMgr_GetObjData()
// inputs   : PkgTheme *pkgTheme_p,const int ithIndex (base0 index)
// purpose  : retrieves a pointer to the actual data bound to a PkgObj
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void *pkgMgr_GetObjData(PkgObj *pkgObj_p)
{
    return pkgObj_p->pkObjContent_p ;
} // end:: pkgMgr_GetObjData()


// --------------------------------------------------------------------
// function: pkgMgr_SetMetadata()
// purpose : assign a generic metadata (string) to object
// inputs  : PkgObj*pkgObj_p,char *pkObjectMetadata_p
// returns : void
// notes   : length of metadata string is limited to POSIX FILENAME_MAX
//           characters.
void pkgMgr_SetObjMetadata(PkgObj*pkgObj_p,char *pkObjectMetadata_p)
{
    strcpy(pkgObj_p->pkObjMetadata, pkObjectMetadata_p);
} // end::  pkgMgr_SetObjMetadata()


// --------------------------------------------------------------------
// function:pkgMgr_GetObjMetadata()
// purpose : retrieve generic metadata (string) from object
// inputs  : PkgObj*pkgObj_p
// returns : char *, string holding metadata, limited to FILENAME_MAX chars
char *pkgMgr_GetObjMetadata(PkgObj*pkgObj_p)
{
    return  pkgObj_p->pkObjMetadata ;
} // end:: pkgMgr_GetObjMetadata()


// ------ NEW ADDITIONS 6/11/2014 --------------------------------------


// --------------------------------------------------------------------
// function:pkgMgr_SetObjShortname()
// purpose : assign CF style shortname (string) to object, max 128-1 char
// inputs  : PkgObj*pkgObj_p
// returns : void
void pkgMgr_SetObjShortname(PkgObj*pkgObj_p,char *newShortname)
{
    strcpy(pkgObj_p->pkObjShortname , newShortname);
} // end:: pkgMgr_SetObjShortname()


// --------------------------------------------------------------------
// function:pkgMgr_GetObjShortname()
// purpose : retrieve a CF style "shortname" from object
// inputs  : PkgObj*pkgObj_p
// returns : char *, string holding 'shortname', limited to 128-1 chars
char *pkgMgr_GetObjShortname(PkgObj*pkgObj_p)
{
    return  pkgObj_p->pkObjShortname ;
} // end:: pkgMgr_GetObjShortname()


// --------------------------------------------------------------------
// function:pkgMgr_SetObjLongname()
// purpose : assign CF style Longname (string) to object, max 128-1 char
// inputs  : PkgObj*pkgObj_p
// returns : void
void pkgMgr_SetObjLongname(PkgObj*pkgObj_p,char *newLongname)
{
    strcpy(pkgObj_p->pkObjLongname , newLongname);
} // end:: pkgMgr_SetObjLongname()


// --------------------------------------------------------------------
// function:pkgMgr_GetObjLongname()
// purpose : retrieve a CF style "Longname" from object
// inputs  : PkgObj*pkgObj_p
// returns : char *, string holding 'Longname', limited to 128-1 chars
char *pkgMgr_GetObjLongname(PkgObj*pkgObj_p)
{
    return  pkgObj_p->pkObjLongname ;
} // end:: pkgMgr_GetObjLongname()


// --------------------------------------------------------------------
// function:pkgMgr_SetObjGenericPath()
// purpose : assign a generic Path (string) to object, max FILENAME_LEN
// inputs  : PkgObj*pkgObj_p
// returns : void
void pkgMgr_SetObjGenericPath(PkgObj*pkgObj_p,char *newGenericPath)
{
    strcpy(pkgObj_p->pkObjGenericPath , newGenericPath);
} // end:: pkgMgr_SetObjGenericPath()


// --------------------------------------------------------------------
// function:pkgMgr_GetObjGenericPath()
// purpose : retrieve a CF style "GenericPath" from object
// inputs  : PkgObj*pkgObj_p
// returns : char *, string holding 'GenericPath', limited to 128-1 chars
char *pkgMgr_GetObjGenericPath(PkgObj*pkgObj_p)
{
    return  pkgObj_p->pkObjGenericPath ;
} // end:: pkgMgr_GetObjGenericPath()

