// --------------------------------------------------------------------
// module   : pkg_mgr_themes.c
// purpose  : prototype implementatin of PkgThemes (middle-layer) 
//            scoping tool services
// revision : 2014-08-13, added  pkgMgr_GetThemes_ChildIndexByName()
// history  : 2014-02-26T19:08:00, added pkgMgr_CreateTheme() and 
//            pkgMgr_ReleaseTheme() to support un-bound themes, and
//            modified pkgMgr_ReleaseAllObjects() to more explicitly 
//            provide a destructor for a theme's PkgObjs
//            and added more bounds,error checking on lookups based
//            on indices or names.
// revision : 2014-02-19T12:58:00,jmg, added pkgThemeReport() prototype
// revision : 2014-02-18T12:58:00,jmg, added initial PkgMg<-->memObj coupling
// revision : 2014-01-15T14:24:00,jmg (original)
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// version  : 1.0
// author   : joe glassy
// assertions : PkgMgr instances must typically define at least (1) PkgTheme, as
//              nothing useful can be done via just a PkgMgr overseer itself.
//              --A PkgTheme may define 0,1 or more PkgObj up to the limit allowed.
//              By default, unless overridden, a PkgTheme creates K PkgObj empty-instances
//              --All PkgThemes have a unique integer ID, >=0, and a text label
//              --All PkgObj have a unique integer ID >=0, and a label
//              --All PkgMgr entities (PkgMgr,PkgTheme,PkgObj) now automatically 
//              store their own DNA signature.
//              --No user data per se is ever directly stored in the PkgMgr complex,
//              since it is simply a theme oriented indexing tool for scoping and provenance
//              e.g. data for any actual variable is always created (and released) by other
//              mechanisms (such as memObj), whereas PkgMgr entities just index these
//              and enable access.
//              --as an indexing service, different PkgThemes or Pkgobj nodes are completely
//              free to reference any given variable. This is analgous to an "SQL View",
//              which enable many different, concurrent "views" of the same information
//              but from discretely different classification schemes.
// dependencies : UM Framework's LOG service (xLogMsg) must be initialized
//               prior, since it is used here.
// notes     :   pkThemeObjCount  ...count of max (candidate) PkgObj available for use
//               pkThemeObjCount ...count of PkgObj actually in use.
// --------------------------------------------------------------------

#include "pkg_mgr.h"
#include "pkg_mgr_prot.h"


// --------------------------------------------------------------------
// function : pkgMgr_InitTheme()
// purpose  : initialize a PkgTheme instance
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// version  : 1.0
// author   : joe glassy
// notes    : Caller may specify nThemes=0 to create these later.
//            This functions does NOT allocate any PkgObj for each 
//            theme.
int pkgMgr_InitTheme(PkgTheme *pkgTheme_p,const int themeId, const int ownerId,
    const char *themeLbl_p,const char *themeCategory_p)
{
    char fxName[256] = {0};
    char uuidBuf[64] = {0};
    
    strcpy(fxName,__func__);
    
    pkgTheme_p->pkThemeId    = themeId   ; //  unique ID of this PkgTheme instance
    pkgTheme_p->pkThemeOwner = ownerId; // id of PkgMgr who owns me
    
    strcpy(pkgTheme_p->pkThemeLabel,themeLbl_p); // Text label identifying this THEME
        
    strcpy(pkgTheme_p->pkThemeUUID,frm_utilGenerateUUID(uuidBuf,1));  // UUID assigned to identify this THEME
    
    strcpy(pkgTheme_p->pkThemeCategory,themeCategory_p);

    // count of max avail PkgObj, of which only a subset may be in use
    pkgTheme_p->pkThemeObjCount = 0; 
    
    // count of actual number of PkgObj in use.
    pkgTheme_p->pkThemeObjCount= 0 ; 
    
    // pointer to array of PkgObj, one per variable
    pkgTheme_p->pkTheme_PkgObjAry_p = NULL   ; 
    
    return 0;
} // end::pkgMgr_InitTheme()


// --------------------------------------------------------------------
// function : pkgMgr_CreateTheme()
// purpose  : Allocates a SINGLE PkgTheme object. This allows for 
//            possibility of an 'anonymous' un-bound PkgTheme not otherwise
//            owned by a PkgMgr overseer.
// inputs   : const char *themeLabel_p -- name of theme.
//            const int themeId -- themes own ID
//            const int themeOwnerId -- optional, ID of owner
//            const char *themeCategory -- optional text label identifying 
//            this category or class of theme.
// returns  : PkgTheme *, ptr to a single PkgTheme instance
// revision : 1.2-2014-06-26T16:20:00, added orig, to support potential
//            use-case of anonymous, un-bound PkgTheme instance.
// history  : n/a
// author   : joe glassy
// notes    : Does NOT create any children PkgObj instances, which must
//            be done separately.
PkgTheme *pkgMgr_CreateTheme(const char *themeLabel_p,
    const int themeId, const int themeOwnerId, const char *themeCategory)
{
 PkgTheme *pThmeNew_p = NULL;   
 char uuidBuffer[64] = {0} ;
 char fxName[256] = {0} ;
 char thLbl[256] ;
 
 
    strcpy(fxName,__func__);
    
    // create a new single, un-bound instance of a PkgTheme
    pThmeNew_p= memCalloc(1, sizeof(PkgTheme));
    
    // assign internal memObj name 
    strcpy(thLbl, themeLabel_p);
    // pass it a non-const label, to avoid complaint 
    if(strlen(themeLabel_p)>= 1)
        memObjSetName(pThmeNew_p,thLbl);
        
    // Assign it a presumably unique ID
    pThmeNew_p->pkThemeId  = themeId;
    // Assign owner ID exactly as supplied
    pThmeNew_p->pkThemeOwner = themeOwnerId ;
    // Assign its self-referential DNA
    sprintf(pThmeNew_p->pkThemeDna,"%p",pThmeNew_p);
    // Reset its count of (child) PkgObj as 0
    pThmeNew_p->pkThemeObjCount =0 ;
    // Set theme level filter to 1 as default to allow use
    pThmeNew_p->pkThemeSelectFilter = 1;
    
    // Supply with a default new unique UUID...
    // char *frm_utilGenerateUUID(char *resultUUID_String,const int displayMode);
    strcpy(pThmeNew_p->pkThemeUUID,frm_utilGenerateUUID(uuidBuffer,1));
    
    // set category label, if supplied
    if(strlen(themeCategory) > 0)
        strcpy(pThmeNew_p->pkThemeCategory,themeCategory); 
    
    // Return ptr to this new single instance of (semi-initialized) PkgTheme
    return pThmeNew_p ;
    
} // end:: pkgMgr_CreateTheme()


// --------------------------------------------------------------------
// function : pkgMgr_CreateThemes()
// purpose  : For one PkgMgr parent, this routine allocates a fixed
//            size collection of nNewThemes PkgTheme objects,
//            which are typically later bound to a PkgMgr owner, and
//            also, if nCountOfPkgObjs[] array is not null and contains
//            valid count of PkgObj, a set of empty PkgObj per theme , are
//            also allocated.
//
// inputs   : PkgMgr *pkgMgr_p,  -- ptr to parent PkgMgr
//            const int nThemes, -- count of new theme containers to create
//            const int *nPkgObjList, int array of size (nThemes) storing counts of PkgObj per theme
//            const char *themeCategory -- text label identifying this theme.
//
// returns  : PkgTheme *, ptr to fixed size array of PkgTheme instances
// revision : 2014-07-01, now traverse nCountOfPkgObjs[] array to minimally
//            validate that it is nNewThemes long
// revision : 2014-06-26T15:38, added pre-check to see if array of
//            PkgTheme already exists. Block attempt to overwrite this!
// revision : 2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : The "nCountOfPkgObjs" array is consulted, to guide the
//            creation of N PkgObj owned by this theme.
PkgTheme *pkgMgr_CreateThemes(PkgMgr *pkgMgr_p,const int nNewThemes,
                const int *nCountOfPkgObjs,const char *themeCategory)
{
 int ithTheme= 0;
 int nThemes = 0; 
 unsigned long thrdId = 0; 
 int  t=0 , p =0,pid=0;
 size_t chkSum =0 ;
 char fxName[256] = {0} ;
 
 char keywordBuffer[256] = {0} ;
 char thisHost[256]      = {0} ;
 char autoLabel[256] = {0} ;
 char uuidBuf[64] = {0} ;
 
 // char thisPID[32] = {0} ;
 PkgTheme *pThm_p  = NULL;
 PkgTheme *pkgThemeAry_p = NULL;
 PkgObj   *pObj_p = NULL;

    strcpy(fxName,__func__);
    
    // <JMG> 2014-07-01, Relexed hard constraint on max n. themes
    // but we do now impose sanity check on the count
    assert(nNewThemes >= 1);
    
    // Traverse the nCountOfPkgObjs[] array to assure it is nNewThemes 
    // long, and warn of any 0 counts. We WANT this to catostrophically
    // fail as early as possible if the array is not correctly sized.
    for(t=0; t< nNewThemes; t++)
        chkSum += (size_t)nCountOfPkgObjs[t] ;
        
    // store as a non-const too
    nThemes = nNewThemes ;
    
    // First check if the array of PkgThemes may already exist
    // and if it does, refuse to accidentally overwrite it
    // <JMG> 2014-06-26, added pre-check!
    if(pkgMgr_p->pkMgrThemeAry_p != NULL)
    {
        // re-evaluate if this should be a fatal error "@F"
        xLogMsg("@E In(%s): WARNING PkgTheme array appears to already exist!~\n");
        return NULL;
    }

    xLogMsg("@A In(%s): Creating %d new themes with a total of %zu PkgObj~\n",
        fxName,nNewThemes,chkSum);

    // Allocate memory for a fixed size array of PkgTheme's
    pkgThemeAry_p = memCalloc(nNewThemes, sizeof(PkgTheme));
    //  BIND (attach) new array PkgTheme's back to owner
    pkgMgr_p->pkMgrThemeAry_p = pkgThemeAry_p ;
    // Update internal count of the themes just created
    pkgMgr_p->pkMgrThemeCount = nNewThemes;
    
    // for now, auto label the array-of-themes object
    sprintf(autoLabel,"%s.ArrayOfThemes.Category_%s", 
            pkgMgr_p->pkMgrLabel,  themeCategory );
    // Name the array-of-themes, may be overriden later
    memObjSetName(pkgThemeAry_p,autoLabel);
        
    // For thread-safe identification, fetch my host and pid
    // and later, will fill in thread id
    pid = getpid();
    gethostname(thisHost,255) ;
    
    // After array of PkgTheme's created, initialize each
    //  to reflect they exist, but are not yet in use
    xLogMsg("@A In(%s): initializing %d empty themes\n",fxName,pkgMgr_p->pkMgrThemeCount);
    
    for(ithTheme=0; ithTheme < pkgMgr_p->pkMgrThemeCount; ithTheme++)
    {
        // fetch pointer to fixed size, flat array-of-structures PkgTheme
        pThm_p = &pkgThemeAry_p[ithTheme];
        
        pThm_p->pkThemeId    = ithTheme; // assign some meaningful ID
        pThm_p->pkThemeOwner = pkgMgr_p->pkMgrId ; // mark with ID of owner

        // provide each theme with a auto-assigned UUID 
        strcpy(pThm_p->pkThemeUUID,frm_utilGenerateUUID(uuidBuf,1)); 
        
        // Note: 'pkThemeDna' signature identifies this 
        // specific 'child' theme node among the collection itself,
        // it is therefore a distinct signature of the one identifying the
        //  top|head (pkgThemeAry_p) container itself.
        sprintf(pThm_p->pkThemeDna,"%p",pThm_p);
        
        // Construct and store my fully-qualified identifer: {host.pid.thread.addr}
        sprintf(pThm_p->pkThemeHostPidThrdAddrId,"%s.%d.%lu.%s",
            thisHost, pid, thrdId, pThm_p->pkThemeDna);
            
        // if a valid nCountOfPkgObjs[] array passed in, use it to 
        // define the number of PkgObj this (ithTheme) will own.
        if(nCountOfPkgObjs != NULL)
            pThm_p->pkThemeObjCount = nCountOfPkgObjs[ithTheme];

        // WARN if the PkgObj array ALREADY appears to exist, to block accidental
        // attempts to over-write a pre-existing collection of PkgObj
        if(pThm_p->pkTheme_PkgObjAry_p != NULL)
        {
            xLogMsg("@F In(%s) ERROR PkgObj cannot re-create prior existing array-of-structures!~\n",
            fxName);
            return NULL;
        }
        
        // Explicitly MARK that the child nodes don't exist yet     
        pThm_p->pkTheme_PkgObjAry_p = NULL; 
        
        // for now, auto label each theme
        sprintf(autoLabel,"%s.Theme-%d",pkgMgr_p->pkMgrLabel, ithTheme);
        
        strcpy(pThm_p->pkThemeLabel, autoLabel);
        strcpy(pThm_p->pkThemeCategory,themeCategory); // category

        // If a valid integer array of counts in nCountOfPkgObjs[] (the n. of PkgObj per theme)
        //   was submitted, use it to establish the array of PkgObj this theme owns, 
        //   then initialize these.
        // <JMG> added 2/20/2014,must not attempt to allocate 0 PkgObj!!
        if(nCountOfPkgObjs != NULL  && nCountOfPkgObjs[ithTheme] > 0 )
        {
            //  Create one-time-only fixed size PkgObj array-of-structures of 
            pThm_p->pkTheme_PkgObjAry_p =(PkgObj *)memCalloc( nCountOfPkgObjs[ithTheme],
                                                        sizeof(PkgObj));
            sprintf(autoLabel,"Theme_%s_PkgObj_CollectionOf_%d",
                pThm_p->pkThemeLabel, nCountOfPkgObjs[ithTheme] );
                
            // Supply a low-level label for the (parent) of allocated object
            memObjSetName(pThm_p->pkTheme_PkgObjAry_p, autoLabel);
            
            // Next, iterate through each PkgObj, marking it with its parents IDs etc.
            //   auto-skip OVER any themes which do not yet own any PkgObj
            //   noting they may own some later
            
            // INDICES: ithTheme --> index of themes 0..n-1
            //          p        --> index of PkgObj 0..n-1
            for(p=0; p < pThm_p->pkThemeObjCount ; p++) 
            {
                // Attach pointer to (p-th) member of the flat array-of-structures
                pObj_p = &pThm_p->pkTheme_PkgObjAry_p[p] ;
                
                // Construct a temporary keyword for this PkgObj (override later)
                
                // NOTE: to later access THIS theme's PkgObj, until it is properly
                // named, we typically must access it via the PkgObj's slot index,
                // since this auto-gen 'keyword' will not be known by the caller
                sprintf(keywordBuffer,"%s_PkgObjDefaultKey_%d",themeCategory,p+1);
                
                // Initialize as much of the PkgObj as we can here...
                // { obj_p,keyword,objID, ownerId,TIVdatatype,categoryCode=-1, rank=0, dims[] }
                pkgMgr_InitObj(pObj_p, keywordBuffer, p+1, pThm_p->pkThemeId,
                    IsNoType, -1, 0, NULL);
                    
                // No actual data variable is bound to this PkgOgj yet so set to NULL
                pObj_p->pkObjContent_p = NULL;
                // Default populate long_name with a UUID until it is overridden
                strcpy(pObj_p->pkObjLongname, frm_utilGenerateUUID(uuidBuf,1));
                    
            } // end:: p loop through the collection of this theme's PkgObj objects
            
        } else { // end:: block if nCountOfPkgObjs[] array was valid
            xLogMsg("@W ~pkgMgr_CreateThemes: WARNING PkgObj count-array NULL or count=0 for theme(%d)\n",
                ithTheme);
        }
    
    } // end:: ithTheme loop

    // Return ptr to head of new, initialized PkgTheme array
   return pkgThemeAry_p ;
   
} // end:: pkgMgr_CreateThemes()


// --------------------------------------------------------------------
// function : pkgMgr_CreatePkgObjArray(n)
// purpose  : Given one existing theme, allocate a fixed size collection 
//            of nNew PkgObj objects for this one theme. Do NOT use
//            this approach is pkgMgr_CreateThemes() was used as 
//            supplied with a count array to already create PkgObj's...
// inputs   : PkgTheme *thisTheme_p -- ptr to theme to work with
//            const int nNewPkgObj -- count of child PkgObj to create
//            for this theme.
// returns  : PkgObj *, ptr to fixed size array of PkgObj
// revision : 2014-07-01T10:26:00-07:00
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy, Lupine Logic Inc.
// notes    : If the theme already had an active set of PkgObj, we now
//            refuse to allow user to create a new set over-writing
//            the prior set. This forces the user to explicitly
//            release any prior set of PkgObj first.
PkgObj *pkgMgr_CreateThemesPkgObjArray(PkgTheme *pThm_p,
            const int nNewPkgObj)
{
 PkgObj *pkgObj_p = NULL;
 int ithObj =0;
 //  int iLen= strlen(themeCategory)  ...placeholder...
 char idBuffer[64]   = {0} ;
 char autoLabel[256] = {0} ;    
 int  nPriorCount = 0 ;
 
    // Before doing ANY of this, check count of PkgObj that
    // (may) already exist!
    nPriorCount = pkgMgr_GetThemeObjectCount(pThm_p);
    
    // REFUSE to create a new set of PkgObj over the top of any 
    // prior existing set of PkgObj. Force user to release them first...
    if( nPriorCount > 0)
    {
        fprintf(stderr,
            "@E pkgMgrCreateThemesPkgObjArray: %d Prior PkgObj exist and still active!~\n",
            nPriorCount);
        return pkgObj_p; // return NULL
    }
    
    // Create fixed size array of (nNewPkgObj) PkgObj instances
    //   as owned by THIS theme, pkgTheme_p.
    if( nNewPkgObj > 0)
    {
        pThm_p->pkTheme_PkgObjAry_p = memCalloc(nNewPkgObj,sizeof(PkgObj));
        // mark with: id, owner, label, categoryDescriptor
        sprintf(autoLabel,"%s.pkgObjArrayOf-%d ~\n",
                pThm_p->pkThemeLabel, nNewPkgObj );

        // add info for tracing later if necessary
        memObjSetName(pThm_p->pkTheme_PkgObjAry_p, autoLabel);
        memObjSetDatatype(pThm_p->pkTheme_PkgObjAry_p, IsCustomUserType);
    
        // Next, initialize each
        for(ithObj =0; ithObj < nNewPkgObj; ithObj++)
        {
                pkgObj_p = &pThm_p->pkTheme_PkgObjAry_p[ithObj];
                sprintf(idBuffer,"PkgObj.placeholder_key_%4.4d",ithObj+1);
                
                // Init can only accomodate so many properties. Assign
                // others using dedicated service functions
                pkgMgr_InitObj(pkgObj_p, 
                    idBuffer, 
                    ithObj+1,   // pkgObj's ID
                    pThm_p->pkThemeId, // owner of pkgObj
                    -1,    // datatype, not known or specified yet
                    -1,    // category, not known or specified yet
                    -1,    // rank, not yet known
                    NULL); // dims, not yet known
                    
        } // end:: ithObj loop
        
        // <JMG> 6/11/2014, update this theme's PkgObj count!
        pkgMgr_SetThemeObjectCount(pThm_p,nNewPkgObj);
        
    } else {
        fprintf(stderr,"@W pkgMgrCreateThemesPkgObjArray: NO PkgObj requested yet~\n");
    }
    
    // returns pointer to the flat array-of-structures as a convenience
    // note: in normal use, we don't normally work with the 'head'
    // array of PkgObj structures, but rather, fetch one at a one
    return pThm_p->pkTheme_PkgObjAry_p ;
    
} // end::pkgMgr_CreateThemesPkgObjArray()
    
// --------------------------------------------------------------------
// function : pkgMgr_ReleaseAllObjects()
// purpose  : free up ALL PkgObj owned by a PkgTheme, by first
//            clearing the member values then releasing memory.
// revision : 2014-06-11T18:16:00, changed free() to memFree()
// revision : 2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_ReleaseAllObjects(PkgTheme *pkgTheme_p)
{
  int nPkgObjCount = pkgMgr_GetThemeObjectCount(pkgTheme_p);

    if( nPkgObjCount >0)
    {
        // NOTE: mirror destructor logic with constructor so
        // release the collection all at once. 
        // TBD: when later converted to array of pointers to PkgObj,
        // we will then release one at a time...
        memFree(pkgTheme_p->pkTheme_PkgObjAry_p);
        // Explicitly reset its count of PkgObj to 0
        pkgMgr_SetThemeObjectCount(pkgTheme_p,0);
    } else {
        fprintf(stderr,
        "@E pkgMgr_ReleaseAllObjects: %d count is invalid!~\n",
            nPkgObjCount);
    }
 
} // end:: pkgMgr_ReleaseAllObjects()
        

// --------------------------------------------------------------------
// function : pkgMgr_ReleaseTheme()
// purpose  : Releases memory for ALL PkgTheme's owned by this PkgMgr
// inputs   : PkgTheme *pThme_p -- ptr to an individual PkgTheme to release
// revision : 2014-06-26T16:50:00, orig, added to support anonymous
//            un-bound themes which may occur as singletons, not 
//            owned by any particular PkgMgr overseer.
// history  : n/a
// author   : joe glassy
// notes    : n/a
void pkgMgr_ReleaseTheme(PkgTheme *pThme_p)
{
    
    if(pThme_p != NULL)
    {
        // If children exist (normally they do) release them first
        if(pThme_p->pkTheme_PkgObjAry_p != NULL)
        {
            memFree(pThme_p->pkTheme_PkgObjAry_p);
        }
        // Finally, release the enclosing parent object instance
        memFree(pThme_p);
    } // end:: if current PkgTheme appears to exist ok
    
} // end:: pkgMgr_ReleaseTheme()



// --------------------------------------------------------------------
// function : pkgMgr_ReleaseThemes()
// purpose  : Releases memory for ALL PkgTheme's owned by this PkgMgr
// revision : 2014-06-11T18:49:00, changed to only release PkgTheme and
//            PkgObj memory and leave overseer PkgMgr intact
// revision : 2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Improved robustness by only releasing PkgObj and PkgTheme
//            memory.
void pkgMgr_ReleaseThemes(PkgMgr *pkgMgr_p)
{
 int ithTheme= 0;
 PkgTheme *pThm_p = NULL;

    
    // For each defined theme, first free the array of PkgObj it owns
    for(ithTheme=0; ithTheme < pkgMgr_p->pkMgrThemeCount ; ithTheme++)
    {
        pThm_p = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
        
        if(pThm_p->pkThemeObjCount >0 && pThm_p->pkTheme_PkgObjAry_p !=NULL )
        {
            xLogMsg("@A pkgMgr_ReleaseThemes: Theme %d(%s) now releasing %d PkgObj!~\n",
            ithTheme, pThm_p->pkThemeLabel, pThm_p->pkThemeObjCount);
            
            memFree(pThm_p->pkTheme_PkgObjAry_p);
            
            pThm_p->pkThemeObjCount = 0 ;
            
        } else {
            xLogMsg("@W pkgMgr_ReleaseThemes: Theme %d(%s) PkgObj count %d so cannot release PkgObj!~\n",
            ithTheme, pThm_p->pkThemeLabel, pThm_p->pkThemeObjCount);           
        }
    }

    // <JMG> 2014-06-30t16:07, added joint test on positive ThemeCount and 
    // that array is actually not NULL prior to attempting free
    if(pkgMgr_p->pkMgrThemeCount > 0 && pkgMgr_p->pkMgrThemeAry_p != NULL)
    {
        xLogMsg("@A pkgMgr_ReleaseThemes: now releasing (%d) themes~\n",
            pkgMgr_p->pkMgrThemeCount);
        // Next, free the middle element, the fixed
        //   size PkgTheme array owned by the PkgMgr overseer
        memFree(pkgMgr_p->pkMgrThemeAry_p);
        pkgMgr_p->pkMgrThemeCount = 0 ;
    }

    // NOTE: a separate release function must be used to release the
    // memory used by the overseer PkgMgr itself
   
} // end:: pkgMgr_ReleaseThemes()


// --------------------------------------------------------------------
// function : pkgMgr_ThemeReport()
// purpose  : Display all PkgThemes
// inputs   : int pkgMgr_ThemeReport(PkgMgr *pkgMgr_p,const int opSwitches,FILE *outf)
// revision : 2020-07-13T17:46, maint only
// revision : 2014-02-19T12:58:00,jmg
// author   : joe glassy
// notes    : option switches: {1,2,3}: 1=PkgMgr level only,
//            2=PkgTheme level only, 3=PkgMgr,PkgTheme,PkgObj levels all included
// notes    : 2/19/2014,This is just an initial stub, and the other reporting levels
//            are not yet implemented, but are planned for soon to follow
//            milestone.
int pkgMgr_ThemeReport(PkgMgr *pkgMgr_p,const int opSwitches,FILE *outf)
{
 int      ithTheme =0;
 PkgTheme *pThm_p = NULL;

    // PkgMgr level
    // PkgTheme level
    for(ithTheme =0; ithTheme < pkgMgr_p->pkMgrThemeCount ; ithTheme++)
    {
        pThm_p = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
        // format: id,owner,label,uuid,dna,memRegDna,hpta,cat
        // 
        fprintf(outf,"%d,%d,%s,%s,%s,%s,%s,%s\n",
            pThm_p->pkThemeId,pThm_p->pkThemeOwner,pThm_p->pkThemeLabel,
            pThm_p->pkThemeUUID, pThm_p->pkThemeDna, pThm_p->pkThemeMemRegDna,
            pThm_p->pkThemeHostPidThrdAddrId,pThm_p->pkThemeCategory);
    } // end:: ithTheme loop
    
    // placeholder return
    return opSwitches; 
    
} // end:: pkgMgrThemeReport()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeByIndex()
// purpose  : retrieves a PkgTheme by base0 index, as owned by PkgMgr
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// returns  : PkgTheme --ptr to discovered package theme object, or NULL if failed.
// author   : joe glassy
// notes    : The referent must have been created prior, and its properties
//            must already be known to this pkgMgr_p.
PkgTheme *pkgMgr_GetThemeByIndex(PkgMgr *pkgMgr_p,const int thisThemeIndex)
{
    PkgTheme *pkgTheme_p = NULL;
    int ithTheme;
    
    for(ithTheme =0; ithTheme < pkgMgr_p->pkMgrThemeCount ; ithTheme++)
    {
        pkgTheme_p = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
        
        if(thisThemeIndex == ithTheme)
            break;
    } // end:: ithTheme loop
 
 return pkgTheme_p ;
 
} // end:: pkgMgr_GetThemeByIndex()
 
 
// --------------------------------------------------------------------
// function : pkgMgr_GetThemeByName()
// purpose  : retrieves a PkgTheme, by name
// returns  : PkgTheme *p --ptr to matched item or NULL if not found
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// returns  : PkgTheme_p -- ptr to found PkgTheme or NULL if failed.
// author   : joe glassy
// notes    : The referent must have been created prior, and its properties
//            must already be known to this pkgMgr_p.
PkgTheme *pkgMgr_GetThemeByName(PkgMgr *pkgMgr_p,const char *themeName_p)
{
    PkgTheme *pkgTheme_p = NULL;
    int ithTheme;
    int hit = 0 ;
    
    for(ithTheme =0; ithTheme < pkgMgr_p->pkMgrThemeCount; ithTheme++)
    {
        pkgTheme_p = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
        
        if(strcmp(pkgTheme_p->pkThemeLabel, themeName_p)==0)
        {
            hit = 1;
            break;
        }
    } // end:: ithTheme loop
    
    if(!hit)
    {
        xLogMsg("@E In(%s) Theme(%s) NOT found in pkgMgr(%s)~\n",
            __func__, themeName_p, pkgMgr_p->pkMgrLabel);
    }
 
    return pkgTheme_p ;
} // end:: pkgMgr_GetThemeByName()


// --------------------------------------------------------------------
// function : pkgMgr_SetThemeCategory()
// purpose  : assign a category label to a theme
// inputs   : PkgTheme *pkgTheme_p,const char *newCategory
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// returns  : void
// author   : joe glassy
// notes    : n/a
void pkgMgr_SetThemeCategory(PkgTheme *pkgTheme_p,char *fmtStr_p,...)
{
  char pkgThemeCategoryString[FILENAME_MAX] = {0} ;
  
  va_list argptr;
  int     fmt_count =0;
  
    va_start(argptr,fmtStr_p);
    fmt_count = vsprintf(pkgThemeCategoryString, fmtStr_p,argptr);
    va_end(argptr);
    // store the sprintf() style formatted label now
    strcpy(pkgTheme_p->pkThemeCategory, pkgThemeCategoryString);

} // end:: pkgMgr_SetThemeCategory()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeCategory()
// purpose  : retrieves a text category label for a theme (PkgTheme)
// inputs   : PkgTheme *pkgTheme_p
// returns  : char * pointer to text category label of the theme
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
char *pkgMgr_GetThemeCategory(PkgTheme *pkgTheme_p)
{
    return pkgTheme_p->pkThemeCategory;
} // end:: pkgMgr_GetThemeCategory()


// --------------------------------------------------------------------
// function : pkgMgr_SetThemeName()
// purpose  : assign a text label or name to a PkgTheme object
// inputs   : PkgTheme *pkgTheme_p,const char *newThemeName
// returns  : void
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_SetThemeName(PkgTheme *pkgTheme_p,char *fmtStr_p,...)
{
  char pkgThemeNameString[FILENAME_MAX] = {0} ;
  
  va_list argptr;
  int     fmt_count =0;
  
    va_start(argptr,fmtStr_p);
    fmt_count = vsprintf(pkgThemeNameString, fmtStr_p,argptr);
    va_end(argptr);
    // load formatted name string into PkgTheme label or name area
    strcpy(pkgTheme_p->pkThemeLabel, pkgThemeNameString);
 
} // end::pkgMgr_SetThemeName()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeName()
// purpose  : get a theme's name
// inputs   : PkgTheme *pkgTheme_p
// returns  : char * pointer to theme's name
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
char *pkgMgr_GetThemeName(PkgTheme *pkgTheme_p)
{
    return pkgTheme_p->pkThemeLabel ;
    
} // end::pkgMgr_GetThemeName()


// --------------------------------------------------------------------
// function : pkgMgr_SetThemeObjectCount()
// purpose  : Set the count of PkgObj owned by THIS theme
// inputs   : PkgTheme *pkgTheme_p,const int newPkgObjCount
// returns  : int , count of PkgObj owned
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void pkgMgr_SetThemeObjectCount(PkgTheme *pkgTheme_p,const int newPkgObjCount)
{
    pkgTheme_p->pkThemeObjCount =  newPkgObjCount ;
    
} // end::pkgMgr_SetThemeObjectCount()


// --------------------------------------------------------------------
// function : pkgMgr_GetThemeObjectCount()
// purpose  : get the count of PkgObj owned by THIS theme
// inputs   : PkgTheme *pkgTheme_p
// returns  : int , count of PkgObj owned
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
int pkgMgr_GetThemeObjectCount(PkgTheme *pkgTheme_p)
{
    return pkgTheme_p->pkThemeObjCount  ;
    
} // end::pkgMgr_GetThemeObjectCount()


// --------------------------------------------------------------------
// function:  pkgMgr_SetThemeMetadata()
// purpose : assign generic metadata string to theme
// inputs  : PkgThem *pThme_p --ptr to theme to work on
//           char *pkThemeMetadata_p --ptr to string to load as metadata
// returns : void
// revised : 2014/02/25
void pkgMgr_SetThemeMetadata(PkgTheme *pThme_p,char *pkThemeMetadata_p)
{
    strcpy(pThme_p->pkThemeMetadata, pkThemeMetadata_p);
} // end::  pkgMgr_SetThemeMetadata()


// --------------------------------------------------------------------
// function:pkgMgr_GetThemeMetadata()
// purpose : retrieve generic metadata string from theme
// inputs  : PkgThem *pThme_p --ptr to theme to work on
// returns : char *ptr to metadata string
// revised : 2014/02/25
char *pkgMgr_GetThemeMetadata(PkgTheme *pThme_p)
{
    return  pThme_p->pkThemeMetadata ;
} // end:: pkgMgr_GetThemeMetadata()


// --------------------------------------------------------------------
// function : pkgMgr_BindObjByName()
// purpose  : Bind raw data to a PkgObj owned by a theme, where the
//            PkgObj is identified by its name (keyword)
// inputs   : PkgTheme *pkgTheme_p,const char *keyword,
//              void *rawData_p
// returns  : int , index of PkgTheme's PkgObj slot matched or -1 if failed
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : Critical that the PkgObj name (keyword) must already have 
//            assigned via pkgMgr_SetThemes_ChildName() or equiv
int pkgMgr_BindObjByName(PkgTheme *pkgTheme_p,const char *keyword,void *rawData_p)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 
        for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
        {
            // fetch pointer to pkgObj
            pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];

            if( strlen(pkgObj_p->pkObjKey) > 0 && 
                strcmp(keyword, pkgObj_p->pkObjKey)== 0 )
            {
                // MATCH FOUND, BIND the raw data to this slot
                pkgObj_p->pkObjContent_p = rawData_p ;
                break;
            }
        } // end:: i loop through pkgObj's owned by this PkgTheme
        
        if( i < 0)
        {
            xLogMsg("@E %s Theme(%s) ERROR No match on PkgObj name %s!~\n",
            (char *)__func__ , pkgTheme_p->pkThemeLabel, keyword);
        }
 // returns base0 index of the slot matched by this name, or -1 if failed
 return i ;
} // end::pkgMgr_BindObjByName()


// ---------------------------------------------------------------------
// function: pkgMgr_SetThemes_ChildMetadata()
// Purpose : assign new metadata to a child PkgObj at index pkgObjIdx
// Inputs  : PkgTheme *pkgTheme_p,const int pkgObjIdx,char *newMetadata_p
// notes   : n/a
void pkgMgr_SetThemes_ChildMetadata(PkgTheme *pkgTheme_p,const int pkgObjIdx,char *newMetadata_p)
{
 PkgObj *pkgObj_p = NULL;
 
    // always make sure pkgObjIdx submitted is in a valid range
    if( pkgObjIdx >=0 && pkgObjIdx < pkgTheme_p->pkThemeObjCount )
    {
        // using this index , fetch the child PkgObj 
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[pkgObjIdx];
        // now, set its metadata
        pkgMgr_SetObjMetadata(pkgObj_p,newMetadata_p);
    } else {
        xLogMsg("@E In(%s) Theme(%s) ERROR child index (%d) is out of range!~\n",
            (char *)__func__ , pkgTheme_p->pkThemeLabel, pkgObjIdx);
    }
    
} // end:: pkgMgr_SetThemes_ChildMetadata()


// ---------------------------------------------------------------------
// function : pkgMgr_GetThemes_ChildMetadata()
// Purpose  : get metadata from a child PkgObj at index pkgObjIdx, or 
//            NULL if not available.
char *pkgMgr_GetThemes_ChildMetadata(PkgTheme *pkgTheme_p,const int pkgObjIdx)
{
 PkgObj *pkgObj_p = NULL;
 char *metadata_p = NULL;
    // always make sure pkgObjIdx submitted is in a valid range
    if( pkgObjIdx >=0 && pkgObjIdx < pkgTheme_p->pkThemeObjCount )
    {
        // using this index , fetch the child PkgObj 
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[pkgObjIdx];
        // now, return its metadata
        metadata_p = pkgObj_p->pkObjMetadata ;
    } else {
        xLogMsg("@E In(%s) ERROR Theme(%s) pkgObj index (%d) is out of range!~\n",
            (char *)__func__ , pkgTheme_p->pkThemeLabel, pkgObjIdx);
    }
    
    return metadata_p;
} // end:: pkgMgr_GetThemes_ChildMetadata()


// ---------------------------------------------------------------------
// function: pkgMgr_SetThemes_ChildName
// Purpose : assign new name 'newObjname_p' to a child PkgObj at index pkgObjIdx
// inputs  : PkgTheme *pkgTheme_p,const int pkgObjIdx,char *newObjname_p
// returns : void
// notes   : Errors are treated here as recoverable
void pkgMgr_SetThemes_ChildName(PkgTheme *pkgTheme_p,const int pkgObjIdx,char *newObjname_p)
{
 PkgObj *pkgObj_p = NULL;
 
    // always make sure pkgObjIdx submitted is in a valid range
    if( pkgObjIdx >= 0 && pkgObjIdx < pkgTheme_p->pkThemeObjCount )
    {
        // using this validated index , fetch the child PkgObj from array of PkgObj
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[pkgObjIdx];
        // now, set its name
        pkgMgr_SetObjName(pkgObj_p,newObjname_p);
    } else {
        xLogMsg("@E Fx(%s) ERROR Theme <%s> Child(%s) has an Invalid OR out-of-range index (%d) or count(%d)!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, newObjname_p, pkgObjIdx, pkgTheme_p->pkThemeObjCount);
    }
    
} // end:: pkgMgr_SetThemes_ChildName()


// ---------------------------------------------------------------------
// function : pkgMgr_GetThemes_ChildName()
// purpose  : retrieve the name of the child PkgObj located at internal
//            base0 index, pkgObjIdx
// inputs   : PkgTheme *pkgTheme_p,
//            const int pkgObjIdx
// revised  : 2014-08-14T11:03
// returns  : ptr to childs existing name located this index 'pkgObjIdx'
// notes    : n/a
char *pkgMgr_GetThemes_ChildName(PkgTheme *pkgTheme_p,const int pkgObjIdx)
{
PkgObj *pkgObj_p = NULL;
char *resultName_p = NULL;
 
    // enforce that the index, pkgObjIdx, submitted is within valid range
    if( pkgObjIdx >=0 && pkgObjIdx < pkgTheme_p->pkThemeObjCount )
    {
        // using this index , fetch the child PkgObj 
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[pkgObjIdx];
        // match, so retrieve the child's name
        resultName_p = pkgMgr_GetObjName(pkgObj_p);
    } else {
        xLogMsg("@E In(%s) ERROR Theme(%s) child at Index(%d) is Invalid OR out-of-range index or count(%d)!~\n",
            (char *)__func__ , pkgTheme_p->pkThemeLabel, pkgObjIdx, pkgTheme_p->pkThemeObjCount);
    }
    
    // return pointer to string, the PkgObj childs name
    return resultName_p;
    
} // end:: pkgMgr_GetThemes_ChildName()


// ---------------------------------------------------------------------
// function : pkgMgr_DoesThemes_ChildNameExist()
// purpose  : Test if a theme's child name exists, e.g this name is
//            owned by and registered to the theme
//            base0 index, pkgObjIdx
// inputs   : PkgTheme *pkgTheme_p,
//            const char *targetChildName_p -- name of child's name to test
// revised  : 2014-09-12T10:48
// returns  : int, boolean, 1= name "ChildNAME" does exist in theme, 0=does NOT exist
// notes    : n/a
int pkgMgr_DoesThemes_ChildNameExist(PkgTheme *pkgTheme_p,const char *targetChildName_p)
{
PkgObj *pkgObj_p = NULL; // ptr to a given child owned by the theme
char *thisName_p = NULL;
int  idx, doesChildExist = 0 ;

    for(idx=0; idx < pkgTheme_p->pkThemeObjCount ; idx++)
    {
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[idx];
        // retrieve the child's name in order to test a match on it
        thisName_p = pkgMgr_GetObjName(pkgObj_p);
        if(strcmp(thisName_p,targetChildName_p)==0)
        {
            doesChildExist = 1;
            break;
        }
    } 
    // return boolean. 1=DOES exist, 0=does NOT exist
    return doesChildExist;
} // end:: pkgMgr_DoesThemes_ChildNameExist()


// ---------------------------------------------------------------------
// function : pkgMgr_GetThemes_ChildListOfNames()
// purpose  : retrieve a stringtable contains a list of ALL child names
//            owned by theme pkgTheme.
// inputs   : PkgTheme *pkgTheme_p -- ptr to theme owning K children
// revised  : 2014-08-14T11:03
// returns  : ptr to childs existing name located this index 'pkgObjIdx'
// notes    : TBD: Later, extend to accept input startIndex,endIndex to
//            allow user to retrieve just a subset of names, or when
//            startIndex and endIndex are both -1, causes a list of ALL
//            names to be returned.
//
char **pkgMgr_GetThemes_ChildListOfNames(PkgTheme *pkgTheme_p)
{
 PkgObj *pkgObj_p = NULL;
 char **resultStrTblList_p = NULL; // ptr to resulting list of ALL child names
 char fxName[256] = {0} ;
 int idx =0;
 
    // step 1: allocate a string table to hold a list of ALL child names
    strcpy(fxName,__func__);
    
    if(pkgTheme_p->pkThemeObjCount < 1)
    {
        xLogMsg("@E In(%s) ERROR Theme(%s) invalid count (%d) of child PkgObj! ~ \n",
            (char *)fxName, pkgTheme_p->pkThemeLabel, pkgTheme_p->pkThemeObjCount);
        return NULL;
    }
    
    resultStrTblList_p = memObjCalloc(pkgTheme_p->pkThemeLabel,
        PKGMGR_MAX_NAMELEN+2, pkgTheme_p->pkThemeObjCount, IsStringTable,fxName,fxName);
    
    // Do NOT use the quick-and-dirty method memCreateStringTbl() here when 
    // we want to name this string table!
    
    // resultStrTblList_p = memCreateStringTbl(pkgTheme_p->pkThemeObjCount,PKGMGR_MAX_NAMELEN+2);
    //   by default, give this memObj the name of the theme
    // WARNING: memObjSetName is NOT yet configured to name a stringtable
    //memObjSetName(resultStrTblList_p, pkgTheme_p->pkThemeLabel);
    
    // step 2: iterate through collection of this 
    //         theme's children, capturing names into list as we go
    for(idx=0; idx < pkgTheme_p->pkThemeObjCount; idx++)
    {
        // using this index , fetch a pointer to the child PkgObj 
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[idx];
        // Found match, so, retrieve and STORE this child's name
        strcpy(resultStrTblList_p[idx], pkgMgr_GetObjName(pkgObj_p));
    } // end:: idx loop through all children

    // Return pointer to table of ALL child names owned by this theme
    // note: remember to release it separately later!
    return resultStrTblList_p;
    
} // end:: pkgMgr_GetThemes_ChildListOfNames()


// ---------------------------------------------------------------------
// purpose: set PkgObj's raw data, based on its index
void *pkgMgr_SetThemes_ChildData(PkgTheme *pkgTheme_p,const int pkgObjIdx,void *rawData_p)
{
 PkgObj *pkgObj_p = NULL;
 
    // always make sure pkgObjIdx submitted is in a valid range
    if( pkgObjIdx >=0 && pkgObjIdx < pkgTheme_p->pkThemeObjCount )
    {
        // using this index , fetch the child PkgObj 
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[pkgObjIdx];
        // now, BIND the new object to it
        pkgObj_p->pkObjContent_p = rawData_p;
        // DONE!
    } else {
        xLogMsg("@E %s ERROR Theme(%s) child index (%d) Out-of-Range!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pkgObjIdx);
    }
 // return pointer to BOUND object, or NULL if not registered
 return rawData_p ;
    
} // end:: pkgMgr_SetThemes_ChildData()


// ---------------------------------------------------------------------
// Purpose: attach raw data to a Theme's child whose name is 'objName'
// Notes  : This is just a wrapper for pkgMgr_BindObjByName()
void pkgMgr_SetThemes_ChildDataByName(PkgTheme *pkgTheme_p,const char *objName,void *rawData_p)
{
    
 pkgMgr_BindObjByName(pkgTheme_p,objName,rawData_p);    
    
} // end:: pkgMgr_SetThemes_ChildDataByName()


// ---------------------------------------------------------------------
// purpose: get PkgObj's raw data, based on its index 
void *pkgMgr_GetThemes_ChildData(PkgTheme *pkgTheme_p,const int pkgObjIdx)
{
 void *rawData_p = NULL;
 PkgObj *pkgObj_p = NULL;
 
    // always make sure pkgObjIdx submitted is in a valid range
    if( pkgObjIdx >=0 && pkgObjIdx < pkgTheme_p->pkThemeObjCount )
    {
        // using this index , fetch the child PkgObj 
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[pkgObjIdx];
        // now, fetch pointer to the BOUND raw data it indexes into
        rawData_p = pkgObj_p->pkObjContent_p ;
    } else {
        xLogMsg("@E In(%s) Theme(%s) ERROR Child index %d Out of range!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pkgObjIdx);
    }
 // return pointer to BOUND object, or NULL if not registered
 return rawData_p ;
    
} // end:: pkgMgr_GetThemes_ChildData()


// ---------------------------------------------------------------------
// purpose: get PkgObj's raw data, based on its index 
// returns: ptr to raw data, or NULL if not found
void *pkgMgr_GetThemes_ChildDataByName(PkgTheme *pkgTheme_p,const char *pObjName)
{
 void *rawData_p = NULL;
 PkgObj *pkgObj_p = NULL;
 int i= 0;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName,pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so attach to its raw data and break out
            rawData_p = pkgObj_p->pkObjContent_p ;
            hit = 1;
            break;
        }
    }
    
 if( !hit )
 {
    xLogMsg("@E %s ERROR Theme(%s) No match for child name %s!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);
 }
 
 // return pointer to BOUND object, or NULL if not registered
 return rawData_p ;
    
} // end:: pkgMgr_GetThemes_ChildDataByName()


// ---------------------------------------------------------------------
// function : pkgMgr_SetThemes_ChildMetadataByName()
// purpose  : set PkgObj's ,metadata, based on child's name
// inputs   : PkgTheme *pkgTheme_p,const char *pObjName,char *newMetadata_p
// returns  : void
// notes    : n/a
void pkgMgr_SetThemes_ChildMetadataByName(PkgTheme *pkgTheme_p,const char *pObjName,char *newMetadata_p)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so assign it some new metadata
            pkgMgr_SetObjMetadata(pkgObj_p, newMetadata_p);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) metadata!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);
 }
    
} // end:: pkgMgr_SetThemes_ChildMetadataByName()


// ---------------------------------------------------------------------
// function : pkgMgr_GetThemes_ChildMetadataByName()
// purpose  : get PkgObj's ,metadata, based on child's name
// returns  : ptr to metadata string, or NULL if not found
// notes    : n/a
char *pkgMgr_GetThemes_ChildMetadataByName(PkgTheme *pkgTheme_p,const char *pObjName)
{
 char *metadata_p= NULL;
 PkgObj *pkgObj_p = NULL;
 int i=0;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so attach to its metadata
            metadata_p = pkgObj_p->pkObjMetadata;
            hit = 1;
            break;
        }
    }
    
 if( !hit )
 {
    xLogMsg("@E %s ERROR Theme(%s) No match for child(%s) metadata!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);
 }
    
 // return pointer
 return metadata_p;
    
} // end:: pkgMgr_GetThemes_ChildMetadataByName()


// ---------------------------------------------------------------------
// function : pkgMgr_SetThemes_ChildGenericPathByName()
// purpose  : set PkgObj's ,generic path, based on child's name
// inputs   : PkgTheme *pkgTheme_p,const char *pObjName,char *newGenPath_p
// returns  : void
// notes    : n/a
void pkgMgr_SetThemes_ChildGenericPathByName(PkgTheme *pkgTheme_p,const char *pObjName,char *newGenPath_p)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so assign new generic path
            pkgMgr_SetObjGenericPath(pkgObj_p, newGenPath_p);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for PkgObj name %s!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);
 }
    
} // end:: pkgMgr_SetThemes_ChildGenericPathByName()


// ---------------------------------------------------------------------
// function : pkgMgr_GetThemes_ChildGenericPathByName()
// purpose  : get PkgObj's ,metadata, based on child's name
// returns  : ptr to metadata string, or NULL if not found
// notes    : n/a
char *pkgMgr_GetThemes_ChildGenericPathByName(PkgTheme *pkgTheme_p,const char *pObjName)
{
 char *genericPath_p = NULL;
 PkgObj *pkgObj_p = NULL;
 int i=0;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so attach its generic path
            genericPath_p = pkgObj_p->pkObjGenericPath ;
            hit = 1;
            break;
        }
    }
    
 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) generic path!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);
 }
    
 // return pointer to object's generic path element
 return genericPath_p;
    
} // end:: pkgMgr_GetThemes_ChildGenericPathByName()

// ---------------------------------------------------------------------
// function : pkgMgr_GetThemes_ChildIndexByName()
// purpose  : Using a child's name, retrieve the encapsulating PkgObj's 
//            base0 internal index where this child resides in the theme.
// inputs   : PkgTheme *pkgTheme_p, -- ptr to theme that owns this child PkgObj
//            const char *pObjName_p -- ptr to string, name to search for
// returns  : int, base0 index of PkgObj assoc. with this name.
// revision : 2020-7-13T18:42, reviewed to re-confirm correctness.
int pkgMgr_GetThemes_ChildIndexByName(PkgTheme *pkgTheme_p,const char *pObjName_p)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 int thisIndex = -1;
 int nCountNotFound = 0 ;
   
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match 
        if(strcmp(pObjName_p, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so store its base0 index and exit
            thisIndex = i ;
            hit = 1;
            break;
        } else {
            nCountNotFound++ ; // tally n. failed attempts
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) index (%d failed searches)! ~\n",
         (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName_p, nCountNotFound);   
 }
 // return presumably valid index 
 return thisIndex ;
    
} // end:: pkgMgr_GetThemes_ChildIndexByName()


// ---------------------------------------------------------------------
// purpose: Set PkgObj's ID, based on child's name
// revised: 2020-07-13, changed for hid_t compat.
void pkgMgr_SetThemes_ChildIdByName(PkgTheme *pkgTheme_p,const char *pObjName,const int64_t newId)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so assign it a fresh ID
            pkgMgr_SetObjId(pkgObj_p, newId);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) ID! %ld~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName,newId);   
 }
    
} // end:: pkgMgr_SetThemes_ChildIdByName()


// ---------------------------------------------------------------------
// purpose: Get PkgObj's ID, based on child's name
int64_t pkgMgr_GetThemes_ChildIdByName(PkgTheme *pkgTheme_p,const char *pObjName)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 int64_t thisId = -1; // <JMG> 2020-07-13, changed for hid_t compat.
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so FETCH its ID
            thisId = pkgMgr_GetObjId(pkgObj_p);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) ID!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);   
 }
 return thisId ;
    
} // end:: pkgMgr_GetThemes_ChildIdByName()


// ---------------------------------------------------------------------
// purpose: Set PkgObj's OWNER ID, based on child's name
void pkgMgr_SetThemes_ChildOwnerByName(PkgTheme *pkgTheme_p,const char *pObjName,const int64_t newOwnerId)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so assign it a fresh ID
            pkgMgr_SetObjOwner(pkgObj_p, newOwnerId);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) new owner ID %ld!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName,newOwnerId);   
 }
    
} // end:: pkgMgr_SetThemes_ChildOwnerByName()


// ---------------------------------------------------------------------
// purpose: Get PkgObj's OWNER, based on child's name
int64_t pkgMgr_GetThemes_ChildOwnerByName(PkgTheme *pkgTheme_p,const char *pObjName)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 int64_t thisOwnerId = -1;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so FETCH its ID
            thisOwnerId = pkgMgr_GetObjOwner(pkgObj_p);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) owner ID %ld!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName,thisOwnerId);   
 }
 return thisOwnerId ;
    
} // end:: pkgMgr_GetThemes_ChildOwnerByName()


// ---------------------------------------------------------------------
// purpose: Set PkgObj's Datatype, based on child's name
void pkgMgr_SetThemes_ChildDatatypeByName(PkgTheme *pkgTheme_p,const char *pObjName,const int newDatatype)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so ASSIGN its datatype
            pkgMgr_SetObjDatatype(pkgObj_p,newDatatype);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) datatype!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);   
 }
    
} // end:: pkgMgr_SetThemes_ChildDatatypeByName()


// ---------------------------------------------------------------------
// purpose: Get PkgObj's Datatype, based on child's name
int pkgMgr_GetThemes_ChildDatatypeByName(PkgTheme *pkgTheme_p,const char *pObjName)
{
 PkgObj *pkgObj_p = NULL;
 int i= -1;
 int hit = 0;
 int datatypeId = -1;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        // next, check for a name match
        if(strcmp(pObjName, pkgMgr_GetObjName(pkgObj_p))==0) 
        {
            // name DID match, so FETCH its datatype
            datatypeId = pkgMgr_GetObjDatatype(pkgObj_p);
            hit = 1;
            break;
        }
    }

 if( !hit )
 {
    xLogMsg("@E In(%s) ERROR Theme(%s) No match for child(%s) datatype!~\n",
            (char *)__func__ ,pkgTheme_p->pkThemeLabel, pObjName);       
 }
 return datatypeId ;
    
} // end:: pkgMgr_GetThemes_ChildDatatypeByName()


// --------------------------------------------------------------------
// function : pkgMgr_ThemeBindToMemObj()
// purpose  : Causes this theme to "bind up" to a current MemObj
//            registry, yielding several coupling options: one is that
//            we create exactly as many PkgObj as there are occupied
//            MemObj slots--this is the global option, but can be risky
//            and overkill. A one-at-a-time, pick-and-choose approach
//            is better, but requires the user to know ahead of time
//            all the names of memObj it will include in the binding
//            operation. This is tricky since some (or most) are now
//            anonyous, have no names, but are identifiable only via DNA.
//       
// inputs   : PkgTheme *pkgThemeDest_p,PkgTheme *pkgThemeSrc_p
// revision : 2014-02-18T17:16,jmg, started PkgMgr <--> memObj integration
// author   : joe glassy
// notes    : NOT YET FINISHED! DO NOT Use yet  Only use this on empty PkgThemes
//
int pkgMgr_ThemeBindToMemRegistry(PkgTheme *pkgTheme_p)
{
 PkgObj *pkgObj_p = NULL;
 uint8 *vPayload_p = NULL;
 char *vName_p = NULL;
 char *vDNA_p = NULL;
 // int nObjBoundCount =0 ;
 int nCandidateMemEntries =0 ;
 int nVerified = 0 ; // count of verified MemObj slots
 size_t vNelem =0;
 int m = 0 ;
 int mSlotIdx =0 ;
 
    // ok, THAT was simple!
    if(pkgTheme_p->pkThemeObjCount ==0)
    {
        pkgTheme_p->pkThemeMemRegistry_p =(MemObj *)memObjExport();
    } else {
        fprintf(stdout,"pkgMgr_ThemeBindToMemRegistry only valid for empty themes!\n");
    }
    
    // Create as many PkgObj entries as there are occupied slots in the registry
    nCandidateMemEntries = memObjSlotsUsed();
    
    for(m=0; m < nCandidateMemEntries; m++)
    {
            vPayload_p = pkgTheme_p->pkThemeMemRegistry_p[m].vContents_p ;
            vDNA_p     = pkgTheme_p->pkThemeMemRegistry_p[m].vDnaLabel ;
            vNelem     = pkgTheme_p->pkThemeMemRegistry_p[m].vNelem;
            // count verified cases
            if( strlen(vDNA_p) > 0 && vNelem > 0 && vPayload_p != NULL)
                    nVerified++ ;
    } // end:: m loop scanning to conform registry integrity  
    
    if(nVerified == nCandidateMemEntries)
    {
        // ok, proceed to the active BONDING step
        // first construct a matching set of PkgObj
        pkgMgr_CreateThemesPkgObjArray(pkgTheme_p,nVerified);
        
        for(m=0; m < nVerified; m++)
        {
            // fetch pointer to current PkgObj
            pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[m];
            // set pointer to this memObj's payload, name, nElem
            vPayload_p = pkgTheme_p->pkThemeMemRegistry_p[m].vContents_p ;
            vName_p    = pkgTheme_p->pkThemeMemRegistry_p[m].vName ;
            vNelem     = pkgTheme_p->pkThemeMemRegistry_p[m].vNelem;
            vDNA_p     = pkgTheme_p->pkThemeMemRegistry_p[m].vDnaLabel;
            
            if( strlen(vDNA_p) > 0 && vNelem > 0 && vPayload_p != NULL)
            {
                // store the referent's slot index ..its ok if "m" != mSlotIdx!
                pkgObj_p->pkObjMemSlotIdx = mSlotIdx ;
                pkgMgr_ObjBind(pkgObj_p, vPayload_p); // bind it!
                
                // next, pass in to store the referents DNA
                strcpy(pkgObj_p->pkObjKey,vName_p);
                // also store referent's DNA
                strcpy(pkgObj_p->pkObjDna, vDNA_p);
                
                mSlotIdx++ ; // increment n. of entries indexed
            } // end:: if block on BIND process
        } // end:: m bonding process
        
    } // end:: if block on verified bind operation

 // return COUNT of n. of memObj bound to this theme    
 return mSlotIdx ;
 
} // end:: pkgMgr_ThemeBindToMemRegistry()


// --------------------------------------------------------------------
// function : pkgMgr_ThemeBindMemByNames()
// purpose  : Causes this theme to "bind up" to a selected subset of 
//            entries in the current MemObj registry, for just those
//            memObj names that match a target IsStringTable of target names
// inputs   : PkgTheme *pkgThemeDest_p,PkgTheme *pkgThemeSrc_p
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : opSwitches =0 , binds to everything. For testing, only
//            a full global bind is performed.
// assert  :  Only use this on empty PkgThemes
int pkgMgr_ThemeBindMemByNames(PkgTheme *pkgTheme_p,const char **targNames_p,const int nTargNames)
{
 PkgObj *pkgObj_p = NULL;
 uint8 *vPayload_p = NULL;
 char *vName_p = NULL;
 char *vDNA_p = NULL;
 
 int nMemObjAvailable = 0;
 int nVerified = 0 ; // count of verified MemObj slots
 size_t vNelem =0;
 int m = 0 ;
 int mSlotIdx =0 ;
 int targIdx =0;
 
    // ok, THAT was simple!
    if(pkgTheme_p->pkThemeObjCount ==0)
    {
        pkgTheme_p->pkThemeMemRegistry_p =(MemObj *)memObjExport();
        
        // Store total memobj in the registry possible to select from
        nMemObjAvailable = memObjSlotsUsed();
        
    } else {
        fprintf(stdout,"pkgMgr_ThemeBindToMemRegistry only valid for empty themes!\n");
    }
    
    // ASSERT: make SURE the memObj registry has already been
    //  bound and there is at 
    // least one entry in it
    if( pkgTheme_p->pkThemeMemRegistry_p == NULL)
    {
         fprintf(stdout,"registry empty! cannot bind anything");
         return 0;
    }
    
    // m is the memObj registry index 
    while(m < nTargNames && m < nMemObjAvailable)
    {
            vPayload_p = pkgTheme_p->pkThemeMemRegistry_p[m].vContents_p ;
            vDNA_p     = pkgTheme_p->pkThemeMemRegistry_p[m].vDnaLabel ;
            vName_p    = pkgTheme_p->pkThemeMemRegistry_p[m].vName ;
            vNelem     = pkgTheme_p->pkThemeMemRegistry_p[m].vNelem;
            
            for(targIdx=0; targIdx < nTargNames; targIdx++)
            {
                if(strcmp( targNames_p[targIdx],  vDNA_p) ==0 ||
                   strcmp( targNames_p[targIdx],  vName_p) )
                   {
                        nVerified++ ;
                        break;
                   } // end:: test on match of vName OR on vDNA
            } 
    } // end:: m loop scanning to conform registry integrity  
    
    if(nVerified > 0)
    {
        // ok, proceed to the active BONDING step
        // first construct a matching set of PkgObj
        pkgMgr_CreateThemesPkgObjArray(pkgTheme_p,nVerified);
        
        while(m < nTargNames && m < nMemObjAvailable)
        {
            vPayload_p = pkgTheme_p->pkThemeMemRegistry_p[m].vContents_p ;
            vDNA_p     = pkgTheme_p->pkThemeMemRegistry_p[m].vDnaLabel ;
            vName_p    = pkgTheme_p->pkThemeMemRegistry_p[m].vName ;
            vNelem     = pkgTheme_p->pkThemeMemRegistry_p[m].vNelem;
            
            for(targIdx=0; targIdx < nTargNames; targIdx++)
            {
                
                if(strcmp(targNames_p[targIdx],  vDNA_p) ==0 &&
                   strcmp(targNames_p[targIdx],  vName_p)==0)
                   {
                    pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[m];
                    // store the referent's slot index ..its ok if "m" != mSlotIdx!
                    pkgObj_p->pkObjMemSlotIdx = mSlotIdx ;
                    pkgMgr_ObjBind(pkgObj_p, vPayload_p); // bind it!
                    // Next, pass in to store the referents DNA
                    strcpy(pkgObj_p->pkObjKey,vName_p);
                    // Also store referent's DNA
                    strcpy(pkgObj_p->pkObjDna, vDNA_p);
                    break;
                   } // end:: test on match of vName OR on vDNA
                   
            } //end: targIdx loop
             
        } // end:: while loop scanning to conform registry integrity  
        
    } // end:: block if nVerified > 0

 // return COUNT of n. of memObj bound to this theme    
 return mSlotIdx ;
 
} // end:: pkgMgr_ThemeBindMemByNames()


// --------------------------------------------------------------------
// function : pkgMgr_ThemeGet_viaMemName()
// purpose  : Fetchs ptr to memObj data,using its name
// inputs   : PkgTheme pkgTheme_p,const char *memObjName_p
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
void *pkgMgr_ThemeGet_viaMemName(PkgTheme *pkgTheme_p,const char *memObjName_p)
{
 PkgObj *pkgObj_p = NULL;
 uint8 *rawData_p = NULL;   
 int i =0;
 
    for(i=0; i < pkgTheme_p->pkThemeObjCount; i++)
    {
        // first, fetch pointer to a given child..
        pkgObj_p = &pkgTheme_p->pkTheme_PkgObjAry_p[i];
        
        // next, check for a name, matching 'key' to 'memObjname'
        if(strcmp(pkgObj_p->pkObjKey, memObjName_p)==0) 
        {
            // name DID match, so attach to its raw data and break out
            rawData_p = pkgObj_p->pkObjContent_p ;
            break;
        }
    } // end:: i loop iterating through PkgObj collections

 // return pointer to raw data, or NULL if failed.
 return rawData_p;
 
} // end:: pkgMgr_ThemeGet_viaMemInterface()

// =====================================================================
// FUTURE CODE -- NOT YET COMPLETED, TESTED, OR INTEGRATED
// =====================================================================

#ifdef PKGMGR_FUTURE
// --------------------------------------------------------------------
// function : pkgMgr_CopyTheme()
// purpose  : copy entire contents of a PkgTheme from src to dest.
// inputs   : PkgTheme *pkgThemeDest_p,PkgTheme *pkgThemeSrc_p
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : n/a
// assert  : Only use this when you are sure pkgThemeDest_p is allocated
//           and EMPTY (IsUnoccupied)
int pkgMgr_CopyTheme(PkgTheme *pkgThemeDest_p,PkgTheme *pkgThemeSrc_p)
{
    // next, set archive individual members...
    pkgThemeDest_p->pkThemeId        = pkgThemeSrc_p->pkThemeId;
    pkgThemeDest_p->pkThemeOwner     = pkgThemeSrc_p->pkThemeOwner;
    pkgThemeDest_p->pkThemeObjCount  = pkgThemeSrc_p->pkThemeObjCount ; 
    pkgThemeDest_p->pkThemeObjCount = pkgThemeSrc_p->pkThemeObjCount;  

    // BIND the array of (child) PkgObj from source to destination
    // This 'overwrites' any prior PkgObj array that might have been
    // defined, so be careful!
    pkgThemeDest_p->pkgObjects_p      = pkgThemeSrc_p->pkgObjects_p ; 
    
    strcpy(pkgThemeDest_p->pkThemeLabel   , pkgThemeSrc_p->pkThemeLabel);
    strcpy(pkgThemeDest_p->pkThemeUUID    , pkgThemeSrc_p->pkThemeUUID);
    strcpy(pkgThemeDest_p->pkThemeCategory, pkgThemeSrc_p->pkThemeCategory);

    return 0;
    
} // end::pkgMgr_CopyTheme()

// --------------------------------------------------------------------
// function : pkgMgr_AddThemes()
// purpose  : "GROW" a collection of themes owned by a PkgMgr, such
//            that the new total is 'newTotalThemes', which is usually
//            greater than the original count of PkgThemes owned.
// inputs   : 
// revision : 1.0-2014-01-15T14:24:00,jmg (original)
// author   : joe glassy
// notes    : If newTotalThemes is <= prior count in n_
// notes    : NOT YET TESTED or finished developing.
int  pkgMgr_AddThemes(PkgMgr *pkgMgr_p,const int newTotalThemes)
{
 PkgObj   *pkgObjBackup_p = NULL ;  // archive collection per theme
 PkgTheme *pkgThemeBackup_p = NULL; // archive collection...
 PkgTheme *pkgThemeSrc_p    = NULL;
 PkgTheme *pkgThemeDest_p   = NULL;
 
 int nThemes  = 0;
 int nExtras  = 0;
 int nObj     = 0;
 int ithTheme = 0;
 int ithObj   = 0;
 
    // If newThemeCount is > current, the GROW the PkgTheme collection
    //  retaining all current instances, adding the rest as empties
    //  marked IsUnoccupied
    if(newTotalThemes <= pkgMgr_p->nMaxThemes)
    {
        return 0 ; // nothing to grow here
    } 
    
    // First, temporarily copy current set of PkgThemes 
    // off to a temp save repository
    if((pkgThemeBackup_p = (PkgTheme *)calloc(pkgMgr_p->nMaxThemes,
                                            sizeof(PkgThemes)))==NULL)
    {
        fprintf(stderr,"@F pkgMgr_AddThemes: failed to archive %d prior themes",
                        nThemes);
        exit(-1);
    } 
        
    // Perform the archive of N prior existing PkgThemes...
    for(ithTheme =0; ithTheme < pkgMgr_p->nMaxThemes; ithTheme++)
    {
        // Set the Source and Destinations for the theme archive
        pkgThemeSrc_p  = &pkgMgr_p->pkMgrThemeAry_p[ithTheme];
        pkgThemeDest_p = &pkgThemeBackup_p[ithTheme];
        // Transfer individual members from source to destination instance
        pkgMgr_CopyTheme(pkgThemeDest_p, pkgThemeSrc_p);
    } // end::ithTheme archive loop
            
    // Next, create the new larger GROWN collection
    if((pkgThemeNew_p=(PkgTheme *)calloc(newTotalThemes,sizeof(PkgThemes)))==NULL)
    {
            fprintf(stderr,"@F pkgMgr_AddThemes: failed to create %d new themes",
                            newTotalThemes);
    } else {
                
        // Finally, copy BACK the prior existing PkgThemes from the
        //     pkgThemeBackup_p collection to the NEW collection
        for(ithTheme =0; ithTheme < pkgMgr_p->nMaxThemes; ithTheme++)   
        {
            // set the Source and Destinations for RESTORE operation
            pkgThemeSrc_p  = &pkgThemeBackup_p[ithTheme];
            pkgThemeDest_p = &pkgThemeNew_p[ithTheme];
            // Transfer individual members from source to destination
            pkgMgr_CopyTheme(pkgThemeDest_p, pkgThemeSrc_p);
            
            // Be sure to FREE UP the original collection now
            pkgMgr_ReleaseTheme(pkgThemeSrc_p);
                    
        } // end:: ithTheme loop
                
        // Next, mark all overflow PkgThemes as IsUnoccupied...
        nExtras = newTotalThemes - pkgMgr_p->nMaxThemes ;
                
        for(ithTheme =pkgMgr_p->nMaxThemes;
            ithTheme < nExtras ; 
            ithTheme++)
        {
            // Set the Source and Destinations for RESTORE operation
            pkgThemeDest_p  = &pkgThemeNew_p->[ithTheme];
            // Blast init in
            memset(pkgThemeDest_p,'\0',sizeof(PkgTheme));
            // MARK all these with an id -1 to indicate IsUnoccupied
            pkgMgr_InitTheme(pkgThemeDest_p,-1,"","");
            
        } // end:: ithTheme loop populating "extras"
                
        // RE-BIND PkgMrg's view, to the brand new collection!
        pkgMgr_p->pkMgrThemeAry_p = pkgTheme_p ;
        // finally, update NEW count
        pkgMgr_p->nMaxThemes = newTotalThemes ;
        // RE-BIND the new set to parent
        pkgMgr_p->pkMgrThemeAry_p = pkgTheme_p ;
            
    } // end:: block to grow PkgTheme count to total of newTotalThemes
    
} // end::pkgMgr_AddThemes()
#endif
