// ---------------------------------------------------------------------
//   module     : l4c_MET_drilldowns.c
//   Purpose    : implement vertical drilldown functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : 7.0, see AppVers macro in l4c_MET.h
//   Revision   : 2015-09-04T16:10, added handler_geos5_QV2M_and_PS() and
//                corrected rescaling logic to use new nearest neighbor
//                rescaleFloatImage() logic to rescale TS 9km into TS 3km
//   Revision   : 2015-08-31T09:02, continued refinements
//   Revision   : 2015-08-26T10:39, reviewed pixel wise numerical constraint 
//                and now explictly added 'lowestValid' in place of hard wired
//                0.0 in test of (pix >= lowestVali && pix < missingValue) test
//                in geos5_fp_MeterologyDrilldown() and related functions.
//   Revision   : 2014-07-22, re-integration into L4_C v5.0.1 in progress
//   Notes      : Several new fields added for D5 representing either new
//                capabilities needed to support the use of F/T in L4_C_DAILY,
//                or (VPD) that were omitted earlier.  New fields are:
//                -- FT_ROLLING_AGE_3KM  ...from Freeze Thaw AGE (uint8,3km)
//                -- vpd (float32, 9KM, "units=Pa") in SMAP_L4_C_MET.h5
//                -- smroot_min (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                -- smroot_max (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                 and new static 3KM float32 layers "smroot_min","smroot_max"

#include "l4c_MET.h"
#include "l4c_MET_prot.h"

// ---------------------------------------------------------------------
// functon : handler_geos5_ts_Fallback()
// purpose : This pre-emptively implements fallback logic that usees
//           GEOS5 FP "TS" to drive Freeze Thaw
//           classification, populating the FT_STATE_UM_M03 memory array
//           using TS derived (deg C) threshold logic. Critical note: if
//           the L3_SM_A is actually present, FT_STATE_UM_M03 will be 
//           correctly then overwritten by L3_SM_A driven logic to assign
//           FT_STATE_UM_M03. this pre-emptive approach is used since
//           its more convenient to initially treat "TS" as just one of
//           the other GEOS5 variables to process, re-grid, initially.
//           Note that FT_STATE_UM_M03 bits 4 and 5 are ONLY adjusted
//           if ftSourceMode = DoUseFT_GEOS5 and are left intact (as 0)
//           if FT_RADAR or L3_SM_A's tsurf are used.
// revised : 2015-09-08T09:35, now also set bit(4,4) to TRUE, and then set
//           bit5 to ftState as an independent confirmation that 
//           GEOS5 TS was used.
// revised : 2015-09-04T16:10:00
// inputs  : SesCtl *sesCtl_p -- ptr to session controls
//           float32 *tsEg9kmImg_p -- ptr to 9KM "TS" image, in Kelvin units
//           char *dayOfYearStr ("001"..."365")
// returns : hid_t dset_id of new dataset to attach provenence to
//
// notes   :  <JMG> 2015-09-08T09:36, also ASSIGN bit4 in FT_STATE_UM_M03 
// to always TRUE  when GEOS5 was used.  And, assigns bit5 to  
// ftStates boolean value (0|1) as an independent confirmation
// when GEOS5 TS was used. Bit4 (and 5) will always thus be FALSE
// when GEOS5 TS was NOT used.
//
// notes   : Accepts a 9KM TS Kelvin as inputs, produces several 
//           outputs: a) 9KM TS deg C verification raster,TS_M09_DEGC_AVG
//                    b) 3KM FT_STATE_UM_M03 bitpacked raster
hid_t handler_geos5_ts_Fallback(SesCtl *sesCtl_p,float32 *tsEg9kmImg_p,char *dayOfYearStr)
{
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package
 
 char fxName[256] = {0} ;
 char actualAttr[32] = {0} ;
 float32 ts9kmFlt32_Pix = 0.0 ;
 float32 *eg3kmImg_p  = NULL;
 float32 *ts3kmDegC_Img_p= NULL;
 
 float32 tsDegC = 0.0 ;
 float32 tsKelvin = 0.0 ;
 float32 tsSum =0.0 ;
 float32 tsMean=0.0 ;
 
 size_t pixEG3KmIndex   = 0 ;
 size_t pixEG9KmIndex   = 0 ;
 size_t nElem9km  = N_9KM_PIX;
 size_t nElem3km  = N_3KM_PIX;
 size_t countUpdates = 0 ;
 size_t nNotTerrestrial = 0; // count of pixels classified as NON terrestrial
 size_t nAreTerrestrial = 0; // count of pixels classified as terrestrial
 size_t nQual =0;
 int g = 0;
 
 uint8  ftStateUm_M03_Pix = 0;
 uint8  ftState =0 ;
 uint8  trueClassifier  = 1;
 uint8  falseClassifier = 0;
 uint8  inSwathClassifier= 0; // const, to allow passing PTR to this for bit assignment
 uint8  pftMask = 0;
 uint8  isInTerrestrial = 0 ;
 
 char  hDatasetname[128] = {0} ;
 char datasourceLbl[512] = {0} ;
 int   status = 0 ;
 
 hid_t thisDsTypeId = H5T_NATIVE_FLOAT ; 
 hid_t dset_id = 0 ;
 
 
        // -------------------------------------------------------------
        strcpy(fxName,__func__);
       
        // retrieve and use user-defined proxy var,"TS" from RIP file. This is
        // a variable in case user later changes it from TS to a similar variable 
        strcpy(hDatasetname,sesCtl_p->geos5ProxyVar);
        
        xLogMsg("@A In(%s) HANDLER for F/T proxy var (%s) thresholding for DOY %s~ \n",
            fxName, hDatasetname, dayOfYearStr);
        
        // clear out to initialize...
        memset(imagePkg_p->ftStateUm_M03_p,'\0',nElem3km);
        
        // While still in 9km, CONVERT 9KM TS (Kelvin) image to 9KM Deg C image
        nQual =0;
        tsSum =0.0 ;
        for(pixEG9KmIndex =0; pixEG9KmIndex < nElem9km; pixEG9KmIndex++)
        {
            // ptr to EASEGRID 9km global mask  
            pftMask = imagePkg_p->pftImg9km_p[pixEG9KmIndex] ;
            if( pftMask )
            {
                tsKelvin = tsEg9kmImg_p[pixEG9KmIndex];
                tsDegC   = tsKelvin - 273.16 ;
                nQual++ ;// increment count of qualifiers
                // re-store TS as deg C
                tsEg9kmImg_p[pixEG9KmIndex] = tsDegC ;
                tsSum += tsDegC ;
            } else {
                tsEg9kmImg_p[pixEG9KmIndex] = -9999.0 ;
            }
        } // end::pixEG9KmIndex loop through 9KM space
        
        // verify: calc mean of TS deg C at 9km 
        tsMean  = tsSum / (float)nQual;
        // UPDATE GRANULE WISE RANGE
        // Next, now that a geographic vMean[] is available on a pixel basis
        // calculate the granule wise min,max for actual_min,actual_max attributes
        nQual =  calculateGranuleActualRange(sesCtl_p, sesCtl_p->geos5ProxyVar,
                 GEOS5_FP_MISSING_VALUE, tsEg9kmImg_p, nElem9km);

        // RESCALE FROM source 9KM TS (degC) image TO a temporary 3KM TS (degC).
        // remember that ts3kmDegC_Img_p is allocated inside this rescaleFloatImage() call
        ts3kmDegC_Img_p = rescaleFloatImage(tsEg9kmImg_p, NROW9KM,NCOL9KM, NROW3KM,NCOL3KM);
        
        // Iterate through 3KM float32 TS image, convert to deg C, and from this,
        // populate the final FT_STATE_UM_M03 3KM uint8 image
        for(pixEG3KmIndex=0; pixEG3KmIndex < nElem3km; pixEG3KmIndex++)
        {
            // Initialize ALL 3k pixel bits as 0 to start
            ftStateUm_M03_Pix = 0 ; 
            ftState = 0 ; // need to set default even for a binary bitvalue
            
            // Initialize ALL 3k pixel bits as 0 to start
            ftStateUm_M03_Pix = 0 ; 
            // Retrieve  TS deg C from 3KM image
            tsDegC = ts3kmDegC_Img_p[pixEG3KmIndex];
            
            // Classify using TERRESTRIAL MASK as IN or OUT of land
            if( tsDegC <= -9999.0)
            {
              isInTerrestrial = 0 ; // FALSE, in ocean, rock, etc
              nNotTerrestrial++;
               
            } else {
               isInTerrestrial = 1 ; // TRUE, on a terrestrial pixel
               nAreTerrestrial++; // count of pixels classified as terrestrial
               // Only here DO we use TS deg C threshold to set as FROZEN or THAWED
               ftState =( tsDegC <= 0.0 ) ? 1 : 0 ;
            }

            // ASSIGN bit0, primary FREEZE/THAW, when isInTerrestrial!
            ftStateUm_M03_Pix = frm_utilAssignBits(0,0,
                  &ftStateUm_M03_Pix, sizeof(uint8), &ftState, sizeof(uint8));
         
            // bit 0 -- defines the freeze thaw state when TSURF is driving the logic
            // bit 1 -- defines the freeze thaw state when ACTIVE RADAR is driving the logic
            // bit 2 -- is the QualityControl est passed in by L3_SM_A. Its always TRUE if GEOS5 used
            // bit 3 -- is the In or Out of swath indicator, where IsInSwath==0 to retain the goofy JPL convention.         
            // FOR GEOS5 TS use-case only:
            // bit 4 -- is set TRUE to indicate the GEOS5 TS was explicitly specified
            // bit 5 -- is the TRUE if InTerrestrial land domain, FALSE otherwise
            // if GEOS5 TS is NOT driving the F/T definition.
            // Set bit1 to TRUE if isInTerrestrial, or FALSE if not
            
            ftStateUm_M03_Pix = frm_utilAssignBits(5,5, 
                &ftStateUm_M03_Pix, sizeof(uint8),&falseClassifier,sizeof(uint8));               

            // Set bit2 of FT_STATE_UM_M03 to ALWAYS  TRUE when GEOS5 TS is used
            ftStateUm_M03_Pix = frm_utilAssignBits(2,2, &ftStateUm_M03_Pix, sizeof(uint8), 
                                    &trueClassifier, sizeof(uint8));
               
            // CRITICAL NOTE: poke bit 3 to FALSE since "InSwath" is defined as 0 
            // and NOT as 1, for GEOS5 FP "TS" mode
            ftStateUm_M03_Pix = (uint8)frm_utilAssignBits(3,3, &ftStateUm_M03_Pix,sizeof(uint8),
                                    &falseClassifier, sizeof(uint8));

            // set InTerrestrial status in bit 5, regardless
            ftStateUm_M03_Pix = frm_utilAssignBits(5,5, 
                &ftStateUm_M03_Pix, sizeof(uint8),&isInTerrestrial,sizeof(uint8));

            if(sesCtl_p->ftSourceMode == DoUse_GEOS5_TS)
            {
              // <JMG> 2015-09-08T09:36, also ASSIGN bit4 in FT_STATE_UM_M03 
              // to always TRUE  when GEOS5 was used.  And, assigns bit5 to  
              // ftStates boolean value (0|1) as an independent confirmation
              // when GEOS5 TS was used. Bit4 (and 5) will always thus be FALSE
              // when GEOS5 TS was NOT used.
              ftStateUm_M03_Pix = frm_utilAssignBits(4,4,
                  &ftStateUm_M03_Pix, sizeof(uint8), &trueClassifier, sizeof(uint8));


              // Explictly CLEAR (set to 0) the bits that indicate TSURF or FT_RADAR were used
              // since they were not in this case.    
              ftStateUm_M03_Pix = frm_utilAssignBits(6,6, &ftStateUm_M03_Pix, sizeof(uint8), 
                                    &falseClassifier, sizeof(uint8));
              ftStateUm_M03_Pix = frm_utilAssignBits(7,7, &ftStateUm_M03_Pix, sizeof(uint8), 
                                    &falseClassifier, sizeof(uint8));
            } // end:: block if GEOS5 TS explicitly specified.

            // RE-LOAD bitpacked uint8 pixel back into final FT_STATE_UM_M03 uint8 3km raster
            imagePkg_p->ftStateUm_M03_p[pixEG3KmIndex] = ftStateUm_M03_Pix;
            
        } //end:: pixEG3kmIndex iteration loop

        xLogMsg("@A In(%s) %zu TERRESTRIAL Pixels with %zu NON-TERRESTRIALs (sum %zu) mean %.2f degC~\n",
         fxName, nAreTerrestrial, nNotTerrestrial, (nAreTerrestrial+nNotTerrestrial), tsMean);
          
        // -------------------------------------------------------------
        // Two new HDF5 datasets are written by default TS_M09_DegC_AVG
        // where proxy var "TS" is used as a preface, and FT_STATE_UM_M03 
        
        // WRITE 9KM TS_M09_DegC_AVG image
        sprintf(hDatasetname,"%s_M09_DegC_AVG",sesCtl_p->geos5ProxyVar);
        thisDsTypeId = H5T_NATIVE_FLOAT ; 
            
        xLogMsg("@A In(%s) CREATED HDF5 variable (%s) now~\n",fxName,hDatasetname);            
            
        dset_id = createMetHdf5Dataset(sesCtl_p,IsEG_Global_9km,
                  hDatasetname, thisDsTypeId, tsEg9kmImg_p);

        // Reference each granule, add provenance metadata, indicating specific source of this data
        xLogMsg("@A In(%s) Adding INPUTPOINTER attributes for %s ~\n",
             fxName, hDatasetname);

        for(g= 0; g < sesCtl_p->nFP_Granules;g++)           
        {
         sprintf(datasourceLbl,"INPUTPOINTER_GEOS5_FP_%-3.3d",g+1) ;
         frm_hdf5_CreateStringAttribute(dset_id, datasourceLbl,sesCtl_p->inputGEOS5_FP_pathList[g]);            
        }

        // Release 3km TS deg C raster, no longer needed
        memFree(ts3kmDegC_Img_p);
        // Release 9km TS deg C raster, no longer needed
        memFree(tsEg9kmImg_p);

 // return status
 return dset_id;    
    
} // end:: geos5_ts_fallbackHandler()


// ---------------------------------------------------------------------
// function: handler_geos5_T2M_Max()
// purpose : Implement special T2M MAX handler logic
// inputs  : SesCtl *sesCtl_p,
//         : char *datasetName,
//         : float32 *eg9kmImg_p,
//         : float32 *vMaxInputGeogImg_p
// returns : dset_id -- HDF5 dataset ID for T2M_MAX
// notes   : Reminder the regrid_FP_to_EG_9km() call is made for T2M_Max here
hid_t handler_geos5_T2M_Max(SesCtl *sesCtl_p,char *datasetName,float32 *varEG9KmOutputImg_p,float32 *vMaxInputGeogImg_p)
{
 char fxName[256] = {0} ;
 char hDatasetname[128] = {0} ;
 char datasourceLbl[512]= {0} ;
 char actualAttr[32] = {0} ;
 hid_t thisDsTypeId = H5T_NATIVE_FLOAT ; 
 hid_t dset_id =0;
 
 // needed for PFT mask image import
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package       
 size_t serialIdx = 0 ;
 size_t nQual =0 ;
 int   g = 0 ;
 
 size_t  nElem9km = N_9KM_PIX ; // 1624*3856
 size_t  nColsGeog  = N_COLS ; // 1152...
 
 float32 chkMinMax = 0.0;
 float32 chkMinSum = 0.0;
 float32 chkMaxMean= 0.0;
 float32 chkMinMean= 0.0;
 
        strcpy(fxName,__func__);

        memset(varEG9KmOutputImg_p, '\0', (nElem9km * sizeof(float32)));
        
        nQual = regrid_FP_to_EG_9km(nColsGeog,
           imagePkg_p->pftImg9km_p, varEG9KmOutputImg_p, vMaxInputGeogImg_p);
        
        // CHECK range of vMaxInputGeogImg_p here...
        chkMinSum = chkMinMean = 0.0 ;
        chkMinMax = GEOS5_FP_MISSING_VALUE+1.0 ; // set as a sentinel
        for(serialIdx =0; serialIdx < nElem9km; serialIdx++)
        {
            /// GEOS5_FP_MISSING_VALUE   9.9999999E14
            if(varEG9KmOutputImg_p[serialIdx] > GEOS5_FP_MISSING_VALUE)
            {
                chkMinSum += varEG9KmOutputImg_p[serialIdx] ;
                if(chkMinMax > GEOS5_FP_MISSING_VALUE) 
                    chkMinMax = varEG9KmOutputImg_p[serialIdx] ;
                else    
                    if(varEG9KmOutputImg_p[serialIdx] >  chkMinMax)
                            chkMinMax = varEG9KmOutputImg_p[serialIdx] ;
            } // end:: block if pixel qualified as not missing
        }
        chkMinMean = chkMinSum / (float32)nElem9km ;
        xLogMsg("@W In(%s) chkMinSum VERIFY vMaxInputGeogImg_p nQual %d Sum %.2f chkMinMax %.2f chkMinMean %.2f\n",
            fxName, nQual , chkMinSum, chkMinMax, chkMinMean);      
        
        sprintf(hDatasetname,"%s_M09_MAX",datasetName);
        thisDsTypeId = H5T_NATIVE_FLOAT ;
        
	    // xLogMsg("@A In(%s) CREATED HDF5 variable (%s)~\n",fxName,hDatasetname);            
        // CREATE OUTPUT DATASET using raster varEG9KmOutputImg_p
        dset_id = createMetHdf5Dataset(sesCtl_p,IsEG_Global_9km,
                         hDatasetname,thisDsTypeId,varEG9KmOutputImg_p);
            
        // Reference each granule, add provenance metadata, indicating specific source of this data
        xLogMsg("@A In(%s) Adding INPUTPOINTER attributes for %s ~\n",
             fxName, datasetName);
             
        for(g= 0; g < sesCtl_p->nFP_Granules; g++)
        {
         sprintf(datasourceLbl,"INPUTPOINTER_GEOS5_FP_%-3.3d",g+1) ;
         frm_hdf5_CreateStringAttribute(dset_id, datasourceLbl,sesCtl_p->inputGEOS5_FP_pathList[g]);
        }

    // return dset_id so it can be closed outside    
    return dset_id ;
    
} // end:: handler_geos5_T2M_Max()


// ---------------------------------------------------------------------
// function: handler_geos5_T2M_Min()
// purpose : Accepting geographic vMin[] raster, regrid it to EG 9KM
//           and write it out to HDF5 output file.
// inputs  :SesCtl *sesCtl_p -- ptr to session controls
//          char *datasetName -- datasetname
//          float32 *varEG9KmOutputImg_p -- OUTPUT EASEGRID 9km raster
//          float32 *vMinInputGeogImg_p -- INPUT geographic vMin[] raster
// returns : hid_t -- dataset ID HDF5 assigned
// notes   : n/a
hid_t handler_geos5_T2M_Min(SesCtl *sesCtl_p,char *datasetName,
               float32 *varEG9KmOutputImg_p,float32 *vMinInputGeogImg_p)
{
 char fxName[256] = {0} ;
 char hDatasetname[128] = {0} ;
 char datasourceLbl[512]= {0} ;
 char actualAttr[32] = {0} ;
 hid_t thisDsTypeId = H5T_NATIVE_FLOAT ; 
 hid_t dset_id =0;
 // needed for PFT mask image import
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ;
 
 size_t serialIdx = 0 ;
 size_t nQual =0 ;
 size_t nElem9km = N_9KM_PIX ; // 1624*3856
 size_t  nColsGeog  = N_COLS ; // 1152...
 int g = 0 ;
 float32 chkMinMax = 0.0;
 float32 chkMinSum = 0.0;
 float32 chkMaxMean= 0.0;
 float32 chkMinMean= 0.0;

        strcpy(fxName,__func__);
   
        // =============================================================
        memset(varEG9KmOutputImg_p, '\0', (nElem9km * sizeof(float32)));

        // Reminder: OUTPUT here is 'varEG9KmOutputImg_p raster        
        nQual = regrid_FP_to_EG_9km(nColsGeog, imagePkg_p->pftImg9km_p,
                              varEG9KmOutputImg_p, vMinInputGeogImg_p);
        
        sprintf(hDatasetname,"%s_M09_MIN",datasetName);
        
        // CHECK range of vMinInputGeogImg_p here...
        chkMinSum = chkMinMean = 0.0 ;
        chkMinMax = GEOS5_FP_MISSING_VALUE+1.0 ; // set as a sentinel
        
        for(serialIdx =0; serialIdx < nElem9km; serialIdx++)
        {
            // GEOS5_FP_MISSING_VALUE   9.9999999E14 (re-verify)
            if(varEG9KmOutputImg_p[serialIdx] > GEOS5_FP_MISSING_VALUE)
            {
                chkMinSum += varEG9KmOutputImg_p[serialIdx] ;
                if(chkMinMax > GEOS5_FP_MISSING_VALUE) 
                    chkMinMax = varEG9KmOutputImg_p[serialIdx] ;
                else    
                    if(varEG9KmOutputImg_p[serialIdx] >  chkMinMax)
                            chkMinMax = varEG9KmOutputImg_p[serialIdx] ;
            } // end:: block if pixel qualified as not missing
        }
        chkMinMean = chkMinSum / (float32)nElem9km ;
        xLogMsg("@W fp_verticalDrilldown::chkMinSum VERIFY Vmin Qual %d Sum %.2f chkMinMax %.2f chkMinMean %.2f\n",
            nQual , chkMinSum, chkMinMax, chkMinMean);
            
        // xLogMsg("@A In(%s) CREATED HDF5 variable (%s)~\n",fxName,hDatasetname);            
            
        // create and WRITE the dataset, whose data is held now in varEG9KmOutputImg_p[]
        thisDsTypeId = H5T_NATIVE_FLOAT ;
        dset_id = createMetHdf5Dataset(sesCtl_p,IsEG_Global_9km,
                         hDatasetname, thisDsTypeId,varEG9KmOutputImg_p);
             
        // Reference each granule, add provenance metadata, indicating specific source of this data
        xLogMsg("@A In(%s) Adding INPUTPOINTER attributes for %s ~\n",
             fxName, datasetName);

        for(g= 0; g < sesCtl_p->nFP_Granules;g++)           
        {
         sprintf(datasourceLbl,"INPUTPOINTER_GEOS5_FP_%-3.3d",g+1) ;
         frm_hdf5_CreateStringAttribute(dset_id, 
               datasourceLbl,sesCtl_p->inputGEOS5_FP_pathList[g]);            
        }
        
    // return dset_id so it can be closed outside    
    return dset_id ;
    
} // end:: handler_geos5_T2M_Min()

// ---------------------------------------------------------------------
// functon : handler_geos5_T2M_Mean()
// purpose : Implement logic to write QV2M and PS vars
// revised : 2015-09-04T11:53:00
// inputs  : SesCtl *sesCtl_p -- ptr to session controls
//           char *datasetName -- T2M for T2M_AVG_M03
//           float32 *varEG9KmOutputImg_p,
//           float32 *vMeanInputGeogImg_p
//           float32 *var9kmImg_p -- ptr to 9KM image for QV2M_AVG,T2M_AVG, or PS_AVG
// returns : hid_t dset_id of new dataset
// notes   : n/a
hid_t handler_geos5_T2M_Mean(SesCtl *sesCtl_p,char *datasetName,float32 *varEG9KmOutputImg_p,float32 *vMeanInputGeogImg_p)
{
 // needed for PFT mask image import
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ;
    
 hid_t dset_id = 0 ;
 hid_t thisDsTypeId = H5T_NATIVE_FLOAT ; 
 char fxName[256] = {0} ;
 char hDatasetname[128] = {0} ;
 char datasourceLbl[512]= {0} ;
 char actualAttr[32] = {0} ;
 size_t nQual = 0 ;
 size_t nElem9km = N_9KM_PIX ; // 1624*3856
 size_t  nColsGeog  = N_COLS ; // 1152...
 int g =0;
 
    strcpy(fxName,__func__);
     
    memset(varEG9KmOutputImg_p, '\0', (nElem9km * sizeof(float32)));
        
    nQual = regrid_FP_to_EG_9km(nColsGeog,
           imagePkg_p->pftImg9km_p, varEG9KmOutputImg_p, vMeanInputGeogImg_p);
  
    // -------------------------------------------------------------
    // CREATE HDF5 dataset for current var T2M_AVG_M09 as a daily-averaged var
    // -------------------------------------------------------------
    sprintf(hDatasetname,"%s_M09_AVG",datasetName);
    thisDsTypeId = H5T_NATIVE_FLOAT ; 
    
    xLogMsg("@A In(%s) CREATED HDF5 variable (%s)~\n",fxName,hDatasetname);
    
    // NOTE: WRITEs the HDF5 dataset and also WRITES DATASET ATTRIBUTES
    dset_id = createMetHdf5Dataset(sesCtl_p,IsEG_Global_9km,
                   hDatasetname, thisDsTypeId, varEG9KmOutputImg_p); 
 
    // Reference each granule, add provenance metadata, indicating specific source of this data
    xLogMsg("@A In(%s) Adding INPUTPOINTER attributes for %s ~\n",
             fxName, hDatasetname);

    for(g= 0; g < sesCtl_p->nFP_Granules;g++)           
    {
         sprintf(datasourceLbl,"INPUTPOINTER_GEOS5_FP_%-3.3d",g+1) ;
         frm_hdf5_CreateStringAttribute(dset_id, 
               datasourceLbl,sesCtl_p->inputGEOS5_FP_pathList[g]);            
    }
 
 return dset_id ;
 
} // end:: handler_geos5_T2M_Mean()


// ---------------------------------------------------------------------
// functon : handler_geos5_QV2M()
// purpose : Implement logic to write QV2M and PS vars
// revised : 2015-09-04T11:53:00
// inputs  : SesCtl *sesCtl_p -- ptr to session controls
//           char *datasetName -- QV2M or PS
//           float32 *var9kmImg_p -- ptr to 9KM image for QV2M_AVG
//           char *dayOfYearStr ("001"..."365")
// returns : hid_t dset_id of new dataset
// notes   : n/a
hid_t handler_geos5_QV2M(SesCtl *sesCtl_p,char *datasetName,float32 *var9kmImg_p,char *dayOfYearStr)
{
 hid_t dset_id = 0 ;
 char fxName[256] = {0} ;
 char hDatasetname[128] = {0} ;
 char datasourceLbl[512]= {0} ;
 char actualAttr[32] = {0} ;
 hid_t thisDsTypeId = H5T_NATIVE_FLOAT ; 
 size_t k =0;
 size_t g =0;
 size_t nPositive =0;
 
 float32 pix =0.0;
 
    strcpy(fxName,__func__);
 
  
    for(k=0; k < N_9KM_PIX ;k++)
    {
        pix = var9kmImg_p[k];
        if( pix > 0.0)
         nPositive++ ;
    }
     
    // -------------------------------------------------------------
    // CREATE HDF5 dataset for current var {QV2M|PS} as a daily-averaged var
    // -------------------------------------------------------------
    sprintf(hDatasetname,"%s_M09_AVG",datasetName);
    thisDsTypeId = H5T_NATIVE_FLOAT ; 
    
    xLogMsg("@A In(%s) CREATED HDF5 variable (%s) with %zu positive pix~\n",
     fxName,hDatasetname, nPositive);
    
    dset_id = createMetHdf5Dataset(sesCtl_p,IsEG_Global_9km,
                   hDatasetname, thisDsTypeId, var9kmImg_p); 
 
    // Reference each granule, add provenance metadata, indicating specific source of this data
    xLogMsg("@A In(%s) Adding INPUTPOINTER attributes for %s ~\n",
             fxName, hDatasetname);

    for(g= 0; g < sesCtl_p->nFP_Granules;g++)           
    {
         sprintf(datasourceLbl,"INPUTPOINTER_GEOS5_FP_%-3.3d",(int)g+1) ;
         frm_hdf5_CreateStringAttribute(dset_id, 
               datasourceLbl,sesCtl_p->inputGEOS5_FP_pathList[g]);            
    }
    
 return dset_id ;
 
} // end:: handler_geos5_QV2M_and_PS()



// -----------------------------------------------------------------------------
// function : geos5_fp_MeteorologyDrilldown(), prior named fp_verticalDrilldown()
//
// Purpose  : for one GEOS-5 FP variable, perform a vertical drilldown
//            through up to 24 hourly files to derive the daily average,
//            and min,max for each named (datasetName) FP variable supplied
//            in the sessions RIP file, then, we regrid the geographic 
//            projection 721x1152 dataset to an EASEGRID 9KM (1624x3856), 
//            and then WRITES the (float32) dataset to SMAP_L4_C_MET.h5 output file
//            This function is therefore called once per FP variable to process.
//
// Inputs   : SesCtl *sesCtl_p -- ptr to session control structure
//          : dayOfYearStr -- '032' day of year 1..366 as a string
//          : datasetName -- name of dataset to process.
// Returns  : status, int (0=good -1=bad)
// revised  : 2015-08-28T15:45, now calc "TS" to deg C conversion directly
//            as part of legacy GEOS5 FP collection of variables
// revised  : 2015-08-24T17:48, reviewed pixel wise numerical constraint 
//            and now explictly added 'lowestValid' in place of hard wired
//            0.0.
// revised  : 2014-10-21T14:28 freshen documenation for ORT/ORR Oct 2014
// notes    : GEOS FP variables include: 
//              L4C_MET_INPUT_VARLIST_FP      =  QV2M
//              L4C_MET_INPUT_VARLIST_FP      =  T2M
//              L4C_MET_INPUT_VARLIST_FP      =  TS
// 
// Notes    : This function calls (local,custom) createMetHdf5Dataset() 
//            for each new dataset defined to go in SMAP_L4_C_MET.h5
// Note     : FP missing value    is : 9.9999999E14
//          : L4_SM missing value i// s : -9999.0
//
int32 geos5_fp_MeteorologyDrilldown(SesCtl *sesCtl_p,char *dayOfYearStr,char *datasetName)
{
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package    
 
 hid_t   fp_file_id   = -1;
 hid_t   dset_id      = -1;
 hid_t   thisDsTypeId = -1 ; // H5T_NATIVE_FLOAT32 etc etc
 hsize_t dims[4]      = {0} ;
 herr_t  status       = -1    ;
 
 int32  grnIdx     = 0 ;
 int32  hourIndex  = 0;
 int32  hitFirstValue =0;
 int32  nQual =0;
 int32  nQualTS = 0;
 int32  dRank =2;
 int32  nMinAdjusts =0;
 int32  nMaxAdjusts =0;

 int32  p = 0, g =0,r =0,c  =0  ;
 int32  repeatFactor3km = 3 ;
 
 int32  serialIdx  =0;
 int32  nRowsGeog =0, nColsGeog =0, nElemGeog =0;
 int32  nRows9km =0, nCols9km =0, nElem9km  =0, nElem3km =0;
 
 
 char hDatasetname[FILENAME_MAX]  = {0}; // HDF5 dataset fully qualified name
 char datasourceLbl[FILENAME_MAX] = {0}; // 
 char *fpCurrPath = NULL;
 char hhmmStr[8] = {0}; // hh:mm
 char yyyyStr[8] = {0}; // yyyy
 char fxName[256] = {0} ;

 size_t nDisqualified = 0; // check only...
 size_t ithGeogPix =0;
 
 float32 *vSum  = NULL; // pixel-wise sums, to compute daily average
 float32 *vMean = NULL; // pixel-wise daily average for a given variable
 float32 *vMin  = NULL; // pixel-wise minimums
 float32 *vMax  = NULL; // pixel-wise maximums
 float32 *dsNativeData_p = NULL;  // 3D : (1,721,1152) input image raster 

 float32 *currVarEG9KmImg_p = NULL;// EASEGRID 9km (output) image raster 
 
 float32 missing_value = GEOS5_FP_MISSING_VALUE; // 9.9999999E14, from FP HDF5 file
 float32 minGranule = 0.0; // to populate actual_min dataset attribute
 float32 maxGranule = 0.0; // to populate actual_max dataset attribute

 // <JMG> 2015-08-24T17:39:00

 // ideally, override for EACH variable , critical when some valid values are negative!!
 float32 lowestValid   = 0.0 ; 
 uint8  ftStateUm_M03_Pix = 0 ; // pixel of UINT8 for freeze thaw assignment
 uint8  frozenClassifier = 1; // basically constant, stored to allow ptr passing in
 uint8  thawedClassifier = 0; // basically constant, stored to allow ptr passing in
 
 float32 pix         = 0.0 ;
 float32 chkMean     = 0.0;
 float32 chkSum      = 0.0 ;
 float32 chk_n       = 0.0;
 float32 allowableMin= 0.0;  // idealy, adjust for EACH variable
 
 // DIAG only, verify correctness of vMin's currVarEG9KmImg_p[] array
 float32 chkMinSum   = 0.0;
 float32 chkMinMax   = 0.0;
 float32 chkMinMean  = 0.0;

 float32 hourlyMin = 0.0;
 float32 hourlyMax = 0.0;
 float32 hourlyMean= 0.0; 
 float32 hrMinArray[24]  = {0};
 float32 hrMaxArray[24]  = {0};
 float32 hrMeanArray[24] = {0};

    strcpy(fxName,__func__);
    
    xLogMsg("\n@A In(%s) Drilldown on FP variable: dataset(%s) DOY[%s]\n",
        fxName, datasetName,dayOfYearStr);
    // ============================================================================
    nRowsGeog = N_ROWS ;     // FP geographic rows, 721 
    nColsGeog = N_COLS ;     // FP geographic columns, 1152 
    nElemGeog = N_FP_NELEM; // 721 * 1152 
    
    nRows9km = NROWS_9KM ; // 1624
    nCols9km = NCOLS_9KM ; // 3856 
    nElem9km = N_9KM_PIX ; // 1624*3856
    
    nElem3km = N_3KM_PIX;  // 4872 * 11568, added 8/26/2015
    
    // Allocate layers for current (variable) pass, 721x1152...
    vSum = memCalloc(nElemGeog,sizeof(float32));
    memObjSetName(vSum,"fp_verticalDrilldown.vSum");
    
    vMean = memCalloc(nElemGeog,sizeof(float32));
    memObjSetName(vMean,"fp_verticalDrilldown.vMean");

    vMin  = memCalloc(nElemGeog,sizeof(float32));
    memObjSetName(vMin,"fp_verticalDrilldown.vMin");
    vMax  = memCalloc(nElemGeog,sizeof(float32));
    memObjSetName(vMax,"fp_verticalDrilldown.vMax");
    
    currVarEG9KmImg_p = memCalloc(nElem9km,sizeof(float32));
    memObjSetName(currVarEG9KmImg_p,"fp_verticalDrilldown.currVarEG9KmImg_p");

    // Allocate memory for 3KM version of GEOS5 FP "TS"
    // REVIEW Where (or if) this is still now used
    // imagePkg_p->tsEg3kmFltImg_p = memCalloc(nElem3km,sizeof(float32));
    //memObjSetName(imagePkg_p->tsEg3kmFltImg_p,"fp_verticalDrilldown.imagePkg_p->tsEg3kmImg_p");

    // Zero out various counters for current variable
    nMaxAdjusts = nMinAdjusts = nQual = hitFirstValue = 0;
    // Reset the whole image wise grand check sum, means
    chkSum = chkMean = chk_n = 0.0 ;
    
    // Iterate through up to 24 hours, e.g. 1-file, per variable, per hourly dataset
    for(grnIdx = 0; grnIdx <  sesCtl_p->nFP_Granules; grnIdx++)
    {
        // <JMG> 7/24/2014, now reference string table
        fpCurrPath = sesCtl_p->inputGEOS5_FP_pathList[grnIdx];
            
        // Extract hhmm part of granulename, to fetch base0 hourIndex (0..23)
        //    for various output labelling
        hourIndex = granuleHHMM_to_HrIndex(fpCurrPath,hhmmStr,yyyyStr);
        
        // xLogMsg("@A FP granule(%s) hourIndex %-2.2d HHMM %s\n",fpCurrPath,hourIndex,hhmmStr);

        // OPEN FP .nc4 as a HDF5 FILE ---------------------------------
        fp_file_id = frm_hdf5_Openfile(fpCurrPath,fxName);

        // READ FP NATIVE DATASET --------------------------------------
        // fp_get_dataset does the following:
        //   Allocate mem, read 3D variable, return as a 1D pointer, dsNativeData_p
        //     which we subsequently interpret as a 2D row-major-order raster
        //     due to its (ignored) 1st dimension of time = '1'
        dsNativeData_p = fp_get_dataset(fp_file_id,datasetName,&dRank,dims);
        
        // CALCULATE NATIVE min,max,sum, and mean for NATIVE instance of variable 
        // for JUST this hourly dataset, across the global FP native spatial domain 721x1152
        hourlyMean = fp_summarizeVariable(0.0, GEOS5_FP_MISSING_VALUE,
                             datasetName, dims, dsNativeData_p, &hourlyMin,&hourlyMax);

        // Next, store hourly min,max,mean for each of 24 hour periods, to check later
        hrMeanArray[grnIdx] = hourlyMean;
        hrMinArray[grnIdx]  = hourlyMin ;
        hrMaxArray[grnIdx]  = hourlyMax ;
        
        // VERTICAL-DRILL-DOWN: Iterate through global FP native spatial domain 721x1152, 
        //  accumulating min,max,mean for each variable across ALL 24 hours worth of data
        for(r=0; r < nRowsGeog; r++)
        {
            for(c=0; c < nColsGeog; c++)
            {
                serialIdx = (r * nColsGeog )+c ;
                pix = dsNativeData_p[serialIdx];
                // for each of (24) hourly obs, accumulate the pixel wise
                // sum for given variable whose name is datasetName
                // Note: missing value is 9.9999999E14
                // <JMG> 2015-08-24, lowestValid defaults to 0.0 BUT, be careful if
                // some valid values might be valid and still negative! 
                // TBD, document which CAN be negative
                
                // if(pix >= lowestValid && pix < missing_value )
                if( pix < missing_value )
                {
                    nQual++ ;
                    vSum[serialIdx] += pix;

                    if(hitFirstValue == 0)
                    {   
                        // Remember to initialize ALL pixel-wise min,max
                        hitFirstValue = 1 ;
                        // Iterate through geographic space 721x1152
                        for(ithGeogPix=0; ithGeogPix < nElemGeog; ithGeogPix++)
                            vMin[ithGeogPix] = vMax[ithGeogPix]= dsNativeData_p[ithGeogPix];
                    } else {
                        if(pix > vMax[serialIdx]) 
                        {
                            vMax[serialIdx] = pix; nMaxAdjusts++;
                        }
                        if( pix < vMin[serialIdx]) 
                        {
                            vMin[serialIdx] = pix; nMinAdjusts++ ;
                        }
                    } // end:: block if valid to adjust min,max 
                } // end:: block if pix < missing_value
                
            } // end:: c column loop
        } // end:: r row loop

        // ASSERT: Information in dsNativeData_p should now all have been transferred to either
        // the daily average vMean[], daily min vMin, or daily max vMax[].
        // so we can release it now
        memFree(dsNativeData_p);
        
        // close this latest hourly FP (NetCDFv4) file 
        status = H5Fclose(fp_file_id);
    } // end:: grnIdx loop
    
    // AFTER All (24) hours of data have now been traversed,
    // Calculate the VERTICAL-DRILL-DOWN pixel wise means for (ALL) variables.
    // Note: nElem  here is (721x1152) the n. pixels in the original geographic domain
    for(serialIdx=0; serialIdx < nElemGeog; serialIdx++)
    {
        vMean[serialIdx] = vSum[serialIdx] / (float32)sesCtl_p->nFP_Granules;
    } // end::serialIdx loop 

    // UPDATE GRANULE WISE RANGE
    // Next, now that a geographic vMean[] is available on a pixel basis
    // calculate the granule wise min,max for actual_min,actual_max attributes
    nQual =  calculateGranuleActualRange(sesCtl_p,datasetName,
                GEOS5_FP_MISSING_VALUE, vMean, nElemGeog);

    if(strcmp(datasetName,"QV2M")==0) xLogMsg("@A In(%s) QV2M DISQUALIFIED pixels %zu ~\n",
              fxName,nDisqualified++ );
    
    xLogMsg("@A In(%s) GRANULE-LEVEL GEOGRAPHIC-VAR %s min %f max %f mean %f %zu Qualified~\n",
     fxName, datasetName, sesCtl_p->granuleMin, sesCtl_p->granuleMax, sesCtl_p->granuleMean,nQual);

#ifdef DIAG_ONLY    
    // generate a 1D vector of valid terrestrial pixel (serial) addresses, store 
    xLogMsg("@A ~main::generating test 1D sparse vector from PFT 9km mask now\n");
    // build a 1D sparse buffer for random sampling of vMean,vMax,vMin
    sesCtl_p->terrAddr =  genTerrestrialAddresses(sesCtl_p->pftImg_p,N_9KM_PIX);
    // use pre-generated global 1D vector of (1664040) terrestrial addresses to pick a valid one to sample from
    g = terrAddr[ R_RANDOM_INT(1000)];  // here we just sample ONCE from initial 1000 terrestrial pixels
    xLogMsg("@A DIAG RandomlySamplingTerrestrialPixel %d\n",g);
    xLogMsg("@A DIAG-CHECK-vMean SamplePixel %d (%s) sum %.1f mean %.5f\n",g,datasetName,vSum[g],vMean[g]);
    xLogMsg("@A DIAG-CHECK-vMax  SamplePixel %d (%s) sum %.1f max  %.5f\n",g,datasetName,vSum[g],vMax[g]);
    xLogMsg("@A DIAG-CHECK-vMax  SamplePixel %d (%s) sum %.1f min  %.5f\n",g,datasetName,vSum[g],vMin[g]);
    // FREE the random test buffer
    memFree(sesCtl_p->terrAddr);
#endif

    // -----------------------------------------------------------------
    // REGRID each geographic variable to output as a EASEGRID 9km image
    // -----------------------------------------------------------------
    
    // REMINDER: INPUT  = vMean
    //           OUTPUT = currVarEG9KmImg_p , used for QV2M,PS,T2M_AVG, etc
    //  The geographic GEOGRAPHIC INPUT (vMean) is reprojected to EASEGRID 9km here
    nQual = regrid_FP_to_EG_9km(nColsGeog,
                     imagePkg_p->pftImg9km_p, currVarEG9KmImg_p, vMean);

    if(strcmp(datasetName,sesCtl_p->geos5ProxyVar)==0) 
    {
        // CALL GEOS5 FP "TS" handler to implement fallback 
        // reminder; the call to regrid_FP_to_EG_9km() is performed inside these 2
        // reminder: this handler WRITES TS_M09_AVG and FT_STATE_UM_M03 rasters
        dset_id = handler_geos5_ts_Fallback(sesCtl_p,currVarEG9KmImg_p, dayOfYearStr);
        
    } else if (strcmp(datasetName,"T2M")== 0) {
        
       // reminder; the call to regrid_FP_to_EG_9km() is performed inside these 3
       dset_id = handler_geos5_T2M_Mean(sesCtl_p,datasetName, currVarEG9KmImg_p, vMean);
       dset_id = handler_geos5_T2M_Max( sesCtl_p,datasetName, currVarEG9KmImg_p, vMax) ;
       dset_id = handler_geos5_T2M_Min( sesCtl_p,datasetName, currVarEG9KmImg_p, vMin) ;
       
    } else {
        
        dset_id = handler_geos5_QV2M(sesCtl_p,datasetName,
                                    currVarEG9KmImg_p, dayOfYearStr);
    }
    
    xLogMsg("@A In(%s) Closing dset_id %d for %s ~\n",
             fxName, dset_id, datasetName);
    H5Dclose(dset_id);
    
    // Release heap memory used locally
    memFree(vSum);
    memFree(vMean);
    memFree(vMin);
    memFree(vMax);
    // release generic float32 9KM raster buffer
    memFree(currVarEG9KmImg_p);

    xLogMsg("@A In(%s) COMPLETED dataset %s in geos5_fp_MeteorologyDrilldown ~ \n",
      fxName, datasetName);

 // returns status
 return status;
} // end::geos5_fp_MeteorologyDrilldown()


// ---------------------------------------------------------------------
// function :l4_sm_verticalDrilldown()
// purpose  : for each L4_SM variable required, perform a vertical 
//            drilldown through up to (8) 3-hour files
//            to derive the daily average equivalent for the variable. 
// inputs   : SesCtl *sesCtl_p,char *dayOfYearStr,char *datasetName
// author   : joe glassy
// revised  : 2014-10-21T20:33:00, doc added
// notes    : missing value is -9999.0
//
// notes    : L4_SM variables processed here are:
//             L4C_MET_INPUT_VARLIST_L4_SM   = sm_surface_wetness 
//             L4C_MET_INPUT_VARLIST_L4_SM   = sm_rootzone_wetness 
//             L4C_MET_INPUT_VARLIST_L4_SM   = radiation_shortwave_downward_flux
//             L4C_MET_INPUT_VARLIST_L4_SM   = soil_temp_layer1
//             L4C_MET_INPUT_VARLIST_L4_SM   = surface_pressure
//             L4C_MET_INPUT_VARLIST_L3_SM_A = surface_flag
//             L4C_MET_INPUT_VARLIST_L3_SM_A = retrieval_qual_flag
//
int32 l4_sm_verticalDrilldown(SesCtl *sesCtl_p,char *dayOfYearStr,char *datasetName)
{
 hid_t  l4_sm_file_id = -1;
 hid_t  dset_id = -1;
 hid_t  dsTypeId = -1 ;
 hsize_t dims[4]   = {0} ;
 herr_t  status  = -1    ;
 
 int32  grnIdx     = 0 ;
 int32  hourIndex  = 0;
 
 int32  dRank = 2;
 int32  g,r,c   ;
 int32  serialIdx  =0;
 int32  n3hrGranules = 8 ;
 int32  nMinAdjusts = 0;
 int32  nMaxAdjusts = 0;
 
 int32  nRows9km = NROW9KM, // n. of 9km rows
        nCols9km = NCOL9KM, // n. of 9km columns
        nElem9km = N_9KM_PIX; // n. of full grid EG 9km pixels
        
 size_t nQual =0;
 size_t nMissingPix = 0 ; // count of pixels NOT qualifying
 
 char hDatasetname[FILENAME_MAX]  = {0};
 char datasourceLbl[FILENAME_MAX] = {0};
 char dLbl[128] = {0} ;
 char *l4_smCurrPath = NULL;
 char hhmmStr[8] = {0}; // hh:mm
 char yyyyStr[8] = {0}; // yyyy
 char fxName[256] = {0};

 int hitFirstValue = 0 ; // boolean, triggers start of min,max scan

 float32 *vMean9km_p = NULL;// pixel-wise daily average for a given variable, EASEGRID 9km (output) image raster 
 float32 *vSum  = NULL; // pixel-wise sums, to compute daily average
 float32 *vMin  = NULL; // pixel-wise minimums
 float32 *vMax  = NULL; // pixel-wise maximums
 
 float32 *dsNativeData_p = NULL;  // 3D : (1,721,1152) input image raster 
 
 float32 missing_value =  L4_SM_MISSING_VALUE; // VERIFY: from L4_SM HDF5 file itself
 float32 pix         = 0.0 ;
 float32 chkMean     = 0.0;
 float32 chkSum      = 0.0 ;
 float32 chk_n       = 0.0;
 // float32 chkMin   = 0.0;

 float32 hourlyMin = 0.0;  // 3-hourly actually
 float32 hourlyMax = 0.0;  // 3-hourly actually 
 float32 hourlyMean= 0.0;  // 3-hourly actually 
 float32 hrMinArray[24]  ={0}; // (8) 3-hour values, left as 24 for now
 float32 hrMaxArray[24]  ={0}; // (8) 3-hour values, left as 24 for now
 float32 hrMeanArray[24] ={0}; // (8) 3-hour values, left as 24 for now


    strcpy(fxName,__func__);
    
    xLogMsg("\n@A In(%s) L4_SM variable: dataset(%s) for %d 3-hourly granules on DOY[%s]\n",
        fxName, datasetName, sesCtl_p->nL4_SM_Granules, dayOfYearStr);
        
    // ============================================================================
    
    nRows9km = NROW9KM   ; // 1624
    nCols9km = NCOL9KM   ; // 3856 
    nElem9km = N_9KM_PIX ; // 1624*3856
    
    // Allocate 9km layers to process current dataset, at 1624x3856 by float32
    vSum = memCalloc(nElem9km,sizeof(float32));
    memObjSetName(vSum,"l4_sm_verticalDrilldown.vSum");
    vMin  = memCalloc(nElem9km,sizeof(float32));
    memObjSetName(vMin,"l4_sm_verticalDrilldown.vMin");
    vMax  = memCalloc(nElem9km,sizeof(float32));
    memObjSetName(vMax,"l4_sm_verticalDrilldown.vMax");
    
    vMean9km_p = memCalloc(nElem9km,sizeof(float32));
    memObjSetName(vMean9km_p,"l4_sm_verticalDrilldown.vMean9km_p");

    // zero out various counters for current variable
    nMaxAdjusts = nMinAdjusts = nQual = hitFirstValue = 0;
    // reset the whole image wise grand check sum, means
    chkSum = chkMean = chk_n = 0.0 ;
    
    // Iterate through up to 24 hours, e.g. 1-file, per variable, per hourly dataset
    for(grnIdx = 0; grnIdx < sesCtl_p->nL4_SM_Granules; grnIdx++)
    {
        // Retrieve (next) L4_SM granule name
        l4_smCurrPath = sesCtl_p->inputL4_SM_pathList [grnIdx] ;
            
        // Extract hhmm (and YYYY) info from granule name, to fetch base0 
        //   hourIndex (0..23) for various output labelling
        hourIndex = l4_sm_granuleHHMM_to_HrIndex(l4_smCurrPath,hhmmStr,yyyyStr);
        
        // xLogMsg("@A ~L4_SM granule(%s) hourIndex %-2.2d HHMM %s\n",l4_smCurrPath,hourIndex,hhmmStr);
        // OPEN current l4_SM HDF5 FILE ---------------------------------------
        l4_sm_file_id = frm_hdf5_Openfile(l4_smCurrPath,fxName);
    
        // READ L4_SM DATASET -----------------------------------------
        // allocate, read this variable, return its data in new pointer, dsNativeData_p
        
        // first, prepend the datasets group, Geophysical_Data
        sprintf(hDatasetname,"/Geophysical_Data/%s",datasetName);
        
        // open, allocate buffer, read dataset, returning data 'dsNativeData_p'
        dsNativeData_p = l4_sm_get_dataset(l4_sm_file_id, hDatasetname,&dRank, dims, &nElem9km);
        
        // reset fully qualified hDatasetname, since we re-unse it below
        strcpy(hDatasetname,"");
        
        // Calculate the min,max,sum,mean for this variable for JUST this current 
        //   3-hour dataset, across the global EASEGRID L4_SM spatial domain 1624x3856
        hourlyMean = l4_sm_summarizeVariable(sesCtl_p->logf,datasetName,dims,dsNativeData_p,&hourlyMin,&hourlyMax);
        
        // Next, store hourly min,max,mean for each of (8) 3-hour periods, to check later
        hrMeanArray[grnIdx] = hourlyMean;
        hrMinArray[grnIdx]  = hourlyMin ;
        hrMaxArray[grnIdx]  = hourlyMax ;
        
        // VERTICAL-DRILL-DOWN: Iterate through global FP native spatial domain 721x1152, 
        //  accumulating min,max,mean for each variable across ALL 24 hours worth of data
        for(r=0; r < nRows9km; r++)
        {
            for(c=0; c < nCols9km; c++)
            {
                serialIdx = (r * nCols9km )+c ;
                pix = dsNativeData_p[serialIdx];
                // Accumulate the pixel wise sum for this variable
                if(pix > missing_value )
                {
                    nQual++ ;
                    vSum[serialIdx] += pix;
                    // TBD: not the most robust logic to trigger initial
                    // min,max scan... 
                    if( hitFirstValue ==0)
                    {
                        hitFirstValue =1 ;
                        // Remember: re-initialize ALL pixel-wise min,max
                        for(g=0;g < nElem9km; g++)
                            vMin[g] = vMax[g]= dsNativeData_p[g];
                    } else {
                        if(pix > vMax[serialIdx]) 
                        {
                            vMax[serialIdx] = pix; nMaxAdjusts++;
                        }
                        if( pix < vMin[serialIdx]) 
                        {
                            vMin[serialIdx] = pix; nMinAdjusts++ ;
                        }
                    } // end:: block if valid to adjust min,max 
                } else { // end:: pix is NOT a missing value..
                    // <JMG> 2017-02-06T17:12, added to fix incorrect missing value problem
                    // where a 0.0 was placed where missing value should be 
                    vMin[serialIdx]       = L4C_MET_MISSING_VALUE;
                    vMax[serialIdx]       = L4C_MET_MISSING_VALUE;
                    nMissingPix++ ;
                }
            } // end:: c column loop
        } // end:: r row loop

        // ASSERT: Information in dsNativeData_p should now all have been transferred to either
        // the daily average vSum[], vMean[], daily min vMin, or daily max vMax[].
        // so we can release it now
        memFree(dsNativeData_p);

        // for diagnostics, display 3-hour wise statistics
        // showHourly_stats(sesCtl_p->logf,sesCtl_p->nL4_SM_Granules, hrMinArray,hrMaxArray,hrMeanArray);
        
        // Close this latest hourly FP file 
        status = H5Fclose(l4_sm_file_id);

    } // end:: grnIdx loop
    
    // 9KM loop: After All (8) 3-hour granules of L4_SM data have been traversed,
    // populate the VERTICAL-DRILL-DOWN pixel wise means -- e.g. each L4_SM var's daily average.
    
    for(serialIdx=0; serialIdx < nElem9km; serialIdx++)
    {
        // <JMG> 2017-02-06 added "if" to correct missing value problem
        if(vMin[serialIdx] > L4C_MET_MISSING_VALUE && vMax[serialIdx] > L4C_MET_MISSING_VALUE)
        {
            vMean9km_p[serialIdx] = vSum[serialIdx] / (float32)sesCtl_p->nL4_SM_Granules;
        } else {
            // <JMG> this was not explicitly assigned in prior logic when it should have been
            // which left 0.0 in pixels where a missing value should have been assigned.
            vMean9km_p[serialIdx] = L4C_MET_MISSING_VALUE;
        }
    }

    // <JMG> 2017-02-08T09:46 removed call to 'calculateGranuleActualRange() which is
    // un-needed and does NOT correctly work with the L4_SM variable handling above.
    // To supply the HDF5 dataset attributes for observed min,max,mean we now use the same
    //  (correct) statistical results produced by l4_sm_SummarizeVariable() which does reflect
    //  the correct min,max,mean values.
    // DEPRECATED: nQual =  calculateGranuleActualRange(sesCtl_p,datasetName, L4_SM_MISSING_VALUE, vMean9km_p, nElem9km);
    // 

#ifdef DIAG_ONLY    
    // use global 1D vector of (1664040) terrestrial addresses to pick a valid one to sample from
    g = terrAddr[ R_RANDOM_INT(100)];  // here we just sample from initial 100 terrestrial pixels
    xLogMsg("\nDIAG RandomlySamplingTerrestrialPixel %d\n",g);
    xLogMsg("DIAG-CHECK-vMean SamplePixel %d (%s) sum %.1f mean %.5f\n",g,datasetName,vSum[g],vMean[g]);
    xLogMsg("DIAG-CHECK-vMin  SamplePixel %d (%s) sum %.1f min  %.5f\n",g,datasetName,vSum[g],vMin[g]);
    xLogMsg("DIAG-CHECK-vMax  SamplePixel %d (%s) sum %.1f max  %.5f\n",g,datasetName,vSum[g],vMax[g]);
    memFree(terrAddr);
#endif
    
    // CREATE current var as a HDF5 dataset
    sprintf(hDatasetname,"%s_M09_AVG",datasetName);
    // convert display form to all upper case
    strcpy(dLbl,hDatasetname);
    l4c_to_upper(dLbl);
    xLogMsg("\n@A In(%s): CREATING EASEGRID 9KM dataset [%s] dspace9km %d ~ \n",
            fxName, dLbl, sesCtl_p->dspace9km);
            
    dsTypeId = H5T_NATIVE_FLOAT ;
    dset_id = createMetHdf5Dataset(sesCtl_p, IsEG_Global_9km, datasetName, dsTypeId, vMean9km_p);
    
    // referencing each granule, add provenance metadata, indicating specific source of this data
    if(dset_id >= 0)
    {
        xLogMsg("\n@A In(%s): for %s Adding INPUTPOINTER provenance from %d granules ~ \n",
            fxName, dLbl, sesCtl_p->nL4_SM_Granules);

        for(g= 0; g < sesCtl_p->nL4_SM_Granules; g++)
        {
            sprintf(datasourceLbl,"INPUTPOINTER_L4_SM_%-3.3d",g+1) ;
            frm_hdf5_CreateStringAttribute(dset_id, datasourceLbl,sesCtl_p->inputL4_SM_pathList[g]);
        }
        H5Dclose(dset_id);
    }

    // release heap memory used locally
    // NOTE: vMean[] not required here since we can store mean value directly in vMean9km_p[] array
    //       added vMin,vMax 
    memFree(vSum);
    memFree(vMin);
    memFree(vMax);
    memFree(vMean9km_p);

 return 0;
} // end::l4_sm_verticalDrilldown()
