// ---------------------------------------------------------------------
//   module     : l4c_MET_drivers.c
//   Purpose    : implement driver level functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : see AppVers macro in l4c_MET.h
//   revised    : v7.0-2015-07-31T09:51:00, V7 fork to support GEOS5 FP "TS" F/T and/or
//                pending new L3_SM_P passive radiometer derived F/T at 3KM.
//   Revision   : 2014-07-22, re-integration into L4_C v5.0.1 in progress
//   Notes      : Several new fields added for D5 representing either new
//                capabilities needed to support the use of F/T in L4_C_DAILY,
//                or (VPD) that were omitted earlier.  New fields are:
//                -- FT_ROLLING_AGE_3KM  ...from Freeze Thaw AGE (uint8,3km)
//                -- vpd (float32, 9KM, "units=Pa") in SMAP_L4_C_MET.h5
//                -- smroot_min (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                -- smroot_max (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                 and new static 3KM float32 layers "smroot_min","smroot_max"
//              in SMAP_L4_C_SOC.h5
//


#include "l4c_MET.h"
#include "l4c_MET_prot.h"

// -----------------------------------------------------------------------------
// function  : driverLogic_GEOS5_FP_Meteorology, prior named driverLogic_FP()
// inputs    : SesCtl *sesCtl_p,
//             char *dayOfYearStr
// purpose   : Implement processing on suite of variables from GEOS5 FP 
// revised   : 2015-08-28T1:12, extended orig geog5_fp_MeteorologyDrilldown()
//             to also now support DoUse_GEOS5_TS to create freeze thaw
// revised   : 2015-07-31T09:53.
int32 driverLogic_GEOS5_FP_Meteorology(SesCtl *sesCtl_p,char *dayOfYearStr)
{
 int varIdx        = 0 ;
 char fxName[256] = {0};
 
    strcpy(fxName,__func__);
    
    xLogMsg("@A In(%s) ACTION: preprocessing %d variables of GEOS-5 FP data from %d granules for DOY [%s]~\n",
              fxName, sesCtl_p->nFP_Variables, sesCtl_p->nFP_Granules,dayOfYearStr);   
            
    // Call main drill-down driver once per variable. Extended to also 
    //  now support GEOS5 FP "TS" derived freeze thaw classifier when in
    //  mode DoUse_GEOS5_TS.
    for(varIdx = 0; varIdx < sesCtl_p->nFP_Variables ; varIdx++)
    {
        geos5_fp_MeteorologyDrilldown(sesCtl_p,dayOfYearStr, sesCtl_p->varList_GEOS5_FP[varIdx]) ;
    } // end:: varIdx loop...

    xLogMsg("@A In(%s) COMPLETED %d variables of GEOS-5 FP data from %d granules for DOY [%s]~\n",
              fxName, sesCtl_p->nFP_Variables, sesCtl_p->nFP_Granules,dayOfYearStr);   

 return 0;
} // end::driverLogic_GEOS5_FP_Meteorology()



// -----------------------------------------------------------------------------
// function: driverLogic_l4_SM()
// purpose : Implement L4_SM processing logic as a vertical drill down
// inputs  : SesCtl *sesCtl_p,char *dayOfYearStr
// returns : int32, varIdx count of granules processed
// notes   : a) iteration across K variables is done in the outside loop
//           b) for each given L4_SM variable, we then perform the spatial
//              drilldown, pixel by pixel, to derive the desired mean.
int32 driverLogic_L4_SM(SesCtl *sesCtl_p,char *dayOfYearStr)
{
int varIdx = 0 ;

    xLogMsg("@A ~ACTION: preprocessing L4_SM data for %d granules DOY [%s]\n",
                    sesCtl_p->nL4_SM_Granules,dayOfYearStr);

    for(varIdx =0; varIdx < sesCtl_p->nL4_SM_Variables; varIdx++)
    {
        l4_sm_verticalDrilldown(sesCtl_p,dayOfYearStr, sesCtl_p->varList_L4_SM[varIdx]);    
    
    } // end:: varIdx loop
    
    // FUTURE: address reading soil moisture from 'gph' in wetness units
    // and then reading 'lmc' granule to retrieve new 
            
 // return count of granules processed for this variable            
 return varIdx;
} // end::driverLogic_L4_SM()


// -----------------------------------------------------------------------------
// function: driverLogic_L3_SM_A()
// Purpose : From the L3_SM_A granule for this day, read the "surface_flag" uint16 variable
//           and extract the Freeze Thaw state binary {0|1} flag from bit (6) 
//           the perform the following actions:

//          1) Rread the EASEGRID_row_index and EASE_column_index uint16 fields.
//            loop through 3km space, noting each pixel for which a valid 
//            row,col index value exists, marking these as IN the current swath,
//            build a uint8 3km "l3smaMaskImg_p" mask, containing this information 
//            if there is sufficient memory.

//          2) Read the uint16 "surface_flag" and using the l3smaMaskImg_p mask,
//           extract just the Freeze-Thaw bitflag from bit (6) for the pixels in the
//           current swath. For the pixels not in the swath, set to 2.

// Inputs : SesCtl *sesCtl_p,
//          char *dayOfYearStr

// history: <JMG> added uint8, 3km 2D FT_ROLLING_AGE_M03 layer,
//          which tracks how many days OLD the Freeze Thaw value "LAST-GOOD" 
//          was, as an integer count {0,1,2,3,4...} etc.
// Notes  : <JMG> 7/30/2014, a) the L3_SM_A's "surface_flag" is now read but NOT
//             passed through to SMAP_L4_C_MET.h5, since the new 
//             FT_STATE_UM_M03 (uint8) written to SMAP_L4_C_MET.h5
//             will take its place. FT_STATE_UM_M03 contains just the
//             (4) bits we need and use: 
//                 -- radar F/T bit,
//                 -- tsurf F/T bit,
//                 -- radar F/T QC bit and
//                 -- "which F/T used" bit 0=tsurf F/T was used, 1=radar F/T used
//
//        : b) we now use surface_flag bit 15 as an "In vs Out of swath" indicator.
//
//        : c) the FT_ROLLING_M03 and FT_ROLLING_AGE_M03 layers in SOC are
//             now maintained only by l4c_DAILY.ex.
 
int32 driverLogic_L3_SM_A(SesCtl *sesCtl_p,char *dayOfYearStr)
{
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package
    
 int32 nRows3km = NROW3KM   ; // 4872 rows
 int32 nCols3km = NCOL3KM   ; // 11568 columns
 int32 nElem3km = N_3KM_PIX ; // NROW3KM * NCOL3KM = 56359296....
 
 int32   grnIdx        = 0 ;
 int32   hourIndex     = 0 ;
 int32   hitFirstValue = 0 ;
 int32   nQual = 0;
 int32   dRank = 2;
 int32   g,r,c   ;
 int32   serialIdx3km =0;
 int32   nAdjusted_FT= 0; // count of adjusted FT pixels
 
 hsize_t dims[4]   = {0} ;
 herr_t  status  = -1    ;

 hid_t   dset_id = -1;
 hid_t   dsType_id = -1 ;
 hid_t   soc_file_id = -1 ;
 
 char   *l3_sm_a_CurrPath = NULL;
 char   datasetName[FILENAME_MAX]  ={0};
 char   hDatasetname[FILENAME_MAX] ={0};
 char   socPath[FILENAME_MAX] ={0};
 char   fxName[256] = {0} ;
 
 // DEPRECATED uint8  *ftRollingLast_p  = NULL; LAST-GOOD F/T 3km array from SOC or MET file
 uint8  *ftStateUm_M03_Img_p = NULL; // <JMG> 2014-07-25T18:41, added new uint8 consolidating old surface_flag info
 uint8  *freezeThawQC_p      = NULL; // new derived FT QC layer, extracted from big uint32 retrieval_qual_flag
 uint16 *surfaceFlagImg_p    = NULL; // "surface_flag" 2D : (4872,11568) input 3KM image raster 
 
 uint8  freezeThawQC_Pix = 2 ; // 0=unreliable F/T, 1=reliable F/T and  2=out-of-swatch
 uint8  freezeThawStatePix =0; // {0|1} if F/T valid and in SWATH otherwise set to 2 
 
 uint16 uPix = 0;    // "surface_flag" pixel, 3Km uint16 2D pixel variable
 uint16 uSurfaceFlagPix = 0;    // "surface_flag" pixel, 3Km uint16 2D pixel variable
 uint8  freezeThawBit   = L3_SM_A_FREEZE_THAW_BIT ; // normally, F/T {0|1} bit is bit 6 
 
    // TASK 1: process the freeze-thaw state flag in surface_flag
    strcpy(fxName,__func__);
    xLogMsg("@A In(%s) ~ACTION: preprocessing L3_SM_A 3KM data for %d granules DOY[%s]\n",
                    fxName, sesCtl_p->nL3_SM_A_Granules,dayOfYearStr);
                    
    dims[0] = NROW3KM  ; //  4872
    dims[1] = NCOL3KM  ; // 11568

    // Pre-fetched L3_SM_A daily path, via NVPcollect using: 
    // nsbGetValueByKey(nsbMgr_p,"L4C_INPUT_MET_L3_SM_A_PATH");
    l3_sm_a_CurrPath = sesCtl_p->inputL3_SM_A_Path ;
        
    // -------------------------------------------------------------------
    // <JMG> 2014-07-23, L3_SM_A (HDF5 file) opened now in l4c_InitSession()
    //  build spatial mask so we restrict output to just the pixels in the current SWATH

    // ALLOC a 3km 2D "uint8" raster buffer to store the current day's F/T QC in.
    // note we later write this as bit x(3?) in FT_STATE_UM_M03 in SMAP_L4_C_MET.h5
    freezeThawQC_p=(uint8 *)memCalloc(nElem3km,sizeof(uint8));
    memObjSetName(freezeThawQC_p,"freezeThawQC_p");

    // First, Read L3_SM_A "surface_flag" DATASET ----------------------------------
    //  Allocate, then read "surface_flag", return its data in new pointer, surfaceFlagImg_p
    //   -- first, prepend the HDF5 dataset "group", Soil_Moisture_Retrieval_Data.
    //  Note that the MACRO named L3_SM_A_FT_DATASETNAME now contains the fully
    //  qualified HDF5 dataset name that includes the 'group' 
    //  so this is now: "/Soil_Moisture_Retrieval_Data/surface_flag"
    
    // Set the HDF5 dataset name as a fully qualified name including the GROUP,
    strcpy(hDatasetname, (char *)L3_SM_A_FT_DATASETNAME);
    
    // Define the simple dataset name "surface_flag" here..
    // a 3KM 2D uint16 field...
    strcpy(datasetName,(char *)L3_SM_A_SURF_FLAG_DATASETNAME); 
    
    // Open, allocate buffer, read dataset, return a 3km 2D (uint16) 'surfaceFlagImg_p' 
    // READ "surface_flag" (3km uint16) from L3_SM_A
    surfaceFlagImg_p = (uint16*)l3_sm_a_get_surface_flag(sesCtl_p, hDatasetname, &dRank, dims);

    // PROCESS surface_flag, extract (4) bits we need, load FT_STATE_UM_M03 array which
    // which is stored in imagePkg_p for later write out
    l3_sm_a_ProcessFreezeThaw(sesCtl_p,surfaceFlagImg_p, freezeThawQC_p);
   
    // FREE the F/T QC FT_QC_M03 array
    memFree(freezeThawQC_p);

    // FREE surface_flag buffer, since FREEZE-THAW info is now in ftRollingState_p buffer
    // free(surfaceFlagImg_p);   check if this was where the memory bug was..
    memFree(surfaceFlagImg_p);   
    
 return 0;
 
} // end::driverLogic_L3_SM_A()



