// ---------------------------------------------------------------------
//   module     : l4c_MET_freeze_thaw.c
//   Purpose    : implement freeze-thaw specific functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : v7+, see AppVers macro in l4c_MET.h
//   Revised    : 2015-08-24T13:58:00, refinements enablined GEOS5-FP TS to substitute for SMAP F/T
//   Revised    : 2015-08-10T11:52:00
//   Revised    : 2014-10-21T10:09, provision test change, to improve quality of FT_STATE_UM_M03
//                prior to its going into L4CMDL
//   Revised    : 2014-07-29, consolidated info from "surface_flag" into FT_STATE_UM_M03
//                new uint8 3km output array for use by L4_C DAILY
//                FT_STATE_UM_M03 holds just the (4) bits of info we use:
//                a) ft_tsurf, b) ft_radar, c) ft_radar_qc, d) ft_in_swath
//                where ft_in_swath is bit 15 from surface_flag, 0=in swath
//                and 1=out-of-swath.
//   Revision   : 2014-07-22, re-integration into L4_C v5.0.1 in progress
//   Notes      : Several new fields added for D5 representing either new
//                capabilities needed to support the use of F/T in L4_C_DAILY,
//                or (VPD) that were omitted earlier.  New fields are:
//                -- FT_ROLLING_AGE_3KM  ...from Freeze Thaw AGE (uint8,3km)
//                -- vpd (float32, 9KM, "units=Pa") in SMAP_L4_C_MET.h5
//                -- smroot_min (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                -- smroot_max (float32, 9KM) now dynamic and back in SMAP_L4_C_SOC.h5
//                 and new static 3KM float32 layers "smroot_min","smroot_max"
//              in SMAP_L4_C_SOC.h5
//


#include "l4c_MET.h"
#include "l4c_MET_prot.h"


// ---------------------------------------------------------------------
// l3_sm_a_ProcessFreezeThaw(SesCtl *sesCtl_p)
// Purpose: perform these tasks:
//          1) READ the 3km (uint32) "retrieval_qual_flag" from 
//          group "Soil_Moisture_Retrieval_Data" in L3_SM_A, to EXTRACT the FT QC bits from it
//          2) PROCESS this latest FT QC, to flag any bad F/T itself
//          3) release local buffers...
// Returns: uint8 *ftStateUm_M03_Img_p -- ptr to 3KM image of bit packed freeze thaw info
// Inputs : SesCtl *sesCtl_p,
//        : uint16 *surfflag_p - ptr to surface_flag uint16 array for use as SWATH mask
//        : uint8  *freezeThawQC_p -- ptr to newest FT QC array, from bit extraction
// Revised: 2014-07-25, consolidated info from "surface_flag" into FT_STATE_UM_M03
//          new uint8 3km output array for use by L4_C DAILY
// Notes  : surface_flag (uint16,3km) pixel values are 55536 when out of SWATH at this time.
//          A macro L3_SM_A_SURFFLAG_OUTSIDE_SWATH defines this value.
//
// Notes  : We retain L3_SM_A (PSD document) logic where 1=Frozen, 0=Thawed
//
// Notes  : Various signatures of bit functions:
//
//        : uint32 frm_utilGetBits(const int32 strt_at,const int32 end_at,uint32 sourceData)
//          uint32 frm_utilRetrieveBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)
//          uint32 frm_utilAssignBits(const int16 startPos,const int16 endPos, void *destData, size_t dataSize,
//                      void *newValueMask, size_t maskSize)
//                    
int l3_sm_a_ProcessFreezeThaw(SesCtl *sesCtl_p,uint16 *surfflag_p,uint8  *freezeThawQC_p)
{
 ImgPkg *imagePkg_p  = l4c_GetImagePackage() ; // local pointer to support image package
    
 hid_t   dset_id = -1;
 hsize_t dims[4]   = {0} ;
 herr_t  ret  = -1    ;
 hid_t   file_id = sesCtl_p->l3smaFile_id ; // inherit file_id from open L3_SM_A file!
 hid_t   soc_file_id = -1 ;
 hid_t   dsTypeId = H5T_NATIVE_UINT16 ;
 uint8   freshen_FT = 0; 
 
 uint32 qaWord = 0 ;  // uint32 retrievalQualFlag pixel
 uint32 inputWord =0; // to enforce input bit manip word to exact size expected
 size_t diagDims[2] = {0} ;
 uint16 surfflag_pix = 0 ;
 int32  nRows3km     = NROW3KM   ; // 4872
 int32  nCols3km     = NCOL3KM   ; // 11568 
 int32  nElem3km     = N_3KM_PIX ; // ...
 int32  serialIdx    = 0;
 int32  whichQcBit   = 4 ; // L3_SM_A PSD  confirms which bit it is
 int32  r,c ;
 
 int32  nNotWithinSwath = 0 ; // count of pixels NOT within swath due to sentinel 32768|32769 occurrence
 int32  nValidQa = 0 ;
 
 size_t nInvalid = 0 ;
 
 uint32 *retrQualFlagImg_p = NULL;  // 2D : uint32, (4872,11568) input 3KM image raster for big QC variable
 uint32 *chkfrqImg_M03_p = NULL;
 uint8  *ftStateUm_M03_Local_p= NULL; // NEW 2014-07-25, FT_STATE_UM_M03 array
 
 uint8 ftStateUm_M03  = 0; // DESTINATION 8-bit word to store in dataset FT_STATE_UM_M03 in SMAP_L4_C_MET.h5
 uint8 ft_in_swath    = 0; // L3_SM_A f/t state from surface temperature,surface_flag bit 15
 // re-verify on ft_tsurf, ft_radar,ft_radar_qc and ft_in_switch whether means TRUE, or 0 means FALSE
 uint8 ft_radar       = 0; // L3_SM_A RADAR f/t state from surface temperature,surface_flag bit 6 
 uint8 ft_tsurf       = 0; // L3_SM_A GEOS5 f/t state from surface temperature,surface_flag bit 7
 uint8 ft_radar_qc    = 0; // L3_SM_A f/t radar QC from retrieval_qual_flag bit 4
 
 uint8 falseClassifier= 0;
 uint8 trueClassifier = 1;
 
 char   datasetName[FILENAME_MAX]  ={0};
 char   hDatasetname[FILENAME_MAX] ={0}; 
 char   hdf5Path[FILENAME_MAX] = {0} ; // L4_C_SOC.h5 file to write F/T, F/T QA vars to.
 char   freezeThawQcDataset[FILENAME_MAX]={0} ; // FT ROLLING...
 char   fxName[256] = {0};
 
    // Define name and alloc DESTINATION RESULT, freeze thaw QC field passed back out..
    strcpy(fxName,__func__);
    strcpy(freezeThawQcDataset,"FT_QC_M03");

    // Define datatype, then allocate and READ the "retrieval_qual_flag" 
    //   from a uint32 3km raster (225Mb of RAM required!) 
    dsTypeId = H5T_NATIVE_UINT32 ;
    // Fetch, store dataset name --F/T QA value: "retrieval_qual_flag"
    strcpy(datasetName, (char *)L3_SM_A_FT_QA_DATASETNAME);
    
    // inherit the pre-allocated FT_STATE_UM_M03 array, to populate here
    // set local convenience pointer to inherit
    ftStateUm_M03_Local_p = imagePkg_p->ftStateUm_M03_p;

    // pre-init
    for(r=0; r < nRows3km; r++)
    {
        for(c=0; c < nCols3km; c++)
        {
            serialIdx = (r* nCols3km)+c ;
            // Fetch next uint16 "surface_flag" pixel
            surfflag_pix = surfflag_p[serialIdx]; 
            // Extract the "in-or-out-of-swath" bitflag
            ft_in_swath = frm_utilGetBits(15,15, surfflag_pix);
            // Initialize to test
            ftStateUm_M03_Local_p[serialIdx] = ft_in_swath;
        } // end:: c loop
    } // end:: r loop

    
    // Allocate buffer for "retrieval_qual_flag", that contains the F/T QC word at bit 4
    retrQualFlagImg_p= (uint32 *)memCalloc(nElem3km, sizeof(uint32));
    memObjSetName(retrQualFlagImg_p,"retrQualFlag_p");
    
    // READ: L3_SM_A, "retrieval_qual_flag"--uint32 L3_SM_A QC contains freeze thaw QC at bit 4
    // int frm_hdf5_ReadData(const hid_t thisFile_id,const hid_t dsTypeId,char *datasetname_p,void *dataBuffer_p,char *callerLbl_p)
    ret =  frm_hdf5_ReadData(file_id, dsTypeId, datasetName, retrQualFlagImg_p, fxName);
    
    xLogMsg("@A In(%s): READ (%s) from file_id %d ret: %d ~\n",
            fxName, datasetName,file_id,ret);   

    // From within this big QC uint32 word, extract the specific 
    //    Freeze Thaw state's QA/QC bit(s)...
    for(r=0; r < nRows3km; r++)
    {
        for(c=0; c < nCols3km; c++)
        {
            // Derive 1D serialized form of 2D 3KM pixel raster location
            serialIdx = (r* nCols3km)+c ;
            
            // Initialize final uint8 at 0, then adjust later
            ftStateUm_M03 = 0 ;
            
            // Fetch next uint16 "surface_flag" pixel
            surfflag_pix = surfflag_p[serialIdx]; 
            
            // Fetch separate uint32  "retrievalQualflag" QC pix to get bit 4
            qaWord = retrQualFlagImg_p[serialIdx];
            
            // NOTE: uint32 frm_utilRetrieveBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)
            // is an alternative way to passively get a range of bits, that works with 32 bit scalars
            
            // Fetch bit 15, (0=InSwath,1=OUT-of-Swath) and evaluate if in swath or not. 
            // JPL says this may or may not be a good SWATH classifier, ok to use for now
            // but not necessarily to rely on. Having InSwath mapped to 0 could not be
            // more confusing, but we can't control that.  
            inputWord = (uint32)surfflag_pix;
            ft_in_swath = frm_utilGetBits(15,15, inputWord);
            
            // READ fetch the bits we need from 3KM surface_flag_m03[] and retrQualFlagImg_p[]
            inputWord = (uint32)qaWord ;
            ft_radar_qc = frm_utilGetBits(4 , 4, inputWord); // from var: 'retrQualFlagImg_p'
            
            // <JMG> 2014-10-21, use new Coastal moved version of bitflag mappings
            inputWord = (uint32)surfflag_pix ;
            ft_radar    = frm_utilGetBits(7 , 7, inputWord); // from surfflag_pix...
            
            // LOGIC clarification: surface_flag bit 8 returns (0|1), as 1=frozen  0=thawed
            inputWord = (uint32)surfflag_pix ;
            ft_tsurf    = frm_utilGetBits(8 , 8, inputWord); // from surfflag_pix...
            
            // REMINDER: for L3_SM_A when pixel is IN SWATH this bit is set to 0 not 1
            // however counter-intuitive that is. 
            if( ft_in_swath == IsInSwath ) // if 0, we ARE in swath
            {
                nNotWithinSwath++ ;         
            } else {
                // Retrieve current uint32 pixel from big 'retrieval_qual' 3KM array
                nValidQa++ ;            
            } // end:: block if valid WITHIN-SWATH pixel
            
            // <JMG> NEW CONSOLIDATION 2014-07-25, LOAD (4) pass-through bits in uint8
#ifdef NOT_YET
            // alternative approach, not currently used,uses frm_utilAssignBits()
            ftStateUm_M03 = frm_utilAssignBits(0,0, &ftStateUm_M03, sizeof(uint8), &ft_tsurf   , sizeof(uint8));
            ftStateUm_M03 = frm_utilAssignBits(1,1, &ftStateUm_M03, sizeof(uint8), &ft_radar   , sizeof(uint8));
            ftStateUm_M03 = frm_utilAssignBits(2,2, &ftStateUm_M03, sizeof(uint8), &ft_radar_qc, sizeof(uint8));
            ftStateUm_M03 = frm_utilAssignBits(3,3, &ftStateUm_M03, sizeof(uint8), &ft_in_swath, sizeof(uint8));
#endif
            // Explanation: we only toggle specified individual bits ON or =1, when the
            //  indicated boolean flag (ft_surf|ft_radar|ft_radar_qc|ft_in_swath) is true or =1,
            //   otherwise, LEAVE the pre-initialized bit set as it was before,typically as 0.
            
            // DoUse_GEOS5_TS, DoUse_ACTIVE_RADAR, DoUse_ACTIVE_TSURF, DoUse_PASSIVE_RADIOMETER
            if( sesCtl_p->ftSourceMode == DoUse_ACTIVE_TSURF)
            {
                // Define bit 0,0 purely on basis of ft_surf's value
                ftStateUm_M03 = frm_utilAssignBits(0,0, &ftStateUm_M03,sizeof(uint8),&ft_tsurf,sizeof(uint8));
                
                // FORCE bit 1 as FALSE here, since only bit 0 (tsurf) is relevant
                ftStateUm_M03 = frm_utilAssignBits(1,1, &ftStateUm_M03,sizeof(uint8),&falseClassifier,sizeof(uint8));
                // ftStateUm_M03 = (uint8)frm_utilSetBits(0,0, &ftStateUm_M03,sizeof(uint8));

               // ASSIGN ft_radar_qc bits, here always TRUE
               ftStateUm_M03 = (uint8)frm_utilSetBits(2,2, &ftStateUm_M03,sizeof(uint8));
               
               // ASSIGN bit (6,6) to TRUE to explicitly indicate that DoUseACTIVE_TSURF was specified.
               ftStateUm_M03 = frm_utilAssignBits(6,6, &ftStateUm_M03,sizeof(uint8),&trueClassifier,sizeof(uint8));
               // set bit(4,4) to false to explicitly indicate that GEOS5 TS was NOT used
               ftStateUm_M03 = frm_utilAssignBits(4,4, &ftStateUm_M03,sizeof(uint8),&falseClassifier,sizeof(uint8));
               // set bit(7,7) to false to explicitly indicate that ACTIVE RADAR was NOT used
               ftStateUm_M03 = frm_utilAssignBits(7,7, &ftStateUm_M03,sizeof(uint8),&falseClassifier,sizeof(uint8));
              
            } else if( sesCtl_p->ftSourceMode == DoUse_ACTIVE_RADAR) {
                
                // <JMG> 2015-09-08, presently, we ECHO ACTIVE_RADAR's F/T state to bit0
                // for backward compatibility with L4CMDL, in case it wants to fetch F/T from bit0
                // regardless of whether ACTIVE RADAR or ACTIVE TSURF drove the F/T classification
                // ECHO F/T state to both bit0 and bit1
                ftStateUm_M03 = frm_utilAssignBits(0,0, &ftStateUm_M03,sizeof(uint8),&ft_radar,sizeof(uint8));
                
                // Define bit 1,1 purely on basis of ft_radar
                ftStateUm_M03 = frm_utilAssignBits(1,1, &ftStateUm_M03,sizeof(uint8),&ft_radar,sizeof(uint8));
                // ftStateUm_M03 = (uint8)frm_utilSetBits(1,1, &ftStateUm_M03,sizeof(uint8));
      
                // ASSIGN bits (2,2) purely on basis of ft_radar_qc as supplied
                ftStateUm_M03 = frm_utilAssignBits(2,2, &ftStateUm_M03,sizeof(uint8),&ft_radar_qc,sizeof(uint8));

               // ASSIGN bit (7,7) to TRUE to explicitly indicate that DoUseACTIVE_RADAR was specified.
               ftStateUm_M03 = frm_utilAssignBits(7,7, &ftStateUm_M03,sizeof(uint8),&trueClassifier,sizeof(uint8));
               // set bit(4,4) to false to explicitly indicate that GEOS5 TS was NOT used
               ftStateUm_M03 = frm_utilAssignBits(4,4, &ftStateUm_M03,sizeof(uint8),&falseClassifier,sizeof(uint8));


            } else if( sesCtl_p->ftSourceMode == DoUse_GEOS5_TS) {

                nInvalid++ ;  // should never get to this branch, but count occurrences if we do
            }

            // <JMG> 2015-09-08T09:54:00 For L3_SM_A -- pass through its verbatim definition of swath status
            // as InSwath(=0), or not in swath(=1)
            ftStateUm_M03 = frm_utilAssignBits(3,3, &ftStateUm_M03,sizeof(uint8),&ft_in_swath,sizeof(uint8));

            // ASSERT: when L3_SM_A is used (RADAR or TSURF), we also explicitly clear (set to 0) bits 4 and 5
            // since bits 4 and 5 are ONLY used when GEOS5 TS is enabled
            ftStateUm_M03 = frm_utilAssignBits(4,4, &ftStateUm_M03,sizeof(uint8),&falseClassifier,sizeof(uint8));
            ftStateUm_M03 = frm_utilAssignBits(5,5, &ftStateUm_M03,sizeof(uint8),&falseClassifier,sizeof(uint8));

            // FINALLY, RELOAD final bitpacked pixel with all (4) bits having potentially adjusted
            ftStateUm_M03_Local_p[serialIdx] = ftStateUm_M03 ;
            
        } // end:: c col loop
    } // end:: r row loop
    
    if(nInvalid >0)
      xLogMsg("@W In(%s) Invalid FT source policy %d (%s) for this handler~\n",
          fxName,sesCtl_p->ftSourceMode,sesCtl_p->ftSourceStr);
    
    // FREE up local "retrieval_qual_flag" -- the big (uint32, 3km) array
    memFree(retrQualFlagImg_p);
    
#ifdef DIAG_ONLY
    // Prior to evaluating change, generate a ftStateUm_M03_Img_p diagnostic image
    xLogMsg("@D In(%s) Generating diag image of FT_STATE_UM_M03 after bit assignment ~\n",fxName);
    diagDims[0] = nRows3km;
    diagDims[1] = nCols3km;
    frm_utilDiagImage("FT_STATE_UM_M03_EVALORR","FT_STATE_UM_M03_DIAG",IsUint8,
                nElem3km, diagDims, ftStateUm_M03_Local_p);
#endif
    // report status
    xLogMsg("@A In(%s): ~Dataset %s PROCESSED Pixels IN-SWATH %d OUT-OF-SWATH %d \n",
            fxName, datasetName, nValidQa, nNotWithinSwath);

    // return status only 
    return 0;
  
} // end:: l3_sm_a_ProcessFreezeThaw()
