// ---------------------------------------------------------------------
//   module     : l4c_MET_meta.c
//   Purpose    : implement basic metadata support functions for l4c_MET.c
//   Author     : joe glassy
//   Version    : see AppVers macro in l4c_MET.h
//   Revision   : 2015-08-03T18:07:00, updated for GEOS5 TS derived F/T when
//                the L3_SM_A is missing.
//   Revision   : 2016-06-10T16:31, <JMG>removed L4_C_MET_SW_VERSION as redundant
//   Revision   : 2014-07-28T14:29:00, re-factored orig metadata writes from
//                l4c_MET_fx.c, l4c_io.c.  Integration into L4_C v5.0.1 in progress
//   Notes      : n/a

#include "l4c_MET.h"
#include "l4c_MET_prot.h"

// --------------------------------------------------------------------
// Function : writeNvpAsAttribute
// Purpose  : using a supported NVP keyword, create a string attribute
// Inputs   : const hid_t file,const char *keyword,char *attrForceName)
// Author   : joe glassy
// Revised  : 2014-10-21T21:23:00,ruby(ASUS,home,jmg)
// Returns  : hid_t attr_id if successful, or -1 if not
// Notes    : The HDF5 attribute written using the keyword as the
//            attributes name and the NVP's 'value' as the attribute value
//      
hid_t l4c_writeNvpAsAttribute(hid_t file_id,const char *keyword,const char *attrForceName)
{
 NsbObj *nsbObj_p = l4c_GetNsbOverseer(); 
    
 char attrValu[2048]= {0} ; 
 char nvpBuf[2048]= {0} ;
 int keyResult = -1 ;
 int status = 0 ;
 hid_t attr_id = 0 ;
  

    if( nsbDoesKeyExist(nsbObj_p,keyword))
    {
        // fetch value of this supported keyword
        strcpy(attrValu, nsbGetValueByKey(nsbObj_p,keyword,nvpBuf));
        
        // Force use of 'attrForceName' instead of 'keyword' as the
        // HDF5 attribute name or label.
        if(attrForceName != NULL && strlen(attrForceName)>= 1)  
        {
            // attr_id = l4c_createHdf5Attribute(file_id,attrForceName, attrValu);
            status = frm_hdf5_WriteAttributeAsType(file_id,"/", attrForceName,IsString,1,attrValu);
            
        } else {
            // use the 'keyword' verbatim as the HDF5 attribute name
            // attr_id = l4c_createHdf5Attribute(file_id,keyword, attrValu);
            status = frm_hdf5_WriteAttributeAsType(file_id,"/", keyword,IsString,1,attrValu);
        }
    } else { // end:: block if keyword IS a supported one
        // notify caller that this keyword was invalid or not supported
        xLogMsg("@E l4c_writeNvpAsAttribute: NO key(%s) found - suppressed attrib write!~\n",
            keyword);
    }
    
 // TBD: must add code now to explicitly retrieve this most
 // recently created unique attr_id 
 // The attr_id returned here is not real, only a placeholder
 return 0 ; 
 
} // end::l4c_writeNvpAsAttribute()


// ---------------------------------------------------------------------
//   function   : l4c_WriteFile_metadata()
//   inputs     : SesCtl *sesCtl_p -- ptr to session controls
//                hid_t file_id    -- file_id HDF5 ID
//   purpose    : write basic file level metadata to SMAP_L4_C_MET.h5 and
//                intermediate SMAP_L4_C_SOC.h5
//   Author     : joe glassy
//   Version    : v5.0.1, see AppVers macro in l4c_MET.h
//   Revision   : 2014-07-28T14:29:00, re-factored orig metadata writes from
//                l4c_MET_fx.c, l4c_io.c.  Integration into L4_C v5.0.1 in progress
//   Notes      : n/a
void l4c_WriteFile_metadata(SesCtl *sesCtl_p,hid_t file_id)
{
 char fxName[256] = {0} ;
 char *inputPtrGranule_p = NULL;
 char inputPtrAttribName[128] = {0} ;
 char uuidString[64] = {0};
 char utcDatetimeStr[64] = {0};
 char localDatetimeStr[64] = {0} ;
 time_t secNow = time(0);
 
 char  ftLegendString[256] = {0} ;
 char  ftLegend0[128] ;
 char  ftLegend1[128] ;
 char  ftLegend2[128] ;
 char  ftLegend3[128] ;
 char  ftLegend4[128] ;
 char  ftLegend5[128] ;
 char  ftLegend6[128] ;
 char  ftLegend7[128] ;
 
 int ithGrn = 0 ;
 
  strcpy(fxName,__func__);
 
  // <JMG> 2015-09-04, TBD: if GEOS5 TS used for F/T, we need a slightly different legend!!
  strcpy(ftLegendString,
    "FT bitflag Legend: all bits (0..8) are used as defined below");
  strcpy(ftLegend0,"bit 0 -- defines the boolean freeze thaw state (0|1) regardless which source defined it.");
  strcpy(ftLegend1,"bit 1 -- RADAR: defines the FT_RADAR boolean freeze thaw state, relevant ONLY when ACTIVE RADAR logic used");
  strcpy(ftLegend2,"bit 2 -- QC: the QualityControl (Good1=,Bad=0) from L3_SM_A. It's always TRUE if GEOS5 or TSURF used");
  strcpy(ftLegend3,"bit 3 -- SWATH: where InSwath=0, OutOfSwath=0 indicator");
  strcpy(ftLegend4,"bit 4 -- is set TRUE only when the GEOS5 TS was explicitly specified");
  strcpy(ftLegend5,"bit 5 -- pixel is InTerrestrial land domain(1), or Not(0)");
  strcpy(ftLegend6,"bit 6 -- TSURF? set TRUE only when L3_SM_A's tsurf is specified for F/T logic, false otherwise");
  strcpy(ftLegend7,"bit 7 -- RADAR? set TRUE only when ACTIVE RADAR is specified for F/T logic, false otherwise");
    
    // capture production time, write: L4_C_PRODUCTION_DATETIME 
    frm_utilGetISO8601_datetime(secNow,localDatetimeStr,utcDatetimeStr);
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_PRODUCTION_DATETIME",utcDatetimeStr);
    
    // Write basic file level metadata
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_SUBSYSTEM_RELEASE", sesCtl_p->SubsysReleaseTag);
    // Assumption is that we write both and whichever is invalid or not current is ignored
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_SW_VERSION", sesCtl_p->swVersion);
        // <JMG> 2016-06-10, removed L4_C_MET_SW_VERSION as redundant
    
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_DATASET_VERSION", "v7.0-R12");  
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_MET_DOI", "DOI-TBD");
    
    frm_hdf5_CreateStringAttribute(file_id,"SMAP_SCIENCE_VERSION_ID",sesCtl_p->ScienceVersID);
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_CRID", sesCtl_p->CRID);
    
    // Explictly convey WHICH Freeze Thaw source was used
    frm_hdf5_CreateStringAttribute(file_id,"L4_C_MET_FT_SOURCE",sesCtl_p->ftSourceStr);
    xLogMsg("@A In(%s) FT_SOURCE method reported as a MET file level attribute: %s ~\n",fxName,sesCtl_p->ftSourceStr);
    
    // Create UUID and store as a file level attribute for now
    //          : displayMode (options:
    //             0 --> displays real UUID as a  packed string
    //             1 --> displays real UUID as a normal hypenated couplet form of output, as from uuid_unparse()
    //             2 --> forces generation of a pseudo UUID and displays it in hyphenated couplet form
    //             3 --> forces generation of a pseudo UUID and displays it as a packed string
    //           or if displayMode==2, produces a dummy random placeholder (not authentic) UUID
    frm_utilGenerateUUID(uuidString,1) ;
    strcpy(sesCtl_p->granuleUUID,uuidString);
    frm_hdf5_CreateStringAttribute(sesCtl_p->met_file_id,"GRANULE_UUID",uuidString);

    // Mirror L4_SM as much as feasible
    frm_hdf5_CreateStringAttribute(file_id,"Title"     ,"SMAP L4 Carbon model (MDL) Data Granule");
    frm_hdf5_CreateStringAttribute(file_id,"Source"    ,"L4_C ATBD, Dr. John Kimball, NTSG, University of Montana");
    frm_hdf5_CreateStringAttribute(file_id,"References","see SMAP L4_C Product Specification Document");
    frm_hdf5_CreateStringAttribute(file_id,"History","file written by NASA GSFC GMAO SMAP L4 production system");
    frm_hdf5_CreateStringAttribute(file_id,"Institution","produced at: NASA GSFC GMAO and developed by University Of Montana, Missoula, MT 59812");
    frm_hdf5_CreateStringAttribute(file_id,"Filename"   ,sesCtl_p->outputMetPath);
    frm_hdf5_CreateStringAttribute(file_id,"Conventions","CF v1.x");
    frm_hdf5_CreateStringAttribute(file_id,"Contact",  "http://www.ntsg.umg.edu" );
    frm_hdf5_CreateStringAttribute(file_id,"File_format","HDF5");
    frm_hdf5_CreateStringAttribute(file_id,"Comment","NASA SMAP L4_C pre-processor l4c_MET.ex");

    // WRITE INPUTPOINTER (provenance sources)
    strcpy(inputPtrAttribName,"L4C_MET_INPUTPOINTER_ANC");
    frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName, "SMAP_L4_C_Ancillary.h5");

    strcpy(inputPtrAttribName,"L4C_MET_INPUTPOINTER_RIP");
    frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName, sesCtl_p->ripPath);

    strcpy(inputPtrAttribName,"L4C_MET_INPUTPOINTER_SOC");
    frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName, "SMAP_L4_C_SOC.h5");

    if(sesCtl_p->ftSourceMode != DoUse_GEOS5_TS)
    {
        strcpy(inputPtrAttribName,"L4C_MET_INPUTPOINTER_L3_SM_A");
        frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName,sesCtl_p->inputL3_SM_A_Path);
    } else {
        strcpy(inputPtrAttribName,"L4C_MET_INPUTPOINTER_GEOS5_FP_TS");
        frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName,"TS-derived-FreezeThaw-State");
    }
    
    for(ithGrn=0; ithGrn < sesCtl_p->nFP_Granules; ithGrn++)
    {
        inputPtrGranule_p = sesCtl_p->inputGEOS5_FP_pathList[ithGrn];
        sprintf(inputPtrAttribName,"L4C_MET_INPUTPOINTER_GEOS_FP_%-2.2d",ithGrn+1);
        frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName,inputPtrGranule_p);
    }
    
    for(ithGrn=0; ithGrn < sesCtl_p->nL4_SM_Granules;  ithGrn++)
    {
        inputPtrGranule_p = sesCtl_p->inputL4_SM_pathList[ithGrn];
        sprintf(inputPtrAttribName,"L4C_MET_INPUTPOINTER_L4_SM_%-2.2d",ithGrn+1);
        frm_hdf5_CreateStringAttribute(file_id, inputPtrAttribName, inputPtrGranule_p);
    }

    // also post the FT_STATE_UM_M03 legend as file level attributes
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag_legend",ftLegendString);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag0", ftLegend0);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag1", ftLegend1);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag2", ftLegend2);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag3", ftLegend3);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag4", ftLegend4);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag5", ftLegend5);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag6", ftLegend6);
    frm_hdf5_CreateStringAttribute(file_id , "FT_STATE_UM_M03_bitflag7", ftLegend7);

    strcpy(fxName,__func__);

} // end:: l4c_WriteFile_metadata()


// ---------------------------------------------------------------------
//   function   : l4c_WriteDataset_metadata()
//   inputs     : SesCtl *sesCtl_p -- ptr to session controls
//                char *datasetName_p -- ptr to dataset's name
//                hid_t dataset_id    -- ptr to dataset HDF5 ID
//   purpose    : write basic dataset level metadata for 
//                all SMAP_L4_C_MET.h5 datasets and intermediate SMAP_L4_C_SOC.h5
//   Author     : joe glassy
//   Version    : v7+, see AppVers macro in l4c_MET.h
//   revision   : 2017-02-08T09:12, cleaned up attrib writes by removing redundant
//                missing value "-9999" writes in string form, to correctly write as a float32
//   revision   : 2015-09-04T13:53, updated for better valid_min,valid_max assignments
//   revision   : 2015-08-03T18:07:00, updated to include GEOS5
//                FP "TS" derived F/T when the L3_SM_A is missing due
//                to failed radar instrument.
//   revision   : 2014-10-21T12:09, for FT_STATE_UM_M03 added bitflag legend and
//                INPUTPOINTER to reveal
//                which L3_SM_A provided info passed forward in FT_STATE_UM_M03
//   history    : v5.0.1-R11E, stubbed in valid_min,max placeholders, plan to update 
//                with RIP supplied values as soon as the L4_C_INPUT_VARIABLE_xx = name,a,b,c,d,e..
//                syntax is also updated in MET RIP file.
//   history    : 2014-07-28T14:29:00, re-factored orig metadata writes from
//                l4c_MET_fx.c, l4c_io.c.  Integration into L4_C v5.0.1 in progress
//   Notes      : TBD 2015-08-03,jmg go back and re-verify that all valid_min,valid_max are
//                correct, as default (0.0<x<100.0) isn't applicable to all, and
//                adjust each variable's valid_max  (these are wrong if set to fillvalue)
void l4c_WriteDataset_metadata(SesCtl *sesCtl_p,char *datasetName_p,hid_t dataset_id)
{
 char fxName[256] = {0} ;
 char  ftLegendString[256] = {0} ;
 char  ftLegend0[128] ;
 char  ftLegend1[128] ;
 char  ftLegend2[128] ;
 char  ftLegend3[128] ;
 char  ftLegend4[128] ;
 char  ftLegend5[128] ;
 char  ftLegend6[128] ;
 char  ftLegend7[128] ;
 
 char inputPtrAttribName[256] ;
 char lbl[128] ={0}; // generic string...
 float32 fltFill = -9999.0 ;
 float32 fltValu = 0.0 ;
 
 // <JMG> 2015-09-04T13:53 now inherit these from sesCtl
 float32 validMin  = sesCtl_p->granuleMin ;
 float32 validMax  = sesCtl_p->granuleMax ;
 float32 validMean = sesCtl_p->granuleMean ;
 
 uint8   bytFill =  254 ;
 uint8   bytValu = 0 ;
 
 
  strcpy(fxName,__func__);
 
  // <JMG> 2015-09-04, TBD: if GEOS5 TS used for F/T, we need a slightly different legend!!
  strcpy(ftLegendString,
    "FT bitflag Legend: all bits (0..8) are used as defined below");
  strcpy(ftLegend0,"bit 0 -- defines the boolean freeze thaw state (0|1) regardless which source defined it.");
  strcpy(ftLegend1,"bit 1 -- RADAR: defines the FT_RADAR boolean freeze thaw state, relevant ONLY when ACTIVE RADAR logic used");
  strcpy(ftLegend2,"bit 2 -- QC: the QualityControl (Good1=,Bad=0) from L3_SM_A. It's always TRUE if GEOS5 or TSURF used");
  strcpy(ftLegend3,"bit 3 -- SWATH: where InSwath=0, OutOfSwath=0 indicator");
  strcpy(ftLegend4,"bit 4 -- is set TRUE only when the GEOS5 TS was explicitly specified");
  strcpy(ftLegend5,"bit 5 -- pixel is InTerrestrial land domain(1), or Not(0)");
  strcpy(ftLegend6,"bit 6 -- TSURF? set TRUE only when L3_SM_A's tsurf is specified for F/T logic, false otherwise");
  strcpy(ftLegend7,"bit 7 -- RADAR? set TRUE only when ACTIVE RADAR is specified for F/T logic, false otherwise");
        
        xLogMsg("@A In(%s) ADDING-DATASET-METADATA for %s~\n",fxName,datasetName_p);
     
        // for now, always write observed_min,max
        frm_hdf5_WriteAttributeAsType(dataset_id,".","observed_min"  ,IsFloat32,1, &validMin);
        frm_hdf5_WriteAttributeAsType(dataset_id,".","observed_max"  ,IsFloat32,1, &validMax);
        frm_hdf5_WriteAttributeAsType(dataset_id,".","observed_mean" ,IsFloat32,1, &validMean);

        // but only write valid_min,valid_max if these appear different and sensible
        if(fabs(validMax - validMin) >0.0)
        {
          // for now, stub observed min,max in for valid_min,valid_max where applicable        
          frm_hdf5_WriteAttributeAsType(dataset_id,".","valid_min"  ,IsFloat32,1, &validMin);
          frm_hdf5_WriteAttributeAsType(dataset_id,".","valid_max"  ,IsFloat32,1, &validMax);
        } // end:: block if a sensible min,max is available
 
        // Create core set of dataset attributes: indicate source of data via inputGranulePath
        // Refactor this to call a separate function to write out the dataset
        // attributes: "units","valid_min","valid_max","shortname","coordinates" etc.
        // but SUPPRESS these for the 3km datasets that need their own lat_3km,lon_3km
        
        if(strstr(datasetName_p,"ROLLING")== NULL && strstr(datasetName_p,"FT_STATE_UM_M03") == NULL)
            frm_hdf5_CreateStringAttribute(dataset_id, "coordinates", "cell_lat,cell_lon");

        if(strstr(datasetName_p,"VPD")!= NULL)    // <JMG> 2/12/2014, now calculated in L4C_MET.ex
        {
            frm_hdf5_CreateStringAttribute(dataset_id, "units","Pa");
            frm_hdf5_CreateStringAttribute(dataset_id, "fill_value","-9999.0");
            
            // id,groupQualifier,attrname,TivType,nElem,void *data)
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);

        } else if (strstr(datasetName_p,"TS")!= NULL)  {  // from: FP
            // <JMG> 2105-08-25T11:21 added for when "TS" may also be processed
            fltValu = 0.0 ;
            frm_hdf5_CreateStringAttribute(dataset_id,"units","C");
            fltValu = 0.0 ;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
            
        } else if (strstr(datasetName_p,"T2M")!= NULL)  {  // from: FP
            fltValu = 0.0 ;
            frm_hdf5_CreateStringAttribute(dataset_id,"units","K");
            fltValu = GEOS5_FP_MISSING_VALUE ;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);

        } else if(strstr(datasetName_p,"QV2M")!=NULL)   { // from: FP
            frm_hdf5_CreateStringAttribute(dataset_id,"units","Kg/Kg");
            fltValu = GEOS5_FP_MISSING_VALUE ;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
            
        } else if(strstr(datasetName_p,"SM_SURFACE_WETNESS")!=NULL) {      // from: L4_SM
            frm_hdf5_CreateStringAttribute(dataset_id,"units","dimensionless");     
            // frm_hdf5_CreateStringAttribute(dataset_id, "fill_value","-9999.0");
            fltValu = L4_SM_MISSING_VALUE  ;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
            
        } else if(strstr(datasetName_p,"SM_ROOTZONE_WETNESS")!=NULL)  {    // from: L4_SM
            frm_hdf5_CreateStringAttribute(dataset_id,"units","dimensionless");
            fltValu = L4_SM_MISSING_VALUE  ;
            // frm_hdf5_CreateStringAttribute(dataset_id, "fill_value","-9999.0");
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
        } else if(strstr(datasetName_p,"RADIATION_SHORTWAVE_DOWNWARD_FLUX")!=NULL)  { // from: L4_SM
            frm_hdf5_CreateStringAttribute(dataset_id,"units","W/m^2");     
            fltValu = L4_SM_MISSING_VALUE  ;
            // <JMG> 2017-02-08T09:10, removed redundant string format for -9999 missing value!
            // frm_hdf5_CreateStringAttribute(dataset_id, "fill_value","-9999.0");
            // we DO want to write out missing value as a float32 
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
        } else if(strstr(datasetName_p,"SOIL_TEMP_LAYER")!=NULL)  { // from: L4_SM        
            frm_hdf5_CreateStringAttribute(dataset_id,"units","K");     
            fltValu = L4_SM_MISSING_VALUE  ;            
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
        } else if(strstr(datasetName_p,"SURFACE_PRESSURE")!=NULL)  { // from: L4_SM
            frm_hdf5_CreateStringAttribute(dataset_id,"units","Pa");        
            // frm_hdf5_CreateStringAttribute(dataset_id, "fill_value","-9999.0");
            fltValu = L4_SM_MISSING_VALUE  ;            
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);
            
        } else if(strstr(datasetName_p,"FT_STATE_UM_M03") !=NULL)   { 
            
            frm_hdf5_CreateStringAttribute(dataset_id, "coordinates", "cell_lat_3km,cell_lon_3km");
            
            // new UM specific image to pass forward the L3_SM_A surface_flag info
            frm_hdf5_CreateStringAttribute(dataset_id,"units","dimensionless");
            bytValu = 0 ;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","valid_min" ,IsUint8,1, &bytValu);
            bytValu = 254 ;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","valid_max" ,IsUint8,1, &bytValu);         
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsUint8,1, &bytFill);
            
            
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag_legend",ftLegendString);
            
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag0",ftLegend0);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag1",ftLegend1);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag2",ftLegend2);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag3",ftLegend3);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag4",ftLegend4);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag5",ftLegend5);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag6",ftLegend6);
            frm_hdf5_CreateStringAttribute(dataset_id,"ft_bitflag7",ftLegend7);
            
            // <JMG> 2014-10-21, added, as its critical for FT_STATE_UM_M03 to pass
            // its particular provenence through!
            if(sesCtl_p->ftSourceMode == DoUse_ACTIVE_RADAR)
            {
               strcpy(inputPtrAttribName,"INPUTPOINTER_L3_SM_A");
                frm_hdf5_CreateStringAttribute(dataset_id, inputPtrAttribName,sesCtl_p->inputL3_SM_A_Path);
            }else if(sesCtl_p->ftSourceMode == DoUse_GEOS5_TS ){
               strcpy(inputPtrAttribName,"INPUTPOINTER_GEOS5_FP");
               frm_hdf5_CreateStringAttribute(dataset_id, inputPtrAttribName,"TS");
            }
               
        } else {
            sprintf(lbl,"%s datasets units are currently undefined",datasetName_p);
            frm_hdf5_CreateStringAttribute(dataset_id,"units",lbl);
            fltFill = -9999.0;
            frm_hdf5_WriteAttributeAsType(dataset_id,".","fill_value",IsFloat32,1, &fltFill);           
        }
        
        frm_hdf5_CreateStringAttribute(dataset_id,"short_name" ,datasetName_p);
        frm_hdf5_CreateStringAttribute(dataset_id,"long_name"  ,datasetName_p);
    
} // end:: l4c_WriteDataset_metadata


