// ---------------------------------------------------------------
// subsystem           : SMAP L4_C
// module              : l4c_MODIS.h
// purpose             : header containing macros, abstract-data-types, and one global CONST
// revision            : v7.3.0.2-2021-10-06T10:01, added function int reorg_VIIRS_Qc()
// revision            : v7.3.0.1-2021-08-09T12:11, added "instrumentId" in SesCtl to allow for
//                       conditional branching between VIIRs and MODIS
// revision            : v7.3.0.1-2021-07-24T15:33
// revision            : v7.3.0.1-2021-07-18T16:24
// revision            : v7.3.0.1-2021-07-14T17:11EvalOnly, this is a study-eval ONLY version for diagnostics
//                       where very limited changes are made in anticipation of the real code transition to VIIRs
// revision            : v7.3.0-2017-06-07T18:33, sync with L4CMDL v7.3.0 and new BPLUT
// revision            : v7.2-2017-03-23T16:48:00, in l4c_MODIS_ses.c, in l4c_CloseSession() 
//                       added calls to nsbReleaseInstance(nsbObj_p), and  pkgMgr_ReleaseMgr(pkgMgr_p)
//                       to properly release all interior heap objects owned by the pkgMgr and nsbOverseers
// revised             : v7.1.0-2017-03-09, incorporates latest changes Lucas made in l4c_MODIS_logic.c
// revised:            : v7.1.0-2017-02-13, modified ImgPkg structure to add qualified
//                       members tile_struct tile_1km_p and tile_struct tile_500m_p
//                        Changed sesCtl_p->modis_Tile_1km_SumPixels to sesCtl_p->modis_Tile_1km_SumPixels
//                        Changed sesCtl_p->modis_ew_Pixels to modis_ew_1km_Pixels and modis_ew_500m_Pixels
//                        Changed sesCtl_p->modis_ns_Pixels to modis_ns_1km Pixels and modis_ns_500m_Pixels
//                       and moved call to get_sinusoid_latlon() to match Lucas science code order.
//
// revised             : v7.1.0-2017-02-09T16:24:00, missing macro "N_QC_RANK 5" now added into l4c_MODIS.h
//                       found in ./gridfpar/src/gridfpar.h, and changed R_ROUND to FRM_API_ROUND
//                       and added newly required l4c_MODIS_FparQaBin.[ch] header and source files
//                       for processing MOD15A2H 500m inputs.
// revised             : v7.1.0-2017-01-26T12:22:00, begin adding support for 500m res (2400x2400) 
//                       tiles from MODIS Collection 6 (MOD15A2H family).
// revised             : v7.0.1-2016-01-04T10:49, relax requirement that MODIS tile count is 286
//                     : since daiy logic triggers use of FPAR Climatology for missing/gap pixels,
//                       and introduced macro L4C_MODIS_OPTIMAL_TILECOUNT = 286.
//                       and deprecated old macro L4C_FPAR_MISSING_VALUE, which was never used
// revised             : v7.0-2015-09-07T20:24:00, version ID sync for v7 codes (no code changes)
// revised             : v7.0-2015-09-06T14:31:00, version ID sync'd with equivalent v7 codes
//                       for L4CMDL, and L4CMET.
// revised             : v6.0-2015-06-10T16:32,verified L4C_SW_VERSION fixes.
// revised             : v6.0-2015-06-01T23:55:00-07:00, prep for V10002 OPS delivery
// revised             : v5.0.1-2014-11-13T14:44, for gppMethodCode replaced number 0|1 with 
//                       enum macro {} in all c module files, to add clarity
// history             : v7.3.0.2-2021-10-06T10:01, added function int reorg_VIIRS_Qc() in tile_loop,
//                       re-assemble output QA/QC bits for VIIRS 500 inputs, prior to resampling to 1km,
//                       to correctly reflect input bits coming from both fparQc and extraQc.
// history             : v7.3.0.1-2021-07-14T10:12EvalOnly, diagnostic study in prep for transition
//                       to VIIRs support.
//    Libraries used: /usr/local/hdfeos5/lib/linux64
// history             : v5.0.1-2014-10-24T14:27:00-07:00, submitted to NCCS for ORT#2,ORR
// history             : v5.0.1-2014-10-15T13:33:03-07:00, found and fixed gridding "smudge" 
//                       the error in l4c_MODIS_io.c:l4c_ReadAncillary() was that dsNameTileRowM01 was 
//                       incorrectly referenced twice where the prior omitted 
//                       dsNameTileColM01 assignment is now correctly in place. Also now
//                       call l4c_MODIS_WriteMetadata.c:l4c_Hdf5_AddFileMetadata(void), which
//                       writes updated ISO req. attributes and replaces the 
//                       deprecated l4c_writeHdf5AttributesToOutput(). Fixes missing PRODUCTION_DATETIME etc
// history             : v5.0.1-2014-10-14T14:22:01-07:00, more refinements for ORT,ORR,code review
//                       including adoption of l4c_MODIS_WriteMetadata() adapted from L4CMDL,
//                       and cloned l4c_MODIS_bplut.c from L4CMDL to L4CMOD in anticipation
//                       of using BPLUT later for retrieval of NDVI params needed to
//                       calculate FPAR early in the pre-processor, to eventually
//                       pass ONLY FPAR to L4CMDL, thus eliminating any need to ever
//                       pass a global 1km EASEGRID of NDVI forward to L4CMDL.
// history             : v5.0.1-2014-08-28T10:26, fine-tuned writes of HDF5 dataset attributes, esp
//                       to correct the "coordinates" syntax, using space (not a comma) delimiters for
//                       "cell_lat cell_lon" specification, to allow Panoply and add other required
//                       CF dataset attributes, using correct datatype, for Panoply.
// history             : v5.0.1-2014-08-26T10:37:00, continuing making more robust.
// history             : v5.0.1-2014-08-21T14:12:00, merged ses init ops to mimic l4c_MET
//                       and modified NDVI handling to read in native int16 and output 
//                       NDVI as uint8, re-scaling {-1.0 <= ndvi <= +1} to 
//                       int {-100 <= ndvi <= +100} to uint8 {0..200} 
//                       such 
//                       NDVI type and embedded blanks in both NDVI field names
// history             : v5.01-2014-08-15T10:11:00-07:00, continued integration for v5.0.1
// version             : v5.0.1 (for D5 and beyond)
// invocation          : ./l4c_MODIS.ex l4c_MODIS.rip
// author              : Joe Glassy, Lupine Logic Inc.
// research code       : Lucas Jones, NTSG
// dependencies        : HDF4,HDF5,HdfEos libraries
// historyn            : 2014-05-01T15:06:00, for D5.0.1
// history             : 2013-11-18T09:34:00-07:00, for D5.0
// history             : 2013-11-18T09:34:00, corrected bug where OUTPUT root dir not prepended to output path
//                     : 2013-11-11T23:05:00, version delivered to NCCS with corrected Makefile
//                       2013-11-11T18:40:00, found,corrected bug in l4c_MODIS_io.c, where tileref.c data was
//                       being read into array, tileref.r. TBD: rename these array variables .r,.c to indicate
//                       they are arrays!!
//                       2013-11-11T11:15:50-07:00, added documentation, minor debug msg clean up
//                       2013-11-10T13:17, l4c_nvpParse_v2 next-gen parser explored...
//                       2013-11-09T16:50:00, corrected several minor indexing bugs in l4c_MODIS_logic.c (tile_loop)..
//                       2013-11-08t14:52,integrated tiv_api into codeset.
//                       2013-10-17T14:11:00, renamed gridfpar.h to l4c_MODIS.h to better conform to production naming protocols
// Notes               : enable macro L4_C, and note optional macros: HAVE_UUID, DIAG_ALGOR.
// Notes                 in structure CONT, taskIdStr[] was missing, Joe added it to allow make process to complete.
//                       rescaled and type-converted all tile_xxx template image files
// notes    : Actual Fpar dataset names in Collection 6 MODIS are:
//             Fpar_500m, Lai_500m, FparLai_QC, FparExtra_QC, 
//             FparStdDev_500m,  LaiStdDev_500m
// notes    : Actual Fpar dataset names in Collection 6 VIIRs are:
//             Fpar_500m, Lai_500m, FparLai_QC, FparExtra_QC, 
//             FparStdDev_500m,  LaiStdDev_500m
// ---------------------------------------------------------------

#ifndef l4c_MODIS_DEF
#define l4c_MODIS_DEF 1

// VERS_TAG identifies both version and release via combined tag
#define L4C_MODIS_VERS_TAG "v7.3.0.2-2021-10-06T10:01"
#define L4C_MODIS_APP_TAG  "l4c_MODIS.ex"

// ----headers----------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <assert.h>

#include <sys/types.h> // for getpid()
#include <unistd.h>    // for gethostname() and getpid()

#include <stddef.h>
#include <stdint.h>

#ifdef HAVE_UUID
#include <uuid/uuid.h>  // for uuid services, uuid_generate() etc..
#endif

// import EASEGRID and related grid constants
#include "l4c_MODIS_gridconst.h"


// import HDF4 and HDFEOS required constants
#include "hdf.h"
#include "mfhdf.h"
// <JMG> Note: this is HDFEOS 2 and does NOT meet needs of the HDFEOS 5 required for VIIRs
#include "HdfEosDef.h"

// import HDF5 
#include "hdf5.h"
#include "hdf5_hl.h"

// <JMG> 2014-10-15T11:15, moved UM FRM and related imports below 
// import of HdfEosDef.h to allow idempotence on PI symbol re-define
// to work as it should 
#include "frm_api.h"
#include "nsb_api.h"
#include "pkg_mgr.h"

// -----------manifest constants----------------------------------

// Set identifing signature
#define L4_C   1

#define L4CMOD_PKGMGR_THEME_COUNT   3

// <JMG> 2016-01-04, introduced Optimal MODIS 
// Sinusoidal tile count as 286, but may run with fewer
#define L4CMOD_OPTIMAL_TILECOUNT 286

// may be 1 to use GMT or 0 to use localtime stamping of log file
#define L4_C_USE_GMT 1

// SERIALIZATION OFFSETS (from Joe)
// 2D offsets: y is outermost, x is innermost (fastest varying) term
// <KAE> 2021-06-06 e.g., given an M x N array (2D), this function maps (X, Y) or
//  column-row pairs to a 1D sequence; the resulting values are described by the
//  array, e.g., in Python 3: numpy.arange(0, M*N).reshape((M,N))
#define M_2D_B0(x,y,n_y) (((x)*n_y)+(y))

// Serialization 
// offset for nested 1km 2D grid within 9km 2D grid 
// indicies x,y correspond to 9km grid, while i,j is the nested cell reference
#define M_2D_N9_B0(x,y,i,j,ny) (((x)*9 + (i))*ny + (y)*9 + (j))

// <LAJ> 2017-02-01 ADDED:
#define M_2D_NESTED_B0(x,y,i,j,nj,ny) (((x)*nj + (i))*ny + (y)*nj + (j))

// <JMG> 2014-10-15T11:14, removed redundant PI def already in frm_api.h
// and in HdfEos.h ...
// borrow definition of PI from math.h if otherwise not defined
// #ifndef PI
// #define PI M_PI
// #endif

// #define L4C_MET_MISSING_VALUE     -9999.0 
// <JMG> 2016-01-04, deprecated old macro L4C_FPAR_MISSING_VALUE, never used

// <JMG> 2021-07-14, Note original FILL_VALUE etc are specific to MODIS not VIIRs
// so we need to further qualify these in order to provide BOTH options
// <JMG> added FILL_VALUE for primary FPAR science variable
#define FILL_VALUE                255
// <JMG> TBD, 2016-01-04, verify if fpar_qa needs a distinct _Fillvalue e.g. 254
#define QA_FILL_VALUE             255
#define NDVI_OUT_OF_DOMAIN_VALUE  -3000

// <JMG> 2016-01-04, check reveals OUT_OF_DOMAIN_VALUE IS used in 
// set_land_mask
#define OUT_OF_DOMAIN_VALUE     254
// <JMG> 2021-07-14, note these are specific to MODIS mode, not VIIRs
#define FPAR_FILL_VALUE   FILL_VALUE
#define BYTE_FILL_VALUE   FILL_VALUE

// Note: these are both now defined in L4CAPI's frm_core.h 
// but allow for local override
#define NULL_CH             '\0'
#define STRING_MATCHED        0

// <JMG> 2017-02-09T15:54, added new required macro, from ./gridfpar/src/gridfpar.h
#define N_QC_RANK 5

// ------------ LEGACY SYMBOLS ---------------------------------------
// symbols below retained from Lucas science code, for backward compatibility...

#define MXSTRLEN     512             // set to pwr of 2, Maximum string length 

#define NROW1KM         14616        // Number of grid rows in EGv2 1km 
#define NCOL1KM         34704        // Number of grid cols in EGv2 1km 
#define NROW9KM          1624        // Number of grid rows in EGv2 9km 
#define NCOL9KM          3856        // Number of grid cols in EGv2 9km 
#define NCELL               9        // Number of nested 1km rows/cols within each 9km row/col

#define NMODISTILE_ROWS_1KM   1200
#define NMODISTILE_COLS_1KM   1200
#define NMODISTILE_CELLS_1KM  (1200*1200)

#define DEG2RAD      0.01745329       /* = PI/180, required by haversine function*/
#define R            6367.45          /* Earth radius [km] for haversine function*/
#define RG           111.195          /* Map constant for sinusoidal projection*/ 

#define FLTCOMPTOL      0.1           // tolerance for float comparison

// --------Enumerated constants ----------------------------------
enum ActionStatusValues {
    ActionSuccess, ActionFailure, ActionPending, ActionStatusUnknown
} ;


enum DatatypeClasses {
 IsClassUndefined, IsClassBoolean, IsClassInteger, IsClassReal, 
 IsClassString,    IsClassStringTable, IsClassScalar, IsClassCustom
} ;

// 0=NDVI, 1=FPAR, 2=FPARCLIM, 3=UNSPECIFIED-or-UNKNOWN
// <JMG> leave defined until switch-out is implemented and tested OK
// DEPRECATE PreProcGppMethodCodes
//   enum PreProcGppMethodCodes {
//      IsNdviSrcForGpp, IsFparSrcForGpp, IsFparClimatologySrcForGpp, IsUnknownSourceForGpp 
//   } ;
// <JMG> 2021-07-14 proposed changes:

// <JMG> 2021-07-18T16:25, change default order to prefer VIIRs first
// PreProcGppMethodCodes = {UseVIIRsInputs,UseMODISInputs,UseFparClimatologyInputs,UseUnspecificInputs
enum PreProcFparMethodCodes {
    UseVIIRsInputs,UseMODISInputs,UseFparClimatologyInputs,UseUnspecificInputs
} ;    

// --------Abstract data types -----------------------------------

// ---------------------------------------------------------------
// ADTs
// ---------------------------------------------------------------

// <JMG> 2014-10-14, inherited from L4CMDL to L4CMOD for consistency
// <JMG> 2014-10-06T16:20:00, added missing ISO elements Joe A reported
// NEE_COMPLETENESS_OMISSION, GPP_COMPLETENESS_OMISSION, RH_COMPLETENESS_OMISSION
// SOC_COMPLETENESS_OMISSION, EC_COMPLETENESS_OMISSION
// NEE_DOMAIN_CONSISTENCY, GPP_DOMAIN_CONSISTENCY,RH_DOMAIN_CONSISTENCY,
// SOC_DOMAIN_CONSISTENCY, EC_DOMAIN_CONSISTENCY
typedef struct tagL4C_ISO_MetaElements {
 float32 nee_CompletenessOmission;
 float32 gpp_CompletenessOmission;
 float32 rh_CompletenessOmission ;
 float32 soc_CompletenessOmission;   
 float32 ec_CompletenessOmission ;
 float32 nee_DomainConsistency   ; 
 float32 gpp_DomainConsistency   ; 
 float32 rh_DomainConsistency    ;   
 float32 soc_DomainConsistency   ; 
 float32 ec_DomainConsistency    ;    
 
 float32 fpar_CompletenessOmission;
 float32 ndvi_CompletenessOmission;
   
 float32 fpar_DomainConsistency    ;    
 float32 ndvi_DomainConsistency    ;    
 // others added later as needed 
} L4C_ISO_MetaElements ;


// <JMG> 2014-10-14, inherited from L4CMDL to L4CMOD for consistency
// <JMG> 2014-09-12T12:36, added, to capture observed (non-fillvalue)
//  metrics and bare-bones univar stats on key output variables.
// Note: this object is GenericMetrics in the UM_Framework, serviced
// by frm_utilSummarizePixel()...
typedef struct tagPostProcMetrics {
 float64 obs_min ;   // observed minimum (non-fillvalue)
 float64 obs_max ;   // observed maximum (non-fillvalue)
 float64 obs_range;  // observed range   (non-fillvalue)
 float64 obs_mean;   // observed mean   (non-fillvalue)
 float64 obs_stddev; // observed std deviation (non-fillvalue)
 float64 obs_stderror;    // observed std error (non-fillvalue)
 float64 obs_pctstderror; // observed percent std error (non-fillvalue)
 float64 obs_coef_var   ; // observed percent std error (non-fillvalue)
 float64 optionScaleFactor; // optional 'scale-by' factor to transform prior to output
 // Below are useful when a valid_min,valid_max bracket is available
 float64 minAllow ;  // optional, value of smallest qualifying pixel value
 float64 maxAllow ;  // optional, value of largest qualifying pixel value
 size_t  obs_CountQualified; // count of non-fillvalue pixels
} PostProcMetrics ;


// <JMG> 2013-10-31, converted tile_v,tile_h,tile_r,tile_c datatypes to 
//                   save wasted memory.
// tile_v --> now uint8, and rescaled +1. Ranges: old -1..14 new: 0..15 so to use, subtract 1..
// tile_h --> now uint8, and rescaled +1. Ranges: old -1..35 new: 0..36 so to use, subtract 1..
// tile_r --> now int16, all prior values -1..1199 left intact
// tile_c --> now int16, all prior values -1..1199 left intact
typedef struct tagTileRef_Struct {
  int16    *r; // was int *r 
  int16    *c; // was int *c 
  uint8    *v; // was int *v 
  uint8    *h; // was int *h
} tileref_struct;

// <JMG> 2021-07-14, Assert: *fpar MAY originate from either 500m or 1km input
typedef struct tagTileStruct {
  uint8 *fpar; // <JMG> 2021-07-14, support as either 500m or 1km 
  uint8 *qc;   // <JMG> 2021-07-14, support as either 500m or 1km 
  uint8 *extra_qc; // <JMG> 2010-07-15, added for future use
  // int16 *ndvi   ;  <JMG> deprecated 2021-07-15
  // int8  *ndvi_qc;  <JMG> deprecated 2021-07-15
  float64   *lat;
  float64   *lon;
  float64   lowright[2];
  float64   upleft[2];
  float64   dx;
  float64   dy;
  char      vstr[MXSTRLEN];
  char      hstr[MXSTRLEN];
  int       v;
  int       h;
  int       row;
  int       col;
  uint32    ij;
} tile_struct;


typedef struct {
  uint8   *fpar    ; // <JMG> specific to just FPAR now
  uint8   *extra_qc; // <JMG> 2021-07-15, added back in for future
  int16   *ndvi    ; // <JMG> 2014-08-19, specific to NDVI now
  uint8   *qc      ; // FPAR QC now separate from NDVI int8 QC
  int8    *ndvi_qc ; // <JMG> 2021-07-15, pending Deprecate
  uint8   *mask    ;
  int     row ;
  int     col ;
  uint32  ij ;
} ease_struct;


// verify where exactly this is used
typedef struct tagContStruct {
  int   print_patch;
} cont_struct;

// Deprecated: log_struct

// Holds variables for controlling program behavior
typedef struct tagControlStruct {
  // Universal timing and logging information
  char       log_basepath[MXSTRLEN];     // root directory basepath for log file
  char       dtBuf[32];                  // Message buffer
  // <JMG> omission broke make process, added 10/17/2013 14:10, guessing at length
  char       taskIdStr[128] ;            // string identifying parallel sub-task 
  // <JMG> added, to migrate MODPATH definition out to an externally defined parameter
  char  mod15RootPath[FILENAME_MAX];  // root directory where MODIS or VIIRS MOD15A2 tiles reside 
  // <JMG> name of 9km mask
  char  mask9kmPath[FILENAME_MAX]; // name of 9km mask: "MCD12Q1_M09_lc_mask" etc
  int   nModisTiles ;  // <JMG> total count of MODIS 10-deg tiles to process 
  // int   nVIIRsTiles ;  // <JMG> 2021-07-14, (add) count of VIIRS tiles to process
} control_struct;


// ImgPkg -- container to hold various pointers to required input and output images
typedef struct tagImgPkg {
  uint8  *pft_M09_maskImg_p; // PFT_M09_mask image
  // Lat,Lon 9km coordinate var buffers, read from SMAP_L4_C_Ancillary.h5
  float32 *latImg_M09_p ; // SMAP_L4_C_LAT_1624_x_3856_M09_flt32
  float32 *lonImg_M09_p ; // SMAP_L4_C_LON_1624_x_3856_M09_flt32
  // Lat,Lon 1km coordinate var buffers, read from SMAP_L4_C_Ancillary.h5
  float32 *latImg_M01_p ; // SMAP_L4_C_LAT_14616x34704_M01_flt32
  float32 *lonImg_M01_p ; // SMAP_L4_C_LON_14616x34704_M01_flt32
  // support legacy structures
  tileref_struct tileref ;
  ease_struct    ease ;
  // tile_struct  tile  <JMG> 2017-02-09 DEPRECATE SOON! now qualified via 1km vs. 500m as below
  tile_struct tile_1km ; // <JMG> 2017-02-10
  tile_struct tile_500m; // <JMG> 2017-02-10
  
  int           nPkgThemes ;       // count of PkgMgr Themes used in this app
  size_t        nElem9km   ;
  size_t        nElem1km   ;
  int32   absFreqs[256] ; // metric, reporting NDVI or FPAR frequencies
  NsbObj *NsbMgrOverseer_p ; // ptr to overseer for NSB (NVP overseer)
  PkgMgr *PkgMgrOverseer_p ; // ptr to overseer for PkgMgr services
} ImgPkg ;


// SesCtl will replace control_struct
typedef struct tagSesCtl {
    // SOC was orig called 'outputSOC' but it is a input-output (append mode) op
    FILE *logf      ;    // ptr to session log file handle
    char  logPath[FILENAME_MAX]     ;

    // COMMON temporal session state terms
    int   runDayOfYear;   // day-of-year index {1..366} applying to THIS run  
    int   runCompositePeriod; // for L4CSYS std compatibility
    int   runModelYear;   // model year YYYY
    int   inputCollectionId ; // dataset Collection identifier 5 or 6 as appropriate
    int   instrumentId ; // distinguish between VIIRS and MODIS, especially for QC handling.
    char  doy_Str[8]     ; // day-of-year string
    char  year_Str[8]    ; // YEAR YYYY string, year pertaining to this run
    char  period_Str[8]  ; // 8-day composite period in string form {01..46}
    char  period16_Str[8]; // 16-day composite period in string form {01..23}
    char  appAction[64];  // L4C_MODIS_RUN_MODE, which 'action' main performs

    char  enableCompletionMsg; // boolean, 0=suppress it 1=enable
    // <JMG> 2021-07-14 (add) to support VIIRs integration
    int   fparMethodCode;    // fparMethodCode is new method, supporting MODIS or VIIRS (no NDVI)
    char  fparMethodStr[16]; // <JMG> 2021-07-14, (add) from RIP, string, read as MODIS|VIIRS
    // OLD LEGACY MODE: if FPAR,gppMethodCode=0, if NDVI, this is 1
    // int   gppMethodCode ;    
    // char  gppMethodStr[16];  // from RIP, string, read as FPAR|NDVI
    // choices for fparResolution are limited to: {500|1000}
    int   fparResolutionMeters ; // <JMG> 2017-01-26, added to support Collection 6 MOD15A2H 500m inputs
    
    int32 modis_ew_1km_Pixels  ; // is 1200 for 1KM MODIS, or is 2400 for 500m MODIS
    int32 modis_ns_1km_Pixels  ; // is 1200 for 1KM MODIS, or is 2400 for 500m MODIS
    int32 modis_ew_500m_Pixels ; // is 2400 for 500m MODIS or Collection 6 VIIRs tiles
    int32 modis_ns_500m_Pixels ; // is 2400 for 500m MODIS or Collection 6+ VIIRs tiles
    int32 modis_Tile_1km_SumPixels ; // <JMG> refactored as runtime variables NOT MACROs
    int32 modis_Tile_500m_SumPixels; // <JMG> refactored as runtime variables NOT MACROs
        
    // COMMON VERSIONING objects
    char  SubsysReleaseTag[128] ; // shared subsys RELEASE identifier
    char  swVersion[64] ; // SOFTWARE Version (tag), local executable
    char  swRelease[64] ; // L4CSYS std now, uses VERS_TAG 
    char  datasetVersion[64]; // L4CSYS std now, from RIP file
    char  productionLevel[16];// L4CSYS std "gmao"|"production" vs. "local" etc
    char  doi_assigned[32] ;  // L4CSYS std, only L4CMDL has 
    
    // char  **ripTbl_p ; // ptr to master RIP NVP string table
    char  l4c_Task_Label[256]; // production run identifier
    char  l4c_Task_UUID[64]  ; // L4C_TASK_UUID
    char  l4c_Platform[128]  ; // L4C_PLATFORM ..name of (host) running this job    
    char  thisHost[256];       // L4CSYS std, hostname from gethostname()
    char  thisUsername[128];
    
    char  granuleUUID[64] ;    // only uses 37 bytes of this for UUID 
    char  ScienceVersID[64];   // match L4_SM output name scheme
    char  CRID[64] ;           // legacy CRID 
    char  frmApiVersion[64];   // UM Framework version
    char  pkgApiVersion[64];   // PkgMgr version
    char  nsbApiVersion[64];   // NSB version
    
    // <JMG> 2014-10-14, to facilitate NDVI direct computation of FPAR
    // at the pre-processor level
    char  inputBplutPath[FILENAME_MAX];
    // <JMG> 2014-10-14, L4CSYS std RIP path name format
    char  inputRipPath[FILENAME_MAX]; 
    // Ancillary root directory
    char  ancRootDir[FILENAME_MAX];
    char  socRootDir[FILENAME_MAX];
    // Input Fpar 8-day root directory
    char  inputFparRootDir[FILENAME_MAX];
    // <JMG> 2021-07-14, Deprecate separate, un-needed NDVI root input directory
    // since only one is active per session anyway
    //   char inputNdviRootDir[FILENAME_MAX]  ; 
    // OUTPUT Fpar root directory
    char  outputFparRootDir[FILENAME_MAX]  ;    

    // log naming support paths
    char  logRootDir[FILENAME_MAX];
    char  logInitialPath[FILENAME_MAX];
    char  logFinalPath[FILENAME_MAX];
    // base paths
    char  outputFparPath[FILENAME_MAX];
    char  inputAncPath[FILENAME_MAX]; 
    char  inputSOCPath[FILENAME_MAX]; // verify to keep, if still used
    // <JMG> 2014-10-14, for future use and L4CSYS compatibility
    char  inputFparClimatologyPath[FILENAME_MAX];
    
    // <JMG> 2014-10-14, inputRipPath replaces deprecated ripNvpPath

    // OUTPUT dataset names, inherited from RIP file
    char  dsOutput_Fpar[32] ;
    char  dsOutput_FparQc[32] ; 
    // INPUT dataaset names inherited from RIP file
    char  dsNameGrid[32]   ;        
    char  dsNameFpar[32]   ;    
    char  dsNameFparQc[32]  ;   
    // <JMG> 2021-07-18T16:05, changed from QcExtra to ExtraQC to match actual name
    char  dsNameFparExtraQc[32];    // <JMG> changed from QcExtra to ExtraQC to match actual name
    char  fparClimDatasetname[64] ; // <JMG> added 2014-08-26T11:01:00
    
    // INPUT dataset names for masks...
    char  dsNamePFT_M09_Mask[32];
    char  dsNameTileRowM09[64]  ;   
    char  dsNameTileColM09[64]  ;   
    char  dsNameTileHoriM09[64] ;   
    char  dsNameTileVertM09[64] ;   
    char  dsNameLatCoordvar[64] ; // SMAP_L4_C_LAT_1624_x_3856_M09_flt32
    char  dsNameLonCoordvar[64] ; // SMAP_L4_C_LON_1624_x_3856_M09_flt32
    char  dsNameLandRowIndex[64]; //
    char  dsNameLandColIndex[64]; //

    // <JMG> added 2014-10-14, common L4CSYS attributes, from L4CMDL
    // governing ATBD and PSD attributes
    char  govATBD_Vers[16] ;
    char  govATBD_Date[16];
    char  govPSD_Edition[16];
    char  govPSD_Date[16] ;

    // New, stringtables storing names of l4c_MODIS variables, one per file
    char **varList_FPAR ;
    char **varList_NDVI ;
    char **input_Fpar_pathList; // string table, stores ALL MODIS FPAR 8-day paths
    // DEPRECATED  **input_Ndvi_pathList
    // deprecated: int  nVIIRsNDVI_granules, nNVDI_granules    
    // <JMG> 2021-07-14, CHANGED to consolidate: nModisFpar_granules to nFparGranules
    int  nFparGranules;       // count of VIIRs (or) MODIS FPAR tiles
    // DEPRECATED:  nVIIRs_granules
    int  nFPAR_Variables;     // count of VIIRs (or) MODIS FPAR dataset (names) 
    int  nPkgThemes ;         // count of PkgMgr Themes used in this app
    int  nTotalNvp  ;         // count, total n. of qualified NVPs
    
    // GLOBAL. Only used in vert drill down, as a test, when macro DIAG_ONLY is defined
    int32  *terrAddr ;   // DIAG ONLY, a 1D vector of valid terrestrial addresses
    int   compressLev;   // gzip compress level 
    uint8 *pftImg_p ;    // TBD: remove here and use ImgPkg's EASEGRID 9km global mask
    time_t sesStrtSec ;  // session starting time, in sec
    time_t sesEndiSec ;  // session ending time, in sec
    double sesElapSec ;  // elapsed seconds for session  
    
    hid_t  anc_file_id ; // for Ancillary file
    hid_t  fparOutput_file_id; // for single output file
    
} SesCtl ;

#endif
