// ---------------------------------------------------------------
// function   : l4c_MODIS_PkgMgr_Init()
// Purpose    : Initializes PkgMgr (master scoping tool index containers) and
//              a collection of child PkgThemes.
// Inputs     : int nThemes (currently 2) - count of at least as many
//               PkgMgr themes we'll use, with one per major ADT 
//               in tcf_struct.h, and others used for pre,post processing
// revised    : 2021-07-14, EVAL/DIAG modifications in prep for VIIRs integration
// revised    : 2021-07-14T14:02, Eval/Diag changes in prep of VIIRs integration.
// revised    : 2017-02-13T15:24, modified to also support MOD15A2H 500m Fpar
// revised    : 2014-08-28T10:58, re-organized themes to isolate INPUT
//              properties in one and OUTPUT properties in the other, 
//              e.g. renamed l4c_PkgInit_MODIS_SetProperties() to 
//              l4c_PkgInit_L4C_MODIS() for symmetry with how this is done
//              in l4c_MET and l4c_DAILY.
// history    : 2014-08-20T17:44, refined l4c_PkgInit_xxx_SetProperties()
// history    : 2014-08-19T15:35:00, placeholders for now
// history    : n/a
// Notes      : CALLED BY: l4c_InitSession()
// Author     : joe glassy

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"


// ---------------------------------------------------------------
// function   : l4c_PkgMgr_Init()
// Purpose    : Initializes PkgMgr (master scoping tool index containers) and
//              a collection of child PkgThemes.
// Inputs     : int nThemes (currently 2) - count of at least as many
//               PkgMgr themes we'll use, with one per major ADT 
//               in tcf_struct.h, and others used for pre,post processing
// revised    : 2021-07-14, EVAL/DIAG modifications in prep for VIIRs integration
// revised    : 2014-08-19T15:35:00, updated for new Ancillary lat,lon 1km
// revised    : 2014-06-04T18:06:00-07:00,
// History    : n/a
// Returns    : int, status
// Notes      : CALLED BY: l4c_InitSession()
// Author     : joe glassy
int l4c_PkgMgr_Init(void)
{
 SesCtl *sesCtl_p     = l4c_GetSesCtl();
 ImgPkg   *imagePkg_p = l4c_GetImagePackage();
 PkgMgr   *pkgMgr_p   = l4c_GetPkgMgrOverseer();
 PkgTheme *pkgTheme_p = NULL; // convenience ptr to middle "parent"
 int status =0;
 int nThemes = L4CMOD_PKGMGR_THEME_COUNT; // usually 3
 
 // PkgMgr: store individual counts of PkgObj's that each defined
 // theme {0..20} owns
 // ASSERT: we always size 'L4C_PKGMGR_THEME_COUNT' (3 here) to be >= the n. of 
 // actually used.
 int  pObjCounts[L4CMOD_PKGMGR_THEME_COUNT] = {0} ;
 
    // <JMG> TBD 08/19/2014 set count of child PkgObj in each of (3) themes
    // THEME INDEX            THEME NAME IS:
    // Note these counts (22,10,10) are not strict, and just must be
    // greater than or equal to number of children actually used.
    pObjCounts[ 0] = 22 ;  // SMAP_L4_C_Ancillary
    pObjCounts[ 1] = 10 ;  // SMAP_L4CMOD
    pObjCounts[ 2] = 10 ;  // SMAP_L4_C_FPAR_8DAY
    
    // Reminder: pkgMgrCreateThemes uses the above counts, in the
    //   pObjCounts[] array, to assign each of nThemes with this number of
    //   child (PkgObj) entries carried in the pObjCounts[] slot
    pkgMgr_CreateThemes(pkgMgr_p, nThemes, pObjCounts, "L4CMOD_PREPROCESSOR");
    
    // First, must assign theme names for each of (3)
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 0);
    pkgMgr_SetThemeName(pkgTheme_p,"SMAP_L4_C_Ancillary");
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 1);
    pkgMgr_SetThemeName(pkgTheme_p,"SMAP_L4CMOD");
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 2);
    pkgMgr_SetThemeName(pkgTheme_p,"SMAP_L4_C_FPAR_8DAY");

    // Second, we can now pre-initialize properties for 
    //   each of (3) themes, referred to by name
    
    // assign all ANCILLARY dataset level properties
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 0);
    l4c_PkgInit_SMAP_L4_C_Ancillary(pkgTheme_p,"SMAP_L4_C_Ancillary");  
    
    // assign all INPUT dataset level properties
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 1);
    l4c_PkgInit_L4_C_MODIS(pkgTheme_p,"SMAP_L4CMOD");
    
    // assign all OUTPUT dataset level properties
    pkgTheme_p = pkgMgr_GetThemeByIndex(pkgMgr_p, 2);
    l4c_PkgInit_SMAP_L4_C_FPAR_8DAY(pkgTheme_p,"SMAP_L4_C_FPAR_8DAY");

    l4c_EventCompletionMsg((char *)__func__);

    return status;
    
} // end:: l4c_PkgMgr_Init()



// ---------------------------------------------------------------
// function : l4c_PkgInit_SMAP_L4_C_Ancillary()
// purpose  : Set properties of indicated theme
// revised  : 2014-08-19T15:42:00, updated to reflect new Ancillary contents
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p

// notes    : TBD double check that the datasets here exactly MATCH
//            those present in the latest Ancillary, as these do NOT
//            match exactly now...

// notes    : CRITICAL: wait to call THIS function until AFTER all
//            themes have been named, and until after all HDF5 files
//            have been opened, so their file_id's are available
// notes    : CALLED BY: l4c_SesInit() 
// notes    : As long as the file_id's don't change this function may 
//             safely be called multiple times, where perhaps not all the
//             files were opened the first time, but were opened by
//             subsequent calls...
/*
   DATASET "L4C_M01_PixFilter_uint8" {
   DATASET "L4C_M01land_PixFilter_uint8" {
   DATASET "MCD12Q1_M01_lc_dom_uint8" {
   DATASET "MCD12Q1_M01_lc_mask_uint8" {
   DATASET "MCD12Q1_M01land_lc_dom_uint8" {
   DATASET "MCD12Q1_M03_lc_mask_uint8" {
   DATASET "MCD12Q1_M09_lc_mask_uint8" {
   DATASET "MCD12Q1_M09land_col_uint16" {
   DATASET "MCD12Q1_M09land_row_uint16" {
   DATASET "MOD12Q1_M09_lc_dom_uint8" {
   DATASET "SMAP_L4_C_LAT_14616_x_34704_M01_flt32" {
   DATASET "SMAP_L4_C_LAT_1624_x_3856_M09_flt32" {
   DATASET "SMAP_L4_C_LON_14616_x_34704_M01_flt32" {
   DATASET "SMAP_L4_C_LON_1624_x_3856_M09_flt32" {
   DATASET "tile_c_M01_1km_ease2_template_int16" {
   DATASET "tile_h_M01_1km_ease2_template_uint8" {
   DATASET "tile_r_M01_1km_ease2_template_int16" {
   DATASET "tile_v_M01_1km_ease2_template_uint8" {

 */
void l4c_PkgInit_SMAP_L4_C_Ancillary(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 NsbObj *nsbMgr_p = l4c_GetNsbOverseer();
 char nvpBuf[FILENAME_MAX];
 char *dsName_p = NULL;
 char fxName[256];
 
    strcpy(fxName,__func__);
    
    pkgMgr_SetThemeName(pkgTheme_p, themeLabel_p);
        
    // these are all HDF5 dataset names in the ancillary .h5 
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, "L4C_M01_PixFilter_uint8");
    
    // note: Currently for pixel filtering THIS IS THE ONE WE USE, M01land_PixFilter 
    // since it is already deflated.
    // ASSERT: the indices shown (1..21) are base 0, so 22 in total are used.
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, "L4C_M01land_PixFilter_uint8");
    
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, "MCD12Q1_M01_lc_dom_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, "MCD12Q1_M01_lc_mask_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, "MCD12Q1_M01land_lc_dom_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "MCD12Q1_M03_lc_mask_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 6, "MCD12Q1_M09_lc_mask_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 7, "MCD12Q1_M09_lc_nmargveg_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 8, "MCD12Q1_M09_lc_nnonveg_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 9, "MCD12Q1_M09_lc_nveg_uint8" );
    pkgMgr_SetThemes_ChildName(pkgTheme_p,10, "MCD12Q1_M09land_col_uint16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,11, "MCD12Q1_M09land_row_uint16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,12, "MCD12Q1_M09_lc_dom.uint8");

    pkgMgr_SetThemes_ChildName(pkgTheme_p,13, "SMAP_L4_C_LAT_1624_x_3856_M09_flt32");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,14, "SMAP_L4_C_LON_1624_x_3856_M09_flt32");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,15, "SMAP_L4_C_LAT_14616_x_34704_M01_flt32");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,16, "SMAP_L4_C_LON_14616_x_34704_M01_flt32");
    
    pkgMgr_SetThemes_ChildName(pkgTheme_p,17, "tile_c_M01_1km_ease2_template_int16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,18, "tile_h_M01_1km_ease2_template_uint8");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,19, "tile_r_M01_1km_ease2_template_int16");
    pkgMgr_SetThemes_ChildName(pkgTheme_p,20, "tile_v_M01_1km_ease2_template_uint8"); 
    // <JMG> 2014-05-19, added one to represent self-referential, file level info
    pkgMgr_SetThemes_ChildName(pkgTheme_p,21, "SMAP_L4_C_Ancillary"); 
    
    l4c_EventCompletionMsg( (char *)__func__);
    
} // end:: l4c_PkgInit_SMAP_L4_C_Ancillary()


// ---------------------------------------------------------------
// function : l4c_PkgInit_SMAP_L4_C_FPAR_8DAY()
// purpose  : Set properties of indicated theme
// revised  : 2014-08-19T15:42:00, updated to reflect new Ancillary contents
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// notes    : Actual Fpar dataset names in Collection 6 MODIS are:
//             Fpar_500m, Lai_500m, FparLai_QC, FparExtra_QC, 
//             FparStdDev_500m,  LaiStdDev_500m
/*
 L4C_MODIS_GRID_NAME      = NPP_Grid_16DAY_1km_VI
 # Equivalent OUTPUT EASEGRID 1k global mosaic output dataset names
 L4C_OUTPUT_DATASET_MODIS_FPAR          = fpar_M01
 L4C_OUTPUT_DATASET_MODIS_FPAR_QC       = fpar_qc_M01
 L4C_OUTPUT_DATASET_VIIRS_NDVI          = ndvi_M01
 L4C_OUTPUT_DATASET_VIIRS_NDVI_QC       = ndvi_qc_M01
*/

// l4c_PkgInit_SMAP_L4_C_FPAR_8DAY() defines BOTH the input AND output
//    objects for Fpar
void l4c_PkgInit_SMAP_L4_C_FPAR_8DAY(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 NsbObj *nsbMgr_p = l4c_GetNsbOverseer();
 char nvpBuf[FILENAME_MAX];
 char *dsName_p = NULL;
 char fxName[256];
 
    strcpy(fxName,__func__);

    // using NSB RIP file references, assign child names and other properties
    // as needed 
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR",nvpBuf);
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 0, dsName_p);
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR_QC",nvpBuf);
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 1, dsName_p);
    // <JMG> 2021-07-14T16:35, add extra QC in anticipation of future use
    // actual dataset name in MODIS Collection 6 is Fpar
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR_EXTRA_QC",nvpBuf);
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 2, dsName_p);
       
    // inherit names of OUTPUT FPAR and NDVI from RIP file
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_OUTPUT_DATASET_FPAR"   ,nvpBuf);
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 3, dsName_p);
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_OUTPUT_DATASET_FPAR_QC",nvpBuf);
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 4, dsName_p);

    // <JMG> 2021-07-14, DEPRECATED all NDVI entries so re-numbered children
    
    // also create a container for properties of the output file itself
    pkgMgr_SetThemes_ChildName(pkgTheme_p, 5, "SMAP_L4_C_FPAR_8DAY");
            
} // end:: l4c_PkgInit_SMAP_L4_C_FPAR_8DAY()


// ---------------------------------------------------------------
// function : l4c_PkgInit_L4_C_MODIS_SetProperties()
// purpose  : Set properties of indicated theme
// revised  : 2021-07-14, EVAL/DIAG provisional, initial mods in prep of
//             VIIRs integration. Note we WILL CHANGE or ADD to these to
//             accomodate the different QA/QC scheme VIIRs uses!
// revised  : 2021-07-14, EVAL/DIAG modifications in prep for VIIRs integration
// revised  : 2017-02-13T15:24, modified to also support MOD15A2H 500m Fpar
// revised  : 2014-08-19T15:42:00, updated to reflect new Ancillary contents
// inputs   : PkgThem *pkgTheme_p
//          : char *themeName_p
// notes    : n/a
/*
 Warning: these DO VARY by product AND resolution 1km vs. 500
 as shown below:
 L4CMOD_GRID_NAME      = NPP_Grid_16DAY_1km_VI
 L4CMOD_GRID_NAME      = MOD_Grid_MOD15A2
 L4CMOD_GRID_NAME      = MOD_Grid_MOD15A2H
 * 
 # Equivalent OUTPUT EASEGRID 1k global mosaic output dataset names
 L4CMOD_OUTPUT_DATASET_FPAR          = fpar_M01
 L4CMOD_OUTPUT_DATASET_FPAR_QC       = fpar_qc_M01
 
        byte Fpar_1km(YDim:MOD_Grid_MOD15A2, XDim:MOD_Grid_MOD15A2) ;
                Fpar_1km:scale_factor = 0.01 ;
                Fpar_1km:scale_factor_err = 0. ;
                Fpar_1km:add_offset = 0. ;
                Fpar_1km:add_offset_err = 0. ;
                Fpar_1km:calibrated_nt = 21 ;
                Fpar_1km:valid_range = '\0', 'd' ;
                Fpar_1km:_FillValue = '\377' ;
                Fpar_1km:long_name = "MOD15A2 MODIS/Terra Gridded 1KM FPAR (8-day composite)" ;
                Fpar_1km:units = "Percent" ;
                Fpar_1km:MOD15A2_FILLVALUE_DOC = "MOD15A2 FILL VALUE LEGEND\n",
    "255 = _Fillvalue, assigned when:\n",
    "    * the MODAGAGG suf. reflectance for channel VIS, NIR was assigned its _Fillvalue, or\n",
    "    * land cover pixel itself was assigned _Fillvalus 255 or 254.\n",
    "254 = land cover assigned as perennial salt or inland fresh water.\n",
    "253 = land cover assigned as barren, sparse vegetation (rock, tundra, desert.)\n",
    "252 = land cover assigned as perennial snow, ice.\n",
    "251 = land cover assigned as \"permanent\" wetlands/inundated marshlands.\n",
    "250 = land cover assigned as urban/built-up.\n",
    "249 = land cover assigned as \"unclassified\" or not able to determine.\n",

        byte FparLai_QC(YDim:MOD_Grid_MOD15A2, XDim:MOD_Grid_MOD15A2) ;
                FparLai_QC:valid_range = '\0', '\376' ;
                FparLai_QC:_FillValue = '\377' ;
                FparLai_QC:long_name = "MOD15A2 MODIS/Terra QC for FPAR and LAI (8-day composite)" ;
                FparLai_QC:units = "class-flag" ;
                FparLai_QC:FparLai_QC_DOC = "\n",
    "FparLai_QC 5 BITFIELDS IN 8 BITWORD\n",
    "MODLAND_QC START 0 END 0 VALIDS 2\n",
    "MODLAND_QC   0 = Good Quality (main algorithm with or without saturation)\n",
    "MODLAND_QC   1 = Other Quality (back-up algorithm or fill value)\n",
    "SENSOR START 1 END 1 VALIDS 2\n",
    "SENSOR       0  = Terra\n",
    "SENSOR       1  = Aqua\n",
    "DEADDETECTOR START 2 END 2 VALIDS 2\n",
    "DEADDETECTOR 0 = Detectors apparently fine for up to 50% of channels 1,2\n",
    "DEADDETECTOR 1 = Dead detectors caused >50% adjacent detector retrieval\n",
    "CLOUDSTATE START 3 END 4 VALIDS 4 (this inherited from Aggregate_QC bits {0,1} cloud state)\n",
    "CLOUDSTATE   00 = 0 Significant clouds NOT present (clear)\n",
    "CLOUDSTATE   01 = 1 Significant clouds WERE present\n",
    "CLOUDSTATE   10 = 2 Mixed cloud present on pixel\n",
    "CLOUDSTATE   11 = 3 Cloud state not defined,assumed clear\n",
    "SCF_QC START 5 END 7 VALIDS 5\n",
    "SCF_QC       000=0 Main (RT) algorithm used, best result possible (no saturation)\n",
    "SCF_QC       001=1 Main (RT) algorithm used, saturation occured. Good, very usable.\n",
    "SCF_QC       010=2 Main algorithm failed due to bad geometry, empirical algorithm used\n",
    "SCF_QC       011=3 Main algorithm faild due to problems other than geometry, empirical algorithm used\n",
    "SCF_QC       100=4 Pixel not produced at all, value coudn\'t be retrieved (possible reasons: bad L1B data, unusable MODAGAGG data)\n",
    "" ;
  
*/
void l4c_PkgInit_L4_C_MODIS(PkgTheme *pkgTheme_p,char *themeLabel_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer();
 PkgTheme *pThme_p = NULL;
 PkgObj   *pObj_p = NULL;
 NsbObj *nsbMgr_p = l4c_GetNsbOverseer();
 
 char nvpBuf[FILENAME_MAX];
 char *dsName_p = NULL;
 char fxName[256] = {0} ;
 size_t dims[2]   = {0} ;
 
    strcpy(fxName,__func__);

    // themeLabel_p typically is: "SMAP_L4_C_MODIS"
    pThme_p = pkgMgr_GetThemeByName(pkgMgr_p,themeLabel_p);
    
    xLogMsg("@A In(%s) Setting properties for theme(%s)\n",fxName,themeLabel_p);
    
     // Next, Assign properties to each child
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR",nvpBuf);
 
    if(sesCtl_p->fparResolutionMeters >= 1000)
    {
      dims[0] = MODIS_1KM_NS_PIXELS ; // 1200...
      dims[1] = MODIS_1KM_EW_PIXELS ; // 1200
      // L4CMOD_INPUT_DATASET_FPAR
      pkgMgr_SetThemes_ChildName(pThme_p, 0, "Fpar_1km");
      // <JMG> 2017-02-13 ASSERT: dsName here is either: Fpar_1km OR Fpar_500m or NVDI_1km etc
      pObj_p = pkgMgr_GetObjByName(pThme_p,dsName_p); 
      pkgMgr_SetObjRank(pObj_p, 2 );
      pkgMgr_SetObjResolution(pObj_p, 1000 ); // 1km sinusoidal
      pkgMgr_SetObjDims(pObj_p,dims);
      // MODIS_1KM_SUM_PIXELS     1440000
      pkgMgr_SetObjNelem(pObj_p, MODIS_1KM_SUM_PIXELS);

    } else {

      dims[0] = MODIS_500M_NS_PIXELS; // 2400...
      dims[1] = MODIS_500M_EW_PIXELS; // 2400
      
      // L4CMOD_INPUT_DATASET_FPAR -- differs if MODIS vs. VIIRs
      if(sesCtl_p->fparMethodCode == UseMODISInputs) 
         pkgMgr_SetThemes_ChildName(pThme_p, 0, "Fpar_500m");
      else if(sesCtl_p->fparMethodCode == UseVIIRsInputs)
         pkgMgr_SetThemes_ChildName(pThme_p, 0, "Fpar");
      
      // <JMG> 2017-02-13 ASSERT: dsName here is either: Fpar_1km OR Fpar_500m or NVDI_1km etc
      pObj_p = pkgMgr_GetObjByName(pThme_p,dsName_p); 
      
      pkgMgr_SetObjRank(pObj_p, 2 );
      pkgMgr_SetObjResolution(pObj_p, 500 ); // 500m sinusoidal
      pkgMgr_SetObjDims(pObj_p,dims);
      pkgMgr_SetObjNelem(pObj_p, MODIS_500M_SUM_PIXELS);
    }

    // L4CMOD_INPUT_DATASET_FPAR_QC, common to BOTH 1km and 500m
    pkgMgr_SetThemes_ChildName(pThme_p, 1, "FparLai_QC");
    pkgMgr_SetThemes_ChildName(pThme_p, 2, "FparExtra_QC");
    pkgMgr_SetThemes_ChildName(pThme_p, 3, "Lai"); // placeholder only
    
    
    // <JMG> 2021-07-14 DEPRECATED all NDVI entries...
    //  L4C_INPUT_DATASET_VIIRS_NDVI -- Currently ONLY 1km NDVI is supported 
    // SPECIAL HANDLING, nsb can't handle embedded blank 'value' strings yet
    //pkgMgr_SetThemes_ChildName(pThme_p, 2, "1km16daysNDVI");
    // L4C_INPUT_DATASET_VIIRS_NDVI_QC
    // pkgMgr_SetThemes_ChildName(pThme_p, 3, "1km16dayspixelreliability");

    // NOTE: output FPAR is always 1km, so it doesn't vary by input resolution
    // L4CMOD_OUTPUT_DATASET_FPAR          = fpar_M01
    // L4CMOD_OUTPUT_DATASET_FPAR_QC       = fpar_qc_M01
    pkgMgr_SetThemes_ChildName(pThme_p, 4, "fpar_M01");
    pkgMgr_SetThemes_ChildName(pThme_p, 5, "fpar_qc_M01");
    
    // pkgMgr_SetThemes_ChildName(pThme_p, 6, "ndvi_M01");
    // pkgMgr_SetThemes_ChildName(pThme_p, 7, "ndvi_qc_M01");
    
    // Assign other pObj properties for Fpar variable    
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR",nvpBuf);        
    pObj_p = pkgMgr_GetObjByName(pThme_p,dsName_p); 
    pkgMgr_SetObjOwner(pObj_p      , -1);
    pkgMgr_SetObjRank(pObj_p, 2 );    
    pkgMgr_SetObjShortname(pObj_p  , dsName_p); 
    pkgMgr_SetObjLongname(pObj_p   , "FPAR fraction of photosyn active radiation"); 
    
    pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8 );    
    pkgMgr_SetObjUnits(pObj_p, "percent (1..100)" );    

    pkgMgr_SetObjFillvalue(pObj_p, 255);
    pkgMgr_SetObjValidMin(pObj_p, 0.0);
    pkgMgr_SetObjValidMax(pObj_p, 100.0);   
    pkgMgr_SetObjMetadata(pObj_p   , "FPAR fraction of photosyn active radiation");
    pkgMgr_SetObjGenericPath(pObj_p, "not-yet-declared");   

    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR_QC",nvpBuf);
    pObj_p = pkgMgr_GetObjByName(pThme_p,dsName_p);
    
        // <JMG> These also vary by Fpar resolution
    if(sesCtl_p->fparResolutionMeters >= 1000)
    {
      dims[0] = MODIS_1KM_NS_PIXELS; // 1200...
      dims[1] = MODIS_1KM_EW_PIXELS; // 1200
      pkgMgr_SetObjRank(pObj_p, 2 );
      pkgMgr_SetObjDims(pObj_p,dims);
      pkgMgr_SetObjResolution(pObj_p, 1000 ); // 1km sinusoidal
      pkgMgr_SetObjNelem(pObj_p, MODIS_1KM_SUM_PIXELS);
    } else {
      dims[0] = MODIS_500M_NS_PIXELS; // 2400...
      dims[1] = MODIS_500M_EW_PIXELS; // 2400
      pkgMgr_SetObjRank(pObj_p, 2 );
      pkgMgr_SetObjDims(pObj_p,dims);
      pkgMgr_SetObjResolution(pObj_p, 500 ); // 500m sinusoidal
      pkgMgr_SetObjNelem(pObj_p, MODIS_500M_SUM_PIXELS);

    }
    
    pkgMgr_SetObjOwner(pObj_p      , -1);
    pkgMgr_SetObjShortname(pObj_p  , dsName_p); 
    pkgMgr_SetObjLongname(pObj_p   , dsName_p); 

    pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8 );    
    
    pkgMgr_SetObjUnits(pObj_p, "units (0,1,2,3)" ); 
    
    pkgMgr_SetObjFillvalue(pObj_p, 255);
    pkgMgr_SetObjValidMin(pObj_p, 0.0);
    pkgMgr_SetObjValidMax(pObj_p, 3.0); 
    pkgMgr_SetObjMetadata(pObj_p   , "not-yet-declared");
    pkgMgr_SetObjGenericPath(pObj_p, "not-yet-declared");   

    // <JMG> 2021-07-14, DEPRECATED all NDVI related entries
    
    dims[0] = NROW1KM ;
    dims[1] = NCOL1KM ;

    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_OUTPUT_DATASET_FPAR"   ,nvpBuf);
    pObj_p = pkgMgr_GetObjByName(pThme_p,dsName_p);
    
    pkgMgr_SetObjOwner(pObj_p      , -1);
    pkgMgr_SetObjShortname(pObj_p  , dsName_p); 
    pkgMgr_SetObjLongname(pObj_p   , dsName_p); 
    pkgMgr_SetObjRank(pObj_p, 2 );
    pkgMgr_SetObjDims(pObj_p,dims);
    pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8 );    
    pkgMgr_SetObjResolution(pObj_p, 1000 ); // 1km sinusoidal
    pkgMgr_SetObjUnits(pObj_p, "percent (0..100)" );    
    pkgMgr_SetObjNelem(pObj_p, NROW1KM * NCOL1KM);
    pkgMgr_SetObjFillvalue(pObj_p, 255);
    pkgMgr_SetObjValidMin(pObj_p, 0.0);
    pkgMgr_SetObjValidMax(pObj_p, 100.0);   
    pkgMgr_SetObjMetadata(pObj_p   , "0.01,0=gain,offset");
    pkgMgr_SetObjGenericPath(pObj_p, "not-yet-declared");
    
    dsName_p = nsbGetValueByKey(nsbMgr_p,"L4CMOD_OUTPUT_DATASET_FPAR_QC",nvpBuf);
    pObj_p = pkgMgr_GetObjByName(pThme_p,dsName_p); 

    pkgMgr_SetObjOwner(pObj_p      , -1);
    pkgMgr_SetObjShortname(pObj_p  , dsName_p); 
    pkgMgr_SetObjLongname(pObj_p   , dsName_p); 
    pkgMgr_SetObjRank(pObj_p, 2 );
    pkgMgr_SetObjDims(pObj_p,dims);
    pkgMgr_SetObjDatatype(pObj_p, (int)DFNT_UINT8 );    
    pkgMgr_SetObjResolution(pObj_p, 1000 ); // 1km GLOBAL sinusoidal here, not TILE
    pkgMgr_SetObjUnits(pObj_p, "units (0,1,2,3)" ); 
    pkgMgr_SetObjNelem(pObj_p, NROW1KM * NCOL1KM);  
    pkgMgr_SetObjFillvalue(pObj_p, -1);
    pkgMgr_SetObjValidMin(pObj_p, 0.0);
    pkgMgr_SetObjValidMax(pObj_p, 3.0); 
    pkgMgr_SetObjMetadata(pObj_p   , "TBD: verify: 0=Good,1=marginal,2=Snow/Ice,3=Cloudy");
    pkgMgr_SetObjGenericPath(pObj_p, "not-yet-declared");

    // <JMG> REMOVED all NDVI entries...
        
    xLogMsg("@A In(%s) Finished setting properties for theme(%s)\n",fxName,themeLabel_p);
        
} // end:: l4c_PkgInit_L4_C_MODIS()

