// ---------------------------------------------------------------
// module       : l4c_MODIS_WriteMetadata.c
// purpose      : Adapted from L4CMDL, implements writes of formal
//                file-level metadata to SMAP_L4_C_FPAR_8DAY.h5 HDF5 file(s)
//                to consolidate all metadata writes to this 1 file.
// project      : SMAP L4C science development.
// dependencies : HDF4 4.2.x, HDF5 v1.8.x or later libraries, L4C FRAMEWORK APIs
// revised      : 2014-06-23T18:10:00, added group-wise handlers.
// history      : 2014-06-23T17:33:00, adding creation of formal metadata
// author       : jglassy
// notes        : All HDF5 groups referenced by these write operations 
//                must pre-exist! note this module is in development and
//                has not yet been added to L4_C makefile. TBD..
// ---------------------------------------------------------------

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"


// ----------------------------------------------------------------------
// function : l4c_Hdf5_AddFormalMetadata(()
// purpose  : Serves as a driver to call a series of group-specific
//            metadata writing handlers, writing attributes into the
//            set of pre-defined groups:
//            {CRID,Config,DatasetIdentification, Extent, 
//             GridSpatialRepresentation,ProcessStep,SeriesIdentification}
// inputs   : const hid_t file_id  --target ID to write info to
// returns  : int,status
// revised  : 2014-10-06T16:08:00, corrected name '_hdf5_' to "_Hdf5_" to
//            match other l4c_Hdf5_xxx functions and added 
//            
// history  : 2014-06-24T18:20:00, orig
// dependency: must call l4c_CreateAllGroups() BEFORE calling this
//             function since all HDF5 groups must already be in place.
// notes    : n/a
int l4c_Hdf5_AddFormalMetadata(const hid_t file_id)
{
    SesCtl *sesCtl_p = l4c_GetSesCtl();
    int status = 0 ;    
    char attribBuf[512] = {0} ;
    char nvpBuf[512]    = {0} ;
    char fxName[256]; 
    // fetch ptr to the RIP name-value-pair stringblock database
    NsbObj *nsbObj_p  = l4c_GetNsbOverseer();
    // fetch ptr to overseer Package Manager instance
    PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer();
    
    
    strcpy(fxName, __func__);
    
    // <JMG> 06/17/2014 work-in-progress
    // NOTE: list of L4_SM CF style subset of high-level attributes Rolf uses:
    
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Contact", "http://www.ntsg.umt.edu");
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Comment", "HDF-5 file format used");
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Conventions", "CF 1.x");
        
    // THIS granule name, here a daily-intermediate
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Filename", sesCtl_p->outputFparPath);
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "History", "file written by l4c_DAILY.ex");
        
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Institution", "University of Montana NTSG");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "References", "See NASA SMAP Level 4 Carbon Product Specification Document");
    // <KAE> 2021-09-27 Added conditional switch for fPAR source
    // <KAE> 2021-10-12 Changed so that the instrument name is written dirctly to this field
    status = frm_hdf5_WriteAttributeCustomString(file_id,
          "/", "Source", sesCtl_p->fparMethodStr);
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/", "Title", "NASA SMAP Level 4 Carbon (L4_C) daily global 9km data product granule");

#ifdef NO_FULL_ISO_REQUIRED_YET 
    // STUB in placeholders for the (2) ISO 19139 XML files that will be
    // added later by a post-processor.
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/Metadata", "iso_19139_dataset_xml", "XML 19139 added added via post-processor");
    status = frm_hdf5_WriteAttributeCustomString(file_id, 
        "/Metadata", "iso_19139_series_xml", "XML 19139 series added via post-processor");
    // Call a succession of subsidery handlers, one per HDF5 group
    // call interface: file_id plus groupFullLabel
    l4c_MetadataWriteGroup_CRID(file_id) ;  
    l4c_MetadataWriteGroup_SeriesIdentification(file_id) ;
    l4c_MetadataWriteGroup_Config(file_id);
    l4c_MetadataWriteGroup_DatasetIdentification(file_id) ;
    l4c_MetadataWriteGroup_Extent(file_id) ;
    l4c_MetadataWriteGroup_GridSpatialRepresentation(file_id) ;
    l4c_MetadataWriteGroup_ProcessStep(file_id) ;
#endif

    // <JMG> 2014-09-03T14:18, TBD yet to add:
    //  l4c_MetadataWriteGroup_AcquisitionInformation() 
    //  and ALL OTHER missing data-source entries

    l4c_EventCompletionMsg( fxName);
    // return status
    return 0 ;
    
} // end:: l4c_Hdf5_AddFormalMetadata()


// ----------------------------------------------------------------------
// function : l4c_Init_ISO_MetadataElements(void)
// purpose  : Initialize master L4C_ISO_MetaElements(), assigning fallback
//            defaults, to later update based on granule level performance.
// inputs   : void
// returns  : int,status
// revised  : 2014-08-13T13:36, changed L4C_PRODUCTION_UTC_DATETIME
//            back to L4C_PRODUCTION_UTC_DATETIME to preserve Joe A's
//            Perl parser logic on post-production ISO metadata injection.
// revised  : 2014-05-21, changed RELEASE_TAG to SUBSYS_RELEASE_TAG
// history  : 2014-05-12T13:37:00,jmg, corrected duplicate UUID tag gen.
// notes    : DomainConsistency: pct of all granule pixels for named variable,
//            falling within acceptable limits
//            CompletenessOmission: pct of all granule pixels for named variable
//            with unacceptable data relative to the total Terrestrial pixels modeled.
void l4c_Init_ISO_MetadataElements(void)
{
 // fetch pointer to master instance of ISO metadata elements   
 L4C_ISO_MetaElements *isoMeta_p = l4c_GetISO_MetaElements();
 
 // CompletenessOmission
 isoMeta_p->nee_CompletenessOmission = 0.0 ;
 isoMeta_p->gpp_CompletenessOmission = 0.0 ;
 isoMeta_p->rh_CompletenessOmission  = 0.0 ;
 isoMeta_p->soc_CompletenessOmission = 0.0 ;   
 isoMeta_p->ec_CompletenessOmission  = 0.0 ;
 // DomainConsistency --pct of all granule pixels within acceptable limits
 isoMeta_p->nee_DomainConsistency    = 100.0 ; 
 isoMeta_p->gpp_DomainConsistency    = 100.0 ; 
 isoMeta_p->rh_DomainConsistency     = 100.0 ;   
 isoMeta_p->soc_DomainConsistency    = 100.0 ; 
 isoMeta_p->ec_DomainConsistency     = 100.0 ;    

 // for L4CMOD, placeholder values...
 isoMeta_p->fpar_CompletenessOmission = 0.0;
 isoMeta_p->ndvi_CompletenessOmission = 0.0;
 isoMeta_p->fpar_DomainConsistency  = 100.0 ;    
 isoMeta_p->ndvi_DomainConsistency  = 100.0 ;    

} // end:: l4c_Init_ISO_MetadataElements()


// ----------------------------------------------------------------------
// function : l4c_Hdf5_AddFileMetadata(()
// purpose  : Add core file level metadata
// inputs   : void
// returns  : int,status
// revised  : 2014-08-13T13:36, changed L4C_PRODUCTION_UTC_DATETIME
//            back to L4C_PRODUCTION_UTC_DATETIME to preserve Joe A's
//            Perl parser logic on post-production ISO metadata injection.
// revised  : 2014-05-21, changed RELEASE_TAG to SUBSYS_RELEASE_TAG
// history  : 2014-05-12T13:37:00,jmg, corrected duplicate UUID tag gen.
// notes    : TBD, plan is to replace these calls with newer
//            frm_hdf5_WriteAttributeByType|CustomString() functions

void l4c_Hdf5_AddFileMetadata(void)
{
SesCtl         *sesCtl_p = l4c_GetSesCtl();
// AncImagePkgAdt *ancImagePkgMasterObj_p = l4c_GetAncImagePkgAdt();
NsbObj *nsb_p = l4c_GetNsbOverseer(); // <JMG> added 09/11/2014...
// PkgMgr *pkgMgr_p = sesCtl_p->l4cPkgMgr_p;
PkgTheme *pThme_p= NULL;

L4C_ISO_MetaElements *isoMeta_p = l4c_GetISO_MetaElements();

// PkgObj   *pObj_p = NULL;
hid_t file_id    = sesCtl_p->fparOutput_file_id;

// char taskUUID[64] = {0}; // buffer storing latest UUID generated
char granuleUUID[64] = {0} ;
char taskCompositePeriodStr[16] = {0} ;
char taskYearStr[16] = {0} ;
char timeStr[64] = {0};
char modelDayOfYear[8] = {0};
char fxName[256] ;
char valueBuffer[FILENAME_MAX] = {0} ;  
char nvpBuf[FILENAME_MAX] = {0};
char *fparTilePath_p = NULL;
char localDatetimeStr[64] ={0} ;
char utcDatetimeStr[64] = {0};
char govDocBuffer[64] = {0}; // for ATBD and PSD version,date etc
char   attrName[256]    = {0}  ;
char   attrValu[1024]    = {0} ;

 hid_t attr_id = -1 ;
 int a =0;
 int  status = -1 ;
time_t secNow = time(0);


    strcpy(fxName,(char *)__func__);
    
    // xLogMsg("@A In(%s) l4c_Hdf5_AddFileMetadata -- in progress, not yet available!\n",fxName);
 
    // not used in l4c_MODIS yet
    // pThme_p  = pkgMgr_GetThemeByName(pkgMgr_p,"L4C_OUTPUTS");


    // ======================================================================
    // write CF style metadata
    // ======================================================================
    status = l4c_Hdf5_AddFormalMetadata(file_id);

    // ======================================================================
    // write local baseline metadata to primary (file_id) ...
    // ======================================================================

    frm_hdf5_WriteAttributeCustomString(file_id,
        "/", "L4C_SUBSYS_RELEASE",sesCtl_p->SubsysReleaseTag);

    xLogMsg("@A In(%s) OUTPUT: L4_C Subsystem Release tag (%s)~\n"  , fxName,sesCtl_p->SubsysReleaseTag);
    xLogMsg("@A In(%s) OUTPUT: L4_C Task label            (%s)~\n"  , fxName,sesCtl_p->l4c_Task_Label);

    // new syntax, uses frm_hdf5_WriteAttributeAsType() :
    strcpy(valueBuffer,sesCtl_p->swVersion);

    // status = frm_hdf5_WriteAttributeAsType(file_id,"/","L4CMOD_SW_VERSION",
    //      IsString, 1, valueBuffer);
    // Note: this is the same as the executable ID tag...
    frm_hdf5_WriteAttributeCustomString(file_id,"/", "L4CMOD_SW_VERSION", sesCtl_p->swVersion);
    
    // <JMG> added 2014/09/11 macro L4C_VERS_TAG written here as
    // L4_C_SW_DATETIME... TBD: standardize across ALL l4c_MET,l4c_MODIS
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4MOD_SW_DATETIME"   ,sesCtl_p->swRelease);
    
    // <JMG> 2014-10-06, inherit from RIP, for consistency across produce + pre-processor outputs
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_DATASET_VERSION" ,sesCtl_p->datasetVersion);

    // <JMG> 2014-10-06, inherit from RIP, or hard-wise since DOI for  for consistency across produce + pre-processor outputs
    // note: for L4CMET, writes L4_C_MET_DOI and for L4CMOD writes L4_C_MOD_DOI, both as "TBD" for ow
    strcpy(valueBuffer, "10.5067/22TFAUSNLO9R" );
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_MDL_DOI" ,valueBuffer);

    // RELEASE tag here is the UM specific system-release identifier that includes the
    //  system version+datetime identifying not just the software but also the combined
    //  suite of static ancillary data (BPLUT,grid lookup tables, etc) that in total make
    //  up the "L4_C system"
     frm_hdf5_WriteAttributeCustomString( file_id,"/", "L4_C_SUBSYSTEM_RELEASE",sesCtl_p->SubsysReleaseTag);
     frm_hdf5_WriteAttributeCustomString( file_id,"/", "L4CMOD_TASK_LABEL"       ,sesCtl_p->l4c_Task_Label);
     frm_hdf5_WriteAttributeCustomString( file_id,"/", "L4CMOD_CRID"             ,sesCtl_p->CRID);

    // <JMG> 2014-09-12T11:39, added, was omitted earlier
    // choices: "local"|"production"
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_PRODUCTION_LEVEL" ,sesCtl_p->productionLevel);

    // PRODUCTION_DATETIME orig format returned was is: "2014-09-11T22:59:21Z"
    // but GMAO wants it in format "2014-09-11T22:59:00.000Z"
    // Joe A, needs the form to look like: 2014-08-26T06:55:28.000Z
    // l4c_GetISO8601_datetime(sesCtl_p->sesStrtSec,localDatetimeStr,utcDatetimeStr);
     frm_utilGetISO8601_datetime(secNow,localDatetimeStr,utcDatetimeStr);
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_PRODUCTION_DATETIME",   utcDatetimeStr);
    
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_HOST_PLATFORM",sesCtl_p->thisHost);
    
    // Do not use sesCtl_p->doy_Str[] as it appears to be contaminated in tcfAction etc.
    // runDayOfYear;  modelDayOfYear[8] = 
    sprintf(modelDayOfYear,"%d",sesCtl_p->runDayOfYear);
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_DAY_OF_YEAR",modelDayOfYear);
    // decide which of these to standardize on
     frm_hdf5_WriteAttributeCustomString( file_id,"/","YEARDAY",modelDayOfYear);

    sprintf(taskCompositePeriodStr,"%d",sesCtl_p->runCompositePeriod);
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_COMPOSITE_PERIOD",taskCompositePeriodStr);
 
    sprintf(taskYearStr,"%d",sesCtl_p->runModelYear);
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_MODEL_YEAR",taskYearStr);
     // BPLUT facility is now present but is not used yet
    if(strlen(sesCtl_p->inputBplutPath) < 1)
        strcpy(valueBuffer, "NOT-YET-USED");
    else
        strcpy(valueBuffer,sesCtl_p->inputBplutPath);
        
     // Note: no L4C_BPLUT is used in L4CMOD but this may be written just as a check-convenience
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_BPLUT_USED", valueBuffer);
 
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_OUTPUT_GRANULE",sesCtl_p->outputFparPath);
     // alias GRANULE_NAME for now, until convention for this is better established
     frm_hdf5_WriteAttributeCustomString( file_id,"/","GRANULE_NAME",sesCtl_p->outputFparPath);
    // generate unique GRANULE_UUID for primary file
    frm_utilGenerateUUID(granuleUUID,1);
     frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_OUTPUT_GRANULE_UUID",granuleUUID);

    // <JMG> 2014-09-11, added governing ATBD and PSD parameters from RIP pre-loaded in CollectNVPs
    // a) governing ATBD and PSD attribute
    // b) governing ATBD and PSD attributes
    if( nsbDoesKeyExist(nsb_p,"L4C_ATBD_VERSION"))
    {
        strcpy(sesCtl_p->govATBD_Vers, nsbGetValueByKey(nsb_p,"L4C_ATBD_VERSION",nvpBuf));
         frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_ATBD_VERSION"  ,sesCtl_p->govATBD_Vers);
    }
    if( nsbDoesKeyExist(nsb_p,"L4C_ATBD_DATE"))
    {
        strcpy(sesCtl_p->govATBD_Date, nsbGetValueByKey(nsb_p,"L4C_ATBD_DATE",nvpBuf));
         frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_ATBD_DATE"  ,sesCtl_p->govATBD_Date);
    }
    if( nsbDoesKeyExist(nsb_p,"L4C_PSD_EDITION"))
    {
        strcpy(sesCtl_p->govPSD_Edition, nsbGetValueByKey(nsb_p,"L4C_PSD_EDITION",nvpBuf));     
         frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_PSD_EDITION"  ,sesCtl_p->govPSD_Edition);       
    }
    if( nsbDoesKeyExist(nsb_p,"L4C_PSD_DATE"))
    {
        strcpy(sesCtl_p->govPSD_Date, nsbGetValueByKey(nsb_p,"L4C_PSD_DATE",nvpBuf));
         frm_hdf5_WriteAttributeCustomString( file_id,"/","L4_C_PSD_DATE"  ,sesCtl_p->govPSD_Date);
    }

    // Write all INPUTPOINTER_xxx entries
    frm_hdf5_WriteAttributeCustomString( file_id,"/","L4CMOD_RUNTIME_INPUTS_USED"  ,sesCtl_p->inputRipPath);
    // not used yet in l4c_MODIS.ex
    // frm_hdf5_WriteAttributeCustomString( file_id,"INPUTPOINTER_FPARCLIM",sesCtl_p->inputFparClimatologyPath);
    frm_hdf5_WriteAttributeCustomString( file_id,"/", "INPUTPOINTER_ANC"  ,sesCtl_p->inputAncPath); 
     
    for(a=0; a < sesCtl_p->nFparGranules; a++)
    {
        // fetch next FPAR 8-DAY path out of tile dictionary
        fparTilePath_p = sesCtl_p->input_Fpar_pathList[a];
        // sprintf(attrName,"INPUTPOINTER_MODIS_%3.3d",a+1);
        // <JMG> 2021-07-18T16:55 changed to _L4CMOD to better distinguish between L4CMDL and others
        // and since 'MODIS' may be confusing when VIIRs is used.
        // <KAE> 2021-10-12 Changed _L4CMOD_ to _FPAR_
        sprintf(attrName,"INPUTPOINTER_FPAR_%3.3d",a+1);
        
        strcpy(attrValu, fparTilePath_p);
        attr_id = l4c_createHdf5Attribute(file_id,attrName, attrValu);
        
    } // end:: a loop through list of all MODIS tiles used as input

    
    // <JMG> 2014-10-06T16:17, added ISO element missing that Joe A reported
    
    xLogMsg("@A In(%s) writing COMPLETENESS_OMISSION and DOMAIN_CONSISTENCY ~\n",fxName);

    // NOTE: the daily intermediate files MAY not require these, but just in case.
    // verify if should be named FPAR_COMPLETENESS_OMISSION,FPAR_DOMAIN_CONSISTENCY
    // DomainConsistency --pct of all granule pixels within acceptable limits
    // TBD: wire in option for forced use of FparClimatology
    
        status = frm_hdf5_WriteAttributeAsType(file_id,"/","FPAR_COMPLETENESS_OMISSION",
                IsFloat32, 1, &isoMeta_p->fpar_CompletenessOmission);
                
        status = frm_hdf5_WriteAttributeAsType(file_id,"/","FPAR_DOMAIN_CONSISTENCY",
            IsFloat32, 1, &isoMeta_p->fpar_DomainConsistency);
 
 l4c_EventCompletionMsg( fxName);
 // done
    
} // end:: l4c_Hdf5_AddFileMetadata()
