// ---------------------------------------------------------------
// function: l4c_MODIS_fx.c
// purpose : Implement misc support functions for l4c_MODIS
// author  : joe glassy
// revised : 2013-11-06T17:04:00-07:00
// history :
// notes   :

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"

#ifdef SUPERCEDED_BY_FRM_API

// -----------------------------------------------------------------------------
// function: l4c_testByteorder()
// purpose : perform boolean test for byte ordering, retrieve platform's endian descriptor
// author  : joe glassy
// revised : 2013-09-01
// returns : 0=Big Endian, 1=Little endian
//  enum order of symbols is CRITICAL here, so these MUST conform to this ordering:
//      enum EndianStates { isMSBfirst, isLSBfirst, isVAXEndian, unknownEndian } ; 
//      char *endianLabels[] = {"big","little"} ;
uint8 l4c_testByteorder(void)
{
 uint8 rtn = 0;
 union {
   int i ;
   char c[sizeof(int)] ;
 } foo ;
 foo.i = 1 ;

 if(foo.c[0] == 1)
   rtn = isLSBfirst ;
 else
   rtn = isMSBfirst ;
 return rtn;
} // end::l4c_testByteorder() 



// -----------------------------------------------------------------------------
//   module     : l4c_GetBits.c
//   revision   : Updated 05/04/1993 @ 15:03:53 
//   purpose    : implements getbits() from K&R C Language, to extact nBits starting at pStart
//                from a input unsigned integer bit word. Input may be 16|32 bits depending on platform
//   limitations: argument unsigned int' is 32 bits on many, if not most, 32 and 64 bit platforms,
//                so it should not be relied upon to represent exactly 16 bits, as on some older PCs.
unsigned int l4c_GetBits(unsigned int xTarget, const unsigned int pStart,const unsigned int nBits)
{
   return((xTarget >> (pStart+1-nBits)) & ~ (~0 << nBits));
} // end::l4c_GetBits()



// -----------------------------------------------------------------------------
// l4c_SetBits() ...was mumSetBits()
// purpose    : sets a range of bits explicitly to '1' in 'data', starting at base0 bit 
//              position 'startPos' and ending at bit position endPos.
// invocation : mumSetBits(strtBit,strtBit+(nBits-1),&myWord,sizeof(myWord));
// history    : from Petr Votava, Joe Glassy from NASA Terra MODIS MUM API (MODIS code)
uint32 l4c_SetBits(int16 startPos, int16 endPos, void *data, size_t dataSize)
{
  int16  byteOrder = l4c_testByteorder(); // renamed from mumTestByteorder()...
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData;

  if(byteOrder == isMSBfirst)
    memcpy(localDataPtr  + (sizeof(uint32) - dataSize), data, dataSize);
  else
    memcpy(localDataPtr, data, dataSize);

  localData = localData | (BIT32[endPos - startPos] << startPos);

  return(localData);
  
} // l4c_SetBits() 


// -----------------------------------------------------------------------------
// function: l4c_RetrieveBits() ..was mumGetBits()
// purpose : from an integer 'data' of size 'dataSize', retrieve bits from 
//           base0 bit positions 'startPos' to 'endPos', returning the result
//           as a 32 bit unsigned integer.
// Notes: (Petr Votava): All byte orderings are normalized to a standard
//           big endian (most significant byte first) ordering.
//
uint32 l4c_RetrieveBits(const int16 startPos,const int16 endPos, void *data,size_t dataSize)
{
  int16  byteOrder = l4c_testByteorder();
  uint32  localData = 0;
  uint8   *localDataPtr = (uint8 *)&localData;

  if(byteOrder == isMSBfirst)
   memcpy(localDataPtr  + (sizeof(uint32) - dataSize), data, dataSize);
  else
   memcpy(localDataPtr, data, dataSize);

  return((uint32)((localData >> startPos) & BIT32[endPos - startPos]));
  
} // end::l4c_RetrieveBits()

// END BLOCK as superceded by UM FRM API equivalent functions
#endif


// ---------------------------------------------------------------
// from: The Practice of Programming, by K&R
// useful and necessary (instead of just strcmp) to store a string table
int l4c_strCompare(const void *p1,const void *p2)
{
  char *v1, *v2 ;
  
   v1 = *(char **)p1 ;
   v2 = *(char **)p2 ;
   
   return strcmp(v1,v2);
   
} // end::l4c_strCompare()


// --------------------------------------------------------------------------
// function : l4c_getPid() 
// Purpose  : retrieve process id
// Inputs   : void
// Returns  : pid_t -- process id, unsigned long via type pid_t
// Author   : joe glassy
// Revised  : 2013-11-04T17:09:00-07:00,wheat
// Notes    : we could also use sys/sysctl.h and sysctl() etc etc.
//
pid_t l4c_getPid(pid_t *parent_pid)
{
    pid_t pid = 0;

    // simplify, just fetch PID using system call 
    pid = getpid();
    // fetch parent process'es PID while we're at it
    *parent_pid = getppid();

    // return the process ID 
    return pid ;
} // end:: l4c_getPid()


// --------------------------------------------------------------------------
// function : l4c_getCtime() 
// Purpose  : retrieve a formatted instance of current time using ctime.
// Inputs   : char *timBuf
// Author   : joe glassy
// Revised  : 2013-11-04T17:09:00-07:00,wheat
char *l4c_getCtime(char *timBuf,const time_t usersTime)
{
 time_t secNow = time(0L);

    if(usersTime > 0)
    {
        sprintf(timBuf,"%s",ctime(&usersTime));
    } else {
        sprintf(timBuf,"%s",ctime(&secNow));
    }

   // trim off the obnoxious last char we never need or want
   timBuf[strlen(timBuf)-1] = NULL_CH;

  // return string containing formatted ctime result
  return timBuf ;
} // end::getCtime()


// ----------------------------------------------------------------
// Purpose: l4c_to_upper(char *targString)
char *l4c_to_upper(char *targString)
{
 int i, sLen = strlen(targString);
 
        for(i=0; i < sLen; i++)
        {
                targString[i] = toupper(targString[i]);
        }
        // return transformed string, as a convenience 
        return targString;
} // end::l4c_to_upper()


// ----------------------------------------------------------------
// Purpose: l4c_to_lower(char *targString)
char *l4c_to_lower(char *targString)
{
 int i, sLen = strlen(targString);
 
        for(i=0; i < sLen; i++)
        {
                targString[i] = tolower(targString[i]);
        }
        // return transformed string, as a convenience 
        return targString;
} // end::l4c_to_lower()


// -----------------------------------------------------------------------------
// function: l4c_GenerateUUID()
// Purpose : generate a type-1 UUID using libuuid's services
// Inputs  : char *resultUUID_String
//         : displayMode (boolean, if 0=displays as packed string, if 1=displays as
//           output from uuid_unparse(), into couplets etc.
//
// Dependencies: -luuid must be linked in and header <uuid/uuid.h> must be available
// on the build platform!
char *l4c_GenerateUUID(char *resultUUID_String,const int displayMode)
{
 int i,k=0;

        strcpy(resultUUID_String,"no-uuid-avail");

// for NCCS had to temporarily comment out...
#ifdef HAVE_UUID
    uuid_t uuid ; 
    uuid_generate(uuid);
    uuid_unparse(uuid, resultUUID_String);
#endif

        if(displayMode == 0)
        {
         for(i=0; i < (int)strlen(resultUUID_String);i++)
          if(resultUUID_String[i] != '-')
                resultUUID_String[k++] = resultUUID_String[i];
         // re-insert term ch
         resultUUID_String[k] = '\0';
        } // end:: displayMode block, if 0
        
   return(resultUUID_String);
        
} // end::l4c_GenerateUUID()

// ----------------------------------------------------------------
// Function : l4c_rowReversal()
// Purpose  : perform a North-to-South "vertical-flip" on rows, since
//            the FP datasets are natively all upside down. Currently not needed due to
//            the way the other reprojection transform logic works.
// Inputs   : float32 *src_p,const int32 nRows,const int32 nCols
// Outputs  : void *, ptr to adjusted buffer
// Author   : joe glassy
void *l4c_rowReversal(float32 *src_p,const int32 nRows,const int32 nCols)
{
 int r;
 int offset1=0;    // initialize this offset for ascending row sequence
 int offset2=(nRows-1)*nCols ; // init for descending row sequence
 int32 nBytes = nCols * sizeof(float32); // establish one row worth of bytes
 float32 *dest_p = NULL;
 
 dest_p = memCalloc(sizeof(float32),(nRows * nCols));
 
    for(r=0; r < nRows; r++)
    {
        memcpy(dest_p+offset2, src_p+offset1, nBytes);
        offset1 += (nCols); // adjust, to move in ascending row order
        offset2 -= (nCols); // adjust, to move in descending row order
    }
    
    // RESULT: adjusted grid in dest_p, so copy back to original source now
    memcpy(src_p,dest_p,nBytes);

 // free up working array
 memFree(dest_p);
 
 // return ptr to adjusted source grid 
 return src_p;

 } // end::l4c_rowReversal()


// ----------------------------------------------------------------
// Function : l4c_strGetBasenamePath()
// Purpose  : from a fully qualified path, extract just the basename
// Inputs   : const char *pathStr_P,char *baseName_p,int pathSepCh
// Returns  : ptr to basename portion of fully qualified file path
// Author   : joe glassy
// Notes    : Linux pathSepCh is '/', Windows is '\'.
char *l4c_strGetBaseNamePath(const char *pathStr_p,char *baseName_p,int pathSepCh)
{
    char *ch_p = NULL;
    
    // check if a prefix portion exists,if so, extract basename from it
    if((ch_p=strrchr(pathStr_p,pathSepCh)+1)==NULL)
    {
      // here, there is no extended prefix portion from which to extract 
      // the basename, so just return ptr to basename itself 
      strcpy(baseName_p,pathStr_p);
    } else {
      strcpy(baseName_p,ch_p);
    }

    // return pointer to just the basename segment of path string
    return baseName_p ;
} // end::l4c_strGetBaseNamePath()


// ----------------------------------------------------------------
// Function : l4c_strGetPathPrefix()
// Purpose  : from a fully qualified path, extract just the prefix portion
// Inputs   : const char *pathStr_p,char *pathPrefix_p,int pathSepCh
// Returns  : ptr to prefix portion of fully qualified file path
// Author   : joe glassy
// Notes    : Linux pathSepCh is '/', Windows is '\'.
// 
char *l4c_strGetPathPrefix(const char *pathStr_p,char *pathPrefix_p,int pathSepCh)
{
 int pathStrLen = 0 ; // length of original (umambiguous) path 
 int prefixLen  = 0 ; // length of path prefix 
 int baseLen    = 0 ; // length of basename segment of path 
 char *baseName_p = NULL; //ptr to basename segment within the overall path

    // First, establish wheather just the bare basename exists or if there is
    //  valid path text that precedes the basename, e.g.
    //     \this\is\prefix\text\in\path\ThisIsTheBaseName.Extension ..for Win2k
    //     /this/is/prefix/text/in/path/ThisIsTheBaseName.Extension ..for Unix
    pathStrLen = strlen(pathStr_p);

    if( (baseName_p=strrchr(pathStr_p,pathSepCh)) !=NULL)
    {
      baseName_p++ ;
      baseLen = (baseName_p!=NULL)? strlen(baseName_p) : 0;
      prefixLen= pathStrLen - baseLen;

      if(prefixLen >=1)
      {
         memset(pathPrefix_p,'\0',prefixLen+1);
         strncpy(pathPrefix_p, pathStr_p, prefixLen);
      }
      pathPrefix_p[prefixLen] = '\0';
    } else {  // ..or else, return an empty string as the prefix 
      strcpy(pathPrefix_p,"");
    }
    // return resulting path prefix fragment
  return pathPrefix_p ;
} // end:: l4c_strGetPathPrefix()


// ----------------------------------------------------------------
// Function : l4c_strExtractExt()
// Purpose  : from a basename or fully qualified path, extract just the extension
//          : if present, or if not present, returns original string
// author   : joe glassy
char *l4c_strExtractExt(char *source,char *dest,const int sep_char)
{
 int sep_symbol = sep_char;

    // if a dummy or invalid separation char passed in, use default
    if(sep_char <= 0)
       sep_symbol = '.' ;

    // at least one sep char found, so identify rightmost '.' symbol
    if(strrchr(source,sep_symbol)!=NULL)
    {
        strcpy(dest,strrchr(source,(int)sep_symbol)+1 );
        return dest ;
    } else {
        // no sep. char found, so return pointer to intact string
        return source;
    }
} // end:: l4c_strExtractExt()


// ----------------------------------------------------------------
// Function : l4c_strExtractRoot()
// Purpose  : from a basename or fully qualified path, extract just the
//          : root portion if present, or if not present, returns original string
// author   : joe glassy
char *l4c_strExtractRoot(char *source,char *dest,const int sep_char)
{
     int sep_symbol = sep_char;

    // if a dummy or invalid separation char passed in, use default
    if(sep_char <= 0)
       sep_symbol = '.';

    // at least one sep char found, so proceed...
    if(strrchr(source,sep_symbol)!=NULL)
    {
        memset(dest, '\0', strlen(source)+1);
        return strncpy(dest, source,
                (strlen(source)-strlen(strrchr(source,sep_symbol))) );
    } else {
    // no sep. char found, so return pointer to intact string
    return source ;
  }
} // end::l4c_strExtractRoot()

// ----------------------------------------------------------------
// function: l4c_calcFrequencies()
// purpose : given an uint8 image of nElem pixels, generate abs frequency
//           table and output it to open file, destf.
int l4c_calcFrequencies(char *taskLabel_p,uint8 *sourceImg_p,const int32 nElem,FILE *destf)
{
    int32 p =0; 
    // nElem usually is EGv2_GLOBAL_1KM_NS_PIXELS * EGv2_GLOBAL_1KM_EW_PIXELS;
    uint16 *frqVar_p= NULL;
    float64 vPix=0.0,vSum =0.0, vMean=0.0,vMin=0.0,vMax=0.0, vN =0.0;
    
    // should eventually work with datatypes up to uint16...
    frqVar_p = (uint16 *)memCalloc(65536,sizeof(uint8));
    
    // construct frequency table (absolute values) of fpar
    for(p=0; p < nElem; p++)
    {
        vPix = (float64)sourceImg_p[p];
        // this is intentionally hard-wired for now since we 
        // want to filter out either OUT_DOMAIN or FILL_VALUE..
        if( vPix < 254.0 )
        {
            vN += 1.0 ;
            vSum += vPix ;
            if(p==0)
            {
                    vMin = vMax = vPix;
            } else {
                    if(vPix < vMin) vMin = vPix;
                    if(vPix > vMax) vMax = vPix;
            }
        } // end:: block if pixel is NOT a fill value or out of domain
        frqVar_p[sourceImg_p[p]]++ ;
    } // end:: p loop through all pixels of image
    
    // calculate simple mean
    vMean = vSum / vN ;
    
    fprintf(destf,"\n@A @D Diagnostic check of [%s] image frequencies~\n",taskLabel_p); 

    for(p=0; p < 256; p++)
        fprintf(destf,"@A @D bin %5.5d freq %d\n",p+1,(int)frqVar_p[p]);

    fprintf(destf,"\n@A @D-DIAG [%s] N(%g) min(%.3lf) max(%.3lf) mean(%.3lf)~\n",
            taskLabel_p, vN, vMin,vMax,vMean);  
    fprintf(destf,"\n@A Diagnostic task (%s) complete~\n",taskLabel_p); 
    
    // release memory
    memFree(frqVar_p);  
    
    // return count
    return (int)vN ;
    
} // end::l4c_calcFrequencies()
