// ---------------------------------------------------------------
// function: l4c_MODIS_io.c
// purpose : Implement I/O for l4c_MODIS
// author  : joe glassy
// revised : 2017-01-26, begin support for MODIS Collection 6 MOD15A2H 500m tiles
// revised : 2015-06-10T16:10, <JMG> changed L4CMOD_SOFTWARE_VERSION to std L4CMOD_SW_VERSION
// revised : 2014-10-15T11:16, gridding global "smudge" error noted, with
//           find+fix task in progess.
// history : 2014-10-14T10:14:02-07:00, more clean up for ORT#2, ORR
// history : 2014-08-19, split out NDVI from FPAR to make more explicit
// history : 2013-11-06T17:04:00-07:00
// history : 2013-11-01
// notes   : tileref.c was incorrect read as tileref.r in readSmapAncillary--corrected 11/11/2013.
// notes   : n/a


#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"


// ----------------------------------------------------------------------------
// function: l4c_read_hdfeos5()
// purpose : Confirm if for VIIRS only as usually MODIS requires HDFEOS2. 
//           The newer VIIRs typically requires HDFEOS5 and NOT the HDFEOS2 required by orig MODIS
// notes   : Reads reads both the FPAR, QC and EXTRA_QC as needed
// revised : 2021-07-24T16:57, added dimswap() call on extra_qc
// revised : 2021-07-14T17:31, for EvalDiag prep of VIIRs integration
// revised : 2021-07-14T17:27, GDopen(),GDattach() is SPECIFIC ONLY to MODIS Coll 6 hdf files
//           that require HDFEOS v2 and NOT the HDFEOS5 library required by VIIRs.
// revised : 2017-02-12T15:06, adjusting tile_p ref for either 500m or 1km case
//           noting that the GRID ID "MOD_GRID_MOD152H is different from MOD_GRID_MOD15A2
// revised : 2017-02-09T17:34, adjust to ALSO allow for reading new 500m tiles,
//           assume "fname[]" here points exactly to the 500m or 1km dataset though

int l4c_read_hdfeos5(char *inputTilePath,tile_struct *tile_p,const int ithTile,const int nTiles)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
  hid_t gridfileid  = 0 ;
  hid_t gridid      = 0 ;
  // 2400 * 2400 totals 5760000
  hid_t i =0 ;
  // 500m requires 2400x2400..
  hid_t nElem = 5760000 ; 
  herr_t retFx = -1 ;
 
  int32     status;
  int32     nx =0,ny =0;
  char fxName[256] = {0} ;
  
  static int ithGranule =0 ;

  // register this functions name for logging
  strcpy(fxName,(char *)__func__); 

  // OPEN the next INPUT FPAR (tile) file in the sequence of 289
  if((gridfileid = HE5_GDopen(inputTilePath, H5F_ACC_RDONLY)) == -1)
  {
    xLogMsg("@F In(%s): HE5_GDopen FAILED on [%s]~\n",fxName,inputTilePath);  
  } else {
    ithGranule++ ;
    // xLogMsg("@A read_hdfeos: SUCCESS-ON-OPEN Granule(%s) %d of %d (%.2lf Pct)~\n", inputTilePath, ithGranule, nTiles, ((float)ithGranule/(float)nTiles)*100.0 );
  }
  
  // ATTACH to the named grid, returning a valid GridID here
  // xLogMsg("@A ATTACH tile %s to grid %s now~\n",inputTilePath,sesCtl_p->dsNameGrid);
  if((gridid =HE5_GDattach(gridfileid,sesCtl_p->dsNameGrid))== -1)
  {
    xLogMsg("@F In(%s) error on HE5_GDattach for %s~\n",fxName, inputTilePath);
  } else {
    // xLogMsg("@A read_hdfeos5: success on HE5_GDattach gridID %d for %s for grid %s~\n",gridid,inputTilePath,sesCtl_p->dsNameGrid);
    status = 0;
  }
       
  // READ: the INPUT named (FPAR) data field, for UseVIIRsInputs or MODIS 500m
  retFx = HE5_GDreadfield(gridid, sesCtl_p->dsNameFpar, NULL, NULL, NULL,tile_p->fpar); 
  
  if(retFx == -1)
  {
     xLogMsg("@E read_hdfeos5:: HE5_GDreadfield ERROR gridID %d dataset [%s] in [%s]~\n",gridid,sesCtl_p->dsNameFpar,inputTilePath);
  } else {
     // xLogMsg("@D read_hdfeos:: GDreadfield SUCCESS gridID %d dataset [%s] in [%s]~\n",gridid,sesCtl_p->dsNameFpar,inputTilePath);
     status = 0;
  }

  // read the INPUT FPAR QC data field
  if((retFx = HE5_GDreadfield(gridid,sesCtl_p->dsNameFparQc,NULL,NULL,NULL,tile_p->qc))== -1)
  {
     xLogMsg("@E read_hdfeos:: GDreadfield ERROR gridID %d dataset [%s] in [%s]~\n",gridid,sesCtl_p->dsNameFparQc,inputTilePath);
  } else {
     status = 0 ;
  }

  // <JMG> 2021-07-18, ADDED READ, assume we will shortly need extra_qc layer as well
  if((retFx = HE5_GDreadfield(gridid,sesCtl_p->dsNameFparExtraQc,NULL,NULL,NULL,tile_p->extra_qc))== -1)
  {
     xLogMsg("@E read_hdfeos:: GDreadfield ERROR gridID %d dataset [%s] in [%s]~\n",gridid,sesCtl_p->dsNameFparExtraQc,inputTilePath);
  } else {
     status = 0 ;
  }


   // READ: the corner coordinates 
   HE5_GDgridinfo(gridid,&nx,&ny,tile_p->upleft,tile_p->lowright);
   
   //   xLogMsg("@A read_hdfeos:: GDgridinfo on %s gridID %d upLeft %.4f %.4f  lowright %.4f %.4F ~\n",
   //   inputTilePath,gridID, tile_p->upleft[0]  ,tile_p->upleft[1], tile_p->lowright[0],tile_p->lowright[1]);
      
   //  detach from grid
   HE5_GDdetach(gridid);
   //  close the HDFEOS 5 as the currently active input file
   HE5_GDclose(gridfileid);

    // <JMG> 2021-07-14 TBD, VERIFY if BOTH MODIS and VIIRs require this dimswap. Default is yes.
    //  flip fpar around to agree with latlon
    //  TBD: verify why this is needed, as the same transformation cannot apparently
    //       be accomplished by adjusting lat/lon instead
    // xLogMsg("@A read_hdfeos:: calling dimswap on (tile.fpar,tile.qc,tile.qc_extra processing [%s]~\n",inputTilePath);

    // <JMG> 2017-02-10T16:58, verify that 1km vs 500m handled correctly.
    // xLogMsg("@A In(%s) calling dimswap on both (fpar,qc) now",fxName);
    dimswap(tile_p->fpar);
    dimswap(tile_p->qc);
    // <JMG> 2021-07-24T16:57, added extra_qc to sync up with parallel logic on qc
    dimswap(tile_p->extra_qc);
  
  return(0);
  
} // end:: read_hdfeos5()

// ----------------------------------------------------------------------------
// function: l4c_read_hdfeos2()
// purpose : reads either the FPAR and its QC , only for MODIS , NOT for VIIRs
// revused : 2021-07-24T16:57, added dimswap() call on extra_qc
// revised : 2021-07-14T17:27, GDopen(),GDattach() is SPECIFIC ONLY to MODIS Coll 6 hdf files
//           that require HDFEOS v2 and NOT the HDFEOS5 library required by VIIRs.
// revised : 2017-02-12T15:06, adjusting tile_p ref for either 500m or 1km case
//           noting that the GRID ID "MOD_GRID_MOD152H is different from MOD_GRID_MOD15A2
// revised : 2017-02-09T17:34, adjust to ALSO allow for reading new 500m tiles,
//           assume "fname[]" here points exactly to the 500m or 1km dataset though
// revision: 2014-08-19T16:38
int l4c_read_hdfeos2(char *inputTilePath, tile_struct *tile_p,const int ithTile,const int nTiles)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 
  int32     hdfID, gridID, status;
  int32     nx =0,ny =0;
  char fxName[256] = {0} ;
  
  static int ithGranule =0 ;

  // register this functions name for logging
  strcpy(fxName,(char *)__func__); 

  // open the next FPAR or NDVI .hdf (tile) file
  if((hdfID=GDopen(inputTilePath,DFACC_READ))== -1)
  {
    xLogMsg("@F read_hdfeos:: GDopen FAILED on [%s]~\n",inputTilePath);  
  } else {
    ithGranule++ ;
    // xLogMsg("@D read_hdfeos: SUCCESS-ON-OPEN Granule(%s) %d of %d (%.2lf Pct)~\n", inputTilePath, ithGranule, nTiles, ((float)ithGranule/(float)nTiles)*100.0 );
  }
  
  // xLogMsg("@A In(%s) read_hdfeos: GDopen on [%s] (%3d of %3d) with ID %d\n",fxName, inputTilePath, ithTile,nTiles, hdfID);

  // ATTACH to the named grid, returning a valid GridID here
  // xLogMsg("@A ATTACH tile %s to grid %s now~\n",inputTilePath,sesCtl_p->dsNameGrid);
  
  if((gridID= GDattach(hdfID,sesCtl_p->dsNameGrid)) == -1)
  {
    xLogMsg("@F In(%s) error on GDattach for %s~\n",fxName, inputTilePath);
  } else {
    // xLogMsg("@D read_hdfeos:: success on GDattach gridID %d for %s for grid %s~\n",gridID,inputTilePath,sesCtl_p->dsNameGrid);
    status = 0;
  }
       
  // read the FPAR data field, same for UseMODISInputs or UseVIIRsInputs
  if((status = GDreadfield(gridID,sesCtl_p->dsNameFpar,NULL,NULL,NULL,tile_p->fpar))== -1)
  {
     xLogMsg("@E read_hdfeos:: GDreadfield ERROR gridID %d dataset [%s] in [%s]~\n",gridID,sesCtl_p->dsNameFpar,inputTilePath);
  } else {
     // xLogMsg("@D read_hdfeos:: GDreadfield SUCCESS gridID %d dataset [%s] in [%s]~\n",gridID,sesCtl_p->dsNameFpar,inputTilePath);
     status = 0;
  }

  // ORIG 2017 logic, read the FPAR QC data field
  if((status = GDreadfield(gridID,sesCtl_p->dsNameFparQc,NULL,NULL,NULL,tile_p->qc))== -1)
  {
     xLogMsg("@E read_hdfeos:: GDreadfield ERROR gridID %d dataset [%s] in [%s]~\n",gridID,sesCtl_p->dsNameFparQc,inputTilePath);
  } else {
     status = 0 ;
  }

  // <JMG> 2021-07-18T16:11 added for consistency with VIIRs, may need legacy extra_qc (tBD confirm)
  if((status = GDreadfield(gridID,sesCtl_p->dsNameFparExtraQc,NULL,NULL,NULL,tile_p->extra_qc))== -1)
  {
     xLogMsg("@E read_hdfeos:: GDreadfield ERROR gridID %d dataset [%s] in [%s]~\n",gridID,sesCtl_p->dsNameFparExtraQc,inputTilePath);
  } else {
     status = 0 ;
  }


   //  READ: the corner coordinates 
   GDgridinfo(gridID,&nx,&ny,tile_p->upleft,tile_p->lowright);
   //   xLogMsg("@A read_hdfeos:: GDgridinfo on %s gridID %d upLeft %.4f %.4f  lowright %.4f %.4F ~\n",
   //   inputTilePath,gridID, tile_p->upleft[0]  ,tile_p->upleft[1], tile_p->lowright[0],tile_p->lowright[1]);
      
   GDdetach(gridID); // detach from the grid
   GDclose(hdfID);   // close the currently active .hdf input file

    // <JMG> 2021-07-14 TBD, VERIFY if BOTH MODIS and VIIRs require this dimswap. Default is yes.
    //  flip fpar around to agree with latlon
    //  TBD: verify why this is needed, as the same transformation cannot apparently
    //       be accomplished by adjusting lat/lon instead
    // xLogMsg("@A read_hdfeos:: calling dimswap on (tile.fpar,tile.qc,tile.qc_extra processing [%s]~\n",inputTilePath);

    // <JMG> 2017-02-10T16:58, verify that 1km vs 500m handled correctly.
    // xLogMsg("@A In(%s) calling dimswap on both (fpar,qc) now",fxName);
    dimswap(tile_p->fpar);
    dimswap(tile_p->qc);
    // <JMG> 2021-07-24T16:57, added extra_qc to sync up with parallel logic on qc
    dimswap(tile_p->extra_qc);
  
  return(0);
  
} // end:: l4c_read_hdfeos2()

// ---------------------------------------------------------------
// function: l4c_AllocateFparEasegrid(void)
// purpose : allocate OUTPUT EASEGRID 1km images for fpar, qc, mask
// revised : 2014-08-19T16:01, split out NDVI from FPAR to make more explicit
// notes   : n/a
int l4c_AllocateFparEasegrid(void)
{
 ImgPkg *imagePkg_p = l4c_GetImagePackage();
 SesCtl *sesCtl_p   = l4c_GetSesCtl();
 char fxName[256] = {0} ;
 long nPixels = -1 ;
 
    strcpy(fxName,(char *)__func__); 

   //  <JMG> 2021-07-14 TBD ASSUME BOTH MODIS|VIIRs cases are same here since output is 1km
   if(sesCtl_p->fparResolutionMeters == 1000)
          nPixels = EGv2_GLOBAL_1KM_PIXELS ;
   else  // <JMG> 2021-07-14 TBD CRUCIAL, verify if need 1km or 500m pixel definition here
          nPixels = EGv2_GLOBAL_1KM_PIXELS ;
          
   xLogMsg("@A In(%s) l4c_AllocateFparEasegrid: allocating ease.fpar 1km as uint8 ~ \n",fxName);
   imagePkg_p->ease.fpar = memObjCalloc("ease.fpar",sizeof(uint8),EGv2_GLOBAL_1KM_PIXELS,IsUint8,fxName,__FILE__);

   xLogMsg("@A In(%s) l4c_AllocateFparEasegrid: allocating ease.qc 1km as uint8 ~ \n",fxName);
   imagePkg_p->ease.qc   = memObjCalloc("ease.qc"  ,sizeof(uint8),EGv2_GLOBAL_1KM_PIXELS,IsUint8,fxName,__FILE__);
    
    return 0;
    
} // end::l4c_AllocateFparEasegrid(void)


// -----------------------------------------------------------------------------
// Function : l4c_createHdf5Attribute()
// Purpose  : create a STRING attribute owned by owner_id.
// Inputs   : const hid_t owner_id,const char *attrLabel,const char *attrValue
// Author   : joe glassy
// Returns  : unique ID of the attribute just created
// htri_t H5Aexists_by_name( hid_t loc_id, const char *obj_name, const char *attr_name, hid_t lapl_id )
hid_t l4c_createHdf5Attribute(const hid_t owner_id,const char *attrLabelName,const char *attrValue)
{
 herr_t ret =0;
 hid_t attr_dspace = -1;
 hid_t attr_type   = -1;
 hid_t attr_id =-1;
 int didCreate = 0 ; // boolean, set to 1 if new attrib created.
 char fxName[256] = {0} ;

    strcpy(fxName,(char *)__func__); 
         
    // FIRST TEST IF ATTRIBUTE Already exists and if it does. here 0 is not considered positive
    if( H5Aexists_by_name(owner_id,".",attrLabelName,H5P_DEFAULT) <= 0) // pos value is TRUE, does exist.
    {
        // Attribute DOES NOT YET EXIST here, so create it
        // xLogMsg("@A In(%s) New attribute (%s) created now on id %d~\n",fxName, attrLabelName, owner_id);
        
        // Create new attributes dataspace property
        attr_dspace = H5Screate(H5S_SCALAR);
        // Create new attributes type as a string. TBD: modify to be able to 
        // create any other datatype as well
        attr_type   = H5Tcopy(H5T_C_S1);
        // set the SIZE of the attribute as the length of the single string + 1
        H5Tset_size(attr_type, strlen(attrValue)+1) ;
        // last, create the empty object instance to hold this HDF5 string attribute
        attr_id = H5Acreate(owner_id, attrLabelName, attr_type, attr_dspace, H5P_DEFAULT, H5P_DEFAULT);
        didCreate = 1 ;
    } else {
        // here, attribute already existed, so just open it
        if((attr_id = H5Aopen(owner_id, attrLabelName, H5P_DEFAULT)) <= 0)
        {
            xLogMsg("@A In(%s) Failed to open pre-existing attribute at file_id %d Label<%s> ~\n",
                fxName, owner_id,attrLabelName);
        }
        // next, retrieve the pre-existing attributes native datatype
        attr_type =  H5Aget_type(attr_id);
    }

    // Last, write the string attribute's value, held in string
    ret = H5Awrite(attr_id, attr_type, attrValue); 
        
    // xLogMsg("@A In(%s) file_id %d Label<%s> = Value<%s> ~\n",fxName, owner_id,attrLabelName,attrValue);

    if(didCreate)
        H5Sclose(attr_dspace);  
        
    // always close afterwards regardless 
    H5Aclose(attr_id);
        
 return attr_id ;
 
} // end::l4c_createHdf5Attribute()


// -----------------------------------------------------------------------------
// Function : writeNvpAsAttribute
// Purpose  : using a supported NVP keyword, create a string attribute
// Inputs   : const hid_t file,const char *keyword,char *attrForceName)
// Author   : joe glassy
// Returns  : hid_t attr_id if successful, or -1 if not
// Notes    : The HDF5 attribute written using the keyword as the
//            attributes name and the NVP's 'value' as the attribute value
hid_t l4c_writeNvpAsAttribute(hid_t file_id,const char *keyword,const char *attrForceName)
{
 NsbObj *nsbObj_p = l4c_GetNsbOverseer(); 
    
 char attrValu[2048]= {0} ; 
 char nvpBuf[2048]= {0} ;
 char fxName[256] = {0} ;
 int keyResult = -1 ;
 hid_t attr_id = -1 ;
 
 
    // search table of supported keywords for this one and if
    // found, return index >=0, or -1 if failed.
    // keyResult = l4c_SearchForSupportedKey(keyword);
    keyResult = 0 ;  // later, use NSB's equiv method for this
    
    if(keyResult >=0)
    {
        // fetch value of this supported keyword
        strcpy(attrValu, nsbGetValueByKey(nsbObj_p,keyword,nvpBuf));
        
        // Force use of 'attrForceName' instead of 'keyword' as the
        // HDF5 attribute name or label.
        if(attrForceName != NULL && strlen(attrForceName)>= 1)  
        {
            attr_id = l4c_createHdf5Attribute(file_id,attrForceName, attrValu);
            
        } else {
            // use the 'keyword' verbatim as the HDF5 attribute name
            attr_id = l4c_createHdf5Attribute(file_id,keyword, attrValu);
        }
        
        xLogMsg("@A l4c_writeNvpAsAttribute: wrote key(%s) value(%s) ~\n",fxName,keyword,attrValu);
        
    } else { // end:: block if keyword IS a supported one
        // notify caller that this keyword was invalid or not supported
        xLogMsg("@E l4c_writeNvpAsAttribute: NO key(%s) found - suppressed attrib write!~\n",
        keyword);
    }
    
 return attr_id;
} // end::l4c_writeNvpAsAttribute()


// ---------------------------------------------------------------
// function: l4c_writeHdf5AttributesToOutput()
// purpose : writes all other SMAP attributes to primary output HDF5 file
// author  : joe glassy
// revised : 2017-01-26T13:39 add write of new FPAR RES RIP keyword to
//           file level attribute for provenance 
// revised : 2014-10-14T11:27:00, refinements for ORT,ORR
// history : 2014-08-28, refinements pending, see below.
// history : 2014-08-20T10;29, for ndvi, add gain,offset
// history : 2013-11-06T17:04:00-07:00
// notes   : TBD 2014-08-28, improve this by better assuring that the
//           exact attribute name and value matches those defined 
//           elsewhere.
// notes   : called by: l4c_writeSmapOutputs() towards end
int l4c_writeHdf5AttributesToOutput(hid_t file_id)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();    
 char  uuidStr[64] = {0} ;
 hid_t attr_id = -1 ;
 int a =0;
 char   attrName[256]    = {0}  ;
 char   attrValu[1024]    = {0} ;
 char   fxName[256]       = {0} ;
 char localDatetimeStr[64]= {0};
 char utcDatetimeStr[64]  = {0};

 char   *fparTilePath_p  = NULL;
 
        
    strcpy(fxName,(char *)__func__); 
    xLogMsg("@A l4c_WriteHdf5AttributesToOutput");

#ifdef HAVE_UUID    
    l4c_GenerateUUID(uuidStr,1);
    // create UUID
    attr_id = l4c_createHdf5Attribute(file_id,"GRANULE_UUID", uuidStr);
#endif  
    
    // Record whether FPAR or NDVI intended for use
    l4c_writeNvpAsAttribute(file_id,"L4C_GPP_SOURCE","L4C_GPP_SOURCE");
    
    // <JMG> 2017-01-26T13:29, add write of new FPAR RES RIP keyword to
    // file level attribute for provenance 
    l4c_writeNvpAsAttribute(file_id,"L4C_FPAR_RESOLUTION_METERS","L4C_FPAR_RESOLUTION_METERS");

    l4c_writeNvpAsAttribute(file_id,"L4C_MODEL_YEAR"       ,"MODEL_YEAR");
    l4c_writeNvpAsAttribute(file_id,"L4C_MODEL_YEARDAY"    ,"YEARDAY");

    // using keyword L4C_OUTPUT_MODIS_PATH, write an attribute named
    // 'GRANULE_NAME' to the output HDF5 file
    l4c_writeNvpAsAttribute(file_id,"L4C_OUTPUT_MODIS_PATH","GRANULE_NAME");

    // for GMAO production always tag with UTC
    l4c_createHdf5Attribute(file_id, "L4_C_PRODUCTION_DATETIME",    utcDatetimeStr);

    l4c_writeNvpAsAttribute(file_id, "L4C_CRID","");    
    l4c_writeNvpAsAttribute(file_id, "L4C_SCIENCE_VERSION_ID"    ,"L4C_SCIENCE_VERSION_ID");
    
    l4c_writeNvpAsAttribute(file_id, "L4C_SUBSYSTEM_RELEASE"     ,"L4C_SUBSYSTEM_RELEASE");
        // <JMG> 2015-06-10T16:05, changed L4C_SOFTWARE_VERSION to std L4C_SW_VERSION
    l4c_writeNvpAsAttribute(file_id, "L4CMOD_SW_VERSION","L4CMOD_SW_VERSION");
    // <JMF> 2021-07-18T16:27 Confirm,verify we want 'INPUT_DATASET_VERSION" here? take out 'INPUT'?
    l4c_writeNvpAsAttribute(file_id, "L4CMOD_DATASET_VERSION","L4CMOD_INPUT_DATASET_VERSION");
    
    l4c_writeNvpAsAttribute(file_id, "L4CMOD_TASK_LABEL"      , "L4CMOD_TASK_LABEL");
    l4c_writeNvpAsAttribute(file_id, "L4CMOD_TASK_UUID"       , "L4CMOD_TASK_UUID");

    l4c_writeNvpAsAttribute(file_id, "L4C_ATBD_VERSION"    , "L4C_ATBD_VERSION");
    l4c_writeNvpAsAttribute(file_id, "L4C_ATBD_DATE"       , "L4C_ATBD_DATE");
    l4c_writeNvpAsAttribute(file_id, "L4C_PSD_EDITION"     , "L4C_PSD_EDITION");
    l4c_writeNvpAsAttribute(file_id, "L4C_PSD_PUBDATE"     , "L4C_PSD_PUBDATE");

    // generate PRODUCTION DATETIME. Note: should express as a ISO-8601 not using ctime semantics
    // l4c_Hdf5_CreateAttributeString( file_id,"L4_C_PRODUCTION_DATETIME",  frm_utilGetCtime(timeStr,sesCtl_p->sesStrtSec));
    l4c_GetISO8601_datetime(sesCtl_p->sesStrtSec,localDatetimeStr,utcDatetimeStr);
    

    l4c_createHdf5Attribute(file_id, "L4_C_HOST_PLATFORM"       , sesCtl_p->thisHost);
    l4c_createHdf5Attribute(file_id, "L4_C_HOST_USER"           , sesCtl_p->thisUsername);
    l4c_createHdf5Attribute(file_id, "RUNTIME_INPUT_PARAMETERS", sesCtl_p->inputRipPath);
    
    // l4c_createHdf5Attribute(file_id,"units","dimensionless");
    for(a=0; a < sesCtl_p->nFparGranules; a++)
    {
        // fetch next FPAR 8-DAY path out of tile dictionary
        fparTilePath_p = sesCtl_p->input_Fpar_pathList[a];
        // <KAE> 2021-09-27 Consistent with changes to l4c_Hdf5_AddFileMetadata()
        //    in l4c_MODIS_WriteMetadata.c, changed "_MODIS_" to "_L4CMOD_" to better
        //    distinguish when VIIRS input fPAR data are used
        sprintf(attrName,"INPUTPOINTER_L4CMOD_%3.3d",a+1);
        
        strcpy(attrValu, fparTilePath_p);
        attr_id = l4c_createHdf5Attribute(file_id,attrName, attrValu);
        
    } // end:: a loop through list of all MODIS tiles used as input
    
    
 return attr_id;
 
} // end::l4c_writeHdf5AttributesToOutput()


// ---------------------------------------------------------------
// function: l4c_readSmapAncillary()
// purpose : reads (5) images from SMAP_L4_C_Ancillary.h5
// author  : joe glassy
// revised : 2013-11-06T17:04:00-07:00
// notes   :
/*
   DATASET "PFT_M09_mask" {
   DATASET "tile_c_M01" {
   DATASET "tile_h_M01" {
   DATASET "tile_r_M01" {
   DATASET "tile_v_M01" {
*/
int l4c_readSmapAncillary(void)
{
 SesCtl *sesCtl_p   = l4c_GetSesCtl();
 NsbObj *nsbObj_p   = l4c_GetNsbOverseer(); 
 ImgPkg *imagePkg_p = l4c_GetImagePackage();
    
 hid_t anc_file_id = -1 ;
 hid_t dsTypeId = H5T_NATIVE_UINT8 ;
 herr_t ret = -1 ;
 char hdf5Path[FILENAME_MAX] = {0} ;
 char datasetName[256] = {0} ;
 char fxName[256] = {0} ;
 char nvpBuf[FILENAME_MAX] ;
 int status = 0 ;
 
    strcpy(fxName,(char *)__func__);   
 
  
    sesCtl_p->anc_file_id  = frm_hdf5_Openfile( sesCtl_p->inputAncPath ,fxName);
    anc_file_id = sesCtl_p->anc_file_id ;
  
    // Allocate memory for tile reference templates. 
    xLogMsg("@A In(%s) readSmapAncillary opening (4) tile mosaic support files~\n",fxName);     
    
    xLogMsg("@A In(%s) allocating (4) tile 1km (14616x34704) (rescaled,type-converted) reference templates~\n",fxName);     
    
    // 9km global land cover {0|1} mask...
    xLogMsg("@A In(%s) allocating ease.mask~\n",fxName);    
    imagePkg_p->ease.mask = (uint8*)memObjCalloc("ease.mask",sizeof(uint8),EGv2_GLOBAL_9KM_PIXELS, IsUint8,fxName,__FILE__);

    xLogMsg("@A In(%s) allocating tileref.v~\n",fxName);    
    imagePkg_p->tileref.v = (uint8*)memObjCalloc("tileref.v",sizeof(uint8),EGv2_GLOBAL_1KM_PIXELS, IsUint8,fxName,__FILE__);

    xLogMsg("@A In(%s) allocating tileref.h~\n",fxName);    
    imagePkg_p->tileref.h = (uint8*)memObjCalloc("tileref.h",sizeof(uint8),EGv2_GLOBAL_1KM_PIXELS, IsUint8,fxName,__FILE__);

    xLogMsg("@A In(%s) allocating tileref.r~\n",fxName);    
    imagePkg_p->tileref.r = (int16*)memObjCalloc("tileref.r",sizeof(int16),EGv2_GLOBAL_1KM_PIXELS, IsInt16,fxName,__FILE__);

    xLogMsg("@A In(%s) allocating tileref.c~\n",fxName);    
    imagePkg_p->tileref.c = (int16*)memObjCalloc("tileref.c",sizeof(int16),EGv2_GLOBAL_1KM_PIXELS, IsInt16,fxName,__FILE__);

    // Read the global EASEGRID 9km mask
    dsTypeId =  H5T_NATIVE_UINT8 ;
    strcpy(datasetName,(char *)sesCtl_p->dsNamePFT_M09_Mask);

    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId, datasetName,(uint8*)imagePkg_p->ease.mask,fxName);
    xLogMsg("@A l4c_readSmapAncillary:: SUCCESS ON read of dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);   

    // Read the HORIZONTAL indices from tile_h_M01
    strcpy(datasetName,(char *)sesCtl_p->dsNameTileHoriM09);
    dsTypeId =  H5T_NATIVE_UINT8 ;
    
    if((ret=H5LTread_dataset(anc_file_id, datasetName, dsTypeId, (uint8*)imagePkg_p->tileref.h))<0)
    {
        xLogMsg("@F l4c_readSmapAncillary:: FAILED to read dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);
    } else {
        xLogMsg("@A l4c_readSmapAncillary:: SUCCESS ON read of dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);
    }
    // Read the VERTICAL indices from tile_v_M01
    strcpy(datasetName,(char *)sesCtl_p->dsNameTileVertM09);
    dsTypeId =  H5T_NATIVE_UINT8 ;  
    if((ret=H5LTread_dataset(anc_file_id, datasetName, dsTypeId, (uint8*)imagePkg_p->tileref.v))<0)
    {
        xLogMsg("@F l4c_readSmapAncillary:: FAILED to read dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);   
    } else {
        xLogMsg("@A l4c_readSmapAncillary:: SUCCESS ON read of dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);   
    }
    
    // Read the TILE ROW indices from tile_r_M01
    strcpy(datasetName,(char *)sesCtl_p->dsNameTileRowM09);
    dsTypeId =  H5T_NATIVE_INT16 ;
    if((ret=H5LTread_dataset(anc_file_id, datasetName, dsTypeId, (int16*)imagePkg_p->tileref.r))<0)
    {
        xLogMsg("@F l4c_readSmapAncillary:: FAILED to read dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);   
    } else {
        xLogMsg("@A l4c_readSmapAncillary:: SUCCESS ON read of dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);   
    }

    // Read the TILE COLUMN indices from tile_c_M01 
    // <JMG> 2014-10-15T11:20:00-07:00, gridding global "smudge" error,
    //  BUG FOUND and FIXED: datasetName was incorrectly assigned
    //  from "dsNameTileRowM09" and SHOULD have been assigned  using dsNameTileColM09 here. Likely
    //  another cut-and-paste error; found (finally) by studying log file.
    //  PRIOR, INCORRECT --> strcpy(datasetName,(char *)sesCtl_p->dsNameTileRowM09);
    strcpy(datasetName,(char *)sesCtl_p->dsNameTileColM09);
    dsTypeId =  H5T_NATIVE_INT16 ;
    // <JMG> 2013-11-11, BUG FOUND and FIXED: tileref.c was previously coded as tileref.r  and should have
    // been tileref.c, due to cut-paste error. Corrected 11/11/2013.
    if((ret=H5LTread_dataset(anc_file_id, datasetName, dsTypeId, (int16*)imagePkg_p->tileref.c))<0)
    {
        xLogMsg("@F l4c_readSmapAncillary:: FAILED to read dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);
    } else {
        xLogMsg("@A l4c_readSmapAncillary:: SUCCESS ON read of dataset %s from file (%s)~\n",datasetName,sesCtl_p->inputAncPath);   
    }

 return 0 ;
} // end:: l4c_readSmapAncillary()


// ---------------------------------------------------------------
// function: l4c_writeHdf5DatasetAttributes()
// inputs  : hid_t out_file_id,
//           hid_t dataset_id,
//           hid_t dsType_id,
//           char *dsName_p
// purpose : this CREATES and WRITES a standard set of dataset attributes
//           out to the named HDF5 dataset in the HDF5 output file
// author  : joe glassy
// revised : 2021-07-15, deprecated gppMethodCode,gppMethodStr usage
// revised : 2014-10-14T12:05:00-07:00, hard wired dataset attributes in
//           l4c_MODIS.ex since RIP CSV style definitions need to wait until
//           after ORT/ORR.
// history : 2021-07-15, changed to use fparMethodStr , deprecated gpp
// history : 2014-08-28T10:31, changed 'coordinates' value to use space
//           delimiters not comma delim, as per CF v1.6+ spec doc.
// history : 2014-08-20T10:25:00, re-check output of CF attributes
// notes   : TBD migrate to use  frm_hdf5_WriteAttributeAsType() to allow
//           numeric dataset attribs to be written as numerics not strings
// notes
// notes   : NDVI attributes are:
//                 1km 16 days NDVI:long_name = "1km 16 days NDVI" ;
//                 1km 16 days NDVI:units = "NDVI" ;
//                 1km 16 days NDVI:valid_range = -2000s, 10000s ;
//                 1km 16 days NDVI:_FillValue = -3000s ;
//                 1km 16 days NDVI:scale_factor = 10000. ;
//                 1km 16 days NDVI:scale_factor_err = 0. ;
//                 1km 16 days NDVI:add_offset = 0. ;
//                 1km 16 days NDVI:add_offset_err = 0. ;
//                 1km 16 days NDVI:calibrated_nt = 5 ;
// 
int l4c_writeHdf5DatasetAttributes(hid_t out_file_id,hid_t dataset_id,hid_t dsType_id,char *dsName_p)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 ImgPkg *imgPkg_p = l4c_GetImagePackage();
 PkgMgr *pkgMgr_p = l4c_GetPkgMgrOverseer();
 NsbObj *nsbMgr_p = l4c_GetNsbOverseer();
 
 hid_t attr_id = -1 ;
 size_t ithPix =0 ;
 float32 ndviSci =0.0 ;
 float32 analyticalSciValue = 0.0 ;
 char fxName[256] = {0} ;
 char attrName[128] = {0} ;
 char attrValu[FILENAME_MAX];
 
 char absFrqAttribName[128] = {0};
 char absFrqBufr[256] = {0} ;

    strcpy(fxName,__func__);
    
    xLogMsg("@A In(%s) posting dataset-attributes file_id(%d) dataset(%s)~\n",
            fxName, out_file_id,dsName_p);
    
    // WRITE DATASET ATTRIBUTES here
    // placeholders... should write to actual datasets object

    attr_id = l4c_createHdf5Attribute(dataset_id, "shortname"    , dsName_p);
    attr_id = l4c_createHdf5Attribute(dataset_id, "longname"     , dsName_p);
    // CF compliance likes this form as well
    attr_id = l4c_createHdf5Attribute(dataset_id, "standard_name", dsName_p);

    // write CF dataset level attributes
    if(strcmp(dsName_p,"1km 16 days NDVI")==0 || strcmp(dsName_p,"ndvi_M01")==0 )
    {
        attr_id = l4c_createHdf5Attribute(dataset_id, "missing_value", "254");
        attr_id = l4c_createHdf5Attribute(dataset_id, "fillvalue"    , "255");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_min"    , "-0.20");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_max"    , "1.0");
        attr_id = l4c_createHdf5Attribute(dataset_id, "units"        , "ndvi");
        attr_id = l4c_createHdf5Attribute(dataset_id, "coordinates"     , "cell_lat cell_lon");
    } else if(strcmp(dsName_p,"1km 16 days pixel reliability")==0 || strcmp(dsName_p,"ndvi_qc_M01")==0) {
        attr_id = l4c_createHdf5Attribute(dataset_id, "missing_value", "-1");
        attr_id = l4c_createHdf5Attribute(dataset_id, "fillvalue"    , "-1");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_min"    , "0");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_max"    , "3");
        attr_id = l4c_createHdf5Attribute(dataset_id, "units"        , "QC -- dimensionless");
        attr_id = l4c_createHdf5Attribute(dataset_id, "coordinates"     , "cell_lat cell_lon");
    } else if(strcmp(dsName_p,"fpar_M01")== 0) {
        attr_id = l4c_createHdf5Attribute(dataset_id, "missing_value", "254");
        attr_id = l4c_createHdf5Attribute(dataset_id, "fillvalue"    , "255");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_min"    , "0");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_max"    , "100");      
        attr_id = l4c_createHdf5Attribute(dataset_id, "units"        , "percent 0..100");       
        attr_id = l4c_createHdf5Attribute(dataset_id, "coordinates"     , "cell_lat cell_lon");
    } else if(strcmp(dsName_p,"fpar_qc_M01") == 0) {
        attr_id = l4c_createHdf5Attribute(dataset_id, "missing_value", "254");
        attr_id = l4c_createHdf5Attribute(dataset_id, "fillvalue"    , "255");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_min"    , "0");
        attr_id = l4c_createHdf5Attribute(dataset_id, "valid_max"    , "3");
        attr_id = l4c_createHdf5Attribute(dataset_id, "units"        , "QC -- dimensionless");
        attr_id = l4c_createHdf5Attribute(dataset_id, "coordinates"     , "cell_lat cell_lon");     
        
    } else if(strcmp(dsName_p,"cell_lat")==0 || strcmp(dsName_p,"cell_lon")==0 ) {
        // <JMG> 2014-08-27T17:47, TBD Fix Pending: change datatype of valid_min,max to
        // a real number, same for fillvalue,missing_value, so that Panoply can use these!
        
        // CF convention disallows missing value and fillvalue here
        // attr_id = l4c_createHdf5Attribute(dataset_id, "fillvalue"    , "-9999.0");

        if(strcmp(dsName_p,"cell_lat")==0)
        {
            attr_id = l4c_createHdf5Attribute(dataset_id, "valid_min"    , "-89.9");
            attr_id = l4c_createHdf5Attribute(dataset_id, "valid_max"    , "+89.9");
            attr_id = l4c_createHdf5Attribute(dataset_id, "units"        , "degrees_north");
            attr_id = l4c_createHdf5Attribute(dataset_id, "longname"     , "latitude");
        }
        if(strcmp(dsName_p,"cell_lon")==0)
        {
            attr_id = l4c_createHdf5Attribute(dataset_id, "valid_min"    , "-179,9");
            attr_id = l4c_createHdf5Attribute(dataset_id, "valid_max"    , "+179.9");
            attr_id = l4c_createHdf5Attribute(dataset_id, "units"        , "degrees_east");
            attr_id = l4c_createHdf5Attribute(dataset_id, "longname"     , "longitude");
        }
        
    }
    
    // write out observed absolute frequencies, but only if FPAR dataset
    if(strcmp(dsName_p,"fpar_M01")==0 || strcmp(dsName_p,"ndvi_M01")==0)
    {
        for(ithPix=0; ithPix < 256 ; ithPix++)
        {
            analyticalSciValue = (float32)ithPix ;
            
            // For labeling only, convert back from digital to
            //   analytical value.The -100 offset reduction scales from
            //   {0..200} --> {-100...+100} then
            //   the /100.0 term scales from {-100...+100} --> {-1.0...+1.0}
            analyticalSciValue = analyticalSciValue * 0.01;
            
            // report only the non-zero positive frequency counts
            if( imgPkg_p->absFreqs[ithPix] > 0)
            {
                // xLogMsg("@A In(%s) %s ~ freqbin(%3d) value(%f) frequency %d\n",
                //  fxName, sesCtl_p->fparMethodStr, ithPix, ndviSci, imgPkg_p->absFreqs[ithPix] );
                sprintf(absFrqAttribName,"%s_FREQ_BIN_%-3.3zu",sesCtl_p->fparMethodStr,ithPix);
                sprintf(absFrqBufr,"%s value(%f) frequency %d",
                            sesCtl_p->fparMethodStr, analyticalSciValue, imgPkg_p->absFreqs[ithPix]);
                // WRITE the custom dataset attribute
                attr_id = l4c_createHdf5Attribute(dataset_id,absFrqAttribName, absFrqBufr);
                
            } // end:: if pixel qualifies
        } // end:: ithPix loop

    } // end:: if FPAR or NDVI output dataset

 return 0 ; 
 
} // end:: l4c_writehdf5DatasetAttributes()



// ---------------------------------------------------------------
// function: l4c_writeHdf5Dataset()
// purpose : this CREATES and WRITES one named HDF5 dataset to HDF5 file
// author  : joe glassy
// revised : 2014-08-19T13:19, re-check output of CF attributes, still 
//           need to supply valid_min,valid_max values though
// revised : 2013-11-06T17:04:00-07:00
// notes   : fpar_M01, fpar_qc_M01 
//  ret = H5Dwrite(dataset_id, dsType_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, dsData_p);
// notes   : NDVI attributes are:
//                 1km 16 days NDVI:long_name = "1km 16 days NDVI" ;
//                 1km 16 days NDVI:units = "NDVI" ;
//                 1km 16 days NDVI:valid_range = -2000s, 10000s ;
//                 1km 16 days NDVI:_FillValue = -3000s ;
//                 1km 16 days NDVI:scale_factor = 10000. ;
//                 1km 16 days NDVI:scale_factor_err = 0. ;
//                 1km 16 days NDVI:add_offset = 0. ;
//                 1km 16 days NDVI:add_offset_err = 0. ;
//                 1km 16 days NDVI:calibrated_nt = 5 ;
// 
int l4c_writeHdf5Dataset(const hid_t out_file_id,char *dsName,
                    const hid_t dsType_id,const hid_t dcpl,const hid_t dspace_id,void *dsData_p)
{
 herr_t ret = -1;
 hid_t  dataset_id = -1;
 hid_t  attr_id = -1;
 
 char attrName[256]  = {0} ;
 char attrValu[1024] = {0} ;
 char fxName[256] = {0};
 
    strcpy(fxName,(char *)__func__);   
 
    dataset_id = H5Dcreate(out_file_id, dsName, dsType_id, dspace_id, H5P_DEFAULT, dcpl, H5P_DEFAULT);
    
    // if used, insert the dcpl in the right places here!
    ret = H5Dwrite(dataset_id, dsType_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, dsData_p); 

    // based on which dsName is current, branch on name and 
    // write out a set of standard CF style dataset attributes such as
    // coordinates, valid_min,valid_max, valid_range,units, longname, standard_name etc
    l4c_writeHdf5DatasetAttributes(out_file_id, dataset_id, dsType_id,dsName);

    H5Dclose(dataset_id);
    
  return ret ;
  
} // end::l4c_writeHdf5Dataset()


// ----------------------------------------------------------------
// function : l4c_readAndWriteLatLon_1kmCoordVars()
// purpose  : open,read lat,lon coordinate vars for EASEGRID 1km global grid
// inputs   : ImgPkg *imagePkg_p -- ptr to container holding lat,lon images
// returns  : int, status from frm_hdf5_WriteData()
// history  : 2014-07-30T10:32:00
// notes    : This is written in as stand-alone a way as possible,
//            although still needs a better way to transmit attribute
//            properties (fillvalue,missing_value,valid_min,valid_max)
//            etc.
int l4c_readAndWriteLatLon_1kmCoordVars(void)
{
 SesCtl *sesCtl_p   = l4c_GetSesCtl();  
 ImgPkg *imagePkg_p = l4c_GetImagePackage();
 char *hdf5Path     = sesCtl_p->inputAncPath ;
 char fxName[256]   = {0};
 char dsName[16]    = {0};
 
 size_t nElem9km    = imagePkg_p->nElem9km ;
 size_t nElem1km    = imagePkg_p->nElem1km ;
 
 int32 status = -1 ;
 herr_t  ret  = -1 ;
 hid_t dsTypeId = H5T_NATIVE_FLOAT ;

 int      rank= 2 ;
 // inherite users RIP setting for compression level 1..9
 int      compressLevel = sesCtl_p->compressLev ;
 
 hid_t    dcpl = -1 ;
 hid_t    dspace1km = -1;
 hsize_t lldims[2] = {0} ;
 hsize_t  chunk_size[2] = {0};
 
 float32  fltFill = -9999.0 ;
 
    strcpy(fxName,__func__);

    lldims[0] = NROW1KM ;
    lldims[1] = NCOL1KM ;
    
    
    // Create 1km image buffers for lat,lon
    imagePkg_p->latImg_M01_p = (float32*)memCalloc(nElem1km, sizeof(float32));
    imagePkg_p->lonImg_M01_p = (float32*)memCalloc(nElem1km, sizeof(float32));
    
    // READ LAT, from SMAP_L4_C_Ancillary.h5, READ float32 lat 1km bin-center image
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId,
      sesCtl_p->dsNameLatCoordvar, (float32 *)imagePkg_p->latImg_M01_p,fxName);

    xLogMsg("@A In(%s): READ latitude  1km dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, sesCtl_p->dsNameLatCoordvar,
                hdf5Path, sesCtl_p->anc_file_id, status);

    // READ LON, from SMAP_L4_C_Ancillary.h5, READ float32 lon 1km bin-center image
    status = frm_hdf5_ReadData(sesCtl_p->anc_file_id, dsTypeId,
      sesCtl_p->dsNameLonCoordvar, (float32 *)imagePkg_p->lonImg_M01_p,fxName);

    xLogMsg("@A In(%s): READ longitude 1km dataset (%s) from (%s) file_id %d Status %d~ \n",
                fxName, sesCtl_p->dsNameLonCoordvar,
                hdf5Path, sesCtl_p->anc_file_id, status);

    // CREATE: local 'dcpl' apparatus for chunking, compression, fillvalue for lat,lon
    dcpl       = H5Pcreate(H5P_DATASET_CREATE);
    dspace1km  = H5Screate_simple(rank, lldims, lldims);
    chunk_size[0] = (lldims[0] / 10);
    chunk_size[1] = (lldims[1] / 10);
    H5Pset_chunk (dcpl, rank, chunk_size);
    H5Pset_deflate(dcpl, sesCtl_p->compressLev);
    H5Pset_fill_value(dcpl, H5T_NATIVE_FLOAT, &fltFill);    
    
    // WRITE out the latitude, longitude coordinates 
    strcpy(dsName,"cell_lat");
    ret = l4c_writeHdf5Dataset(sesCtl_p->fparOutput_file_id, dsName, dsTypeId,dcpl,dspace1km,(float32 *)imagePkg_p->latImg_M01_p);  
    
    strcpy(dsName,"cell_lon");
    ret = l4c_writeHdf5Dataset(sesCtl_p->fparOutput_file_id, dsName, dsTypeId,dcpl,dspace1km,(float32 *)imagePkg_p->lonImg_M01_p);      
    
    // <JMG> TBD 2014-08-27Y17:51, add other CF mandatory dataset attributes required
    // of 'cell_lat' and 'cell_lon' so that Panoply can correctly use these!
    // see: http://cfconventions.org/Data/cf-documents/requirements-recommendations/requirements-recommendations-1.6.html
    // see: http://www.unidata.ucar.edu/software/netcdf/conventions.html
    // float lat(lat) ;
    // lat:long_name = "latitude" ;
    // lat:units = "degrees_north" ;
    // lat:standard_name = "latitude" ;
    // lon:units = "degrees_east" 
    
    xLogMsg("@A In(%s): WROTE lat and lon 1km datasets to new output file_id %d~ \n",
                fxName, sesCtl_p->fparOutput_file_id);
    
    ret = H5Pclose(dcpl);
    ret = H5Sclose(dspace1km);
        
    H5Fflush(sesCtl_p->fparOutput_file_id, H5F_SCOPE_LOCAL); // flush...
    
    // Free lat,lon bin center coordinate variable images
    memFree(imagePkg_p->latImg_M01_p);
    memFree(imagePkg_p->lonImg_M01_p);
    
  // return status
  return status;
  
} // end::readAndWriteLatLon_1kmCoordVars()

#ifdef IS_DEPRECATED
// ---------------------------------------------------------------
// function: l4c_NDVI_RescaleConversion()
// purpose : Converts the NDVI global 1km EASEGRID int16 native
//           image (field:___) to a final uint8 format by applying
//           the (gain,offset) required, rescaling the original int16
//           data from valid(min=x,max=x) to represent the analytical NDVI
//           (-1.0 <=ndvi<=+1.0) to rescaled range (0..200) suitable for
//           storing as a uint8 (0..255) image.
// inputs  : void
// author  : joe glassy
// revised : 2014-08-21T10:48, refactored from l4c_writeSmapOutputs()
// notes   : fpar_M01, when fpar_qc_M01 written if gppMethodCode==1, or 
//           ndvi_M01, when ndvi_qc_M01 written if gppMethodCode==0
// notes   : When gppMethodCode==1, fpar_M01 and fpar_qc_M01 are written to HDF5
//           When gppMethodCode==0, ndvi_M01 and ndvi_qc_M01 are written to HDF5
//
uint8 *l4c_NDVI_RescaleConversion(void)
{
 SesCtl *sesCtl_p = l4c_GetSesCtl();
 ImgPkg *imgPkg_p = l4c_GetImagePackage();
 // fetch pointer to primary EASEGRID image package
 ease_struct *ease_p = &imgPkg_p->ease ; // set convenience pointers
 
 char   fxName[256] = {0} ;
 uint8   *outputNdviImg_p = NULL;
 uint8   ndviPix = 0 ;
 int32   ndviAbsFrq[255] = {0}; // unit bin abs frequency counts
 size_t  nInvalid = 0 ; // count of invalid pixels (e.g. <= -3000 etc)
 size_t  nValid   = 0 ;
 size_t  nValidBins=0 ;
 size_t  nValidUpdates=0;
 size_t  ithPix = 0 ;
 float32 ndviSci  = 0.0 ;
 float32 ndviGain = 10000.0; // gain for orig NDVI 16 day
 float32 ndviOffset=0.0;     // offset... ignore
 
 float32 ndvi_valid_min = -2000.0 ; // valid min for int16 range
 float32 ndvi_valid_max = 10000.0 ; // valid max for int16 range
 
    strcpy(fxName, __func__);
        
    xLogMsg("@D In(%s) Allocating NDVI uint8 image of %zu pixels~\n",
                    fxName, imgPkg_p->nElem1km);

    // Step 1: allocate resulting uint8 global 1km image
    outputNdviImg_p = (uint8 *)memCalloc(sizeof(uint8),imgPkg_p->nElem1km);
    // assign descriptive name
    memObjSetName(outputNdviImg_p,"final_Ndvi_EGv2_1km_Image_uint8");

    xLogMsg("@D In(%s) Rescaling int16 NDVI to uint8 image~\n",fxName);
        
    // iterate through 507 odd pixels. This loop is slow...  :)
    for(ithPix =0; ithPix < imgPkg_p->nElem1km; ithPix++)
    {
        outputNdviImg_p[ithPix] = BYTE_FILL_VALUE ;
        // Capture original int16 digital value -3000 to 10000
        ndviSci = (float32)ease_p->ndvi[ithPix] ;
        // filter on valid pixels
        if(ndviSci >= ndvi_valid_min && ndviSci <= ndvi_valid_max)
        {
            // rescale min from -2000 to -0.20, max from 10000 to 1.0
            // next bump (-0.20....1.0) *100 --> (-20...100)
            // finally, add offset +100, scaling (-20...100) --> (80..200)
            ndviSci = ((ndviSci / ndviGain) * 100.0) + 100.0 ;
            // verify if this round step is necessary...
            ndviPix = (uint8)FRM_API_ROUND(ndviSci) ;
            // track statistics to make sure we've scaled correctly
            // accumulate absolute frequencies scaled into 0..255 range
            if( ndviSci > 0.0 && ndviSci < 256.0)
            {
                imgPkg_p->absFreqs[ndviPix]++ ; // increment frequency count
                nValidUpdates++ ;       // increment count of valid updates
                // LOAD the final rescaled NDVI into final output image
                outputNdviImg_p[ithPix] = ndviPix ;
            } // end:: if block, where valid analytical science NDVI...

            nValid++ ;
        } else {
            nInvalid++ ; // count invalid pixels
        }
        // update consule every 10 million pixels to show progress..(ca 50 times total)
        if( ithPix % 10000000 == 0)
        {
            fprintf(stdout,".");
            // fprintf(stdout,"@A PIX %9zu nValid %zu nValidUpdates %zu nInvalid %zu\n",ithPix, nValid, nValidUpdates, nInvalid);
            fflush(stdout);
        }
    } // end:: ithPix loop assigning fill value to output image

#ifdef DIAG_ONLY    
    xLogMsg("@D In(%s) Summarizing NDVI frequencies ~\n",fxName);
    for(ithPix=0; ithPix < 256 ; ithPix++)
    {
        ndviSci = (float32)ithPix ;
        // convert back from digital to analytical value.The
        // -100 offset reduction scales from {0..200} --> {-100...+100} then
        // the /100.0 term scales from {-100...+100} --> {-1.0...+1.0}
        
        // report only the non-zero positive frequency counts
        if( imgPkg_p->absFreqs[ithPix] > 0)
        {
            ndviSci = (ndviSci - 100.0) / 100.0 ;   
            xLogMsg("@A In(%s) NDVI~ bin(%3d) value(%f) frequency %d\n",
                    fxName, ithPix, ndviSci, imgPkg_p->absFreqs[ithPix] );
            nValidBins++ ;
        }
    }
#endif
            
    xLogMsg("@A In(%s) DIAG n. valid bins(%zu) n. bin updates (%zu) qualifying pixels (%zu) n. non-qualifying pixels (%zu) ~\n",
                    fxName, nValidBins, nValidUpdates, nValid, nInvalid);


 // return pointer to populated final NDVI global 1km 2D image in uint8 form
 return outputNdviImg_p ;
    
} // end:: l4c_NDVI_RescaleConversion()
// DEPRECATED ALL OF FUNCTION
#endif    

// ---------------------------------------------------------------
// function: l4c_writeSmapOutputs()
// purpose : opens new output HDF5 file, writes global moasic 
//           of FPAR or NDVI outputs to the daily HDF5 file
// inputs  : void
// author  : joe glassy
// revised : 2021-07-15T17:15, changed gppMethodCode to fparMethodCode and Str
// revised : 2014-08-19T12:40:00, distinguish on fparMethodCode now for
//           FPAR vs. VIIRS (verify TBD)
// revised : 2013-11-06T17:04:00-07:00
// notes   : fpar_M01, fpar_qc_M01 written if gppMethodCode==1, or 
//           ndvi_M01, ndvi_qc_M01 written if gppMethodCode==0

//  ret = H5Dwrite(dataset_id, dsType_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, dsData_p);
int l4c_writeSmapOutputs(void)
{
    SesCtl *sesCtl_p       = l4c_GetSesCtl() ;
    ImgPkg *imagePkg_p     = l4c_GetImagePackage();
    ease_struct *ease_p = &imagePkg_p->ease ; // set convenience pointers

    herr_t ret = -1 ;
    hid_t  out_file_id = -1;
    
    hid_t  dsType_id = -1;
    hid_t  plist1km ;
    hid_t  dspace1km ;
    
    // hid_t  plist9km ;
    hid_t  dcpl = -1 ;
    int32  rank= 2 ;
    int32  ithPix =0 ;
    hsize_t chunk_size[2] ;
    hsize_t  dims[2] = {0} ;;
    int32 compressLevel = sesCtl_p->compressLev ;
    
    int16 shortFill = -3000 ;
    int8  int8Fill  = -1 ;
    
    uint8   byteFill= FPAR_FILL_VALUE ;
    uint8   *outNdviUint8_p = NULL; // ptr to NDVI output uint8 image
    
    void *dsData_p = NULL;
    
    char dsName[256] = {0} ; // Fpar_1km or Ndvi_1km 
    char dsNameQc[256] = {0} ;// Fpar_1
    char fxName[256];
    
    strcpy(fxName,(char *)__func__);   
        
    dsType_id  = H5T_NATIVE_UINT8 ; // all (3) output vars share this datatype...
    
    dcpl = plist1km  = H5Pcreate(H5P_DATASET_CREATE);

    dims[0] = NROW1KM; // was: EGv2_GLOBAL_1KM_NS_PIXELS 
    dims[1] = NCOL1KM; // was: EGv2_GLOBAL_1KM_EW_PIXELS 
    
    // TBD: investigate if a more efficient chunking geometry could be found!
    chunk_size[0] = (dims[0] / 10);
    chunk_size[1] = (dims[1] / 10);
    
    // Recap: output opened prior in InitSession
    out_file_id = sesCtl_p->fparOutput_file_id ;
    
    // ---------------------------------------------------------------
    xLogMsg("@A In(%s) writing (new) outputs to HDF5 (%s)\n",fxName,sesCtl_p->outputFparPath);

    dspace1km  = H5Screate_simple(rank, dims, dims);
    
    xLogMsg("@A (%s) Generating (%s) OUTPUT to file (%s) with new id %d dcpl %d dspace %d~\n",
            fxName, sesCtl_p->fparMethodStr, sesCtl_p->outputFparPath, out_file_id,dcpl, dspace1km);

    H5Pset_chunk (dcpl, rank, chunk_size);
    H5Pset_deflate(dcpl, compressLevel);
    H5Pset_fill_value(dcpl, H5T_NATIVE_UINT8, &byteFill);

    // <JMG> 2014-08-19T12:07, adjust dataset name dsName to reflect whether this
    // is a FPAR or an NDVI output!!
    
    // <JMG> 2021-07-14 DEPRECATED branch for IsNdviSrcForGpp
        strcpy(dsName,   sesCtl_p->dsOutput_Fpar);
        strcpy(dsNameQc, sesCtl_p->dsOutput_FparQc);
        dsType_id= H5T_NATIVE_UINT8 ;
        H5Pset_fill_value(dcpl, H5T_NATIVE_UINT8, &byteFill);
        
        // WRITE fpar_M01 output field using source array ease_p->fpar
        ret = l4c_writeHdf5Dataset(out_file_id,dsName,dsType_id,dcpl,dspace1km,ease_p->fpar );
        xLogMsg("@A In(%s) created FPAR dataset (%s) ret %d~\n",fxName, dsName,ret);

        // WRITE fpar_qc_M01 using source array 'ease_p->qc'    
        dsType_id = H5T_NATIVE_UINT8 ;
        H5Pset_fill_value(dcpl, H5T_NATIVE_UINT8, &byteFill);
        ret = l4c_writeHdf5Dataset(out_file_id,dsNameQc,dsType_id,dcpl,dspace1km,ease_p->qc);
        xLogMsg("@A In(%s) created FPAR QC dataset (%s) ret %d~\n",fxName, dsNameQc,ret);
    
    // DIAG_ALGOR: accumulate frequencies
#ifdef DIAG_ALGOR
    l4c_calcFrequencies(dsName,ease_p->fpar,
            EGv2_GLOBAL_1KM_PIXELS, l4c_GetLogfileHandle());
#endif  
    
    
    // DEPRECATED fpar_qc_extra_M01, as it has NOT been used in l4c_DAILY
    // WRITE fpar_qc_extra_M01 output field
    // strcpy(dsName,"fpar_qc_extra_M01");
    // dsData_p = ease_p->qc_extra;
    // ret = l4c_writeHdf5Dataset(out_file_id,dsName,dsType_id,dcpl,dspace1km,dsData_p);
    // xLogMsg("@A In(%s) created FPAR dataset (%s) ret %d~\n",fxName, dsName,ret); 

    // Write all other misc (unofficial metadata) attributes to FPAR 8_DAY output HDF5 file
    // PRIOR, deprecated: l4c_writeHdf5AttributesToOutput(out_file_id);
    // NEW, updated to produce more required ISO attributes
    l4c_Hdf5_AddFileMetadata();

    // close everything in the right order
    ret = H5Pclose(dcpl);
    xLogMsg("@A l4c_writeSmapOutputs: closed dcpl %d with ret %d~\n",dcpl,ret);
    // close shared 1km dataspace 
    ret = H5Sclose(dspace1km);
    xLogMsg("@A l4c_writeSmapOutputs: closed dspace1km %d with ret %d~\n",dspace1km,ret);
    
    // flush now, but close later in CloseSession
    H5Fflush(sesCtl_p->fparOutput_file_id,H5F_SCOPE_LOCAL);

 return (int)out_file_id;
 
} // end:: l4c_writeSmapOutputs()
