// ---------------------------------------------------------------
// module       : L4C_MODIS_rip.c
// purpose      : implement parsing of all RIP (runtime input parameter) 
//                and related runtime input management functions.
// project      : SMAP L4C science development.
// dependencies : HDF4, hdfeos, HDF5 libraries.
// revision     : v7.3.0.1-2021-08-09T12:11, use SesCtl "instrumentId" to support
//                conditional branching between VIIRS and MODIS.
// revision     : v7.x-2021-07-14T11:50, Eval/Diag changes to prep for VIIRs
// revision     : v7.1.0-2017-01-26T12:22, add support for MODIS Collection 6 MOD15A2H 500m inputs
//                introduced in header fparResolutionMeters, and introduced RIP 
//                RIP keyword "L4C_FPAR_RESOLUTION_METERS" as {500|1000}
// revision     : 2016-01-04, introduced macro L4CMOD_OPTIMAL_TILECOUNT = 286 to make
//                explicit. Can run with fewer tiles, triggering FPAR CLIMATOLOGY fill in
// revision     : 2014-10-14T15:08:00, for ORT#2,ORR updates
// revision     : 2014-08-15T13:02:00 for v5.0.1
// revision     : 2014-3-11-18T09:36:00, corrected bug where OUTPUT root dir not prepended to output path
// history      : 2013-11-18T09:34:00, corrected bug where OUTPUT root dir not prepended to output path
//              : 2013-11-08T08:32:00, initial version delivered to NCCS
// author       : jglassy
// notes        : Reminder: session log facility is now in L4CMOD_log.c, AS xLogMsg()
//
// ---------------------------------------------------------------

#include "l4c_MODIS.h"
#include "l4c_MODIS_prot.h"


// <JMG> 2014-10-14, adopted from L4CMDL, for future use
// 46 8-day periods in one year
// FIELDS: YEARDAY,IDX_8DAYPERIOD,IDX_16DAYPERIOD,FparClimDatasetName
static char *fparClimatologyNameLut[46] = {
"001,01,01,fpar_clim_M01_day001_per01",
"009,02,01,fpar_clim_M01_day009_per02",
"017,03,02,fpar_clim_M01_day017_per03",
"025,04,02,fpar_clim_M01_day025_per04",
"033,05,03,fpar_clim_M01_day033_per05",
"041,06,03,fpar_clim_M01_day041_per06",
"049,07,04,fpar_clim_M01_day049_per07",
"057,08,04,fpar_clim_M01_day057_per08",
"065,09,05,fpar_clim_M01_day065_per09",
"073,10,05,fpar_clim_M01_day073_per10",
"081,11,06,fpar_clim_M01_day081_per11",
"089,12,06,fpar_clim_M01_day089_per12",
"097,13,07,fpar_clim_M01_day097_per13",
"105,14,07,fpar_clim_M01_day105_per14",
"113,15,08,fpar_clim_M01_day113_per15",
"121,16,08,fpar_clim_M01_day121_per16",
"129,17,09,fpar_clim_M01_day129_per17",
"137,18,09,fpar_clim_M01_day137_per18",
"145,19,10,fpar_clim_M01_day145_per19",
"153,20,10,fpar_clim_M01_day153_per20",
"161,21,11,fpar_clim_M01_day161_per21",
"169,22,11,fpar_clim_M01_day169_per22",
"177,23,12,fpar_clim_M01_day177_per23",
"185,24,12,fpar_clim_M01_day185_per24",
"193,25,13,fpar_clim_M01_day193_per25",
"201,26,13,fpar_clim_M01_day201_per26",
"209,27,14,fpar_clim_M01_day209_per27",
"217,28,14,fpar_clim_M01_day217_per28",
"225,29,15,fpar_clim_M01_day225_per29",
"233,30,15,fpar_clim_M01_day233_per30",
"241,31,16,fpar_clim_M01_day241_per31",
"249,32,16,fpar_clim_M01_day249_per32",
"257,33,17,fpar_clim_M01_day257_per33",
"265,34,17,fpar_clim_M01_day265_per34",
"273,35,18,fpar_clim_M01_day273_per35",
"281,36,19,fpar_clim_M01_day281_per36",
"289,37,20,fpar_clim_M01_day289_per37",
"297,38,21,fpar_clim_M01_day297_per38",
"305,39,22,fpar_clim_M01_day305_per39",
"313,40,23,fpar_clim_M01_day313_per40",
"321,41,24,fpar_clim_M01_day321_per41",
"329,42,25,fpar_clim_M01_day329_per42",
"337,43,26,fpar_clim_M01_day337_per43",
"345,44,27,fpar_clim_M01_day345_per44",
"353,45,28,fpar_clim_M01_day353_per45",
"361,46,29,fpar_clim_M01_day361_per46"
} ;


// ---------------------------------------------------------------------
// function :l4c_ApplyRootDirToPaths()
// Purpose  : Apply standard root-directory prepend rules, and apply 
//            any optional file-specific or generic ROOTDIR to any
//            associated BASE PATHS following these rules. All pre-qualifies
//            existence of any input files.
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls
// returns  : void
// notes    : This must be called AFTER l4c_SesCollectNvps() has parsed
//            the RIP file and stored all releavant paths in the 
//            appropriate sesCtl (sesCtl_p) members.
// notes    : called by: l4c_SesCollectNvps() 
//
void l4c_ApplyRootDirToPaths(SesCtl *sesCtl_p)
{
 char fxName[256] = {0};
 
 int ithGrn = 0 ;
 int nGranuleCount = 0 ;
  
 strcpy(fxName,__func__);

    // Step 1a: Do one-time-only path qualifications.
    // note: re-verified, this works correctly for Ancillary
    frm_ioQualifyPathWithRoot(sesCtl_p->ancRootDir , sesCtl_p->inputAncPath);
   
    // Step 1b: Verify that input files exist at the indicated path locations 
    if( !frm_ioDoesFileExist(sesCtl_p->inputAncPath))
        xLogMsg("@F In(%s) input file(%s) FAILS TO EXIST at named location~ \n",
            fxName, sesCtl_p->inputAncPath);
    else
        xLogMsg("@A In(%s) Ancillary file(%s) correctly identified at named location~ \n",
            fxName, sesCtl_p->inputAncPath);
    
    xLogMsg("@A In(%s) %d FPAR granules identified for use\n",
        fxName, sesCtl_p->nFparGranules);

    // re-verified, this works correctly for scalar instances
    frm_ioQualifyPathWithRoot(sesCtl_p->outputFparRootDir, sesCtl_p->outputFparPath);

    xLogMsg("@A In(%s) FPAR Output ROOTDIR (%s) to qualify [%s]~ \n",fxName,
            sesCtl_p->outputFparRootDir, sesCtl_p->outputFparPath);
    
    xLogMsg("@A In(%s) FPAR Input ROOTDIR (%s) to qualify ~\n",fxName,
            sesCtl_p->inputFparRootDir);
          
    // Step 2: Perform path pre-qualifications on list-types: multiple-granule file types
    // note that the FPAR list has already been sorted, string-table expanded,
    // so we are ready now for pre-pend of root dir
    if(sesCtl_p->nFparGranules >= 2)
    {
        
        xLogMsg("@A In(%s) FPAR Granule FULL LIST of (%d)~\n",fxName,
                sesCtl_p->nFparGranules);
        
        for(ithGrn=0; ithGrn < sesCtl_p->nFparGranules;  ithGrn++)
        {
            frm_ioQualifyPathWithRoot(sesCtl_p->inputFparRootDir,
                    sesCtl_p->input_Fpar_pathList[ithGrn] );
                    
            if( !frm_ioDoesFileExist(sesCtl_p->input_Fpar_pathList[ithGrn]))
            {
                xLogMsg("@F In(%s) INPUT GRANULE FAILS-TO-EXIST at location (%s)~ \n",
                    fxName,sesCtl_p->input_Fpar_pathList[ithGrn] );
            }

        } // end:: ithGrn for FPAR loop
    } // end:: if block when at least (2) FPAR granules were present

    frm_utilEventCompletionMsg( __func__, sesCtl_p->sesStrtSec, 
            sesCtl_p->enableCompletionMsg);
    
} // end::l4c_ApplyRootDirToPaths()


// ---------------------------------------------------------------------
// function :l4c_SesCollectCoreNvps()
// Purpose  : collect "CORE" standardized NVPs from RIP file, loading
//             into destination vars
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls
//          : NsbObj *nsb_p    -- ptr to NSB overseer
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// revised  : 2014-10-15T13:50, changed in 2021 to L4CMOD_DATASET_VERSION and productionLevel
//            assignment alongside L4CMOD_SW_VERSION etc
// notes    : TBD, wrap in a mutex lock to make thread safe
void  l4c_SesCollectCoreNvps(SesCtl *sesCtl_p)
{
 // inherit NSB overseer
 NsbObj *nsbMgr_p   = l4c_GetNsbOverseer();
 // inherit pkgMgr_p 
 // PkgMgr   *pkgMgr_p = l4c_GetPkgMgrOverseer();
 char    nvpBuf[FILENAME_MAX] = {0};    
 char    fxName[256] = {0} ; 
 

    // register this entry
    strcpy(fxName,(char *)__func__);
    xLogMsg("@A In(%s) organizing all l4c_MODIS runtime input parameters (RIP)~ \n",fxName);
    
    // fetch: Science Version ID, CRID, release tag, software sw version, "action" operator etc
    strcpy(sesCtl_p->ScienceVersID, nsbGetValueByKey(nsbMgr_p,"L4C_SCIENCE_VERSION_ID",nvpBuf));
    
    strcpy(sesCtl_p->CRID ,nsbGetValueByKey(nsbMgr_p,"L4C_CRID",nvpBuf));

    // Note: a "RELEASE_TAG" here refers to the combination of a 
    // tightly-coupled collection of required ANCILLARY DATA 
    // (BPLUT,SMAP_L4_C_Ancillary.h5,etc) making up a release instance
    strcpy(sesCtl_p->SubsysReleaseTag,nsbGetValueByKey(nsbMgr_p,"L4C_SUBSYSTEM_RELEASE",nvpBuf));
    
    // <JMG> 2014-10-14 updpate to match L4CMDL, L4CMET
    strcpy(sesCtl_p->swVersion     ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_SW_VERSION",nvpBuf));
    strcpy(sesCtl_p->datasetVersion,nsbGetValueByKey(nsbMgr_p,"L4CMOD_DATASET_VERSION",nvpBuf));
    
    // populates appAction only for backwards compatibility with TCF SPIN 
    strcpy(sesCtl_p->appAction      ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_RUN_MODE",nvpBuf));
    // alias for now...
    strcpy(sesCtl_p->productionLevel,nsbGetValueByKey(nsbMgr_p,"L4CMOD_RUN_MODE",nvpBuf));

    nsbGetValueByKey(nsbMgr_p,"L4CMOD_MODEL_YEARDAY",nvpBuf);
    strcpy(sesCtl_p->doy_Str, nvpBuf);
    
    sesCtl_p->runDayOfYear = (int)atoi(nvpBuf);
    strcpy(sesCtl_p->year_Str, nvpBuf);
    
    nsbGetValueByKey(nsbMgr_p,"L4CMOD_MODEL_YEAR"   ,nvpBuf);
    sesCtl_p->runModelYear = (int)atoi(nvpBuf);

    strcpy(sesCtl_p->l4c_Task_Label, nsbGetValueByKey(nsbMgr_p,"L4CMOD_TASK_LABEL",nvpBuf));
    strcpy(sesCtl_p->l4c_Task_UUID , nsbGetValueByKey(nsbMgr_p,"L4CMOD_TASK_UUID",nvpBuf));

} // end:: l4c_SesCollectCoreNvps()


// ---------------------------------------------------------------------
// function :l4c_SesCollectNvps()
// Purpose  : collect application specific NVPs from RIP file, 
//            loading each into a destination session variable
// Inputs   : SesCtl *sesCtl_p -- ptr to session controls structure defined
//            in l4c_MODIS.h header.
//          : NsbObj *nsb_p    -- ptr to NSB overseer
// returns  : NsbMgr *, ptr to name-value-pair string-block NSB overseer
// revised  : 20210-7-18T16:13, fix ups to support FparExtraQC parsing modernization
// notes    : TBD, wrap in a mutex lock to make thread safe
// notes: Default VIIRS grid name is: "VNP_Grid_VNP15A2H" 
// notes: Default collection 6 modis grid name is: TBD
//
void  l4c_SesCollectNvps(SesCtl *sesCtl_p)
{
 // inherit NSB overseer
 NsbObj *nsbMgr_p   = l4c_GetNsbOverseer();
 // inherit pkgMgr_p 
 // PkgMgr   *pkgMgr_p = l4c_GetPkgMgrOverseer();
 // ...and pThme_p used for FparClimatology special case handling 
 char    nvpBuf[FILENAME_MAX] = {0};    
 char    fxName[256] = {0} ; 
 int ithGrn = 0 ;
 
 float64 numericBuf = 0.0 ; // size generic numeric scalar to as per API (nsb) requirements
 
    // register this entry
    strcpy(fxName,(char *)__func__);

    if(nsbDoesKeyExist(nsbMgr_p,"L4C_L4CMDL_DOI"))
    {
        strcpy(sesCtl_p->doi_assigned,nsbGetValueByKey(nsbMgr_p,"L4C_L4CMDL_DOI",nvpBuf));
    }
    
    // L4C_RUN_MODE should eventually be changed to L4C_PRODUCTION_LEVEL to match?
    if(nsbDoesKeyExist(nsbMgr_p,"L4CMOD_RUN_MODE"))
    {
        strcpy(sesCtl_p->productionLevel, nsbGetValueByKey(nsbMgr_p,"L4CMOD_RUN_MODE",nvpBuf));
    }

    // <JMG> 2017-01-26T12:33 added support for MODIS Collection 6 MOD15A2H 500m inputs
    sesCtl_p->fparResolutionMeters = 1000 ;  // first, set default at 1KM MODIS
    if(nsbDoesKeyExist(nsbMgr_p,"L4CMOD_FPAR_RESOLUTION_METERS"))
    {
        sesCtl_p->fparResolutionMeters = nsbGetNumericByKey(nsbMgr_p,"L4CMOD_FPAR_RESOLUTION_METERS",&numericBuf);
    } else {
         xLogMsg("@F In(%s) MISSING REQUIRED Input Resolution Meters parameter!",fxName);
    }

    // SET DEFAULTS, as these are ALWAYS needed
    sesCtl_p->modis_Tile_1km_SumPixels = MODIS_1KM_SUM_PIXELS;
    sesCtl_p->modis_ew_1km_Pixels      = MODIS_1KM_EW_PIXELS ;
    sesCtl_p->modis_ns_1km_Pixels      = MODIS_1KM_NS_PIXELS ;
    // SET these regardless now, in case FPAR 500m resolution inputs are used
    sesCtl_p->modis_Tile_500m_SumPixels = MODIS_500M_SUM_PIXELS;
    sesCtl_p->modis_ew_500m_Pixels      = MODIS_500M_EW_PIXELS ;
    sesCtl_p->modis_ns_500m_Pixels      = MODIS_500M_NS_PIXELS ;

    // Intentionally preset sentinel value as -1 forcing assignment to valid {0|1}
    sesCtl_p->fparMethodCode = -1; 
    sesCtl_p->inputCollectionId = 6 ; // default, set as 5 or 6 or whatever is most prevalent

    // Legacy choices: for source of Fpar were: strings, "FPAR"|"NDVI" NOW ARE "MODIS|VIIRS" 
    //  where possibly in future a "CLIM" may be supported to force FparClimatology usage.
    
    // DEPRECATED: gppMethodStr FROM "L4C_GPP_SOURCE" replaced by fparMethodStr
    // strcpy(sesCtl_p->fparMethodStr,nsbGetValueByKey(nsbMgr_p,"L4CMOD_FPAR_SOURCE",nvpBuf));
    
    // <JMG> 7/26/2021, L4CMOD_FPAR_SOURCE revised to L4CMOD_INSTRUMENT so INSTRUMENT supercedes now
    // to assigned to fparMethodStr
    strcpy(sesCtl_p->fparMethodStr,nsbGetValueByKey(nsbMgr_p,"L4CMOD_INSTRUMENT",nvpBuf));
    // <JMG> 2021-07-26, force 'VIIRs' to upper case 'VIIRS' or 'MODIS' etc.
    frm_strToUpper(sesCtl_p->fparMethodStr);
   
    // <JMG> COLLECTION_USED modified,shortened, deprecated now use L4CMOD_COLLECTION
    
    // <JMG> 2021-07-26 revise name as per Arthur E. meeting to juse _COLLECTION
    if(nsbDoesKeyExist(nsbMgr_p,"L4CMOD_COLLECTION"))
    {
        sesCtl_p->inputCollectionId = nsbGetNumericByKey(nsbMgr_p,"L4CMOD_COLLECTION",&numericBuf);
    }
    
    // <JMG> 2021-07-14, refactor for clear MODIS vs. VIIRs support
    if( strcmp(sesCtl_p->fparMethodStr,"MODIS")==0)
    {
        sesCtl_p->fparMethodCode = UseMODISInputs; // pull code from enum
        // <JMG> 2021-08-09, added "instrumentId" for clarity
        sesCtl_p->instrumentId   = UseMODISInputs;
        
        // set default grid names, later allow user to override
        if(sesCtl_p->inputCollectionId == 5)
            strcpy(sesCtl_p->dsNameGrid,"MOD_Grid_MOD15A2");
        else if(sesCtl_p->inputCollectionId == 6)
            strcpy(sesCtl_p->dsNameGrid,"MOD_Grid_MOD15A2H");
        else
            xLogMsg("@F In(%s) inputCollectionId supports only 5 and 6 now",fxName);    
            
    } else if(strcmp(sesCtl_p->fparMethodStr,"VIIRS")==0) {
        sesCtl_p->fparMethodCode =   UseVIIRsInputs ; // pull from enum
        // <JMG> 2021-08-09, for clarity in VIIRS vs. MODIS QC  handling.
        sesCtl_p->instrumentId   = UseVIIRsInputs;    // pull from enum
        
        strcpy(sesCtl_p->dsNameGrid, "VNP_Grid_VNP15A2H" );
    }

    // Allow user override of input grid name here    
    if(nsbDoesKeyExist(nsbMgr_p,"L4CMOD_GRID_NAME"))
    {
       strcpy(sesCtl_p->dsNameGrid,nsbGetValueByKey(nsbMgr_p,"L4CMOD_GRID_NAME",nvpBuf));    
    }
    
    // Retrieve L4C_OUTPUT_DATASET_xxx names. These are all specified regardless
    // of which exact ones are used.
    strcpy(sesCtl_p->dsOutput_Fpar    ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_OUTPUT_DATASET_FPAR"   ,nvpBuf));
    strcpy(sesCtl_p->dsOutput_FparQc  ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_OUTPUT_DATASET_FPAR_QC",nvpBuf));

    // INPUTs   
    strcpy(sesCtl_p->dsNameFpar       ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR"   ,nvpBuf));
    strcpy(sesCtl_p->dsNameFparQc     ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR_QC",nvpBuf));
    // <JMG> 2021-07-18T16:07, re-enable as ExtraQc (renamed QcExtra to ExtraQc to match actual name
    strcpy(sesCtl_p->dsNameFparExtraQc,nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_DATASET_FPAR_EXTRA_QC",nvpBuf));
    // DEPRECATED all NDVI..
    
    strcpy(sesCtl_p->dsNamePFT_M09_Mask,nsbGetValueByKey(nsbMgr_p,"L4C_ANC_DATASET_PFT_M09_MASK"    ,nvpBuf));
    
    strcpy(sesCtl_p->dsNameTileRowM09  ,nsbGetValueByKey(nsbMgr_p,"L4C_ANC_DATASET_TILE_ROW_M01"    ,nvpBuf));
    strcpy(sesCtl_p->dsNameTileColM09  ,nsbGetValueByKey(nsbMgr_p,"L4C_ANC_DATASET_TILE_COL_M01"    ,nvpBuf));
    strcpy(sesCtl_p->dsNameTileHoriM09 ,nsbGetValueByKey(nsbMgr_p,"L4C_ANC_DATASET_TILE_HORI_HH_M01",nvpBuf));
    strcpy(sesCtl_p->dsNameTileVertM09 ,nsbGetValueByKey(nsbMgr_p,"L4C_ANC_DATASET_TILE_VERT_VV_M01",nvpBuf));
    
    // L4C_ANC_DATASET_PFT_M09_MASK =  MCD12Q1_M09_lc_mask_uint8
    strcpy(sesCtl_p->dsNamePFT_M09_Mask, nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_PFT_M09_MASK",nvpBuf));
    
    // latitude, longitude 1km coord vars used here:
    // L4C_ANC_DATASET_LATITUDE_1KM_COORD  = SMAP_L4_C_LAT_14616_x_34704_M01_flt32
    // L4C_ANC_DATASET_LONGITUDE_1KM_COORD = SMAP_L4_C_LON_14616_x_34704_M01_flt32
    // Note that the 9km LAT,LON may also be defined but are not yet used in L4CMOD
    strcpy(sesCtl_p->dsNameLatCoordvar,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_LATITUDE_1KM_COORD",nvpBuf));
    strcpy(sesCtl_p->dsNameLonCoordvar,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_LONGITUDE_1KM_COORD",nvpBuf));
    
    // MCD12Q1_M09land_col_uint16
    strcpy(sesCtl_p->dsNameLandRowIndex,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_M09LAND_ROW_INDEX",nvpBuf)); 
    // MCD12Q1_M09land_row_uint16
    strcpy(sesCtl_p->dsNameLandColIndex,nsbGetValueByKey(nsbMgr_p, "L4C_ANC_DATASET_M09LAND_COL_INDEX",nvpBuf)); 

    // Retrieve all ROOTDIR name-value-pairs. 
    // TBD, make these more robust, to better handle optional use-case when they
    //  are NOT specified or used, due to use of separate, fully-qualified pathnames
    // L4CMOD_INPUT_ROOTDIR
    
    strcpy(sesCtl_p->inputFparRootDir,nsbGetValueByKey(nsbMgr_p, "L4CMOD_INPUT_ROOTDIR",nvpBuf));
    xLogMsg("@A FPAR MODIS or VIIRs) INPUT ROOT directory (%s)~\n",sesCtl_p->inputFparRootDir);
    
    // Deprecated:    sesCtl_p->inputNdviRootDir 
    
    strcpy(sesCtl_p->logRootDir,"");
    strcpy(sesCtl_p->logRootDir,       nsbGetValueByKey(nsbMgr_p, "L4CMOD_OUTPUT_LOG_ROOTDIR",nvpBuf));

    strcpy(sesCtl_p->outputFparRootDir,nsbGetValueByKey(nsbMgr_p, "L4CMOD_OUTPUT_ROOTDIR",nvpBuf));
    
    strcpy(sesCtl_p->ancRootDir       ,nsbGetValueByKey(nsbMgr_p,"L4CMOD_INPUT_ANCILLARY_ROOTDIR",nvpBuf));

    // Retrieve all basepath name value pairs
    strcpy(sesCtl_p->inputAncPath  ,nsbGetValueByKey(nsbMgr_p, "L4CMOD_INPUT_ANCILLARY_PATH",nvpBuf));
    
    strcpy(sesCtl_p->outputFparPath,nsbGetValueByKey(nsbMgr_p, "L4CMOD_OUTPUT_PATH",nvpBuf));

    // query L4CMOD_INPUT_GRANULE_COUNT too, via DoCountStrictMatch
    sesCtl_p->nFparGranules = nsbGetKeywordCount(nsbMgr_p,"L4CMOD_INPUT_GRANULE_NAME",DoCountStrictMatch);
    
    xLogMsg("@A In(%s) Counted (%d) qualified INPUT FPAR granule names~ \n",fxName,sesCtl_p->nFparGranules);    
    
    // L4CMOD_OPTIMAL_TILECOUNT is typically 286 here, L4CMOD_OPTIMAL_TILECOUNT
    
    if(sesCtl_p->nFparGranules < L4CMOD_OPTIMAL_TILECOUNT)
    {
        xLogMsg("@W In(%s) WARNING Possibly insufficient FPAR granule count %d (of 286 expected) for a full global moasic!~ \n",
                fxName,sesCtl_p->nFparGranules);
    } 
    
    // Retrieve list of all nModisTiles (e.g.286 odd) FPAR 8-DAY tiles parsed from RIP file
    //  legacy: read_list(list,fnames,&ntiles);
    //  new   : uses name-value-pair sll mechanism to fetch all (286) tile paths into stringtable **fparTileList
    // string table storing (286) FPAR 8-day paths
    sesCtl_p->input_Fpar_pathList = nsbGetMultiValueByKey(nsbMgr_p,"L4CMOD_INPUT_GRANULE_NAME",
                                         &sesCtl_p->nFparGranules);

    // Next, sort string table of FPAR paths, using qsort(), using natural lexical ordering
    qsort(sesCtl_p->input_Fpar_pathList, sesCtl_p->nFparGranules, 
            sizeof(sesCtl_p->input_Fpar_pathList[0]), l4c_strCompare);
    
    // Next, be SURE to explicitly EXPAND the width of entries in the orig string table, 
    // prior to pre-pend of root dir, allowing the destination stringtable to accept wider pre-pend results!
    
    sesCtl_p->input_Fpar_pathList = frm_strExpandStringTbl(sesCtl_p->nFparGranules,
            strlen(sesCtl_p->inputFparRootDir), sesCtl_p->input_Fpar_pathList );
            
    /*
    for(ithGrn=0; ithGrn < sesCtl_p->nFparGranules; ithGrn++)
     xLogMsg("@A DIAG-PATH (%s)~ \n",sesCtl_p->input_Fpar_pathList[ithGrn]);
    */
    
    // Adjust ALL PATHS to PRE-PEND ROOTDIR if a relevant ROOT is supplied!
    xLogMsg("@A In(%s) Applying root directory (%s) to %d qualified input path~ \n",
        fxName, sesCtl_p->inputFparRootDir, sesCtl_p->nFparGranules);    
        
    l4c_ApplyRootDirToPaths(sesCtl_p);
    
} // end::l4c_SesCollectNvps()


// --------------------------------------------------------------------
// purpose  : Fetch the HDF5 dataset name in SMAP_L4_C_FparClimatology.h5
//            matching this YEARDAY and 8-day PERIOD
// function : l4c_SetFparClimatologyDatasetName()
// inputs   : char *lutPath_p -- name
// returns  : int, status 0 if ok -1 if failed
// outputs  : char *dsetName_p -- FparClimatology dataset name requested
//            that correctly matches the 'compositePeriodIdx' supplied.
// dependencies: LUT named  fparClimatologyNameLut must exist, and is 
//          : usually defined as a module level static array of structures,
//            where in this case, it is defined in THIS module file.
// notes    : may change to return char *dsetName_p 
int l4c_SetFparClimatologyDatasetName(const int compositePeriodIdx,char *lutPath_p,char *dsetName_p)
{
 // SesCtl *sesCtl_p = l4c_GetSesCtl();
 char fxName[256]= {0};
 char **doyPeriodLutStrTbl_p = NULL;
 char lutRecdStr[64] = {0} ;
 char dsetBuffer[64] = {0} ;
 char *periodStr_p =NULL;
 int ithLine=0 ;
 int nLines =0 ; // count of lines read as returned ...
 int lutRecPeriodIndex =0; 
 int maxWidthCharFound =0;
 
    strcpy(fxName,__func__);
    
    // LUT data now supplied via a static hard wired 
    // string table instead of being read from a file as below.
    
    // doyPeriodLutStrTbl_p = frm_ioReadTextIntoTable(lutPath_p, 
    //  DoIgnoreComments, &maxWidthCharFound, &nLines);
    
    // now uses static string table method
    nLines = 46;
    // set stringtable pointer to point to built in static table
    doyPeriodLutStrTbl_p = fparClimatologyNameLut;
    
    // scan for match on sesCtl_p->runCompositePeriod
    
    for(ithLine=0; ithLine < nLines; ithLine++)
    {
        // extract the period index at end of record
        strcpy(lutRecdStr,doyPeriodLutStrTbl_p[ithLine] );
        
        periodStr_p = strstr(lutRecdStr,"_per")+strlen("_per");
        lutRecPeriodIndex = atoi(periodStr_p);
        // if invalid, skip ahead
        if(periodStr_p == NULL || lutRecPeriodIndex > 46)
        {
            xLogMsg("@E In(%s) error reading LUT(%s) at line %d ~\n",
                fxName,lutPath_p,ithLine);
            continue;
        }
    
        // on match, fetch whole 4th token as the
        // FparClimatology dataset name for this YEARDAY PERIOD combo
        if( compositePeriodIdx == lutRecPeriodIndex)
        {
            // fetch 4th token, identified as 3rd base0 index
            frm_strGetToken(dsetBuffer,lutRecdStr,3,",");
            if(strlen(dsetBuffer)>0)
            {
                    strcpy(dsetName_p,dsetBuffer);
                    break;
            } // end if valid result
        } // end:: block on correct match
    } // end:: ithLine

    // memFreeStringTbl(doyPeriodLutStrTbl_p,nLines);
    
 // also returns base0 index of line for matching LUT record
 return ithLine;    
} // end:: l4c_SetFparClimatologyDatasetName()
